
! A version with non-blocking receive and blocking send. Doesn't need 
! EAGER_LIMIT set for communication.

subroutine pressure_gmg
	
!   parallel iterative pressure-solver for 3D large domains.
!   with multigrid solver in y-z
!   (C) 2015, Marat Khairoutdinov


use vars
use params, only: dowallx, dowally, docolumn, dolatlon, ggr, earth_factor
implicit none

integer, parameter :: nx_s=nx_gl/nsubdomains ! width of the x-slabs
integer, parameter :: ny_s=ny_gl/nsubdomains ! width of the y-slabs

! Slabs:
real fx(nx_gl, ny_s, nzm) ! slab for x-pass Fourier coefs
real fy(ny_gl, nx_s, nzm) ! slab for y-pass Fourier coefs
real(8) gx(nx_gl+2, ny_s) ! array to perform FFT in x

! Message buffers:
real bufx1(nx, ny_s, nzm)
real bufx2(nx, ny_s, nzm, max(1,nsubdomains_x))
real bufy1(nx_s, ny, nzm)
real bufy2(nx_s, ny, nzm, max(1,nsubdomains_y))

! FFT stuff:
real(8) work(max((nx_gl+3)*(ny_s+1),(nx_s+1)*(ny_gl+2)))
real(8) trigxi(3*nx_gl/2+1)
integer ifaxi(100)

! iterative solver's solver coefficients:
real(8) alpha
real(8) yy(ny_gl+1),pp(nz),zi8(nz)
real(8) ff(ny_gl,nzm)
real(8) gg(ny_gl+1,nzm+1)
real(8) xi,xnx,ddx,pii,eps,resnorm,error
real(8) ay1(ny_gl,nzm),cy1(ny_gl,nzm)
real(8) az1(ny_gl,nzm),cz1(ny_gl,nzm),b1(ny_gl,nzm)
real zonalv(ny_gl,nzm), zonalw(ny_gl,nzm)
integer niter,unit,coarse_size,v_cyc,pre_smooth,post_smooth

integer reqs_in(nsubdomains)
integer i, j, k, id, m, n, it, jt, tag, ii,jj,jb,jc,kb,kc
integer irank, rnk
integer n_in, count
logical flag(nsubdomains)
integer jwall

! for wrapping p
real buff_ew1(ny,nzm), buff_ns1(nx,nzm)
real buff_ew2(ny,nzm), buff_ns2(nx,nzm)
integer rf, tagrf
logical waitflag

! Make sure that the grid is suitable for the solver:

if(.not.dowally) then
  if(masterproc) print*,'pressure_big: iterative solver requires a wall in y (dowall=.true.). STOP'
  call task_abort
end if
if(mod(nx_gl,nsubdomains).ne.0) then
  if(masterproc) print*,'pressure_big: nx_gl/nsubdomains is not round number. STOP'
  call task_abort
endif
if(mod(ny_gl,nsubdomains).ne.0) then
  if(masterproc) print*,'pressure_big: ny_gl/nsubdomains is not round number. STOP'
  call task_abort
endif
if(dowallx) then
  if(masterproc) print*,'pressure_big: dowallx cannot be used with it. STOP'
  call task_abort
end if

!-----------------------------------------------------------------

if(docolumn) return
if(RUN2D) then
  print*,'pressure3D() cannot be called for 2D domains. Quitting...'
  call task_abort()
endif


!==========================================================================
!  Compute the r.h.s. of the Poisson equation for pressure

call t_startf ('press_rhs')

call press_rhs()

call t_stopf ('press_rhs')
! variable p will also be used as grid decomposition placeholder between transposes
!   for the fourier coefficients

!==========================================================================
!   Form the vertical slabs (x-z) of right-hand-sides of Poisson equation 
!   for the FFT - one slab per a processor.

call t_startf ('press_fft1')

call transpose_x(fx)

!==========================================================================
! Perform Fourier transformation n x-direction for a slab:
! This routing produces  N+2 coefficients for input array of langth N.
! However, two of those (2, and N+2) are identical zeros, so there is no point to
! send them around during transpose. That's why the temporary array gx is used to
! conform to requirements of fft991_crm subroutine.

call fftfax_crm(nx_gl,ifaxi,trigxi)

do k=1,nzm
  gx(1:nx_gl,1:ny_s) = fx(1:nx_gl,1:ny_s,k)
  call fft991_crm(gx,work,trigxi,ifaxi,1,nx_gl+2,nx_gl,ny_s,-1)
  fx(1,1:ny_s,k) = gx(1,1:ny_s)
  fx(2:nx_gl,1:ny_s,k) = gx(3:nx_gl+1,1:ny_s)
end do

call task_barrier()

!=========t=================================================================
!   Form blocks again from x-z slabs   

call transpose_x_inv(fx)

call t_stopf ('press_fft1')
call t_startf ('press_solver')
!==========================================================================
!   Form the vertical slabs (y-z) of Fourier coefs  
!   for the FFT - in y, one slab per a processor.

call transpose_y(fy)

!==========================================================================
! solve 2D elliptic problem by iterations:

if(dolatlon) then
 unit = 0
else
 unit = 1
end if
coarse_size=8
v_cyc=15
pre_smooth=1
post_smooth=1
error_max = 0.
niter_max = 0
pii = acos(-1.d0)
xnx=nx_gl
it=rank*nx_s
pp(1)=presi(1)*100./ggr
zi8(1) = zi(1)
do k=2,nz
 pp(k) = pp(k-1)-rho(k-1)*(zi(k)-zi(k-1))
 zi8(k) = zi(k)
end do
do j=1,ny_gl+1
 yy(j) = (j-ny_gl/2-1.)*dy
end do
ddx=1./(dx*dx)
eps = 1.e-4
if(.not.gmg_initialized) then   
  call neu_gmg_init(ny_gl,nzm,coarse_size,yy,zi8,unit,dble(earth_factor),pp)
  if(nstep.eq.1) then
    pfy = 0
  end if
  gmg_initialized = .true.
end if
do i=1,nx_s
    id=(i+1+it-0.1)/2.
    xi=id
    do k=1,nzm
     ff(:,k)=fy(:,i,k)*mu_gl(:)**2
    end do
    alpha = (2.d0-2.d0*cos(2.d0*pii/xnx*xi))*ddx
    call neu_gmg_solve(alpha,ff,eps,v_cyc,pre_smooth,post_smooth,pfy(:,:,i),resnorm,niter)
    niter_max = max(niter_max, niter)
    error_max = max(error_max, resnorm)
    do k=1,nzm
     do j=1,ny_gl
        fy(j,i,k)=pfy(j,k,i)
     end do
    end do
end do

!==========================================================================
!   Form blocks again from y-z slabs   

call transpose_y_inv(fy)

call t_stopf ('press_solver')
call t_startf ('press_fft2')
!==========================================================================
!   Form the vertical slabs (x-z) of Fourier coefs
!   for the inverse FFT - in x, one slab per a processor.

call transpose_x(fx)

! Perform inverse Fourier transform n x-direction for a slab:

 call fftfax_crm(nx_gl,ifaxi,trigxi)

 do k=1,nzm
  gx(1,1:ny_s) = fx(1,1:ny_s,k)
  gx(2,:) = 0.
  gx(3:nx_gl+1,1:ny_s) = fx(2:nx_gl,1:ny_s,k)
  gx(nx_gl+2,:) = 0.
  call fft991_crm(gx,work,trigxi,ifaxi,1,nx_gl+2,nx_gl,ny_s,1)
  fx(1:nx_gl,1:ny_s,k) = gx(1:nx_gl,1:ny_s)
 end do

call task_barrier()

call transpose_x_inv(fx)

call t_stopf ('press_fft2')
call t_startf ('press_grad')
!==========================================================================
!  Update the pressure fields in the subdomains
!

! when we cut back on the ffts wrap p here - look to sib for the model
!DD temporary measure for dompi
  if(dompi) then
!DD custom build a wrap that sends from the north and east edges, and receives at the
!DD    south and west edges

      if(rank==rankee) then
        p(0,1:ny,1:nzm) = p(nx,1:ny,1:nzm)
      else
        call task_receive_float(buff_ew2(:,:),ny*nzm,reqs_in(1))
          buff_ew1(1:ny,1:nzm) = p(nx,1:ny,1:nzm)
          call task_bsend_float(rankee,buff_ew1(:,:),ny*nzm,1)
          waitflag = .false.
        do while (.not.waitflag)
            call task_test(reqs_in(1),waitflag,rf,tagrf)
        end do
        call task_barrier()
        p(0,1:ny,1:nzm) = buff_ew2(1:ny,1:nzm)
      endif

      if(rank==ranknn) then
         p(:,1-YES3D,1:nzm) = p(:,ny,1:nzm)
      else
             call task_receive_float(buff_ns2(:,:),nx*nzm,reqs_in(1))
             buff_ns1(1:nx,1:nzm) = p(1:nx,ny,1:nzm)
             call task_bsend_float(ranknn,buff_ns1(:,:),nx*nzm,1)
             waitflag = .false.
         do while (.not.waitflag)
               call task_test(reqs_in(1),waitflag,rf,tagrf)
             end do
         call task_barrier()
         p(1:nx,1-YES3D,1:nzm) = buff_ns2(1:nx,1:nzm)
      endif

  else
    p(0,:,1:nzm) = p(nx,:,1:nzm)
    p(:,1-YES3D,1:nzm) = p(:,ny,1:nzm)
  endif
!DD end ugly wrap code.

!==========================================================================
!  Ad2d pressure gradient term to the rhs of the momentum equation:

call press_grad()

call t_stopf ('press_grad')
!==========================================================================
!==========================================================================
!==========================================================================

contains

!==========================================================================
   subroutine transpose_x(f)

! transpose from blocks to x-z slabs

      REAL, INTENT(OUT) :: f(nx_gl, ny_s, nzm)
      
      irank = rank-mod(rank,nsubdomains_x)  

      n_in = 0
      do m = irank, irank+nsubdomains_x-1

        if(m.ne.rank) then

          n_in = n_in + 1
          call task_receive_float(bufx2(:,:,:,n_in),nx*ny_s*nzm,reqs_in(n_in))
          flag(n_in) = .false.

        end if

      end do ! m

      do m = irank, irank+nsubdomains_x-1

        if(m.ne.rank) then

          n = m-irank

          bufx1(:,:,:) = p(1:nx,n*ny_s+1:n*ny_s+ny_s,1:nzm)
          call task_bsend_float(m,bufx1(:,:,:),nx*ny_s*nzm, 33) 

        endif

      end do ! m


! don't sent a buffer to itself, just fill directly.

      n = rank-irank
      call task_rank_to_index(rank,it,jt)
      f(1+it:nx+it,1:ny_s,1:nzm) = p(1:nx,n*ny_s+1:n*ny_s+ny_s,1:nzm)


      ! Fill slabs when receive buffers are full:

      count = n_in
      do while (count .gt. 0)
        do m = 1,n_in
         if(.not.flag(m)) then
      	    call task_test(reqs_in(m), flag(m), rnk, tag)
              if(flag(m)) then 
            	 count=count-1
                 call task_rank_to_index(rnk,it,jt)	  
                 f(1+it:nx+it,1:ny_s,1:nzm) = bufx2(1:nx,1:ny_s,1:nzm,m)
              endif   
          endif
         end do
      end do
      call task_barrier()
   end subroutine transpose_x
   
!==========================================================================
   subroutine transpose_x_inv(f)

! transpose from x-z slabs to blocks

      REAL, INTENT(IN) :: f(nx_gl, ny_s, nzm)
      
      irank = rank-mod(rank,nsubdomains_x)
      n_in = 0
      do m = irank, irank+nsubdomains_x-1

        if(m.ne.rank) then

          n_in = n_in + 1
          call task_receive_float(bufx2(:,:,:,n_in),nx*ny_s*nzm,reqs_in(n_in))
          flag(n_in) = .false.

        endif

      end do ! m

      do m = irank, irank+nsubdomains_x-1

        if(m.ne.rank) then

          call task_rank_to_index(m,it,jt)
          bufx1(:,:,:) = f(1+it:it+nx,1:ny_s,1:nzm)
          call task_bsend_float(m,bufx1(:,:,:),nx*ny_s*nzm, 33)

        endif

      end do ! m

! don't sent a buffer to itself, just fill directly.

      n = rank-irank
      call task_rank_to_index(rank,it,jt)
      p(1:nx,n*ny_s+1:n*ny_s+ny_s,1:nzm) = f(1+it:nx+it,1:ny_s,1:nzm)  

! Fill slabs when receive buffers are full:

      count = n_in
      do while (count .gt. 0)
        do m = 1,n_in
         if(.not.flag(m)) then
              call task_test(reqs_in(m), flag(m), rnk, tag)
              if(flag(m)) then
                 count=count-1
                 n = rnk-irank
                 p(1:nx,n*ny_s+1:n*ny_s+ny_s,1:nzm) = bufx2(1:nx,1:ny_s,1:nzm,m)
              endif
         endif
        end do
      end do

      call task_barrier()
   end subroutine transpose_x_inv

!==========================================================================
   subroutine transpose_y(f)

! transpose from blocks to y-z slabs

      REAL, INTENT(OUT) :: f(ny_gl, nx_s, nzm)
      
      irank = rank / nsubdomains_y  

      n_in = 0
      do m = irank, nsubdomains-1, nsubdomains_x

        if(m.ne.rank) then

          n_in = n_in + 1
          call task_receive_float(bufy2(:,:,:,n_in),ny*nx_s*nzm,reqs_in(n_in))
          flag(n_in) = .false.
          
        else
! don't sent a buffer to itself, just fill directly.

          n = mod(rank,nsubdomains_y)
          call task_rank_to_index(rank,it,jt)
          do i = 1,nx_s	  
            f(1+jt:ny+jt,i,1:nzm) = p(n*nx_s+i,1:ny,1:nzm)
          enddo

        end if

      end do ! m

      irank = nsubdomains_y*mod(rank,nsubdomains_x)
      do m = irank, irank+nsubdomains_y-1

        if(m.ne.rank) then

          n = m-irank

          bufy1(:,:,:) = p(n*nx_s+1:n*nx_s+nx_s,1:ny,1:nzm)
          call task_bsend_float(m,bufy1(:,:,:),ny*nx_s*nzm, 33) 

        endif

      end do ! m


      ! Fill slabs when receive buffers are full:

      count = n_in
      do while (count .gt. 0)
        do m = 1,n_in
         if(.not.flag(m)) then
      	    call task_test(reqs_in(m), flag(m), rnk, tag)
              if(flag(m)) then 
            	 count=count-1
                 call task_rank_to_index(rnk,it,jt)
                 do i = 1,nx_s	  
                   f(1+jt:ny+jt,i,1:nzm) = bufy2(i,1:ny,1:nzm,m)
                 enddo
              endif   
          endif
         end do
      end do

      call task_barrier()
   end subroutine transpose_y
   
!==========================================================================
   subroutine transpose_y_inv(f)

! transpose from y-z slabs to blocks

      REAL, INTENT(IN) :: f(ny_gl, nx_s, nzm)
      
      n_in = 0
      irank = nsubdomains_y*mod(rank,nsubdomains_x)
      do m = irank, irank+nsubdomains_y-1

        if(m.ne.rank) then

          n_in = n_in + 1
          call task_receive_float(bufy2(:,:,:,n_in),ny*nx_s*nzm,reqs_in(n_in))
          flag(n_in) = .false.
          
        else

! don't sent a buffer to itself, just fill directly.

          n = rank-irank
          call task_rank_to_index(rank,it,jt)
          do i = 1,nx_s
            p(n*nx_s+i,1:ny,1:nzm) = f(1+jt:ny+jt,i,1:nzm) 
          enddo 

        endif

      end do ! m

      irank = rank / nsubdomains_y  
      do m = irank, nsubdomains-1, nsubdomains_x

        if(m.ne.rank) then

          call task_rank_to_index(m,it,jt)
          do i = 1,nx_s
            bufy1(i,:,:) = f(1+jt:jt+ny,i,1:nzm)
          enddo
          call task_bsend_float(m,bufy1(:,:,:),ny*nx_s*nzm, 33)

        endif

      end do ! m

! Fill slabs when receive buffers are full:

      irank = nsubdomains_y*mod(rank,nsubdomains_x)
      count = n_in
      do while (count .gt. 0)
        do m = 1,n_in
         if(.not.flag(m)) then
              call task_test(reqs_in(m), flag(m), rnk, tag)
              if(flag(m)) then
                 count=count-1
                 n = rnk-irank
                 p(n*nx_s+1:n*nx_s+nx_s,1:ny,1:nzm) = bufy2(1:nx_s,1:ny,1:nzm,m)
              endif
         endif
        end do
      end do

      call task_barrier()
   end subroutine transpose_y_inv

end 

