module simple_ocean

!------------------------------------------------------------
! Purpose:
!
! A collection of routines used to specify fixed 
! or compute interactive SSTs, like slab-ocean model, etc.
!
! Author: Marat Khairoutdinov
! Based on dynamic ocean impelemntation from the UW version of SAM.
!------------------------------------------------------------

use grid
implicit none

public set_sst     ! set SST 
public sst_evolve ! evolve SST according to a model set by the ocean_type

CONTAINS


SUBROUTINE set_sst()

use vars, only: sstxy,t00,landmask
use params, only: tabs_s, delta_sst, ocean_type, sst_mean, pi, &
                   rad_earth, readsst, earth_factor

! parameters of the sinusoidal SST destribution 
! along the X for Walker-type simulatons( ocean-type = 1):

real(8) tmpx(nx), lx, ly, yy, phi
real(4) mask(nx_gl,ny_gl)
integer i,j, it,jt,nx1,ny1

if(readsst) return

select case (ocean_type)

   case(0) ! fixed constant SST

     do j=1,ny
       do i=1,nx
         if(landmask(i,j).eq.0) sstxy(i,j) = tabs_s - t00
       end do
     end do
     sst_mean = tabs_s

   case(1) ! Sinusoidal distribution along the x-direction:

     lx = float(nx_gl)*dx
     do i = 1,nx
        tmpx(i) = float(mod(rank,nsubdomains_x)*nx+i-1)*dx
     end do
     do j=1,ny
       do i=1,nx
         if(landmask(i,j).eq.0) sstxy(i,j) = tabs_s-delta_sst*cos(2.*pi*tmpx(i)/lx) - t00
       end do
     end do
   
   case(2) ! Sinusoidal distribution along the y-direction:
     
     call task_rank_to_index(rank,it,jt)
     
     lx = float(ny_gl)*dy
     do j=1,ny
        yy = dy*(j+jt-(ny_gl+YES3D-1)/2-1)
       do i=1,nx
         if(landmask(i,j).eq.0) sstxy(i,j) = tabs_s+delta_sst*(2.*cos(pi*yy/lx)-1.) - t00
       end do
     end do 


   case(3) !  distribution along the y-direction:

     call task_rank_to_index(rank,it,jt)

     ly = float(ny_gl)*dy
     do j=1,ny
        yy = dy*(j+jt-(ny_gl+YES3D-1)/2-1)
       do i=1,nx
         if(landmask(i,j).eq.0.) sstxy(i,j) = tabs_s+delta_sst/(1+(yy/ly/0.4)**6)-t00
       end do
     end do

   case(4)

     call task_rank_to_index(rank,it,jt)

     lx = float(nx_gl)*dx
     ly = 0.5*float(ny_gl)*dy
     do j=1,ny
        yy = dy*(j+jt-(ny_gl+YES3D-1)/2-1)
       do i=1,nx
        if(landmask(i,j).eq.0.) then
         ! approx SST disribution in January averaged over 60E-90E
         sstxy(i,j) = tabs_s+delta_sst*cos(pi*min(1.,abs(yy+333333.)/ly)**1.2)-t00
         tmpx(i) = float(mod(rank,nsubdomains_x)*nx+i-1)*dx
         sstxy(i,j) = sstxy(i,j)-0.5*exp(-(2.*yy/ly)**2)*delta_sst*cos(2.*pi*tmpx(i)/lx)
        end if
       end do
     end do

   case(5) !  Aquaplanet Experiment (APE) "Control" (Neale and Hoskins 2001)

     call task_rank_to_index(rank,it,jt)

     lx = float(nx_gl)*dx
     ly = 0.5*float(ny_gl)*dy
     do j=1,ny
        yy = dy*(j+jt+0.5-(ny_gl+YES3D-1)/2-1)
        phi = pi/180.*yy/(2.*pi*rad_earth/earth_factor)*360.
       do i=1,nx
         if(landmask(i,j).eq.0.) then
           if(abs(phi).le.pi/3.) then
               sstxy(i,j) = tabs_s+delta_sst*(1-sin(1.5*phi)**2)-t00
           else
               sstxy(i,j) = tabs_s - t00
           end if
         end if
       end do
     end do

   case(6) !  Aquaplanet Experiment (APE) "QOBS" (Neale and Hoskins 2001)

     call task_rank_to_index(rank,it,jt)

     lx = float(nx_gl)*dx
     ly = 0.5*float(ny_gl)*dy
     do j=1,ny
        yy = dy*(j+jt+0.5-(ny_gl+YES3D-1)/2-1)
        phi = pi/180.*yy/(2.*pi*rad_earth/earth_factor)*360.
       do i=1,nx
         if(landmask(i,j).eq.0.) then
           if(abs(phi).le.pi/3.) then
              sstxy(i,j) = tabs_s+delta_sst*(1-0.5*(sin(1.5*phi)**2+sin(1.5*phi)**4))-t00
           else
              sstxy(i,j) = tabs_s - t00
           end if
         end if
       end do
     end do

   case(7) !  Aquaplanet Experiment (APE) FLAT 

     call task_rank_to_index(rank,it,jt)

     lx = float(nx_gl)*dx
     ly = 0.5*float(ny_gl)*dy
     do j=1,ny
        yy = dy*(j+jt+0.5-(ny_gl+YES3D-1)/2-1)
        phi = pi/180.*yy/(2.*pi*rad_earth/earth_factor)*360.
       do i=1,nx
         if(landmask(i,j).eq.0.) then
           if(abs(phi).le.pi/3.) then
             sstxy(i,j) =  tabs_s+delta_sst*(1-sin(1.5*phi)**4)-t00
           else
             sstxy(i,j) = tabs_s - t00
           end if
         end if
       end do
     end do

   case default

     if(masterproc) then
         print*, 'unknown ocean type in set_sst. Exitting...'
         call task_abort
     end if

end select

end subroutine set_sst



SUBROUTINE sst_evolve
 use vars, only: sstxy, t00, fluxbt, fluxbq, rhow, qocean_xy, landmask, sst_xy
 use params, only: dodynamicocean, cp, lcond, tabs_s, ocean_type, dossthomo, &
                   depth_slab_ocean, Szero, deltaS, timesimpleocean, &
                   sst_climo, tau_ocean, dosstclimo, sst_mean
 use rad, only: swnsxy, lwnsxy
 use terrain, only: k_terra
 use slm_vars, only: seaicemask

 real, parameter :: rhor = 1000. ! density of water (kg/m3)
 real, parameter :: cw = 4187.   ! Liquid Water heat capacity = 4187 J/kg/K
 real factor_cp, factor_lc, qoceanxy
 real tmpx(nx), lx
 real(8) sss(2),ssss(2),sstnudge
 integer i,j

   if(dodynamicocean) then
      if(time.lt.timesimpleocean) return
      lx = float(nx_gl)*dx
      do i = 1,nx
        tmpx(i) = float(mod(rank,nsubdomains_x)*nx+i-1)*dx
      end do

      ! Use forward Euler to integrate the differential equation
      ! for the ocean mixed layer temperature: dT/dt = S - E.
      ! The source: CPT?GCSS WG4 idealized Walker-circulation 
      ! RCE Intercomparison proposed by C. Bretherton.
      do j=1,ny
         do i=1,nx
           if(landmask(i,j).eq.0.and.seaicemask(i,j).eq.0) then
             if(dosstclimo) then
               qoceanxy = -(sst_mean-sst_climo)/tau_ocean*rhor*cw*depth_slab_ocean
             else
               qoceanxy = Szero + deltaS*abs(2.*tmpx(i)/lx - 1)
             end if
	     qocean_xy(i,j) = qocean_xy(i,j) + qoceanxy

             factor_cp = rhow(k_terra(i,j))*cp
             factor_lc = rhow(k_terra(i,j))*lcond
             sstxy(i,j) = sstxy(i,j) &
                   + dtn*(swnsxy(i,j)          & ! SW Radiative Heating
                   - lwnsxy(i,j)               & ! LW Radiative Heating
                   - factor_cp*fluxbt(i,j)     & ! Sensible Heat Flux
                   - factor_lc*fluxbq(i,j)     & ! Latent Heat Flux
                   + qoceanxy)            & ! Ocean Heating
                   /(rhor*cw*depth_slab_ocean)         ! Convert W/m^2 Heating to K/s
                
           end if
         end do
      end do

     if(dossthomo.or.dosstclimo) then
        sss = 0.
        do j=1,ny
         do i=1,nx
           if(landmask(i,j).eq.0) then
            sss(1) = sss(1) + sstxy(i,j)
            sss(2) = sss(2) + 1
           end if
         end do
        end do
        if(dompi) then
            call task_sum_real8(sss,ssss,2)
            sss = ssss
        end if ! dompi
        if(sss(2).eq.0.) then
           if(masterproc) print*,'no ocean points found in dynamic ocean!'
           call task_abort
        end if
        sst_mean = sss(1) / sss(2) + t00
        if(dossthomo) then
          sstxy = sst_mean - t00
          tabs_s = sst_mean
        end if
     end if

   end if ! dodynamic ocean

   do j=1,ny
      do i=1,nx
        if(landmask(i,j).eq.0.and.seaicemask(i,j).eq.0) then
          sst_xy(i,j)=sst_xy(i,j)+sstxy(i,j)*dtfactor   ! 2D output
        end if
      end do
   end do

end subroutine sst_evolve


end module simple_ocean
