module ocn_seaice
implicit none

! simple sea-ice model
! (c) Marat Khairtoutdinov, Apr 2019

!single-slab of ice that grows in thickness only to keep the sea-water
!immediately below at local freezing temperature (depending on salinity).
!because extremely low conductivity of ice, no heat diffusion in the ice is allowed.
!the surface has no sensible heat flux, that is, the surface temperature is equal to 
!the environment. The evaporation flux is not neglected though, which
!also affects the ice thickness. 

real, parameter :: rho_ice = 934. ! kg/m3, density of ice
real, parameter :: cp_ice = 2108. ! J/kgK,  specific heat of ice    
real, parameter :: cp_wat = 3992. ! J/kgK,  specific heat of water    
real, parameter :: L_fr = 0.3336e6 ! J/kg,  latent heat of freezing/melting    

contains


subroutine ocn_seaice_init(h, t_wat, h_wat, s_wat, rho_wat, t_ice)

! creat new sea-ice if water temperature is below freezing

 real, intent(in out) :: h ! m, ice thickness
 real, intent(in out) :: t_wat ! oC, water temperature
 real, intent(in out) :: s_wat ! g/kg, salinity of water
 real, intent(in) :: rho_wat ! kg/m3, density  of water
 real, intent(in) :: h_wat ! m, depth of upper layer of water
 real, intent(in out) :: t_ice ! oC, mean ice temperature

 real t_fr, h1, h_old

 if(h.gt.0.) return

 t_fr = -s_wat/35.*1.8

! check if water temperature is below freezing
! if yes, add some ice keeping water temperature at freezing

 if(t_wat.lt.t_fr) then
 ! add ice until tmperature of water reaches freezing:
  h1 = cp_wat*h_wat*(t_fr-t_wat)/(L_fr+cp_wat*(t_fr-t_wat) &
                      -cp_ice*(t_fr-t_ice))*rho_wat/rho_ice
  t_wat = t_fr
  t_ice = t_fr
  h = h1
  s_wat = s_wat + h/(h_wat-h)*rho_ice/rho_wat*s_wat
 end if

end subroutine ocn_seaice_init


subroutine ocn_seaice_proc(h, t_wat, h_wat, s_wat, rho_wat, t_ice)
 
! grow/melt an existing sea-ice
! for new ice, use seaice_init

 real, intent(in out) :: h ! m, ice thickness
 real, intent(in out) :: t_wat ! oC, water temperature
 real, intent(in out) :: s_wat ! g/kg, salinity of water
 real, intent(in) :: rho_wat ! kg/m3, density  of water
 real, intent(in) :: h_wat ! m, depth of upper layer of water 
 real, intent(in out) :: t_ice ! oC, mean ice temperature

 real t_fr, h1, h_old

 h_old = h
 t_fr = -s_wat/35.*1.8

! First check if water temperature is below freezing
! if yes, add some ice keeping water temperature at freezing

 if(t_wat.lt.t_fr) then
 ! add ice until tmperature of water reaches freezing:
  h1 = cp_wat*h_wat*(t_fr-t_wat)/(L_fr+cp_wat*(t_fr-t_wat) &
                      -cp_ice*(t_fr-t_ice))*rho_wat/rho_ice
  t_wat = t_fr
  t_ice = (t_ice*h+t_fr*h1)/(h+h1)
  h = h + h1
 elseif(h.gt.0.) then
  ! otherwise melt ice and cool water 
  ! how much melting of ice it would take to cool water to freezing temperature
  h1 = cp_wat*h_wat*(t_wat-t_fr)/(L_fr-cp_ice*(t_ice-t_fr))*rho_wat/rho_ice
  if(h.le.h1) then
    ! ice is too thin, that is, all is melted
    t_wat = t_wat - h/h1*(t_wat-t_ice)
    t_ice = t_fr
    h = 0.
  else
    t_wat = t_fr
    h = h - h1
  end if
 else
  t_ice = t_fr
 end if

 s_wat = s_wat + (h-h_old)/(h_wat-h)*s_wat

end subroutine ocn_seaice_proc


end module ocn_seaice
