module ocn_sgs

! module ocn_for original SAM subgrid-scale SGS closure (Smagorinsky or 1st-order TKE)
! Marat Khairoutdinov, 2012

use ocn_grid, only: nx,nxp1,ny,nyp1,YES3D,nzm,nz,dimx1_s,dimx2_s,dimy1_s,dimy2_s 
use ocn_params, only: dosgs
implicit none

!----------------------------------------------------------------------
! Required definitions:

!!! prognostic scalar (need to be advected arround the grid):

integer, parameter :: nsgs_fields = 1   ! total number of prognostic sgs vars

real sgs_field(dimx1_s:dimx2_s, dimy1_s:dimy2_s, nzm, nsgs_fields)

!!! sgs diagnostic variables that need to exchange boundary information (via MPI):

integer, parameter :: nsgs_fields_diag = 2   ! total number of diagnostic sgs vars

! diagnostic fields' boundaries:
integer, parameter :: dimx1_d=0, dimx2_d=nxp1, dimy1_d=1-YES3D, dimy2_d=nyp1

real sgs_field_diag(dimx1_d:dimx2_d, dimy1_d:dimy2_d, nzm, nsgs_fields_diag)

logical:: advect_sgs = .false. ! advect prognostics or not, default - not (Smagorinsky)
logical, parameter:: do_sgsdiag_bound = .true.  ! exchange boundaries for diagnostics fields

! SGS fields that output by default (if =1).
integer, parameter :: flag_sgs3Dout(nsgs_fields) = (/0/)
integer, parameter :: flag_sgsdiag3Dout(nsgs_fields_diag) = (/0,0/)

real fluxbsgs (nx, ny, 1:nsgs_fields) ! surface fluxes 
real fluxtsgs (nx, ny, 1:nsgs_fields) ! top boundary fluxes 

!!! these arrays may be needed for output statistics:

real sgswle(nz,1:nsgs_fields)  ! resolved vertical flux
real sgswsb(nz,1:nsgs_fields)  ! SGS vertical flux
real sgsadv(nz,1:nsgs_fields)  ! tendency due to vertical advection
real sgslsadv(nz,1:nsgs_fields)  ! tendency due to large-scale vertical advection
real sgsdiff(nz,1:nsgs_fields)  ! tendency due to vertical diffusion

real tkmax(0:ny,nzm)  ! maximum allowed coef for diffusion
real tkmin(0:ny,nzm)  ! minimum coef for diffusion
!------------------------------------------------------------------
! internal (optional) definitions:

! make aliases for prognostic variables:

real tke(dimx1_s:dimx2_s, dimy1_s:dimy2_s, nzm)   ! SGS TKE
equivalence (tke(dimx1_s,dimy1_s,1),sgs_field(dimx1_s,dimy1_s,1,1))

! make aliases for diagnostic variables:

real tk  (dimx1_d:dimx2_d, dimy1_d:dimy2_d, nzm) ! SGS eddy viscosity
real tkh (dimx1_d:dimx2_d, dimy1_d:dimy2_d, nzm) ! SGS eddy conductivity
equivalence (tk(dimx1_d,dimy1_d,1), sgs_field_diag(dimx1_d, dimy1_d,1,1))
equivalence (tkh(dimx1_d,dimy1_d,1), sgs_field_diag(dimx1_d, dimy1_d,1,2))


real grdf_x(0:ny,nzm)! grid factor for eddy diffusion in x
real grdf_y(0:ny,nzm)! grid factor for eddy diffusion in y
real grdf_z(nzm)! grid factor for eddy diffusion in z

logical:: dosmagor   ! if true, then use ocn_Smagorinsky closure

! Local diagnostics:

real tkesbbuoy(nz), tkesbshear(nz),tkesbdiss(nz), tkesbdiff(nz)

CONTAINS

! required microphysics subroutines and function:
!----------------------------------------------------------------------
!!! Read microphysics options from prm (namelist) file

subroutine ocn_sgs_setparm()

  use ocn_grid, only: case, caseid, masterproc
  implicit none

  integer ierr, ios, ios_missing_namelist, place_holder

  !======================================================================
  NAMELIST /OCN_SGS_TKE/ &
       dosmagor ! Diagnostic Smagorinsky closure

  NAMELIST /BNCUIODSBJCB/ place_holder

  dosmagor = .false. ! default 

  !----------------------------------
  !  Read namelist for microphysics options from prm file:
  !------------
  open(55,file='./CASES/'//trim(case)//'/prm_ocn', status='old',form='formatted')

  read (UNIT=55,NML=BNCUIODSBJCB,IOSTAT=ios_missing_namelist)
  rewind(55) !note that one must rewind before searching for new namelists

  read (55,OCN_SGS_TKE,IOSTAT=ios)

  advect_sgs = .not.dosmagor

  if (ios.ne.0) then
     !namelist error checking
     if(ios.ne.ios_missing_namelist) then
        write(*,*) '****** ERROR: bad specification in OCN_SGS_TKE namelist'
        rewind(55)
        read (55,OCN_SGS_TKE)
        call task_abort()
     end if
  end if
  close(55)
   ! write namelist values out to file for documentation
   if(masterproc) then
      open(unit=55,file='./OUT_STAT/'//trim(case)//'_'//trim(caseid)//'.nml', form='formatted', position='append')
      write (unit=55,nml=OCN_SGS_TKE,IOSTAT=ios)
      write(55,*) ' '
      close(unit=55)
   end if

  ! END UW ADDITION
  !======================================================================

end subroutine ocn_sgs_setparm

!----------------------------------------------------------------------
!!! Initialize sgs:


subroutine ocn_sgs_init()

  use ocn_grid, only: nrestart, dx, dy, dz, adz, mu, masterproc, ady, adzw
  integer j,k

  if(nrestart.eq.0) then

     sgs_field = 0.
     sgs_field_diag = 0.

     fluxbsgs = 0.
     fluxtsgs = 0.

  end if

  if(masterproc) then
     if(dosmagor) then
        write(*,*) 'Smagorinsky SGS Closure'
     else
        write(*,*) 'Prognostic TKE 1.5-order SGS Closure'
     end if
  end if

  do k=1,nzm
     do j=0,ny
      grdf_x(j,k) = min(16.,(dx*mu(j))**2/(adz(k)*dz)**2)
      grdf_y(j,k) = min(16.,(dy*ady(j))**2/(adz(k)*dz)**2)
     end do
     grdf_z(k) = 1.
  end do

  sgswle = 0.
  sgswsb = 0.
  sgsadv = 0.
  sgsdiff = 0.
  sgslsadv = 0.

end subroutine ocn_sgs_init

!----------------------------------------------------------------------
!!! make some initial noise in sgs:
!
subroutine ocn_setperturb_sgs(ptype)

use ocn_vars, only: s0, z
integer, intent(in) :: ptype
integer i,j,k

select case (ptype)

  case(0)

     do k=1,nzm
      do j=1,ny
       do i=1,nx
         if(k.ge.nzm-4.and..not.dosmagor) then
            tke(i,j,k)=0.04*(5-k)
         endif
       end do
      end do
     end do

  case(1)

  case default

end select

end subroutine ocn_setperturb_sgs

!----------------------------------------------------------------------
!!! Estimate Courant number limit for SGS
!

subroutine ocn_kurant_sgs(cfl)

use ocn_grid, only: dt, dx, dy, dz, adz, adzw, imu, ady
use ocn_params, only: doimplicitdiff
implicit none

real, intent(out) :: cfl

integer j,k
real tkhxmax(nzm),tkhymax(nzm),tkhmax(nzm)

tkhxmax(:) = 0.
tkhymax(:) = 0.
do k = 1,nzm
 tkhmax(k) = maxval(tkh(1:nx,1:ny,k))
 do j = 1,ny
  tkhxmax(k) = max(tkhxmax(k),maxval(tkh(:,j,k)*grdf_x(j,k)*imu(j)**2))
  tkhymax(k) = max(tkhymax(k),maxval(tkh(:,j,k)*grdf_y(j,k)/ady(j)**2))
 end do
end do

cfl = 0.
if(doimplicitdiff) then
  do k=1,nzm
   cfl = max(cfl,        &
     tkhxmax(k)*dt/dx**2, &
     YES3D*tkhymax(k)*dt/dy**2)
  end do
else
  do k=1,nzm
    cfl = max(cfl,        &
     tkhmax(k)*grdf_z(k)*dt/(dz*adzw(k))**2, &
     tkhxmax(k)*dt/dx**2, &
     YES3D*tkhymax(k)*dt/dy**2)
  end do
end if

end subroutine ocn_kurant_sgs


!----------------------------------------------------------------------
!!! compute sgs diffusion of momentum:
!
subroutine ocn_sgs_mom()

   call ocn_diffuse_mom()

end subroutine ocn_sgs_mom

!----------------------------------------------------------------------
!!! compute sgs diffusion of scalars:
!
subroutine ocn_sgs_scalars()

  use ocn_vars
  use ocn_tracers
  use ocn_params, only: dotracers
  implicit none

    real dummy(nz)
    real fluxbtmp(nx,ny), fluxttmp(nx,ny) !bloss
    integer k

  call t_startf ('ocn_diffuse_scalars')

      call ocn_diffuse_scalar(t,fluxbt,fluxtt,tkh,tdiff,twsb, &
                           t2lediff,t2lediss,twlediff,.true.)   

      call ocn_diffuse_scalar(s,fluxbs,fluxts,tkh,sdiff,swsb, &
                           s2lediff,s2lediss,swlediff,.true.)
      where(s.lt.0.) s=0.    

      if(advect_sgs) then
         call ocn_diffuse_scalar(tke,fzero,fzero,tk*3.,dummy,sgswsb, &
                                    dummy,dummy,dummy,.false.)
         where(tke.lt.0.) tke=0.    
      end if


      if(dotracers) then

        call ocn_tracers_flux()

        do k = 1,ntracers

          fluxbtmp = fluxbtr(:,:,k)
          fluxttmp = fluxttr(:,:,k)
          call ocn_diffuse_scalar(tracer(:,:,:,k),fluxbtmp,fluxttmp, tkh,&
               trdiff(:,k),trwsb(:,k), &
               dummy,dummy,dummy,.false.)
          where(tracer(:,:,:,k).lt.0.) tracer(:,:,:,k)=0.    

        end do

      end if

  call t_stopf ('ocn_diffuse_scalars')


end subroutine ocn_sgs_scalars

!----------------------------------------------------------------------
!!! compute sgs processes (beyond advection):
!
subroutine ocn_sgs_proc()

   use ocn_grid, only: nstep, dt, icycle,dx, dy, dz, adz, mu, ady, adzw
   use ocn_params, only : doimplicitdiff, tk_factor
   real cx, cy, cz
   integer k, j

!    SGS TKE equation:

   if(dosgs) then
     call ocn_tke_full()

     do k=1,nzm
       if(doimplicitdiff) then
        cz = 1.e20
       else
        cz = (dz*min(adzw(k),adzw(k+1)))**2/dt/grdf_z(k)
       end if
       do j=0,ny
        ! compute maximum horizontal eddy coef for horizontal diffusion
        cx = (dx*mu(j))**2/dt
        cy = (dy*ady(j))**2/dt
        tkmax(j,k) = 0.09/(1./cx+1./cy+1./cz)  ! maximum value of eddy visc/cond
        tkmin(j,k) = tk_factor*tkmax(j,k)  ! minimum value of eddy visc/cond
       end do
    end do
   end if
end subroutine ocn_sgs_proc

!----------------------------------------------------------------------
!!! Diagnose arrays nessesary for dynamical core and statistics:
!
subroutine ocn_sgs_diagnose()
! None 

end subroutine ocn_sgs_diagnose


!----------------------------------------------------------------------
!!!! Collect microphysics history statistics (vertical profiles)
!
subroutine ocn_sgs_statistics()
  
  use ocn_vars
  use ocn_hbuffer, only: ocn_hbuf_put, ocn_hbuf_avg_put
  use ocn_params, only : lcond

  real tmp(2), factor_xy 
  real tkz(nzm), tkhz(nzm)
  integer i,j,k,n
  character(LEN=6) :: statname  !bloss: for conditional averages

  call t_startf ('sgs_statistics')

  factor_xy = 1./float(nx*ny)

  do k=1,nzm
    tkz(k) = 0.
    tkhz(k) = 0.
    do j=1,ny
    do i=1,nx
      tkz(k)=tkz(k)+tk(i,j,k)
      tkhz(k)=tkhz(k)+tkh(i,j,k)
    end do
    end do
  end do

  call ocn_hbuf_avg_put('TKES',tke,dimx1_s,dimx2_s,dimy1_s,dimy2_s,nzm,1.)

  call ocn_hbuf_put('TK',tkz,factor_xy)
  call ocn_hbuf_put('TKH',tkhz,factor_xy)

!---------------------------------------------------------
! SGS TKE Budget:

         call ocn_hbuf_put('ADVTRS',sgswle(:,1),factor_xy)
         call ocn_hbuf_put('BUOYAS',tkesbbuoy,factor_xy)
         call ocn_hbuf_put('SHEARS',tkesbshear,factor_xy)
         call ocn_hbuf_put('DISSIPS',tkesbdiss,factor_xy)

  call t_stopf ('sgs_statistics')

end subroutine ocn_sgs_statistics

!----------------------------------------------------------------------
! called when stepout() called

subroutine ocn_sgs_print()

 call ocn_fminmax_print('tke:',tke,dimx1_s,dimx2_s,dimy1_s,dimy2_s,nzm)
 call ocn_fminmax_print('tk:',tk,0,nxp1,1-YES3D,nyp1,nzm)
 call ocn_fminmax_print('tkh:',tkh,0,nxp1,1-YES3D,nyp1,nzm)

end subroutine ocn_sgs_print

!----------------------------------------------------------------------
!!! Initialize the list of sgs statistics 
!
subroutine ocn_sgs_hbuf_init(namelist,deflist,unitlist,status,average_type,count,sgscount)
character(*) namelist(*), deflist(*), unitlist(*)
integer status(*),average_type(*),count,sgscount

end subroutine ocn_sgs_hbuf_init


end module ocn_sgs



