module ocn_toplayer

! ocean top layer with variable depth as an interface to the atmosphere
! Marat Khairoutdinov, Dec 2017

use ocn_grid
implicit none

! prognostic fields:
real h(0:nx+1,0:ny+1)    ! layer thickness
real hu(0:nx+1,0:ny+1)   ! layer's x-momentum (h*u)
real hv(0:nx+1,0:ny+1)   ! layer's y momentum (h*v) 

! diagnostic fields

real us(nx+1,ny)  ! velocity in x
real vs(nx,ny+1)  ! velocity in y


real h0(0:nx+1,0:ny+1)  ! bottom topography height (relative to layer's bottom)

! layers topographic  functions
real terraul(0:nx+1,0:ny+1),terraur(0:nx+1,0:ny+1)
real terravl(0:nx+1,0:ny+1),terravr(0:nx+1,0:ny+1)

! misc:

 real, parameter :: eps = 1.e-9 ! padding to avoid division by zero

! MPI stuff:

real buffer(0:nx+1,0:ny+1,4) ! buffer for subdomain boundary exchange
integer nbuff  ! number of 2D fields in buffer

CONTAINS

!--------------------------------------------------------------------------------------------
!--------------------------------------------------------------------------------------------

subroutine toplayer_init(h_init, h0_init, u_init, v_init)

 real, intent(in) :: h_init(nx,ny)  ! initial top layer thickness (m, layers top is zero ocean level)
 real, intent(in) :: h0_init(nx,ny) ! initial relative height of topography (m, relative to zero ocean level)
 real, intent(in) :: u_init(nx,ny)  ! initial top layer x-velocity (m/s)
 real, intent(in) :: v_init(nx,ny)  ! initial top layer y_velocity (m/s)
 integer i,j
 
 h(1:nx,1:ny) = h_init(1:nx,1:ny)
 h0(1:nx,1:ny) = h(1:nx,1:ny) + h0_init(1:nx,1:ny)
 
 hu(1:nx,1:ny) = h_init(1:nx,1:ny)*u_init(1:nx,1:ny)
 hv(1:nx,1:ny) = h_init(1:nx,1:ny)*v_init(1:nx,1:ny)
 
 if(dompi) then
  buffer(:,:,1) = h(:,:)
  buffer(:,:,2) = h0(:,:)
  buffer(:,:,3) = hu(:,:)
  buffer(:,:,4) = hv(:,:)
  nbuff=4
  call ocn_task_exchange(buffer,0,nx+1,0,ny+1,nbuff,1,1,1,1,0)
  h(:,:)  = buffer(:,:,1)
  h0(:,:) = buffer(:,:,2)
  hu(:,:) = buffer(:,:,3)
  hv(:,:) = buffer(:,:,4)
  call task_barrier()
 else
  call ocn_bound_exchange(h,0,nx+1,0,ny+1,1,1,1,1,1,0)
  call ocn_bound_exchange(h0,0,nx+1,0,ny+1,1,1,1,1,1,0)
  call ocn_bound_exchange(hu,0,nx+1,0,ny+1,1,1,1,1,1,0)
  call ocn_bound_exchange(hv,0,nx+1,0,ny+1,1,1,1,1,1,0)
 endif
 
 do j=1,ny
  do i=1,nx
    terraul(i,j) = min(1.,max(0.,1.+(h0(i,j)-h0(i-1,j))/(h(i,j)+eps)))
    terraur(i,j) = min(1.,max(0.,1.+(h0(i,j)-h0(i+1,j))/(h(i,j)+eps)))
    terravl(i,j) = min(1.,max(0.,1.+(h0(i,j)-h0(i,j-1)+eps)/(h(i,j)+eps)))
    terravr(i,j) = min(1.,max(0.,1.+(h0(i,j)-h0(i,j+1)+eps)/(h(i,j)+eps)))
  end do
 end do
 if(dompi) then
  buffer(:,:,1) = terraul(:,:)
  buffer(:,:,2) = terraur(:,:)
  buffer(:,:,3) = terravl(:,:)
  buffer(:,:,4) = terravr(:,:)
  nbuff=4
  call ocn_task_exchange(buffer,0,nx+1,0,ny+1,nbuff,1,1,1,1,0)
  terraul(:,:) = buffer(:,:,1)
  terraur(:,:) = buffer(:,:,2)
  terravl(:,:) = buffer(:,:,3)
  terravr(:,:) = buffer(:,:,4)
  call task_barrier()
 else
  call ocn_bound_exchange(terraul,0,nx+1,0,ny+1,1,1,1,1,1,0)
  call ocn_bound_exchange(terraur,0,nx+1,0,ny+1,1,1,1,1,1,0)
  call ocn_bound_exchange(terravl,0,nx+1,0,ny+1,1,1,1,1,1,0)
  call ocn_bound_exchange(terravr,0,nx+1,0,ny+1,1,1,1,1,1,0)
 endif

end subroutine toplayer_init

!--------------------------------------------------------------------------------------------
!--------------------------------------------------------------------------------------------

subroutine toplayer_step

! move the ocean toplayer one step in time

use ocn_params, only: ggr

 real, parameter :: hmin = 0.05 ! enforced minimum thinckness

 real fhx(nx+1,ny)
 real fux(nx+1,ny)
 real fvx(nx+1,ny)
 
 real fhy(nx,ny+1)
 real fuy(nx,ny+1)
 real fvy(nx,ny+1)
 
 integer i,j
 real rdx, rdy, c
 
! advection velocities:

  c= sqrt(ggr*maxval(h))
  do j=1,ny
   do i=1,nx+1
     us(i,j) = (terraul(i,j)*max(-c,min(c,hu(i,j)/max(h(i,j),hmin)))+ &
               terraur(i-1,j)*max(-c,min(c,hu(i-1,j)/max(h(i-1,j),hmin)))) &
               /(terraul(i,j)+terraur(i-1,j)+eps)
    end do
  end do
  do j=1,ny+1
   do i=1,nx
     vs(i,j) = (terravl(i,j)*max(-c,min(c,hv(i,j)/max(h(i,j),hmin)))+ &
               terravr(i,j-1)*max(-c,min(c,hv(i,j-1)/max(h(i,j-1),hmin)))) &
               /(terravl(i,j)+terravr(i,j-1)+eps)
    end do
  end do

! Advection (donor cell):

  do j=1,ny
   do i=1,nx+1
     fhx(i,j) = min(0.,us(i,j))*h(i,j)*terraul(i,j)+max(0.,us(i,j))*h(i-1,j)*terraur(i-1,j)
     fux(i,j) = min(0.,us(i,j))*hu(i,j)*terraul(i,j)+max(0.,us(i,j))*hu(i-1,j)*terraur(i-1,j)
     fvx(i,j) = min(0.,us(i,j))*hv(i,j)*terraul(i,j)+max(0.,us(i,j))*hv(i-1,j)*terraur(i-1,j)
   end do
  end do
  do j=1,ny+1
   do i=1,nx
     fhy(i,j) = min(0.,vs(i,j))*h(i,j)*terravl(i,j)+max(0.,vs(i,j))*h(i,j-1)*terravr(i,j-1)
     fuy(i,j) = min(0.,vs(i,j))*hu(i,j)*terravl(i,j)+max(0.,vs(i,j))*hu(i,j-1)*terravr(i,j-1)
     fvy(i,j) = min(0.,vs(i,j))*hv(i,j)*terravl(i,j)+max(0.,vs(i,j))*hv(i,j-1)*terravr(i,j-1)
   end do
  end do


  rdx = dtn/dx
  rdy = dtn/dy
  do j=1,ny
   do i=1,nx
     h(i,j) = h(i,j) - rdx*(fhx(i+1,j)-fhx(i,j)) - rdy*(fhy(i,j+1)-fhy(i,j))
     hu(i,j) = hu(i,j) - rdx*(fux(i+1,j)-fux(i,j)) - rdy*(fuy(i,j+1)-fuy(i,j))
     hv(i,j) = hv(i,j) - rdx*(fvx(i+1,j)-fvx(i,j)) - rdy*(fvy(i,j+1)-fvy(i,j))
    end do
  end do

! Boundaries:

  if(dompi) then
     buffer(:,:,1) = h(:,:)
     nbuff=1
     call ocn_task_exchange(buffer,0,nx+1,0,ny+1,nbuff,1,1,1,1,0)
     h(:,:) = buffer(:,:,1)
     call task_barrier()
  else
     call ocn_bound_exchange(h,0,nx+1,0,ny+1,1,1,1,1,1,0)
  endif 

! cell boundaries functions:

  do j=1,ny
   do i=1,nx
     terraul(i,j) = min(1.,max(0.,1.+(h0(i,j)-h0(i-1,j))/(h(i,j)+eps)))
     terraur(i,j) = min(1.,max(0.,1.+(h0(i,j)-h0(i+1,j))/(h(i,j)+eps)))
     terravl(i,j) = min(1.,max(0.,1.+(h0(i,j)-h0(i,j-1)+eps)/(h(i,j)+eps)))
     terravr(i,j) = min(1.,max(0.,1.+(h0(i,j)-h0(i,j+1)+eps)/(h(i,j)+eps)))
   end do
  end do

! Boundaries

  if(dompi) then
    buffer(:,:,1) = terraul(:,:)
    buffer(:,:,2) = terraur(:,:)
    buffer(:,:,3) = terravl(:,:)
    buffer(:,:,4) = terravr(:,:)
    nbuff=4
    call ocn_task_exchange(buffer,0,nx+1,0,ny+1,nbuff,1,1,1,1,0)
    terraul(:,:) = buffer(:,:,1)
    terraur(:,:) = buffer(:,:,2)
    terravl(:,:) = buffer(:,:,3)
    terravr(:,:) = buffer(:,:,4)
    call task_barrier()
  else
   call ocn_bound_exchange(terraul,0,nx+1,0,ny+1,1,1,1,1,1,0)
   call ocn_bound_exchange(terraur,0,nx+1,0,ny+1,1,1,1,1,1,0)
   call ocn_bound_exchange(terravl,0,nx+1,0,ny+1,1,1,1,1,1,0)
   call ocn_bound_exchange(terravr,0,nx+1,0,ny+1,1,1,1,1,1,0)
  endif 

! Gradient force:

  rdx = dtn*ggr/dx
  rdy = dtn*ggr/dy
  do j=1,ny
    do i=1,nx
     hu(i,j) = hu(i,j) - rdx*( &
        h(i,j)*(h(i+1,j)-h(i,j)+h0(i+1,j)-h0(i,j))*terraur(i,j) + &
        h(i,j)*(h(i,j)-h(i-1,j)+h0(i,j)-h0(i-1,j))*terraul(i,j)) / &
        (terraur(i,j)+terraul(i,j)+eps)
     hv(i,j) = hv(i,j) - rdy*( &
        h(i,j)*(h(i,j+1)-h(i,j)+h0(i,j+1)-h0(i,j))*terravr(i,j) + &
        h(i,j)*(h(i,j)-h(i,j-1)+h0(i,j)-h0(i,j-1))*terravl(i,j)) / &
        (terravr(i,j)+terravl(i,j)+eps)
    end do
  end do

! Boundaries

  if(dompi) then
    buffer(:,:,1) = hu(:,:)
    buffer(:,:,2) = hv(:,:)
    nbuff=2
    call ocn_task_exchange(buffer,0,nx+1,0,nx+1,nbuff,1,1,1,1,0)
    hu(:,:) = buffer(:,:,1)
    hv(:,:) = buffer(:,:,2)
    call task_barrier()
  else
    call ocn_bound_exchange(hu,0,nx+1,0,ny+1,1,1,1,1,1,0)
    call ocn_bound_exchange(hv,0,nx+1,0,ny+1,1,1,1,1,1,0)
  endif

end subroutine toplayer_step

end module ocn_toplayer
