module ocn_vars

use ocn_grid
use vars, only: fcory, fcorzy, latitude, longitude

implicit none
!--------------------------------------------------------------------
! prognostic variables:

real, allocatable :: u   (:,:,:) ! x-wind
real, allocatable :: v   (:,:,:) ! y-wind
real, allocatable :: w   (:,:,:) ! z-wind
real, allocatable :: t   (:,:,:) ! conservative temperature (deg C)
real, allocatable :: s   (:,:,:) ! absolute salinity (g/kg)

!--------------------------------------------------------------------
! diagnostic variables:

real, allocatable ::    p    (:,:,:,:) ! perturbation pressure (from Poison eq) 
real, allocatable ::    p3D  (:,:,:) ! sea pressure (dBar) 
real, allocatable ::    tabs (:,:,:) ! temperature (oC)
real, allocatable ::    rhoo (:,:,:) ! local water density (kg/m3)
real, allocatable ::    rhoo0(:,:,:) ! density reference profile (depends on basin) (kg/m3)

!--------------------------------------------------------------------
! time-tendencies for prognostic variables

real, allocatable :: dudt   (:,:,:,:)
real, allocatable :: dvdt   (:,:,:,:)
real, allocatable :: dwdt   (:,:,:,:)

! for forward-in-time diffusion:

real, allocatable :: dudtd  (:,:,:)
real, allocatable :: dvdtd  (:,:,:)
real, allocatable :: dwdtd  (:,:,:)

! shortwave radiation heating rate:

real, allocatable :: qrad(:,:,:)

!----------------------------------------------------------------
! Temporary storage array:

! storage of velocity on previous time step (for advection of scalars):
real, allocatable :: u1   (:,:,:) ! x-wind
real, allocatable :: v1   (:,:,:) ! y-wind
real, allocatable :: w1   (:,:,:) ! z-wind
!----------------------------------------------------------------
! Solution of geometric multigrid solver

real(8), allocatable :: pfy(:,:,:)
!------------------------------------------------------------------
! fluxes at the top and bottom of the domain:

real fluxbu (nx, ny), fluxbv (nx, ny), fluxbt (nx, ny), fluxbs (nx, ny)
real fluxtu (nx, ny), fluxtv (nx, ny), fluxtt (nx, ny), fluxts (nx, ny)
real fzero  (nx, ny)
                
!-----------------------------------------------------------------
! profiles 

real   t0(nzm), s0(nzm), tabs0(nzm), u0(nzm), v0(nzm), p0(nzm), rho0(nzm), &
       tg0(nzm), sg0(nzm), ug0(nzm), vg0(nzm), sp0(nzm), t01(nzm), s01(nzm)
!----------------------------------------------------------------
! "observed" (read from snd file) surface characteristics 

real  sst1obs, lhobs, shobs

!-----------------------------------------------------------------
! reference vertical profiles:
 
real   rho(nzm)	  ! base density at pressure levels,kg/m3 
real   rhow(nz)   ! base density at vertical velocity levels,kg/m3
real   bet(nzm)	  ! = ggr/rho
real   stend(nzm) ! Large-scale tendency for salinity
real   ttend(nzm) ! Large-scale tendency for temp.

!---------------------------------------------------------------------
! Large-scale and surface forcing:

integer nlsf	! number of large-scale forcing profiles
integer nrfc	! number of radiative forcing profiles
integer nsfc	! number of surface forcing units
integer nsnd	! number of observed soundings
integer nzlsf	! number of large-scale forcing profiles
integer nzrfc	! number of radiative forcing profiles
integer nzsnd	! number of observed soundings

real, allocatable :: dsls(:,:) ! Large-scale tendency for salinity
real, allocatable :: dtls(:,:) ! Large-scale tendency for temp.
real, allocatable :: ugls(:,:) ! Large-scale wind in X-direction
real, allocatable :: vgls(:,:) ! Large-scale wind in Y-direction
real, allocatable :: wgls(:,:) ! Large-scale subsidence velocity,m/s
real, allocatable :: pres0ls(:)! Surface pressure, mb
real, allocatable :: zls(:,:)  ! Height
real, allocatable :: pls(:,:)  ! Pressure
real, allocatable :: dayls(:)  ! Large-scale forcing arrays time (days) 
real, allocatable :: dtrfc(:,:)! Radiative tendency for pot. temp.
real, allocatable :: dayrfc(:) ! Radiative forcing arrays time (days) 
real, allocatable :: prfc(:,:) ! Pressure/Height
real, allocatable :: sstsfc(:) ! SSTs
real, allocatable :: shsfc(:)   ! Sensible heat flux,W/m2
real, allocatable :: lhsfc(:)  ! Latent heat flux,W/m2
real, allocatable :: tauxsfc(:) ! Surface zonal drag,m2/s2
real, allocatable :: tauysfc(:) ! Surface meridional drag,m2/s2
real, allocatable :: daysfc(:) ! Surface forcing arrays time (days) 
real, allocatable :: usnd(:,:) ! Observed zonal wind
real, allocatable :: vsnd(:,:) ! Observed meriod wind
real, allocatable :: tsnd(:,:) ! Observed Abs. temperature
real, allocatable :: ssnd(:,:) ! Observed Moisture
real, allocatable :: zsnd(:,:) ! Height
real, allocatable :: psnd(:,:) ! Pressure
real, allocatable :: daysnd(:) ! number of sounding samples

real precobs(nx,ny,2) ! prescribed surface preciitation (kg/s/m2)
real dayprecobs(2) ! time arrays (days)
real roffobs(nx,ny,2) ! prescribed river runoff (kg/s/m2)
real dayroffobs(2) ! time arrays (days)
real qobs(nx,ny,2) ! prescribed water vapor above surface  (W/m2)
real dayqobs(2) ! time arrays (days)
real tobs(nx,ny,2) ! prescribed temperature above surface (W/m2)
real daytobs(2) ! time arrays (days)
real lhfobs(nx,ny,2) ! prescribed surface lhf (W/m2)
real daylhfobs(2) ! time arrays (days)
real shfobs(nx,ny,2) ! prescribed surface shf (W/m2)
real dayshfobs(2) ! time arrays (days)
real lwobs(nx,ny,2) ! prescribed surface net lw flux (W/m2)
real daylwobs(2) ! time arrays (days)
real swobs(nx,ny,2) ! prescribed surface net sw flux (W/m2)
real dayswobs(2) ! time arrays (days)
real uobs(nx,ny,2) ! prescribed surface drag in x (N/m2)
real dayuobs(2) ! time arrays (days)
real vobs(nx,ny,2) ! prescribed surface drag in y (N/m2)
real dayvobs(2) ! time arrays (days)
real tauxobs(nx,ny,2) ! prescribed surface drag in x (N/m2)
real daytauxobs(2) ! time arrays (days)
real tauyobs(nx,ny,2) ! prescribed surface drag in y (N/m2)
real daytauyobs(2) ! time arrays (days)
 
!---------------------------------------------------------------------
!  Horizontally varying stuff (as a function of xy)
!
real elevation(nx,ny) ! actual surface elevation, m
real elevationg(nx,ny) ! surface elevation truncated to grid levelas, m
integer:: basins(nx,ny) = 1 ! connected ocean basins

real q_atm(nx,ny) ! water vapor above surface
real t_atm(nx,ny) ! temperature above surface
real u_atm(nx,ny) ! u-wind above surface
real v_atm(nx,ny) ! v_wind above surface

!---------------------------------------------------------------------
! seaice variables

real ::  h_seaice(nx,ny) = 0. ! sea-ice thickness (m)
real ::  t_seaice(nx,ny) = 0. ! seaice temperature (oC)

!---------------------------------------------------------------------
! 2D output dtuff:
real prec_xy(nx,ny) ! precip. rate
real roff_xy(nx,ny) ! river runoff
real taux_xy(nx,ny) ! stress in x
real tauy_xy(nx,ny) ! stress in y
real shf_xy(nx,ny) ! sensible heat flux
real lhf_xy(nx,ny) ! latent heat flux
real seaicefrac_xy(nx,ny) ! seaice fraction
real seaicethic_xy(nx,ny) ! seaice thickness
real u100_xy(nx,ny) ! u-wind at 100 m
real v100_xy(nx,ny) ! v-wind at 100 m
real w100_xy(nx,ny) ! w-wind at 100 m
real u300_xy(nx,ny) ! u-wind at 100 m
real v300_xy(nx,ny) ! v-wind at 100 m
real w300_xy(nx,ny) ! w-wind at 100 m
real u500_xy(nx,ny) ! u-wind at 100 m
real v500_xy(nx,ny) ! v-wind at 100 m
real w500_xy(nx,ny) ! w-wind at 100 m
real usfc_xy(nx,ny) ! u-wind at at the surface
real vsfc_xy(nx,ny) ! v-wind at the surface
real tsfc_xy(nx,ny) ! temperature  near the surface
real ssfc_xy(nx,ny) ! salinity  near the surface
real sst_min_xy(nx,ny) ! minimum surface temperature
real sst_max_xy(nx,ny) ! maximum surface temperature
real lwns_xy(nx,ny) ! net lw at SFC
real swns_xy(nx,ny) ! net sw at SFC
real uatm_xy(nx,ny) ! u-wind at above the surface
real vatm_xy(nx,ny) ! v-wind at above the surface
real tatm_xy(nx,ny) ! t at above the surface
real qatm_xy(nx,ny) ! q at above the surface
real cflz_max_xy(nx,ny) ! maximum cfl in vertical
real cflh_max_xy(nx,ny) ! maximum cfl in horizonal
real cfl_max_xy(nx,ny) ! maximum cfl 
real zcflz_max_xy(nx,ny) ! height of maximum cfl in vertical
real zcflh_max_xy(nx,ny) ! height of maximum cfl in horizonal
real zcfl_max_xy(nx,ny) ! height of maximum cfl 

real, parameter :: t00 = 273.15   ! freezing temperature 

!----------------------------------------------------------------------
!	Vertical profiles of quantities sampled for statitistics purposes:

real &
    twle(nz), swle(nz), twsb(nz), swsb(nz), precflux(nz), &
    uwle(nz), uwsb(nz), vwle(nz), vwsb(nz), &
    radlwup(nz), radlwdn(nz), radswup(nz), radswdn(nz), &
    radqrlw(nz), radqrsw(nz), w_max, u_max, s_acld, s_acldcold, s_ar, s_arthr, s_sst, &
    s_acldl, s_acldm, s_acldh,  ncmn, nrmn, z_inv, z_cb, z_ct, z_cbmn, z_ctmn, &
    z2_inv, z2_cb, z2_ct, cwpmean, cwp2, precmean, prec2, precmax, nrainy, ncloudy, &
    s_acldisccp, s_acldlisccp, s_acldmisccp, s_acldhisccp, s_ptopisccp, &
    s_acldmodis, s_acldlmodis, s_acldmmodis, s_acldhmodis, s_ptopmodis, &
    s_acldmisr, s_ztopmisr, s_relmodis, s_reimodis, s_lwpmodis, s_iwpmodis, &
    s_tbisccp, s_tbclrisccp, s_acldliqmodis, s_acldicemodis, &
    s_cldtauisccp,s_cldtaumodis,s_cldtaulmodis,s_cldtauimodis,s_cldalbisccp, &
    s_flns,s_flnt,s_flntoa,s_flnsc,s_flntoac,s_flds,s_fsns, &
    s_fsnt,s_fsntoa,s_fsnsc,s_fsntoac,s_fsds,s_solin, s_grndflx,& 
    tkeleadv(nz), tkelepress(nz), tkelediss(nz), tkelediff(nz),tkelebuoy(nz), &
    t2leadv(nz),t2legrad(nz),t2lediff(nz),t2leprec(nz),t2lediss(nz), &
    s2leadv(nz),s2legrad(nz),s2lediff(nz),s2leprec(nz),s2lediss(nz), &
    twleadv(nz),twlediff(nz),twlepres(nz),twlebuoy(nz),twleprec(nz), &
    swleadv(nz),swlediff(nz),swlepres(nz),swlebuoy(nz),swleprec(nz), &
    momleadv(nz,3),momlepress(nz,3),momlebuoy(nz,3), &
    momlediff(nz,3),tadv(nz),tdiff(nz),sadv(nz),sdiff(nz)


! register functions:


        integer, external :: lenstr, bytes_in_rec

end module ocn_vars
