!==========================================================================
pure subroutine gsw_add_barrier (input_data, long, lat, long_grid, &
                              lat_grid, dlong_grid, dlat_grid, output_data)
!==========================================================================
!
!  Adds a barrier through Central America (Panama) and then averages
!  over the appropriate side of the barrier
! 
!  data_in      : data                                            [unitless]
!  long         : Long of data in decimal degs east               [ 0 ... +360 ]
!  lat          : Lat of data in decimal degs north               [-90 ... +90 ]
!  longs_grid   : Long of regular grid in decimal degs east       [ 0 ... +360 ]
!  lats_grid    : Lat of regular grid in decimal degs north       [-90 ... +90 ]
!  dlongs_grid  : Long difference of regular grid in decimal degs [ deg long ]
!  dlats_grid   : Lat difference of regular grid in decimal degs  [ deg lat ]
!
! output_data   : average of data depending on which side of the 
!                 Panama canal it is on                           [unitless]
!--------------------------------------------------------------------------

use gsw_mod_toolbox, only : gsw_util_indx

use gsw_mod_saar_data

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: long, lat, long_grid, lat_grid, dlong_grid
real (r8), intent(in) :: dlat_grid
real (r8), intent(in), dimension(4) :: input_data
real (r8), intent(out), dimension(4) :: output_data

integer, dimension(4) :: above_line
integer :: k, nmean, above_line0, kk
real (r8) :: r, lats_line, data_mean

call gsw_util_indx(longs_pan,npan,long,k)                !   the long/lat point
r = (long-longs_pan(k))/(longs_pan(k+1)-longs_pan(k))
lats_line = lats_pan(k) + r*(lats_pan(k+1)-lats_pan(k))

if (lats_line.le.lat) then
   above_line0 = 1
else
   above_line0 = 0
end if

call gsw_util_indx(longs_pan,npan,long_grid,k)       ! the 1 and 4 long/lat pts
r = (long_grid-longs_pan(k))/(longs_pan(k+1)-longs_pan(k))
lats_line = lats_pan(k) + r*(lats_pan(k+1)-lats_pan(k))

if (lats_line.le.lat_grid) then
   above_line(1) = 1
else
   above_line(1) = 0
end if

if (lats_line.le.lat_grid+dlat_grid) then
   above_line(4) = 1
else
   above_line(4) = 0
end if

call gsw_util_indx(longs_pan,npan,long_grid+dlong_grid,k)    ! the 2 & 3 points 
r = (long_grid+dlong_grid-longs_pan(k))/(longs_pan(k+1)-longs_pan(k))
lats_line = lats_pan(k) + r*(lats_pan(k+1)-lats_pan(k))

if (lats_line.le.lat_grid) then
   above_line(2) = 1
else
   above_line(2) = 0
end if

if (lats_line.le.lat_grid+dlat_grid) then
   above_line(3) = 1
else
   above_line(3) = 0
end if

nmean = 0 
data_mean = 0.0_r8

do kk = 1,4
   if ((abs(input_data(kk)).le.100._r8).and.above_line0.eq.above_line(kk)) then
      nmean = nmean + 1
      data_mean = data_mean + input_data(kk)
   end if
end do

if (nmean .eq. 0)then
   data_mean = 0.0_r8    !errorreturn
else
   data_mean = data_mean/nmean
endif

do kk = 1,4
   if ((abs(input_data(kk)).ge.1e10_r8).or.above_line0.ne.above_line(kk)) then
      output_data(kk) = data_mean
   else
      output_data(kk) = input_data(kk)
   end if
end do

return
end subroutine

!--------------------------------------------------------------------------
!==========================================================================
pure subroutine gsw_add_mean (data_in, data_out)
!==========================================================================
!
! Replaces NaN's with non-nan mean of the 4 adjacent neighbours
!
! data_in   : data set of the 4 adjacent neighbours   
!
! data_out : non-nan mean of the 4 adjacent neighbours     [unitless]
!--------------------------------------------------------------------------

use gsw_mod_kinds

implicit none

real (r8), intent(in), dimension(4) :: data_in
real (r8), intent(out), dimension(4) :: data_out

integer :: k, nmean

real (r8) :: data_mean

nmean = 0
data_mean = 0.0_r8

do k = 1,4
   if (abs(data_in(k)).le.100.0_r8) then
      nmean = nmean + 1
      data_mean = data_mean + data_in(k)
   end if
end do

if(nmean.eq.0)then
   data_mean = 0.0_r8    !error return
else
   data_mean = data_mean/nmean
endif

do k = 1,4
   if(abs(data_in(k)).le.100.0_r8) then
      data_out(k) = data_in(k)
   else
      data_out(k) = data_mean
   end if
end do

return
end subroutine

!--------------------------------------------------------------------------
!==========================================================================
elemental function gsw_adiabatic_lapse_rate_from_ct (sa, ct, p) 
!==========================================================================

! Calculates the adiabatic lapse rate from Conservative Temperature
!
! sa     : Absolute Salinity                                 [g/kg]
! ct     : Conservative Temperature                          [deg C]
! p      : sea pressure                                      [dbar]
! 
! gsw_adiabatic_lapse_rate_from_ct : adiabatic lapse rate    [K/Pa]
!--------------------------------------------------------------------------

use gsw_mod_toolbox, only : gsw_gibbs, gsw_pt_from_ct, gsw_pt_from_t

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sa, ct, p 

real (r8) :: gsw_adiabatic_lapse_rate_from_ct

real (r8) :: pt0, t

integer, parameter :: n0=0, n1=1, n2=2
real (r8), parameter :: pr0 = 0.0_r8

pt0 = gsw_pt_from_ct(sa,ct)
t = gsw_pt_from_t(sa,pt0,pr0,p)

gsw_adiabatic_lapse_rate_from_ct = -gsw_gibbs(n0,n1,n1,sa,t,p) / &
                                        gsw_gibbs(n0,n2,n0,sa,t,p)

return
end function

!--------------------------------------------------------------------------
!==========================================================================
elemental function gsw_adiabatic_lapse_rate_ice (t, p)
!==========================================================================
!
!  Calculates the adiabatic lapse rate of ice.
!
!  t  =  in-situ temperature (ITS-90)                             [ deg C ]
!  p  =  sea pressure                                              [ dbar ]
!         ( i.e. absolute pressure - 10.1325 dbar ) 
!
!    Note.  The output is in unit of degress Celsius per Pa,
!      (or equivilently K/Pa) not in units of K/dbar. 
!--------------------------------------------------------------------------

use gsw_mod_toolbox, only : gsw_gibbs_ice

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: t, p

real (r8) :: gsw_adiabatic_lapse_rate_ice

gsw_adiabatic_lapse_rate_ice = -gsw_gibbs_ice(1,1,t,p)/gsw_gibbs_ice(2,0,t,p)

return
end function

!--------------------------------------------------------------------------
!==========================================================================
elemental function gsw_alpha (sa, ct, p)
!==========================================================================
!
!  Calculates the thermal expansion coefficient of seawater with respect to 
!  Conservative Temperature using the computationally-efficient expression
!  for specific volume in terms of SA, CT and p (Roquet et al., 2014).
!   
!  SA  =  Absolute Salinity                                        [ g/kg ]
!  CT  =  Conservative Temperature (ITS-90)                       [ deg C ]
!  p   =  sea pressure                                             [ dbar ]
!         ( i.e. absolute pressure - 10.1325 dbar )
!
!  alpha  =  thermal expansion coefficient                          [ 1/K ]
!            with respect to Conservative Temperature
!--------------------------------------------------------------------------

use gsw_mod_toolbox, only : gsw_specvol

use gsw_mod_teos10_constants, only : gsw_sfac, offset

use gsw_mod_specvol_coefficients

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sa, ct, p

real (r8) :: gsw_alpha

real (r8) :: xs, ys, z, v_ct_part

xs = sqrt(gsw_sfac*sa + offset)
ys = ct*0.025_r8
z = p*1e-4_r8

v_ct_part = a000 + xs*(a100 + xs*(a200 + xs*(a300 + xs*(a400 + a500*xs)))) &
         + ys*(a010 + xs*(a110 + xs*(a210 + xs*(a310 + a410*xs))) &
         + ys*(a020 + xs*(a120 + xs*(a220 + a320*xs)) + ys*(a030 &
         + xs*(a130 + a230*xs) + ys*(a040 + a140*xs + a050*ys )))) &
         + z*(a001 + xs*(a101 + xs*(a201 + xs*(a301 + a401*xs))) &
         + ys*(a011 + xs*(a111 + xs*(a211 + a311*xs)) + ys*(a021 &
         + xs*(a121 + a221*xs) + ys*(a031 + a131*xs + a041*ys))) &
         + z*(a002 + xs*(a102 + xs*(a202 + a302*xs)) + ys*(a012 &
         + xs*(a112 + a212*xs) + ys*(a022 + a122*xs + a032*ys)) &
         + z*(a003 + a103*xs + a013*ys + a004*z))) 

gsw_alpha = 0.025_r8*v_ct_part/gsw_specvol(sa,ct,p)

return
end function

!--------------------------------------------------------------------------
!==========================================================================
elemental function gsw_alpha_on_beta (sa, ct, p)
!==========================================================================
! 
!  Calculates alpha divided by beta, where alpha is the thermal expansion
!  coefficient and beta is the saline contraction coefficient of seawater 
!  from Absolute Salinity and Conservative Temperature.  This function uses
!  the computationally-efficient expression for specific volume in terms of 
!  SA, CT and p (Roquet et al., 2014).
!
!  SA  =  Absolute Salinity                                        [ g/kg ]
!  CT  =  Conservative Temperature (ITS-90)                       [ deg C ]
!  p   =  sea pressure                                             [ dbar ]
!         ( i.e. absolute pressure - 10.1325 dbar )
!
!  alpha_on_beta  =  thermal expansion coefficient with respect to 
!                    Conservative Temperature divided by the saline 
!                    contraction coefficient at constant Conservative
!                    Temperature                           [ kg g^-1 K^-1 ]
!--------------------------------------------------------------------------

use gsw_mod_teos10_constants, only : gsw_sfac, offset

use gsw_mod_specvol_coefficients

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sa, ct, p

real (r8) :: gsw_alpha_on_beta

real (r8) :: xs, ys, z, v_ct_part, v_sa_part

xs = sqrt(gsw_sfac*sa + offset)
ys = ct*0.025_r8
z = p*1e-4_r8

v_ct_part = a000 + xs*(a100 + xs*(a200 + xs*(a300 + xs*(a400 + a500*xs)))) &
         + ys*(a010 + xs*(a110 + xs*(a210 + xs*(a310 + a410*xs))) &
         + ys*(a020 + xs*(a120 + xs*(a220 + a320*xs)) + ys*(a030 &
         + xs*(a130 + a230*xs) + ys*(a040 + a140*xs + a050*ys )))) &
         + z*(a001 + xs*(a101 + xs*(a201 + xs*(a301 + a401*xs))) &
         + ys*(a011 + xs*(a111 + xs*(a211 + a311*xs)) + ys*(a021 &
         + xs*(a121 + a221*xs) + ys*(a031 + a131*xs + a041*ys))) &
         + z*(a002 + xs*(a102 + xs*(a202 + a302*xs)) + ys*(a012 &
         + xs*(a112 + a212*xs) + ys*(a022 + a122*xs + a032*ys)) &
         + z*(a003 + a103*xs + a013*ys + a004*z))) 

v_sa_part = b000 + xs*(b100 + xs*(b200 + xs*(b300 + xs*(b400 + b500*xs)))) &
       + ys*(b010 + xs*(b110 + xs*(b210 + xs*(b310 + b410*xs))) &
       + ys*(b020 + xs*(b120 + xs*(b220 + b320*xs)) + ys*(b030 &
       + xs*(b130 + b230*xs) + ys*(b040 + b140*xs + b050*ys)))) &
       + z*(b001 + xs*(b101 + xs*(b201 + xs*(b301 + b401*xs))) &
       + ys*(b011 + xs*(b111 + xs*(b211 + b311*xs)) + ys*(b021 &
       + xs*(b121 + b221*xs) + ys*(b031 + b131*xs + b041*ys))) &
       + z*(b002 + xs*(b102 + xs*(b202 + b302*xs))+ ys*(b012 &
       + xs*(b112 + b212*xs) + ys*(b022 + b122*xs + b032*ys)) &
       + z*(b003 +  b103*xs + b013*ys + b004*z)))
 
gsw_alpha_on_beta = -(v_ct_part*xs)/(20.0_r8*gsw_sfac*v_sa_part)
         
return
end function

!--------------------------------------------------------------------------
!==========================================================================
elemental function gsw_alpha_wrt_t_exact (sa, t, p) 
!==========================================================================
!
! Calculates thermal expansion coefficient of seawater with respect to 
! in-situ temperature
!
! sa     : Absolute Salinity                               [g/kg]
! t      : insitu temperature                              [deg C]
! p      : sea pressure                                    [dbar]
!
! gsw_alpha_wrt_t_exact : thermal expansion coefficient    [1/K]
!                         wrt (in-situ) temperature
!--------------------------------------------------------------------------

use gsw_mod_toolbox, only : gsw_gibbs

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sa, t, p 

real (r8) :: gsw_alpha_wrt_t_exact

integer, parameter :: n0=0, n1=1

gsw_alpha_wrt_t_exact = gsw_gibbs(n0,n1,n1,sa,t,p)/gsw_gibbs(n0,n0,n1,sa,t,p)

return
end function
!==========================================================================
elemental function gsw_alpha_wrt_t_ice (t, p)
!==========================================================================
!
!  Calculates the thermal expansion coefficient of ice with respect to  
!  in-situ temperature.
!   
!  t  =  in-situ temperature (ITS-90)                             [ deg C ]
!  p  =  sea pressure                                              [ dbar ]
!         ( i.e. absolute pressure - 10.1325 dbar )
!
!  alpha_wrt_t_ice  =  thermal expansion coefficient of ice with respect      
!                      to in-situ temperature                       [ 1/K ]
!--------------------------------------------------------------------------

use gsw_mod_toolbox, only : gsw_gibbs_ice

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: t, p

real (r8) :: gsw_alpha_wrt_t_ice

gsw_alpha_wrt_t_ice = gsw_gibbs_ice(1,1,t,p)/gsw_gibbs_ice(0,1,t,p)

return
end function

!--------------------------------------------------------------------------
!==========================================================================
elemental function gsw_beta (sa, ct, p)
!==========================================================================
!
!  Calculates the saline (i.e. haline) contraction coefficient of seawater  
!  at constant Conservative Temperature using the computationally-efficient
!  expression for specific volume in terms of SA, CT and p 
!  (Roquet et al., 2014).
!   
!  SA  =  Absolute Salinity                                        [ g/kg ]
!  CT  =  Conservative Temperature (ITS-90)                       [ deg C ]
!  p   =  sea pressure                                             [ dbar ]
!         ( i.e. absolute pressure - 10.1325 dbar )
!
!  beta  =  saline contraction coefficient                         [ kg/g ]
!           at constant Conservative Temperature
!--------------------------------------------------------------------------

use gsw_mod_toolbox, only : gsw_specvol

use gsw_mod_teos10_constants, only : gsw_sfac, offset

use gsw_mod_specvol_coefficients

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sa, ct, p

real (r8) :: gsw_beta

real (r8) :: xs, ys, z, v_sa_part

xs = sqrt(gsw_sfac*sa + offset)
ys = ct*0.025_r8
z = p*1e-4_r8

v_sa_part = b000 + xs*(b100 + xs*(b200 + xs*(b300 + xs*(b400 + b500*xs)))) &
       + ys*(b010 + xs*(b110 + xs*(b210 + xs*(b310 + b410*xs))) &
       + ys*(b020 + xs*(b120 + xs*(b220 + b320*xs)) + ys*(b030 &
       + xs*(b130 + b230*xs) + ys*(b040 + b140*xs + b050*ys)))) &
       + z*(b001 + xs*(b101 + xs*(b201 + xs*(b301 + b401*xs))) &
       + ys*(b011 + xs*(b111 + xs*(b211 + b311*xs)) + ys*(b021 &
       + xs*(b121 + b221*xs) + ys*(b031 + b131*xs + b041*ys))) &
       + z*(b002 + xs*(b102 + xs*(b202 + b302*xs))+ ys*(b012 &
       + xs*(b112 + b212*xs) + ys*(b022 + b122*xs + b032*ys)) &
       + z*(b003 +  b103*xs + b013*ys + b004*z)))
 
gsw_beta = -v_sa_part*0.5_r8*gsw_sfac/(gsw_specvol(sa,ct,p)*xs)

return
end function

!--------------------------------------------------------------------------
!==========================================================================
elemental function gsw_beta_const_t_exact (sa, t, p)  
!==========================================================================
!
! Calculates saline (haline) contraction coefficient of seawater at 
! constant in-situ temperature.
!
! sa     : Absolute Salinity                               [g/kg]
! t      : in-situ temperature                             [deg C]
! p      : sea pressure                                    [dbar]
! 
! gsw_beta_const_t_exact : haline contraction coefficient  [kg/g]
!--------------------------------------------------------------------------

use gsw_mod_toolbox, only : gsw_gibbs

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sa, t, p  

real (r8) :: gsw_beta_const_t_exact

integer, parameter :: n0=0, n1=1

gsw_beta_const_t_exact = -gsw_gibbs(n1,n0,n1,sa,t,p)/gsw_gibbs(n0,n0,n1,sa,t,p)

return
end function

!--------------------------------------------------------------------------
!==========================================================================
elemental function gsw_c_from_sp (sp, t, p)       
!==========================================================================
!
!  Calculates conductivity, C, from (SP,t,p) using PSS-78 in the range 
!  2 < SP < 42.  If the input Practical Salinity is less than 2 then a 
!  modified form of the Hill et al. (1986) fomula is used for Practical 
!  Salinity.  The modification of the Hill et al. (1986) expression is to
!  ensure that it is exactly consistent with PSS-78 at SP = 2.
!
!  The conductivity ratio returned by this function is consistent with the
!  input value of Practical Salinity, SP, to 2x10^-14 psu over the full 
!  range of input parameters (from pure fresh water up to SP = 42 psu).  
!  This error of 2x10^-14 psu is machine precision at typical seawater 
!  salinities.  This accuracy is achieved by having four different 
!  polynomials for the starting value of Rtx (the square root of Rt) in 
!  four different ranges of SP, and by using one and a half iterations of 
!  a computationally efficient modified Newton-Raphson technique (McDougall 
!  and Wotherspoon, 2012) to find the root of the equation.  
!
!  Note that strictly speaking PSS-78 (Unesco, 1983) defines Practical
!  Salinity in terms of the conductivity ratio, R, without actually
!  specifying the value of C(35,15,0) (which we currently take to be
!  42.9140 mS/cm).
!
! sp     : Practical Salinity                               [unitless]
! t      : in-situ temperature [ITS-90]                     [deg C]
! p      : sea pressure                                     [dbar]
!
! c      : conductivity                                     [ mS/cm ]
!--------------------------------------------------------------------------

use gsw_mod_toolbox, only : gsw_hill_ratio_at_sp2

use gsw_mod_teos10_constants, only : gsw_c3515

use gsw_mod_sp_coefficients

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sp, t, p       

real (r8) :: gsw_c_from_sp

real (r8) :: t68, ft68, x, rtx, dsp_drtx, sqrty
real (r8) :: part1, part2, hill_ratio, sp_hill_raw, sp_est
real (r8) :: rtx_old, rt, aa, bb, cc, dd, ee, ra,r, rt_lc, rtxm

real (r8), parameter :: p0  =  4.577801212923119e-3_r8
real (r8), parameter :: p1  =  1.924049429136640e-1_r8
real (r8), parameter :: p2  =  2.183871685127932e-5_r8
real (r8), parameter :: p3  = -7.292156330457999e-3_r8
real (r8), parameter :: p4  =  1.568129536470258e-4_r8
real (r8), parameter :: p5  = -1.478995271680869e-6_r8
real (r8), parameter :: p6  =  9.086442524716395e-4_r8
real (r8), parameter :: p7  = -1.949560839540487e-5_r8
real (r8), parameter :: p8  = -3.223058111118377e-6_r8
real (r8), parameter :: p9  =  1.175871639741131e-7_r8
real (r8), parameter :: p10 = -7.522895856600089e-5_r8
real (r8), parameter :: p11 = -2.254458513439107e-6_r8
real (r8), parameter :: p12 =  6.179992190192848e-7_r8
real (r8), parameter :: p13 =  1.005054226996868e-8_r8
real (r8), parameter :: p14 = -1.923745566122602e-9_r8
real (r8), parameter :: p15 =  2.259550611212616e-6_r8
real (r8), parameter :: p16 =  1.631749165091437e-7_r8
real (r8), parameter :: p17 = -5.931857989915256e-9_r8
real (r8), parameter :: p18 = -4.693392029005252e-9_r8
real (r8), parameter :: p19 =  2.571854839274148e-10_r8
real (r8), parameter :: p20 =  4.198786822861038e-12_r8

real (r8), parameter :: q0  =  5.540896868127855e-5_r8
real (r8), parameter :: q1  =  2.015419291097848e-1_r8
real (r8), parameter :: q2  = -1.445310045430192e-5_r8 
real (r8), parameter :: q3  = -1.567047628411722e-2_r8
real (r8), parameter :: q4  =  2.464756294660119e-4_r8
real (r8), parameter :: q5  = -2.575458304732166e-7_r8
real (r8), parameter :: q6  =  5.071449842454419e-3_r8
real (r8), parameter :: q7  =  9.081985795339206e-5_r8
real (r8), parameter :: q8  = -3.635420818812898e-6_r8
real (r8), parameter :: q9  =  2.249490528450555e-8_r8
real (r8), parameter :: q10 = -1.143810377431888e-3_r8
real (r8), parameter :: q11 =  2.066112484281530e-5_r8
real (r8), parameter :: q12 =  7.482907137737503e-7_r8
real (r8), parameter :: q13 =  4.019321577844724e-8_r8
real (r8), parameter :: q14 = -5.755568141370501e-10_r8
real (r8), parameter :: q15 =  1.120748754429459e-4_r8
real (r8), parameter :: q16 = -2.420274029674485e-6_r8
real (r8), parameter :: q17 = -4.774829347564670e-8_r8
real (r8), parameter :: q18 = -4.279037686797859e-9_r8
real (r8), parameter :: q19 = -2.045829202713288e-10_r8
real (r8), parameter :: q20 =  5.025109163112005e-12_r8

real (r8), parameter :: s0  =  3.432285006604888e-3_r8
real (r8), parameter :: s1  =  1.672940491817403e-1_r8
real (r8), parameter :: s2  =  2.640304401023995e-5_r8
real (r8), parameter :: s3  =  1.082267090441036e-1_r8
real (r8), parameter :: s4  = -6.296778883666940e-5_r8
real (r8), parameter :: s5  = -4.542775152303671e-7_r8
real (r8), parameter :: s6  = -1.859711038699727e-1_r8
real (r8), parameter :: s7  =  7.659006320303959e-4_r8
real (r8), parameter :: s8  = -4.794661268817618e-7_r8
real (r8), parameter :: s9  =  8.093368602891911e-9_r8
real (r8), parameter :: s10 =  1.001140606840692e-1_r8 
real (r8), parameter :: s11 = -1.038712945546608e-3_r8
real (r8), parameter :: s12 = -6.227915160991074e-6_r8
real (r8), parameter :: s13 =  2.798564479737090e-8_r8
real (r8), parameter :: s14 = -1.343623657549961e-10_r8
real (r8), parameter :: s15 =  1.024345179842964e-2_r8
real (r8), parameter :: s16 =  4.981135430579384e-4_r8
real (r8), parameter :: s17 =  4.466087528793912e-6_r8
real (r8), parameter :: s18 =  1.960872795577774e-8_r8
real (r8), parameter :: s19 = -2.723159418888634e-10_r8
real (r8), parameter :: s20 =  1.122200786423241e-12_r8

real (r8), parameter :: u0  =  5.180529787390576e-3_r8
real (r8), parameter :: u1  =  1.052097167201052e-3_r8
real (r8), parameter :: u2  =  3.666193708310848e-5_r8
real (r8), parameter :: u3  =  7.112223828976632_r8
real (r8), parameter :: u4  = -3.631366777096209e-4_r8
real (r8), parameter :: u5  = -7.336295318742821e-7_r8
real (r8), parameter :: u6  = -1.576886793288888e+2_r8
real (r8), parameter :: u7  = -1.840239113483083e-3_r8
real (r8), parameter :: u8  =  8.624279120240952e-6_r8
real (r8), parameter :: u9  =  1.233529799729501e-8_r8
real (r8), parameter :: u10 =  1.826482800939545e+3_r8
real (r8), parameter :: u11 =  1.633903983457674e-1_r8
real (r8), parameter :: u12 = -9.201096427222349e-5_r8
real (r8), parameter :: u13 = -9.187900959754842e-8_r8
real (r8), parameter :: u14 = -1.442010369809705e-10_r8
real (r8), parameter :: u15 = -8.542357182595853e+3_r8
real (r8), parameter :: u16 = -1.408635241899082_r8
real (r8), parameter :: u17 =  1.660164829963661e-4_r8
real (r8), parameter :: u18 =  6.797409608973845e-7_r8
real (r8), parameter :: u19 =  3.345074990451475e-10_r8
real (r8), parameter :: u20 =  8.285687652694768e-13_r8

t68 = t*1.00024_r8
ft68 = (t68 - 15.0_r8)/(1.0_r8 + k*(t68 - 15.0_r8))

x = sqrt(sp)

!--------------------------------------------------------------------------
! Finding the starting value of Rtx, the square root of Rt, using four 
! different polynomials of SP and t68.  
!--------------------------------------------------------------------------

if (sp.ge.9.0_r8) then

    rtx = p0 + x*(p1 + p4*t68 + x*(p3 + p7*t68 + x*(p6  &
        + p11*t68 + x*(p10 + p16*t68 + x*p15))))  &
        + t68*(p2+ t68*(p5 + x*x*(p12 + x*p17) + p8*x  &
        + t68*(p9 + x*(p13 + x*p18)+ t68*(p14 + p19*x + p20*t68))))

else if (sp.ge.0.25_r8.and.sp.lt.9.0_r8) then

    rtx = q0 + x*(q1 + q4*t68 + x*(q3 + q7*t68 + x*(q6  &
        + q11*t68 + x*(q10 + q16*t68 + x*q15))))  &
        + t68*(q2+ t68*(q5 + x*x*(q12 + x*q17) + q8*x  &
        + t68*(q9 + x*(q13 + x*q18)+ t68*(q14 + q19*x + q20*t68))))

else if (sp.ge.0.003_r8.and.sp.lt.0.25_r8) then

    rtx = s0 + x*(s1 + s4*t68 + x*(s3 + s7*t68 + x*(s6  &
        + s11*t68 + x*(s10 + s16*t68 + x*s15))))  &
        + t68*(s2+ t68*(s5 + x*x*(s12 + x*s17) + s8*x  &
        + t68*(s9 + x*(s13 + x*s18)+ t68*(s14 + s19*x + s20*t68))))

else

    rtx = u0 + x*(u1 + u4*t68 + x*(u3 + u7*t68 + x*(u6  &
        + u11*t68 + x*(u10 + u16*t68 + x*u15))))  &
        + t68*(u2+ t68*(u5 + x*x*(u12 + x*u17) + u8*x  &
        + t68*(u9 + x*(u13 + x*u18)+ t68*(u14 + u19*x + u20*t68))))

end if

!--------------------------------------------------------------------------
! Finding the starting value of dSP_dRtx, the derivative of SP with respect
! to Rtx.  
!--------------------------------------------------------------------------
dsp_drtx =  a1 + (2.0_r8*a2 + (3.0_r8*a3 + &
                              (4.0_r8*a4 + 5.0_r8*a5*rtx)*rtx)*rtx)*rtx  &
    + ft68*(b1 + (2.0_r8*b2 + (3.0_r8*b3 + &
                              (4.0_r8*b4 + 5.0_r8*b5*rtx)*rtx)*rtx)*rtx)

if (sp.lt.2.0_r8) then
    x = 400.0_r8*(rtx*rtx)
    sqrty = 10.0_r8*rtx
    part1 = 1.0_r8 + x*(1.5_r8 + x) 
    part2 = 1.0_r8 + sqrty*(1.0_r8 + sqrty*(1.0_r8 + sqrty))
    hill_ratio = gsw_hill_ratio_at_sp2(t)
    dsp_drtx = dsp_drtx  &
        + a0*800.0_r8*Rtx*(1.5_r8 + 2.0_r8*x)/(part1*part1)  &
        + b0*ft68*(10.0_r8 + sqrty*(20.0_r8 + 30.0_r8*sqrty))/(part2*part2)
    dsp_drtx = hill_ratio*dsp_drtx
end if

!--------------------------------------------------------------------------
! One iteration through the modified Newton-Raphson method (McDougall and 
! Wotherspoon, 2012) achieves an error in Practical Salinity of about 
! 10^-12 for all combinations of the inputs.  One and a half iterations of 
! the modified Newton-Raphson method achevies a maximum error in terms of 
! Practical Salinity of better than 2x10^-14 everywhere. 
!
! We recommend one and a half iterations of the modified Newton-Raphson
! method. 
!
! Begin the modified Newton-Raphson method.  
!--------------------------------------------------------------------------
    sp_est = a0 + (a1 + (a2 + (a3 + (a4 + a5*rtx)*rtx)*rtx)*rtx)*rtx &
        + ft68*(b0 + (b1 + (b2+ (b3 + (b4 + b5*rtx)*rtx)*rtx)*rtx)*rtx)
    if (sp_est .lt. 2.0_r8) then
        x = 400.0_r8*(rtx*rtx)
        sqrty = 10.0_r8*rtx
        part1 = 1.0_r8 + x*(1.5_r8 + x) 
        part2 = 1.0_r8 + sqrty*(1.0_r8 + sqrty*(1.0_r8 + sqrty))
        sp_hill_raw = sp_est - a0/part1 - b0*ft68/part2
        hill_ratio = gsw_hill_ratio_at_sp2(t)
        sp_est = hill_ratio*sp_hill_raw
    end if
 
    rtx_old = rtx
    rtx = rtx_old - (sp_est - sp)/dsp_drtx
    
    rtxm = 0.5_r8*(rtx + rtx_old)      ! This mean value of Rtx, Rtxm, is the  
!                 value of Rtx at which the derivative dSP_dRtx is evaluated.
    
    dsp_drtx = a1 + (2.0_r8*a2 + (3.0_r8*a3 + (4.0_r8*a4 + &
                     5.0_r8*a5*rtxm)*rtxm)*rtxm)*rtxm&
       + ft68*(b1 + (2.0_r8*b2 + (3.0_r8*b3 + (4.0_r8*b4 + &
                     5.0_r8*b5*rtxm)*rtxm)*rtxm)*rtxm)
    if (sp_est .lt. 2.0_r8) then
        x = 400.0_r8*(rtxm*rtxm)
        sqrty = 10.0_r8*rtxm
        part1 = 1.0_r8 + x*(1.5_r8 + x) 
        part2 = 1.0_r8 + sqrty*(1.0_r8 + sqrty*(1.0_r8 + sqrty))
        dsp_drtx = dsp_drtx  &
            + a0*800.0_r8*rtxm*(1.5_r8 + 2.0_r8*x)/(part1*part1)  &
            + b0*ft68*(10.0_r8 + sqrty*(20.0_r8 + 30.0_r8*sqrty))/(part2*part2)
        hill_ratio = gsw_hill_ratio_at_sp2(t)
        dsp_drtx = hill_ratio*dsp_drtx
    end if

!--------------------------------------------------------------------------
! The line below is where Rtx is updated at the end of the one full 
! iteration of the modified Newton-Raphson technique.
!--------------------------------------------------------------------------
    rtx = rtx_old - (sp_est - sp)/dsp_drtx
!--------------------------------------------------------------------------
! Now we do another half iteration of the modified Newton-Raphson  
! technique, making a total of one and a half modified N-R iterations.
!-------------------------------------------------------------------------- 
    sp_est = a0 + (a1 + (a2 + (a3 + (a4 + a5*rtx)*rtx)*rtx)*rtx)*rtx  &
        + ft68*(b0 + (b1 + (b2+ (b3 + (b4 + b5*rtx)*rtx)*rtx)*rtx)*rtx)
    if (sp_est .lt. 2.0_r8) then
        x = 400.0_r8*(rtx*rtx)
        sqrty = 10.0_r8*rtx
        part1 = 1.0_r8 + x*(1.5_r8 + x) 
        part2 = 1.0_r8 + sqrty*(1.0_r8 + sqrty*(1.0_r8 + sqrty))
        sp_hill_raw = sp_est - a0/part1 - b0*ft68/part2
        hill_ratio = gsw_hill_ratio_at_sp2(t)
        sp_est = hill_ratio*sp_hill_raw
    end if
    rtx = rtx - (sp_est - sp)/dsp_drtx

!--------------------------------------------------------------------------
! Now go from Rtx to Rt and then to the conductivity ratio R at pressure p.
!--------------------------------------------------------------------------
rt = rtx*rtx
aa  = d3 + d4*t68
bb  = 1.0_r8 + t68*(d1 + d2*t68)
cc  = p*(e1 + p*(e2 + e3*p))
! rt_lc (i.e. rt_lower_case) corresponds to rt as defined in 
! the UNESCO 44 (1983) routines.
rt_lc = c0 + (c1 + (c2 + (c3 + c4*t68)*t68)*t68)*t68

dd  = bb - aa*rt_lc*rt
ee  = rt_lc*rt*aa*(bb + cc)
ra = sqrt(dd*dd + 4.0_r8*ee) - dd
r  = 0.5_r8*ra/aa

! The dimensionless conductivity ratio, R, is the conductivity input, C,
! divided by the present estimate of C(SP=35, t_68=15, p=0) which is 
! 42.9140 mS/cm (=4.29140 S/m^). 

gsw_c_from_sp = gsw_c3515*r      

return
end function

!--------------------------------------------------------------------------
!==========================================================================
elemental function gsw_cabbeling (sa, ct, p)
!==========================================================================
!
!  Calculates the cabbeling coefficient of seawater with respect to  
!  Conservative Temperature.  This function uses the computationally-
!  efficient expression for specific volume in terms of SA, CT and p
!  (Roquet et al., 2014).
!   
!  SA  =  Absolute Salinity                                        [ g/kg ]
!  CT  =  Conservative Temperature (ITS-90)                       [ deg C ]
!  p   =  sea pressure                                             [ dbar ]
!         ( i.e. absolute pressure - 10.1325 dbar )
!
!  cabbeling  =  cabbeling coefficient with respect to            [ 1/K^2 ]
!                Conservative Temperature.                    
!--------------------------------------------------------------------------

use gsw_mod_toolbox, only : gsw_alpha_on_beta, gsw_rho
use gsw_mod_toolbox, only : gsw_specvol_first_derivatives
use gsw_mod_toolbox, only : gsw_specvol_second_derivatives

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sa, ct, p

real (r8) :: gsw_cabbeling

real (r8) :: alpha_ct, alpha_on_beta, alpha_sa, beta_sa, rho
real (r8) :: v_sa, v_ct, v_sa_sa, v_sa_ct, v_ct_ct

call gsw_specvol_first_derivatives(sa,ct,p,v_sa,v_ct)
     
call gsw_specvol_second_derivatives(sa,ct,p,v_sa_sa,v_sa_ct,v_ct_ct)

rho = gsw_rho(sa,ct,p)
     
alpha_ct = rho*(v_ct_ct - rho*v_ct*v_ct)

alpha_sa = rho*(v_sa_ct - rho*v_sa*v_ct)

beta_sa = -rho*(v_sa_sa - rho*v_sa*v_sa)

alpha_on_beta = gsw_alpha_on_beta(sa,ct,p)

gsw_cabbeling = alpha_ct + &
                alpha_on_beta*(2.0_r8*alpha_sa - alpha_on_beta*beta_sa)

return
end function

!--------------------------------------------------------------------------
!==========================================================================
elemental function gsw_chem_potential_water_ice (t, p)
!==========================================================================
! 
!  Calculates the chemical potential of water in ice from in-situ
!  temperature and pressure.
!
!  t  =  in-situ temperature (ITS-90)                             [ deg C ]
!  p  =  sea pressure                                              [ dbar ]
!         ( i.e. absolute pressure - 10.1325 dbar )
!
!  chem_potential_water_ice  =  chemical potential of ice          [ J/kg ]
!--------------------------------------------------------------------------

use gsw_mod_toolbox, only : gsw_gibbs_ice

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: t, p

real (r8) :: gsw_chem_potential_water_ice

gsw_chem_potential_water_ice = gsw_gibbs_ice(0,0,t,p)

return
end function

!--------------------------------------------------------------------------
!==========================================================================
elemental function gsw_chem_potential_water_t_exact (sa, t, p)
!==========================================================================
!
!  Calculates the chemical potential of water in seawater.
!
!  SA  =  Absolute Salinity                                        [ g/kg ]
!  t   =  in-situ temperature (ITS-90)                            [ deg C ]
!  p   =  sea pressure                                             [ dbar ]
!         ( i.e. absolute pressure - 10.1325 dbar ) 
!
!  chem_potential_water_t_exact  =  chemical potential of water in seawater
!                                                                   [ J/g ]
!--------------------------------------------------------------------------

use gsw_mod_teos10_constants, only : gsw_sfac

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sa, t, p

real (r8) :: gsw_chem_potential_water_t_exact

real (r8) :: g03_g, g08_g, g_sa_part, x, x2, y, z

real (r8), parameter :: kg2g = 1e-3_r8

x2 = gsw_sfac*sa
x = sqrt(x2)
y = t*0.025_r8
z = p*1e-4_r8

g03_g = 101.342743139674_r8 + z*(100015.695367145_r8 + &
    z*(-2544.5765420363_r8 + z*(284.517778446287_r8 + &
    z*(-33.3146754253611_r8 + (4.20263108803084_r8 - 0.546428511471039_r8*z)*z)))) + &
    y*(5.90578347909402_r8 + z*(-270.983805184062_r8 + &
    z*(776.153611613101_r8 + z*(-196.51255088122_r8 + (28.9796526294175_r8 - 2.13290083518327_r8*z)*z))) + &
    y*(-12357.785933039_r8 + z*(1455.0364540468_r8 + &
    z*(-756.558385769359_r8 + z*(273.479662323528_r8 + z*(-55.5604063817218_r8 + 4.34420671917197_r8*z)))) + &
    y*(736.741204151612_r8 + z*(-672.50778314507_r8 + &
    z*(499.360390819152_r8 + z*(-239.545330654412_r8 + (48.8012518593872_r8 - 1.66307106208905_r8*z)*z))) + &
    y*(-148.185936433658_r8 + z*(397.968445406972_r8 + &
    z*(-301.815380621876_r8 + (152.196371733841_r8 - 26.3748377232802_r8*z)*z)) + &
    y*(58.0259125842571_r8 + z*(-194.618310617595_r8 + &
    z*(120.520654902025_r8 + z*(-55.2723052340152_r8 + 6.48190668077221_r8*z))) + &
    y*(-18.9843846514172_r8 + y*(3.05081646487967_r8 - 9.63108119393062_r8*z) + &
    z*(63.5113936641785_r8 + z*(-22.2897317140459_r8 + 8.17060541818112_r8*z))))))))

g08_g = x2*(1416.27648484197_r8 + &
    x*(-2432.14662381794_r8 + x*(2025.80115603697_r8 + &
    y*(543.835333000098_r8 + y*(-68.5572509204491_r8 + &
    y*(49.3667694856254_r8 + y*(-17.1397577419788_r8 + 2.49697009569508_r8*y))) - 22.6683558512829_r8*z) + &
    x*(-1091.66841042967_r8 - 196.028306689776_r8*y + &
    x*(374.60123787784_r8 - 48.5891069025409_r8*x + 36.7571622995805_r8*y) + 36.0284195611086_r8*z) + &
    z*(-54.7919133532887_r8 + (-4.08193978912261_r8 - 30.1755111971161_r8*z)*z)) + &
    z*(199.459603073901_r8 + z*(-52.2940909281335_r8 + (68.0444942726459_r8 - 3.41251932441282_r8*z)*z)) + &
    y*(-493.407510141682_r8 + z*(-175.292041186547_r8 + (83.1923927801819_r8 - 29.483064349429_r8*z)*z) + &
    y*(-43.0664675978042_r8 + z*(383.058066002476_r8 + z*(-54.1917262517112_r8 + 25.6398487389914_r8*z)) + &
    y*(-10.0227370861875_r8 - 460.319931801257_r8*z + y*(0.875600661808945_r8 + 234.565187611355_r8*z))))) + &
    y*(168.072408311545_r8))

g_sa_part = 8645.36753595126_r8 + &
    x*(-7296.43987145382_r8 + x*(8103.20462414788_r8 + &
    y*(2175.341332000392_r8 + y*(-274.2290036817964_r8 + &
    y*(197.4670779425016_r8 + y*(-68.5590309679152_r8 + 9.98788038278032_r8*y))) - 90.6734234051316_r8*z) + &
    x*(-5458.34205214835_r8 - 980.14153344888_r8*y + &
    x*(2247.60742726704_r8 - 340.1237483177863_r8*x + 220.542973797483_r8*y) + 180.142097805543_r8*z) + &
    z*(-219.1676534131548_r8 + (-16.32775915649044_r8 - 120.7020447884644_r8*z)*z)) + &
    z*(598.378809221703_r8 + z*(-156.8822727844005_r8 + (204.1334828179377_r8 - 10.23755797323846_r8*z)*z)) + &
    y*(-1480.222530425046_r8 + z*(-525.876123559641_r8 + (249.57717834054571_r8 - 88.449193048287_r8*z)*z) + &
    y*(-129.1994027934126_r8 + z*(1149.174198007428_r8 + z*(-162.5751787551336_r8 + 76.9195462169742_r8*z)) + &
    y*(-30.0682112585625_r8 - 1380.9597954037708_r8*z + y*(2.626801985426835_r8 + 703.695562834065_r8*z))))) + &
    y*(1187.3715515697959_r8)

gsw_chem_potential_water_t_exact = kg2g*(g03_g + g08_g - 0.5_r8*x2*g_sa_part)

return
end function

!--------------------------------------------------------------------------
!==========================================================================
elemental function gsw_cp_ice (t, p)
!==========================================================================
! 
!  Calculates the isobaric heat capacity of seawater.
!
!  t   =  in-situ temperature (ITS-90)                            [ deg C ]
!  p   =  sea pressure                                             [ dbar ]
!          ( i.e. absolute pressure - 10.1325 dbar )
!
!  gsw_cp_ice  =  heat capacity of ice                       [J kg^-1 K^-1]
!--------------------------------------------------------------------------

use gsw_mod_toolbox, only : gsw_gibbs_ice

use gsw_mod_teos10_constants, only : gsw_t0

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: t, p

real (r8) :: gsw_cp_ice

gsw_cp_ice = -(t + gsw_t0)*gsw_gibbs_ice(2,0,t,p)

end function

!--------------------------------------------------------------------------
!==========================================================================
elemental subroutine gsw_ct_first_derivatives (sa, pt, ct_sa, ct_pt)
!==========================================================================
!
!  Calculates the following two derivatives of Conservative Temperature
!  (1) CT_SA, the derivative with respect to Absolute Salinity at 
!      constant potential temperature (with pr = 0 dbar), and
!   2) CT_pt, the derivative with respect to potential temperature
!      (the regular potential temperature which is referenced to 0 dbar)
!      at constant Absolute Salinity.
!
!  SA  =  Absolute Salinity                                        [ g/kg ]
!  pt  =  potential temperature (ITS-90)                          [ deg C ]   
!         (whose reference pressure is 0 dbar)
!
!  CT_SA  =  The derivative of Conservative Temperature with respect to 
!            Absolute Salinity at constant potential temperature 
!            (the regular potential temperature which has reference 
!            sea pressure of 0 dbar).    
!            The CT_SA output has units of:                     [ K/(g/kg)]
!  CT_pt  =  The derivative of Conservative Temperature with respect to 
!            potential temperature (the regular one with pr = 0 dbar)
!            at constant SA. CT_pt is dimensionless.           [ unitless ]
!--------------------------------------------------------------------------

use gsw_mod_teos10_constants, only : gsw_cp0, gsw_sfac, gsw_t0

use gsw_mod_toolbox, only : gsw_gibbs_pt0_pt0

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sa, pt
real (r8), intent(out), optional :: ct_sa, ct_pt

real (r8) :: abs_pt, g_sa_mod, g_sa_t_mod, x, y_pt

abs_pt = gsw_t0 + pt 

if (present(ct_pt)) ct_pt = -(abs_pt*gsw_gibbs_pt0_pt0(sa,pt))/gsw_cp0

if (.not. present(ct_sa)) return

!--------------------------------------------------------------------------

x = sqrt(gsw_sfac*sa)
y_pt = 0.025_r8*pt

g_sa_t_mod = 1187.3715515697959_r8 + &
        x*(-1480.222530425046_r8 + x*(2175.341332000392_r8 + x*(-980.14153344888_r8 + 220.542973797483_r8*x) + &
        y_pt*(-548.4580073635929_r8 + y_pt*(592.4012338275047_r8 + y_pt*(-274.2361238716608_r8 + 49.9394019139016_r8*y_pt)))) + &
        y_pt*(-258.3988055868252_r8 + y_pt*(-90.2046337756875_r8 + y_pt*10.50720794170734_r8))) + &
        y_pt*(3520.125411988816_r8  + y_pt*(-1351.605895580406_r8 + &
        y_pt*(731.4083582010072_r8  + y_pt*(-216.60324087531103_r8 + 25.56203650166196_r8*y_pt))))
g_sa_t_mod = 0.5_r8*gsw_sfac*0.025_r8*g_sa_t_mod
   
g_sa_mod = 8645.36753595126_r8 + &
        x*(-7296.43987145382_r8 + x*(8103.20462414788_r8 + &
        y_pt*(2175.341332000392_r8 + y_pt*(-274.2290036817964_r8 + &
        y_pt*(197.4670779425016_r8 + y_pt*(-68.5590309679152_r8 + 9.98788038278032_r8*y_pt)))) + &
        x*(-5458.34205214835_r8 - 980.14153344888_r8*y_pt + &
        x*(2247.60742726704_r8 - 340.1237483177863_r8*x + 220.542973797483_r8*y_pt))) + &
        y_pt*(-1480.222530425046_r8 + &
        y_pt*(-129.1994027934126_r8 + &
        y_pt*(-30.0682112585625_r8 + y_pt*(2.626801985426835_r8 ))))) + &
        y_pt*(1187.3715515697959_r8 + &
        y_pt*(1760.062705994408_r8 + y_pt*(-450.535298526802_r8 + &
        y_pt*(182.8520895502518_r8 + y_pt*(-43.3206481750622_r8 + 4.26033941694366_r8*y_pt)))))
g_sa_mod = 0.5_r8*gsw_sfac*g_sa_mod   

ct_sa = (g_sa_mod - abs_pt*g_sa_t_mod)/gsw_cp0

return
end subroutine

!--------------------------------------------------------------------------
!==========================================================================
elemental subroutine gsw_ct_first_derivatives_wrt_t_exact (sa, t, p, &
                                       ct_sa_wrt_t, ct_t_wrt_t, ct_p_wrt_t)
!==========================================================================
!
!  Calculates the following three derivatives of Conservative Temperature.
!  These derivatives are done with respect to in-situ temperature t (in the
!  case of CT_T_wrt_t) or at constant in-situ tempertature (in the cases of
!  CT_SA_wrt_t and CT_P_wrt_t).  
!   (1) CT_SA_wrt_t, the derivative of CT with respect to Absolute Salinity 
!       at constant t and p, and
!   (2) CT_T_wrt_t, derivative of CT with respect to in-situ temperature t 
!       at constant SA and p. 
!   (3) CT_P_wrt_t, derivative of CT with respect to pressure P (in Pa) at  
!       constant SA and t.    
!
!  This function uses the full Gibbs function. Note that this function
!  avoids the NaN that would exist in CT_SA_wrt_t at SA = 0 if it were
!  evaluated in the straightforward way from the derivatives of the Gibbs 
!  function function.
!   
!  SA  =  Absolute Salinity                                        [ g/kg ]
!  t   =  in-situ temperature (ITS-90)                            [ deg C ]
!  p   =  sea pressure                                             [ dbar ]
!         ( i.e. absolute pressure - 10.1325 dbar)
!
!  CT_SA_wrt_t  =  The first derivative of Conservative Temperature with 
!                  respect to Absolute Salinity at constant t and p.     
!                                              [ K/(g/kg)]  i.e. [ K kg/g ]
!  CT_T_wrt_t  =  The first derivative of Conservative Temperature with 
!                 respect to in-situ temperature, t, at constant SA and p.     
!                                                              [ unitless ]
!  CT_P_wrt_t  =  The first derivative of Conservative Temperature with 
!                 respect to pressure P (in Pa) at constant SA and t. 
!                                                                  [ K/Pa ]
!--------------------------------------------------------------------------

use gsw_mod_toolbox, only : gsw_gibbs, gsw_pt0_from_t

use gsw_mod_teos10_constants, only : gsw_cp0, gsw_sfac, gsw_t0, rec_db2pa

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sa, t, p
real (r8), intent(out), optional :: ct_p_wrt_t, ct_sa_wrt_t, ct_t_wrt_t

real (r8) :: g_sa_mod, g_sa_t_mod, pt0, x, y, y_pt, z

pt0 = gsw_pt0_from_t(sa,t,p)

if (present(ct_sa_wrt_t)) then

    x = sqrt(gsw_sfac*sa)
    y = 0.025_r8*t
    y_pt = 0.025_r8*pt0
    z = rec_db2pa*p   ! the input pressure (p) is sea pressure in units of dbar

    g_sa_t_mod = 1187.3715515697959_r8 + z*(1458.233059470092_r8 + &
            z*(-687.913805923122_r8 + z*(249.375342232496_r8 + &
            z*(-63.313928772146_r8 + 14.09317606630898_r8*z)))) + &
            x*(-1480.222530425046_r8 + x*(2175.341332000392_r8 + &
            x*(-980.14153344888_r8 + 220.542973797483_r8*x) + &
            y*(-548.4580073635929_r8 + y*(592.4012338275047_r8 + &
            y*(-274.2361238716608_r8 + 49.9394019139016_r8*y))) - &
            90.6734234051316_r8*z) + z*(-525.876123559641_r8 + &
            (249.57717834054571_r8 - 88.449193048287_r8*z)*z) + &
            y*(-258.3988055868252_r8 + z*(2298.348396014856_r8 + &
            z*(-325.1503575102672_r8 + 153.8390924339484_r8*z)) + &
            y*(-90.2046337756875_r8 - 4142.8793862113125_r8*z + &
            y*(10.50720794170734_r8 + 2814.78225133626_r8*z)))) + &
            y*(3520.125411988816_r8 + y*(-1351.605895580406_r8 + &
            y*(731.4083582010072_r8 + y*(-216.60324087531103_r8 + &
            25.56203650166196_r8*y) + z*(-2381.829935897496_r8 + &
            (597.809129110048_r8 - 291.8983352012704_r8*z)*z)) + &
            z*(4165.4688847996085_r8 + z*(-1229.337851789418_r8 + &
            (681.370187043564_r8 - 66.7696405958478_r8*z)*z))) + &
            z*(-3443.057215135908_r8 + z*(1349.638121077468_r8 + &
            z*(-713.258224830552_r8 + &
            (176.8161433232_r8 - 31.68006188846728_r8*z)*z))))
    g_sa_t_mod = 0.5_r8*gsw_sfac*0.025_r8*g_sa_t_mod
   
    g_sa_mod = 8645.36753595126_r8 + &
            x*(-7296.43987145382_r8 + x*(8103.20462414788_r8 + &
            y_pt*(2175.341332000392_r8 + y_pt*(-274.2290036817964_r8 + &
            y_pt*(197.4670779425016_r8 + y_pt*(-68.5590309679152_r8 + &
            9.98788038278032_r8*y_pt)))) + &
            x*(-5458.34205214835_r8 - 980.14153344888_r8*y_pt + &
            x*(2247.60742726704_r8 - 340.1237483177863_r8*x + &
            220.542973797483_r8*y_pt))) + &
            y_pt*(-1480.222530425046_r8 + &
            y_pt*(-129.1994027934126_r8 + &
            y_pt*(-30.0682112585625_r8 + y_pt*(2.626801985426835_r8 ))))) + &
            y_pt*(1187.3715515697959_r8 + &
            y_pt*(1760.062705994408_r8 + y_pt*(-450.535298526802_r8 + &
            y_pt*(182.8520895502518_r8 + y_pt*(-43.3206481750622_r8 + &
            4.26033941694366_r8*y_pt)))))
    g_sa_mod = 0.5_r8*gsw_sfac*g_sa_mod   

    ct_sa_wrt_t = (g_sa_mod - (gsw_t0+pt0)*g_sa_t_mod)/gsw_cp0

end if

if (present(ct_t_wrt_t)) &
    ct_t_wrt_t = -(gsw_t0+pt0)*gsw_gibbs(0,2,0,sa,t,p)/gsw_cp0

if (present(ct_p_wrt_t)) &
    ct_p_wrt_t = -(gsw_t0+pt0)*gsw_gibbs(0,1,1,sa,t,p)/gsw_cp0
                       
return
end subroutine

!--------------------------------------------------------------------------
!==========================================================================
elemental function gsw_ct_freezing (sa, p, saturation_fraction, poly)
!==========================================================================
!
!  Calculates the Conservative Temperature at which seawater freezes.
!
!  sa  =  absolute Salinity                                        [ g/kg ]
!  p   =  sea pressure                                             [ dbar ]
!         ( i.e. absolute pressure - 10.1325 dbar ) 
!  saturation_fraction = the saturation fraction of dissolved air in 
!                        seawater
!
!  ct_freezing = Conservative Temperature at freezing of seawater [ deg C ]
!                That is, the freezing temperature expressed in
!                terms of Conservative Temperature (ITS-90).                
!--------------------------------------------------------------------------

use gsw_mod_toolbox, only : gsw_ct_freezing_exact, gsw_ct_freezing_poly

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sa, p, saturation_fraction
logical, intent(in), optional :: poly

real (r8) :: gsw_ct_freezing

logical :: do_poly

if (present(poly)) then
    do_poly = poly
else
    do_poly = .false.
end if

if (do_poly) then
    gsw_ct_freezing = gsw_ct_freezing_poly(sa,p,saturation_fraction)
else
    gsw_ct_freezing = gsw_ct_freezing_exact(sa,p,saturation_fraction)
end if

return
end function

!--------------------------------------------------------------------------
!==========================================================================
elemental function gsw_ct_freezing_exact (sa, p, saturation_fraction)
!==========================================================================
!
!  Calculates the Conservative Temperature at which seawater freezes.  The 
!  Conservative Temperature freezing point is calculated from the exact 
!  in-situ freezing temperature which is found by a modified Newton-Raphson
!  iteration (McDougall and Wotherspoon, 2013) of the equality of the 
!  chemical potentials of water in seawater and in ice.
!
!  An alternative GSW function, gsw_CT_freezing_poly, it is based on a 
!  computationally-efficient polynomial, and is accurate to within -5e-4 K 
!  and 6e-4 K, when compared with this function.
!
!  SA  =  Absolute Salinity                                        [ g/kg ]
!  p   =  sea pressure                                             [ dbar ]
!         ( i.e. absolute pressure - 10.1325 dbar ) 
!  saturation_fraction = the saturation fraction of dissolved air in 
!                        seawater
!
!  CT_freezing = Conservative Temperature at freezing of seawater [ deg C ]
!--------------------------------------------------------------------------

use gsw_mod_toolbox, only : gsw_t_freezing_exact
use gsw_mod_toolbox, only : gsw_ct_from_t

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sa, p, saturation_fraction

real (r8) :: gsw_ct_freezing_exact

real (r8) :: t_freezing

t_freezing = gsw_t_freezing_exact(sa,p,saturation_fraction)
gsw_ct_freezing_exact = gsw_ct_from_t(sa,t_freezing,p)

return
end function

!--------------------------------------------------------------------------
!==========================================================================
elemental subroutine gsw_ct_freezing_first_derivatives (sa, p, &
                          saturation_fraction, ctfreezing_sa, ctfreezing_p)
!==========================================================================
!
!  Calculates the first derivatives of the Conservative Temperature at
!  which seawater freezes, with respect to Absolute Salinity SA and
!  pressure P (in Pa).  
!
!  SA  =  Absolute Salinity                                        [ g/kg ]
!  p   =  sea pressure                                             [ dbar ]
!         ( i.e. absolute pressure - 10.1325 dbar )
!  saturation_fraction = the saturation fraction of dissolved air in 
!                        seawater
!
!  CTfreezing_SA = the derivative of the Conservative Temperature at
!                  freezing (ITS-90) with respect to Absolute Salinity at
!                  fixed pressure              [ K/(g/kg) ] i.e. [ K kg/g ]
!
!  CTfreezing_P  = the derivative of the Conservative Temperature at
!                  freezing (ITS-90) with respect to pressure (in Pa) at
!                  fixed Absolute Salinity                         [ K/Pa ]
!--------------------------------------------------------------------------

use gsw_mod_toolbox, only : gsw_ct_first_derivatives_wrt_t_exact
use gsw_mod_toolbox, only : gsw_t_freezing_first_derivatives
use gsw_mod_toolbox, only : gsw_t_freezing_exact

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sa, p, saturation_fraction
real (r8), intent(out), optional :: ctfreezing_sa, ctfreezing_p

real (r8) :: tf_sa, tf_p, ct_sa_wrt_t, ct_t_wrt_t, ct_p_wrt_t, tf

tf = gsw_t_freezing_exact(sa,p,saturation_fraction)

if (present(ctfreezing_sa) .and. present(ctfreezing_p)) then

    call gsw_t_freezing_first_derivatives(sa,p,saturation_fraction, &
                                          tfreezing_sa=tf_sa,tfreezing_p=tf_p)
    call gsw_ct_first_derivatives_wrt_t_exact(sa,tf,p, &
          ct_sa_wrt_t=ct_sa_wrt_t,ct_t_wrt_t=ct_t_wrt_t,ct_p_wrt_t=ct_p_wrt_t)

    ctfreezing_sa = ct_sa_wrt_t + ct_t_wrt_t*tf_sa
    ctfreezing_p  = ct_p_wrt_t  + ct_t_wrt_t*tf_p

else if (present(ctfreezing_sa) .and. .not. present(ctfreezing_p)) then

    call gsw_t_freezing_first_derivatives(sa,p,saturation_fraction, &
                                          tfreezing_sa=tf_sa)
    call gsw_ct_first_derivatives_wrt_t_exact(sa,tf,p, &
          ct_sa_wrt_t=ct_sa_wrt_t,ct_t_wrt_t=ct_t_wrt_t)

    ctfreezing_sa = ct_sa_wrt_t + ct_t_wrt_t*tf_sa

else if (.not. present(ctfreezing_sa) .and. present(ctfreezing_p)) then

    call gsw_t_freezing_first_derivatives(sa,p,saturation_fraction, &
                                          tfreezing_p=tf_p)
    call gsw_ct_first_derivatives_wrt_t_exact(sa,tf,p, &
          ct_t_wrt_t=ct_t_wrt_t,ct_p_wrt_t=ct_p_wrt_t)

    ctfreezing_p  = ct_p_wrt_t  + ct_t_wrt_t*tf_p

end if

return
end subroutine

!--------------------------------------------------------------------------
!==========================================================================
elemental subroutine gsw_ct_freezing_first_derivatives_poly (sa, p, &
                          saturation_fraction, ctfreezing_sa, ctfreezing_p)
!==========================================================================
!
!  Calculates the first derivatives of the Conservative Temperature at
!  which seawater freezes, with respect to Absolute Salinity SA and
!  pressure P (in Pa) of the comptationally efficient polynomial fit of the
!  freezing temperature (McDougall et al., 2014).
!
!  SA  =  Absolute Salinity                                        [ g/kg ]
!  p   =  sea pressure                                             [ dbar ]
!         ( i.e. absolute pressure - 10.1325 dbar ) 
!  saturation_fraction = the saturation fraction of dissolved air in 
!                        seawater
!
!  CTfreezing_SA = the derivative of the Conservative Temperature at
!                  freezing (ITS-90) with respect to Absolute Salinity at
!                  fixed pressure              [ K/(g/kg) ] i.e. [ K kg/g ]
!
!  CTfreezing_P  = the derivative of the Conservative Temperature at
!                  freezing (ITS-90) with respect to pressure (in Pa) at
!                  fixed Absolute Salinity                         [ K/Pa ]
!--------------------------------------------------------------------------

use gsw_mod_teos10_constants, only : gsw_sso

use gsw_mod_freezing_poly_coefficients

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sa, p, saturation_fraction
real (r8), intent(out), optional :: ctfreezing_sa, ctfreezing_p

real (r8) :: p_r, sa_r, x

real (r8), parameter :: d = -a - a*b - 2.4_r8*b/gsw_sso
real (r8), parameter :: e = 2.0_r8*a*b/gsw_sso

sa_r = sa*1e-2_r8
x = sqrt(sa_r)
p_r = p*1e-4_r8

if (present(ctfreezing_sa)) ctfreezing_sa = &
    (c1 + x*(1.5_r8*c2 + x*(2.0_r8*c3 + x*(2.5_r8*c4 + x*(3.0_r8*c5 &
        + 3.5_r8*c6*x)))) + p_r*(c10 + x*(1.5_r8*c11 + x*(2.0_r8*c13 &
        + x*(2.5_r8*c16 + x*(3.0_r8*c19 + 3.5_r8*c22*x)))) & 
        + p_r*(c12 + x*(1.5_r8*c14 + x*(2.0_r8*c17 + 2.5_r8*c20*x)) &
        + p_r*(c15 + x*(1.5_r8*c18 + 2.0_r8*c21*x)))))*1e-2_r8 &
        - saturation_fraction*1e-3_r8*(d - sa*e)

if (present(ctfreezing_p)) ctfreezing_p = &
    (c7 + sa_r*(c10 + x*(c11 + x*(c13 + x*(c16 + x*(c19 + c22*x))))) &
        + p_r*(2.0_r8*c8 + sa_r*(2.0_r8*c12 + x*(2.0_r8*c14 + x*(2.0_r8*c17 &
        + 2.0_r8*c20*x))) + p_r*(3.0_r8*c9 + sa_r*(3.0_r8*c15 + x*(3.0_r8*c18 &
        + 3.0_r8*c21*x)))))*1e-8_r8

return
end subroutine

!--------------------------------------------------------------------------
!==========================================================================
elemental function gsw_ct_freezing_poly (sa, p, saturation_fraction)
!==========================================================================
!
!  Calculates the Conservative Temperature at which seawater freezes.
!  The error of this fit ranges between -5e-4 K and 6e-4 K when compared 
!  with the Conservative Temperature calculated from the exact in-situ 
!  freezing temperature which is found by a Newton-Raphson iteration of the 
!  equality of the chemical potentials of water in seawater and in ice.  
!  Note that the Conservative temperature freezing temperature can be found
!  by this exact method using the function gsw_CT_freezing.
!
!  SA  =  Absolute Salinity                                        [ g/kg ]
!  p   =  sea pressure                                             [ dbar ]
!         ( i.e. absolute pressure - 10.1325 dbar ) 
!  saturation_fraction = the saturation fraction of dissolved air in 
!                        seawater
!
!  CT_freezing = Conservative Temperature at freezing of seawater [ deg C ]
!                That is, the freezing temperature expressed in
!                terms of Conservative Temperature (ITS-90).                
!--------------------------------------------------------------------------

use gsw_mod_teos10_constants, only : gsw_sso

use gsw_mod_freezing_poly_coefficients

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sa, p, saturation_fraction

real (r8) :: gsw_ct_freezing_poly

real (r8) :: p_r, sa_r, x

sa_r = sa*1e-2_r8
x = sqrt(sa_r)
p_r = p*1e-4_r8

gsw_ct_freezing_poly = c0 &
    + sa_r*(c1 + x*(c2 + x*(c3 + x*(c4 + x*(c5 + c6*x))))) &
    + p_r*(c7 + p_r*(c8 + c9*p_r)) + sa_r*p_r*(c10 + p_r*(c12 &
    + p_r*(c15 + c21*sa_r)) + sa_r*(c13 + c17*p_r + c19*sa_r) &
    + x*(c11 + p_r*(c14 + c18*p_r) + sa_r*(c16 + c20*p_r + c22*sa_r)))

! Adjust for the effects of dissolved air 
gsw_ct_freezing_poly = gsw_ct_freezing_poly - saturation_fraction* &
                 (1e-3_r8)*(2.4_r8 - a*sa)*(1.0_r8 + b*(1.0_r8 - sa/gsw_sso))

return
end function
!==========================================================================
elemental function gsw_ct_from_enthalpy (sa, h, p)
!==========================================================================
!
!  Calculates the Conservative Temperature of seawater, given the Absolute 
!  Salinity, specific enthalpy, h, and pressure p.
!
!  SA  =  Absolute Salinity                                        [ g/kg ]
!  h   =  specific enthalpy                                        [ J/kg ]
!  p   =  sea pressure                                             [ dbar ]
!         ( i.e. absolute pressure - 10.1325d0 dbar ) 
!
!  CT  =  Conservative Temperature ( ITS-90)                      [ deg C ]
!--------------------------------------------------------------------------

use gsw_mod_toolbox, only : gsw_ct_freezing_exact, gsw_enthalpy
use gsw_mod_toolbox, only : gsw_enthalpy_first_derivatives

use gsw_mod_error_functions, only : gsw_error_code

use gsw_mod_teos10_constants, only : gsw_cp0

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sa, h, p

real (r8) :: gsw_ct_from_enthalpy

real (r8) :: ct, ct_freezing, ct_mean, ct_old, f, h_freezing
real (r8) :: h_ct, h_40

real (r8), parameter :: ct_40 = 40.0_r8

character (*), parameter :: func_name = "gsw_ct_from_enthalpy"

ct_freezing = gsw_ct_freezing_exact(sa,p,0.0_r8)

h_freezing = gsw_enthalpy(sa,ct_freezing,p)
if (h .lt. h_freezing - gsw_cp0) then
    ! The input, seawater enthalpy h, is less than the enthalpy at the
    ! freezing temperature, i.e. the water is frozen.
    gsw_ct_from_enthalpy = gsw_error_code(1,func_name)
    return
endif

h_40 = gsw_enthalpy(sa,ct_40,p)
if (h .gt. h_40) then
    ! The input seawater enthalpy is greater than the enthalpy when CT is 40C
    gsw_ct_from_enthalpy = gsw_error_code(2,func_name)
    return
endif

! first guess of ct
ct = ct_freezing + (ct_40 - ct_freezing)*(h - h_freezing)/(h_40 - h_freezing)
call gsw_enthalpy_first_derivatives(sa,ct,p,h_ct=h_ct)

!--------------------------------------------------------------------------
! Begin the modified Newton-Raphson iterative procedure 
!--------------------------------------------------------------------------

ct_old = ct
f = gsw_enthalpy(sa,ct_old,p) - h
ct = ct_old - f/h_ct
ct_mean = 0.5_r8*(ct + ct_old)
call gsw_enthalpy_first_derivatives(sa,ct_mean,p,h_ct=h_ct)
ct = ct_old - f/h_ct

ct_old = ct
f = gsw_enthalpy(sa,ct_old,p) - h
ct = ct_old - f/h_ct

! After 1.5d0 iterations of this modified Newton-Raphson iteration,
! the error in CT is no larger than 4x10^-13 degrees C, which 
! is machine precision for this calculation. 

gsw_ct_from_enthalpy = ct

return
end function

!--------------------------------------------------------------------------
!==========================================================================
elemental function gsw_ct_from_enthalpy_exact (sa, h, p)
!==========================================================================
!
!  Calculates the Conservative Temperature of seawater, given the Absolute 
!  Salinity, specific enthalpy, h, and pressure p.
!
!  SA  =  Absolute Salinity                                        [ g/kg ]
!  h   =  specific enthalpy                                        [ J/kg ]
!  p   =  sea pressure                                             [ dbar ]
!         ( i.e. absolute pressure - 10.1325d0 dbar ) 
!
!  CT  =  Conservative Temperature ( ITS-90)                      [ deg C ]
!--------------------------------------------------------------------------

use gsw_mod_toolbox, only : gsw_ct_freezing_exact, gsw_enthalpy_ct_exact
use gsw_mod_toolbox, only : gsw_enthalpy_first_derivatives_ct_exact

use gsw_mod_error_functions, only : gsw_error_code

use gsw_mod_teos10_constants, only : gsw_cp0

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sa, h, p

real (r8) :: gsw_ct_from_enthalpy_exact

real (r8) :: ct, ct_freezing, ct_mean, ct_old, f, h_freezing
real (r8) :: h_ct, h_40

real (r8), parameter :: ct_40 = 40.0_r8

character (*), parameter :: func_name = "gsw_ct_from_enthalpy_exact"

ct_freezing = gsw_ct_freezing_exact(sa,p,0.0_r8)

h_freezing = gsw_enthalpy_ct_exact(sa,ct_freezing,p)
if (h .lt. h_freezing - gsw_cp0) then
    ! The input, seawater enthalpy h, is less than the enthalpy at the
    ! freezing temperature, i.e. the water is frozen.
    gsw_ct_from_enthalpy_exact = gsw_error_code(1,func_name)
    return
endif

h_40 = gsw_enthalpy_ct_exact(sa,ct_40,p)
if (h .gt. h_40) then
    ! The input seawater enthalpy is greater than the enthalpy when CT is 40C
    gsw_ct_from_enthalpy_exact = gsw_error_code(2,func_name)
    return
endif

! First guess of ct
ct = ct_freezing + (ct_40 - ct_freezing)*(h - h_freezing)/(h_40 - h_freezing)
call gsw_enthalpy_first_derivatives_ct_exact(sa,ct,p,h_ct=h_ct)

!--------------------------------------------------------------------------
! Begin the modified Newton-Raphson iterative procedure 
!--------------------------------------------------------------------------

ct_old = ct
f = gsw_enthalpy_ct_exact(sa,ct_old,p) - h
ct = ct_old - f/h_ct
ct_mean = 0.5_r8*(ct + ct_old)
call gsw_enthalpy_first_derivatives_ct_exact(sa,ct_mean,p,h_ct=h_ct)
ct = ct_old - f/h_ct

! After 1 iteration of this modified Newton-Raphson iteration,
! the error in CT is no larger than 5x10^-14 degrees C, which 
! is machine precision for this calculation. 

gsw_ct_from_enthalpy_exact = ct

return
end function

!--------------------------------------------------------------------------
!=========================================================================
elemental function gsw_ct_from_entropy (sa, entropy)
!=========================================================================
!
!  Calculates Conservative Temperature with entropy as an input variable.  
!
!  SA       =  Absolute Salinity                                   [ g/kg ]
!  entropy  =  specific entropy                                   [ deg C ]
!
!  CT  =  Conservative Temperature (ITS-90)                       [ deg C ]
!--------------------------------------------------------------------------

use gsw_mod_toolbox, only : gsw_ct_from_pt, gsw_pt_from_entropy

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sa, entropy

real (r8) :: gsw_ct_from_entropy

real (r8) :: pt

pt = gsw_pt_from_entropy(sa,entropy)
gsw_ct_from_entropy = gsw_ct_from_pt(sa,pt)

return
end function

!--------------------------------------------------------------------------
!==========================================================================
elemental function gsw_ct_from_pt (sa, pt) 
!==========================================================================
!
! Calculates Conservative Temperature from potential temperature of seawater  
!
! sa      : Absolute Salinity                              [g/kg]
! pt      : potential temperature with                     [deg C]
!           reference pressure of 0 dbar
!
! gsw_ct_from_pt : Conservative Temperature                [deg C]
!--------------------------------------------------------------------------

use gsw_mod_teos10_constants, only : gsw_cp0, gsw_sfac

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sa, pt 

real (r8) :: gsw_ct_from_pt

real (r8) :: pot_enthalpy, x2, x, y

x2 = gsw_sfac*sa
x = sqrt(x2)
y = pt*0.025_r8        ! normalize for F03 and F08

pot_enthalpy =  61.01362420681071_r8 + y*(168776.46138048015_r8 + &
               y*(-2735.2785605119625_r8 + y*(2574.2164453821433_r8 + &
               y*(-1536.6644434977543_r8 + y*(545.7340497931629_r8 + &
               (-50.91091728474331_r8 - 18.30489878927802_r8*y)*y))))) + &
               x2*(268.5520265845071_r8 + y*(-12019.028203559312_r8 + &
               y*(3734.858026725145_r8 + y*(-2046.7671145057618_r8 + &
               y*(465.28655623826234_r8 + (-0.6370820302376359_r8 - &
               10.650848542359153_r8*y)*y)))) + &
               x*(937.2099110620707_r8 + y*(588.1802812170108_r8 + &
               y*(248.39476522971285_r8 + (-3.871557904936333_r8 - &
               2.6268019854268356_r8*y)*y)) + &
               x*(-1687.914374187449_r8 + x*(246.9598888781377_r8 + &
               x*(123.59576582457964_r8 - 48.5891069025409_r8*x)) + &
               y*(936.3206544460336_r8 + &
               y*(-942.7827304544439_r8 + y*(369.4389437509002_r8 + &
               (-33.83664947895248_r8 - 9.987880382780322_r8*y)*y))))))

gsw_ct_from_pt = pot_enthalpy/gsw_cp0

return
end function

!--------------------------------------------------------------------------
!==========================================================================
elemental subroutine gsw_ct_from_rho (rho, sa, p, ct, ct_multiple)
! =========================================================================
!
!  Calculates the Conservative Temperature of a seawater sample, for given
!  values of its density, Absolute Salinity and sea pressure (in dbar).
!
!  rho  =  density of a seawater sample (e.g. 1026 kg/m^3)       [ kg/m^3 ]
!   Note. This input has not had 1000 kg/m^3 subtracted from it.
!     That is, it is 'density', not 'density anomaly'.
!  SA   =  Absolute Salinity                                       [ g/kg ]
!  p    =  sea pressure                                            [ dbar ]
!          ( i.e. absolute pressure - 10.1325 dbar )
!
!  CT  =  Conservative Temperature  (ITS-90)                      [ deg C ]
!  CT_multiple  =  Conservative Temperature  (ITS-90)             [ deg C ]
!    Note that at low salinities, in brackish water, there are two possible
!      Conservative Temperatures for a single density.  This programme will
!      output both valid solutions.  To see this second solution the user 
!      must call the programme with two outputs (i.e. [CT,CT_multiple]), if
!      there is only one possible solution and the programme has been 
!      called with two outputs the second variable will be set to NaN.
!--------------------------------------------------------------------------

use gsw_mod_toolbox, only : gsw_ct_freezing_poly, gsw_ct_maxdensity, gsw_rho
use gsw_mod_toolbox, only : gsw_rho_alpha_beta, gsw_specvol

use gsw_mod_error_functions, only : gsw_error_code, gsw_error_limit

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: rho, sa, p
real (r8), intent(out) :: ct
real (r8), intent(out), optional :: ct_multiple

integer :: number_of_iterations
real (r8) :: a, alpha_freezing, alpha_mean, b, c, ct_a, ct_b, ct_diff
real (r8) :: ct_freezing, ct_max_rho, ct_mean, ct_old
real (r8) :: delta_ct, delta_v, factor, factorqa, factorqb
real (r8) :: rho_40, rho_extreme, rho_freezing, rho_max, rho_mean
real (r8) :: rho_old, sqrt_disc, top, v_ct, v_lab

character (*), parameter :: func_name = "gsw_ct_from_rho"

! alpha_limit is the positive value of the thermal expansion coefficient
! which is used at the freezing temperature to distinguish between
! salty and fresh water.
real (r8), parameter :: alpha_limit = 1e-5_r8

! rec_half_rho_TT is a constant representing the reciprocal of half the
! second derivative of density with respect to temperature near the
! temperature of maximum density.
real (r8), parameter :: rec_half_rho_tt = -110.0_r8

rho_40 = gsw_rho(sa,40.0_r8,p)
if (rho .lt. rho_40) then
    ct = gsw_error_code(1,func_name)
    if (present(ct_multiple)) ct_multiple = ct
    return
end if

ct_max_rho = gsw_ct_maxdensity(sa,p)
rho_max = gsw_rho(sa,ct_max_rho,p)
rho_extreme = rho_max

! Assumes that the seawater is always unsaturated with air
ct_freezing = gsw_ct_freezing_poly(sa,p,0.0_r8)

call gsw_rho_alpha_beta(sa,ct_freezing,p,rho=rho_freezing,alpha=alpha_freezing)

! reset the extreme values
if (ct_freezing .gt. ct_max_rho) rho_extreme = rho_freezing

if (rho .gt. rho_extreme) then
    ct = gsw_error_code(2,func_name)
    if (present(ct_multiple)) ct_multiple = ct
    return
end if

if (alpha_freezing .gt. alpha_limit) then

    ct_diff = 40.0_r8 - ct_freezing
    top = rho_40 - rho_freezing + rho_freezing*alpha_freezing*ct_diff
    a = top/(ct_diff*ct_diff)
    b = -rho_freezing*alpha_freezing
    c = rho_freezing - rho
    sqrt_disc = sqrt(b*b - 4*a*c)
    ct = ct_freezing + 0.5_r8*(-b - sqrt_disc)/a

else

    ct_diff = 40.0_r8 - ct_max_rho
    factor = (rho_max - rho)/(rho_max - rho_40)
    delta_ct = ct_diff*sqrt(factor)
    
    if (delta_ct .gt. 5.0_r8) then

        ct = ct_max_rho + delta_ct

    else
     
        ! Set the initial value of the quadratic solution roots.
        ct_a = ct_max_rho + sqrt(rec_half_rho_tt*(rho - rho_max))       
        do number_of_iterations = 1, 7
            ct_old = ct_a
            rho_old = gsw_rho(sa,ct_old,p)
            factorqa = (rho_max - rho)/(rho_max - rho_old)
            ct_a = ct_max_rho + (ct_old - ct_max_rho)*sqrt(factorqa)
        end do

        if (ct_freezing - ct_a .lt. 0.0_r8) then
            ct = gsw_error_code(3,func_name)
            if (present(ct_multiple)) ct_multiple = ct
            return
        end if

        ct = ct_a
        if (.not. present(ct_multiple)) return

        ! Set the initial value of the quadratic solution roots.
        ct_b = ct_max_rho - sqrt(rec_half_rho_tt*(rho - rho_max))    
        do number_of_iterations = 1, 7
            ct_old = ct_b
            rho_old = gsw_rho(sa,ct_old,p)
            factorqb = (rho_max - rho)/(rho_max - rho_old)
            ct_b = ct_max_rho + (ct_old - ct_max_rho)*sqrt(factorqb)
        end do

        ! After seven iterations of this quadratic iterative procedure,
        ! the error in rho is no larger than 4.6x10^-13 kg/m^3.

        if (ct_freezing - ct_b .lt. 0.0_r8) then
            ct = gsw_error_code(4,func_name)
            if (present(ct_multiple)) ct_multiple = ct
            return
        end if

        ct_multiple = ct_b
        return

    end if

end if

! Begin the modified Newton-Raphson iterative method

v_lab = 1.0_r8/rho
call gsw_rho_alpha_beta(sa,ct,p,rho=rho_mean,alpha=alpha_mean)
v_ct = alpha_mean/rho_mean

do number_of_iterations = 1, 3
    ct_old = ct
    delta_v = gsw_specvol(sa,ct_old,p) - v_lab
    ct = ct_old - delta_v/v_ct
    ct_mean = 0.5_r8*(ct + ct_old)
    call gsw_rho_alpha_beta(sa,ct_mean,p,rho=rho_mean,alpha=alpha_mean)
    v_ct = alpha_mean/rho_mean
    ct = ct_old - delta_v/v_ct 
end do

! After three iterations of this modified Newton-Raphson iteration,
! the error in rho is no larger than 1.6x10^-12 kg/m^3.

if (present(ct_multiple)) ct_multiple = gsw_error_code(5,func_name)

return
end subroutine

!--------------------------------------------------------------------------
!==========================================================================
elemental function gsw_ct_from_t (sa, t, p) 
!==========================================================================
!   
! Calculates Conservative Temperature from in-situ temperature
!
! sa     : Absolute Salinity                               [g/kg]
! t      : in-situ temperature                             [deg C]
! p      : sea pressure                                    [dbar]
!
! gsw_ct_from_t : Conservative Temperature                 [deg C]
!--------------------------------------------------------------------------

use gsw_mod_toolbox, only : gsw_ct_from_pt, gsw_pt0_from_t

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sa, t, p 

real (r8) :: gsw_ct_from_t

real (r8) :: pt0

pt0 = gsw_pt0_from_t(sa,t,p)
gsw_ct_from_t = gsw_ct_from_pt(sa,pt0)

return
end function

!--------------------------------------------------------------------------
!==========================================================================
elemental function gsw_ct_maxdensity (sa, p)
!==========================================================================
!
!  Calculates the Conservative Temperature of maximum density of seawater. 
!  This function returns the Conservative temperature at which the density
!  of seawater is a maximum, at given Absolute Salinity, SA, and sea 
!  pressure, p (in dbar).
!
!  SA =  Absolute Salinity                                         [ g/kg ]
!  p  =  sea pressure                                              [ dbar ]
!        ( i.e. absolute pressure - 10.1325 dbar ) 
!
!  CT_maxdensity  =  Conservative Temperature at which            [ deg C ]
!                    the density of seawater is a maximum for
!                    given Absolute Salinity and pressure.
!--------------------------------------------------------------------------

use gsw_mod_toolbox, only : gsw_alpha

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sa, p

real (r8) :: gsw_ct_maxdensity

integer :: number_of_iterations
real (r8) :: alpha, ct, ct_mean, ct_old, dalpha_dct

real (r8), parameter :: dct = 0.001_r8

ct = 3.978_r8 - 0.22072_r8*sa         ! the initial guess of ct.

dalpha_dct = 1.1e-5_r8                ! the initial guess for dalpha_dct.

do number_of_iterations = 1, 3
    ct_old = ct
    alpha = gsw_alpha(sa,ct_old,p)
    ct = ct_old - alpha/dalpha_dct
    ct_mean = 0.5_r8*(ct + ct_old)
    dalpha_dct = (gsw_alpha(sa,ct_mean+dct,p) &
                  - gsw_alpha(sa,ct_mean-dct,p))/(dct + dct)
    ct = ct_old - alpha/dalpha_dct
end do

! After three iterations of this modified Newton-Raphson (McDougall and 
! Wotherspoon, 2012) iteration, the error in CT_maxdensity is typically no
! larger than 1x10^-15 degress C.  

gsw_ct_maxdensity = ct

return
end function

!--------------------------------------------------------------------------
!==========================================================================
elemental subroutine gsw_ct_second_derivatives (sa, pt, ct_sa_sa, ct_sa_pt, &
                                                ct_pt_pt)
!==========================================================================
!
!  Calculates the following three, second-order derivatives of Conservative 
!  Temperature
!   (1) CT_SA_SA, the second derivative with respect to Absolute Salinity  
!       at constant potential temperature (with p_ref = 0 dbar),
!   (2) CT_SA_pt, the derivative with respect to potential temperature
!       (the regular potential temperature which is referenced to 0 dbar)
!       and Absolute Salinity, and
!   (3) CT_pt_pt, the second derivative with respect to potential 
!       temperature (the regular potential temperature which is referenced 
!       to 0 dbar) at constant Absolute Salinity. 
!
!  SA  =  Absolute Salinity                                        [ g/kg ]
!  pt  =  potential temperature (ITS-90)                          [ deg C ]   
!         (whose reference pressure is 0 dbar)
!
!  CT_SA_SA  =  The second derivative of Conservative Temperature with 
!               respect to Absolute Salinity at constant potential 
!               temperature (the regular potential temperature which 
!               has reference sea pressure of 0 dbar).  
!               CT_SA_SA has units of:                     [ K/((g/kg)^2) ]
!  CT_SA_pt  =  The derivative of Conservative Temperature with 
!               respect to potential temperature (the regular one with 
!               p_ref = 0 dbar) and Absolute Salinity.   
!               CT_SA_pt has units of:                        [ 1/(g/kg) ]
!  CT_pt_pt  =  The second derivative of Conservative Temperature with 
!               respect to potential temperature (the regular one with 
!               p_ref = 0 dbar) at constant SA.   
!               CT_pt_pt has units of:                              [ 1/K ]
!--------------------------------------------------------------------------

use gsw_mod_toolbox, only : gsw_ct_first_derivatives

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sa, pt
real (r8), intent(out), optional :: ct_sa_sa, ct_sa_pt, ct_pt_pt

real (r8) :: ct_pt_l, ct_pt_u, ct_sa_l, ct_sa_u, pt_l, pt_u, sa_l, sa_u

real (r8), parameter :: dsa = 1e-3_r8, dpt = 1e-2_r8

if (present(ct_sa_sa)) then

    sa_l = max(sa - dsa, 0.0_r8)
    sa_u = sa + dsa

    call gsw_ct_first_derivatives(sa_l,pt,ct_sa=ct_sa_l)
    call gsw_ct_first_derivatives(sa_u,pt,ct_sa=ct_sa_u)

    ct_sa_sa = (ct_sa_u - ct_sa_l)/(sa_u - sa_l)

end if

if (present(ct_sa_pt) .or. present(ct_pt_pt)) then

    pt_l = pt - dpt
    pt_u = pt + dpt

    if (present(ct_sa_pt) .and. present(ct_pt_pt)) then

        call gsw_ct_first_derivatives(sa,pt_l,ct_sa_l,ct_pt_l)
        call gsw_ct_first_derivatives(sa,pt_u,ct_sa_u,ct_pt_u)

        ct_sa_pt = (ct_sa_u - ct_sa_l)/(pt_u - pt_l)
        ct_pt_pt = (ct_pt_u - ct_pt_l)/(pt_u - pt_l)

    else if (present(ct_sa_pt) .and. .not. present(ct_pt_pt)) then

        call gsw_ct_first_derivatives(sa,pt_l,ct_sa=ct_sa_l)
        call gsw_ct_first_derivatives(sa,pt_u,ct_sa=ct_sa_u)

        ct_sa_pt = (ct_sa_u - ct_sa_l)/(pt_u - pt_l)

    else if (.not. present(ct_sa_pt) .and. present(ct_pt_pt)) then

        call gsw_ct_first_derivatives(sa,pt_l,ct_pt=ct_pt_l)
        call gsw_ct_first_derivatives(sa,pt_u,ct_pt=ct_pt_u)

        ct_pt_pt = (ct_pt_u - ct_pt_l)/(pt_u - pt_l)

    end if

end if

return
end subroutine

!--------------------------------------------------------------------------
!==========================================================================
elemental function gsw_deltasa_atlas (p, long, lat)
!==========================================================================
!
! Calculates the Absolute Salinity Anomaly atlas value, deltaSA_atlas.
!
! p      : sea pressure                                    [dbar]
! long   : longiture                                       [deg E]     
! lat    : latitude                                        [deg N]
!
! gsw_deltasa_atlas : Absolute Salinity Anomaly atlas value    [g/kg]
!--------------------------------------------------------------------------

use gsw_mod_toolbox, only : gsw_add_barrier, gsw_add_mean, gsw_util_indx

use gsw_mod_error_functions, only : gsw_error_code, gsw_error_limit

use gsw_mod_saar_data

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: p, long, lat

real (r8) :: gsw_deltasa_atlas

integer :: indx0, indy0, indz0, k, ndepth_max

real (r8), dimension(4) :: dsar, dsar_old
real (r8) :: dlong, dlat
real (r8) :: p0_original, sa_upper, sa_lower 
real (r8) :: r1, s1, t1, p_tmp, long360

character (*), parameter :: func_name = "gsw_deltasa_atlas"

long360 = long
if (long360.lt.0.0_r8) long360 = long360 + 360.0_r8

indx0 = floor(1.0_r8 + (nx-1)*(long360-longs_ref(1)) / &
                (longs_ref(nx)-longs_ref(1)))

indy0 = floor(1.0_r8 + (ny-1)*(lat-lats_ref(1)) / (lats_ref(ny)-lats_ref(1)))

if ((indx0.ge.1 .and. indx0.le.nx) .and. (indy0.ge.1 .and. indy0.le.ny)) then
   if (indx0.eq.nx) indx0 = nx-1
   if (indy0.eq.ny) indy0 = ny-1
else
   ! This is to catch any out-of-range or nonsense lat/long input (including
   ! NaN, +Inf, -Inf etc). Note: NaNs will not satisfy any "if" conditional
   ! so will be trapped by the "else".
   gsw_deltasa_atlas = gsw_error_code(2,func_name)
   return
end if

! Look for the maximum valid "ndepth_ref" value around our point.
! Note: invalid "ndepth_ref" values are NaNs (a hangover from the codes
! Matlab origins), but we have replaced the NaNs with a value of "9e90",
! hence we need an additional upper-limit check in the code below so they
! will not be recognised as valid values.
ndepth_max = -1
do k = 1, 4
   if ((ndepth_ref(indy0+delj(k),indx0+deli(k)).gt.0) .and. &
       (ndepth_ref(indy0+delj(k),indx0+deli(k)).lt.99)) &
      ndepth_max = max(ndepth_max,ndepth_ref(indy0+delj(k),indx0+deli(k)))
end do

! If we are a long way from the ocean then there will be no valid "ndepth_ref"
! values near the point (ie. surrounded by NaNs) - so deltasa_atlas = 0.0
if (ndepth_max.eq.-1) then
   gsw_deltasa_atlas = 0.0_r8 
   return
end if 

p0_original = p
p_tmp = p
if (p_tmp.gt.p_ref(ndepth_max)) p_tmp = p_ref(ndepth_max)
call gsw_util_indx(p_ref,nz,p_tmp,indz0)
    
dlong = longs_ref(indx0+1) - longs_ref(indx0)
dlat = lats_ref(indy0+1) - lats_ref(indy0)

r1 = (long360-longs_ref(indx0)) / dlong
s1 = (lat-lats_ref(indy0)) / dlat
t1 = (p_tmp-p_ref(indz0)) / (p_ref(indz0+1)-p_ref(indz0))

do k = 1, 4
   dsar(k) = delta_sa_ref(indz0,indy0+delj(k),indx0+deli(k))
end do

if (  longs_pan(1).le.long360 .and. long360.le.longs_pan(npan)-0.001_r8 .and. &
    lats_pan(npan).le.lat     .and.     lat.le.lats_pan(1)) then
   dsar_old = dsar
   call gsw_add_barrier(dsar_old,long360,lat,longs_ref(indx0), &
                        lats_ref(indy0),dlong,dlat,dsar)
else if (abs(sum(dsar)).ge.1e10_r8) then 
   dsar_old = dsar
   call gsw_add_mean(dsar_old,dsar)
end if

sa_upper = (1.0_r8-s1)*(dsar(1) + r1*(dsar(2)-dsar(1))) + s1*(dsar(4) + &
                r1*(dsar(3)-dsar(4)))

do k = 1,4
   dsar(k) = delta_sa_ref(indz0+1,indy0+delj(k),indx0+deli(k))
end do

if (  longs_pan(1).le.long360 .and. long360.le.longs_pan(npan)-0.001_r8 .and. &
    lats_pan(npan).le.lat     .and.     lat.le.lats_pan(1)) then
   dsar_old = dsar
   call gsw_add_barrier(dsar_old,long360,lat,longs_ref(indx0), &
                        lats_ref(indy0),dlong,dlat,dsar)
else if (abs(sum(dsar)).ge.1e10_r8) then 
   dsar_old = dsar
   call gsw_add_mean(dsar_old,dsar)
end if

sa_lower = (1.0_r8-s1)*(dsar(1) + r1*(dsar(2)-dsar(1))) + s1*(dsar(4) + &
                r1*(dsar(3)-dsar(4)))
if (abs(sa_lower).ge.1e10_r8) sa_lower = sa_upper
gsw_deltasa_atlas = sa_upper + t1*(sa_lower-sa_upper)

if (abs(gsw_deltasa_atlas).ge.1e10_r8) &
        gsw_deltasa_atlas = gsw_error_code(3,func_name)
  
return
end function

!--------------------------------------------------------------------------
!==========================================================================
elemental function gsw_deltasa_from_sp (sp, p, long, lat) 
!==========================================================================
!
! Calculates Absolute Salinity Anomaly, deltaSA, from Practical Salinity, SP. 
!
! sp     : Practical Salinity                              [unitless]
! p      : sea pressure                                    [dbar]
! long   : longitude                                       [deg E]     
! lat    : latitude                                        [deg N]
!
! gsw_deltasa_from_sp : Absolute Salinty Anomaly           [g/kg]
!--------------------------------------------------------------------------

use gsw_mod_toolbox, only : gsw_sa_from_sp, gsw_sr_from_sp

use gsw_mod_error_functions, only : gsw_error_code, gsw_error_limit

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sp, p, long, lat 

real (r8) :: gsw_deltasa_from_sp

character (*), parameter :: func_name = "gsw_deltasa_from_sp"

gsw_deltasa_from_sp = gsw_sa_from_sp(sp,p,long,lat) - gsw_sr_from_sp(sp)

if (gsw_deltasa_from_sp.gt.gsw_error_limit) &
        gsw_deltasa_from_sp = gsw_error_code(1,func_name)

return
end function

!--------------------------------------------------------------------------
!==========================================================================
elemental function gsw_dilution_coefficient_t_exact (sa, t, p)
!==========================================================================
!
!  Calculates the dilution coefficient of seawater.  The dilution 
!  coefficient of seawater is defined as the Absolute Salinity times the 
!  second derivative of the Gibbs function with respect to Absolute 
!  Salinity, that is, SA.*g_SA_SA.
!
!  SA =  Absolute Salinity                                         [ g/kg ]
!  t  =  in-situ temperature (ITS-90)                             [ deg C ]
!  p  =  sea pressure                                              [ dbar ]
!        ( i.e. absolute pressure - 10.1325 dbar ) 
!
!  dilution_coefficient_t_exact  =  dilution coefficient   [ (J/kg)(kg/g) ]
!--------------------------------------------------------------------------

use gsw_mod_teos10_constants, only : gsw_sfac

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sa, t, p

real (r8) :: gsw_dilution_coefficient_t_exact

real (r8) :: g08, x, x2, y, z

x2 = gsw_sfac*sa
x = sqrt(x2)
y = t*0.025_r8
z = p*1e-4_r8      ! note.the input pressure (p) is sea pressure in units of dbar.

g08 = 2.0_r8*(8103.20462414788_r8 + &
          y*(2175.341332000392_r8 + &
              y*(-274.2290036817964_r8 + &
                  y*(197.4670779425016_r8 + &
                      y*(-68.5590309679152_r8 + 9.98788038278032_r8*y))) - &
          90.6734234051316_r8*z) + &
              1.5_r8*x*(-5458.34205214835_r8 - 980.14153344888_r8*y + &
                  (4.0_r8/3.0_r8)*x*(2247.60742726704_r8 - &
                  340.1237483177863_r8*1.25_r8*x + 220.542973797483_r8*y) + &
              180.142097805543_r8*z) + &
          z*(-219.1676534131548_r8 + &
              (-16.32775915649044_r8 - 120.7020447884644_r8*z)*z))
    
g08 = x2*g08 + & 
          x*(-7296.43987145382_r8 + &
              z*(598.378809221703_r8 + &
                  z*(-156.8822727844005_r8 + &
                      (204.1334828179377_r8 - 10.23755797323846_r8*z)*z)) + &
              y*(-1480.222530425046_r8 + &
                  z*(-525.876123559641_r8 + &
                      (249.57717834054571_r8 - 88.449193048287_r8*z)*z) + &
                  y*(-129.1994027934126_r8 + &
                      z*(1149.174198007428_r8 + &
                          z*(-162.5751787551336_r8 + 76.9195462169742_r8*z)) + &
                  y*(-30.0682112585625_r8 - 1380.9597954037708_r8*z + &
                      y*(2.626801985426835_r8 + 703.695562834065_r8*z))))) + &
      11625.62913253464_r8 + 1702.453469893412_r8*y
    
gsw_dilution_coefficient_t_exact = 0.25_r8*gsw_sfac*g08

! Note that this function avoids the singularity that occurs at SA = 0 if
! the straightforward expression for the dilution coefficient of seawater,
! SA*g_SA_SA is simply evaluated as SA.*gsw_gibbs(2,0,0,SA,t,p). 

return
end function

!--------------------------------------------------------------------------

!==========================================================================
elemental function gsw_dynamic_enthalpy (sa, ct, p)
!==========================================================================
!
!  Calculates dynamic enthalpy of seawater using the computationally-
!  efficient expression for specific volume in terms of SA, CT and p
!  (Roquet et al., 2014).  Dynamic enthalpy is defined as enthalpy minus
!  potential enthalpy (Young, 2010). 
!
!  SA  =  Absolute Salinity                                        [ g/kg ]
!  CT  =  Conservative Temperature (ITS-90)                       [ deg C ]
!  p   =  sea pressure                                             [ dbar ]
!         ( i.e. absolute pressure - 10.1325 dbar )
!
!  dynamic_enthalpy  =  dynamic enthalpy                           [ J/kg ]
!--------------------------------------------------------------------------

use gsw_mod_teos10_constants, only : db2pa, gsw_sfac, offset

use gsw_mod_specvol_coefficients

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sa, ct, p

real (r8) :: gsw_dynamic_enthalpy

real (r8) :: dynamic_enthalpy_part, xs, ys, z

xs = sqrt(gsw_sfac*sa + offset)
ys = ct*0.025_r8
z = p*1e-4_r8

dynamic_enthalpy_part =  z*(h001 + xs*(h101 + xs*(h201 + xs*(h301 + xs*(h401 &
    + xs*(h501 + h601*xs))))) + ys*(h011 + xs*(h111 + xs*(h211 + xs*(h311 &
    + xs*(h411 + h511*xs)))) + ys*(h021 + xs*(h121 + xs*(h221 + xs*(h321 &
    + h421*xs))) + ys*(h031 + xs*(h131 + xs*(h231 + h331*xs)) + ys*(h041 &
    + xs*(h141 + h241*xs) + ys*(h051 + h151*xs + h061*ys))))) + z*(h002 &
    + xs*(h102 + xs*(h202 + xs*(h302 + xs*(h402 + h502*xs)))) + ys*(h012 &
    + xs*(h112 + xs*(h212 + xs*(h312 + h412*xs))) + ys*(h022 + xs*(h122 &
    + xs*(h222 + h322*xs)) + ys*(h032 + xs*(h132 + h232*xs) + ys*(h042 &
    + h142*xs + h052*ys)))) + z*(h003 + xs*(h103 + xs*(h203 + xs*(h303 &
    + h403*xs))) + ys*(h013 + xs*(h113 + xs*(h213 + h313*xs)) + ys*(h023 &
    + xs*(h123 + h223*xs) + ys*(h033 + h133*xs + h043*ys))) + z*(h004 &
    + xs*(h104 + h204*xs) + ys*(h014 + h114*xs + h024*ys) + z*(h005 &
    + h105*xs + h015*ys + z*(h006 + h007*z))))))

gsw_dynamic_enthalpy = dynamic_enthalpy_part*db2pa*1e4_r8

return
end function

!--------------------------------------------------------------------------
!==========================================================================
elemental function gsw_enthalpy (sa, ct, p)
!==========================================================================
!
!  Calculates specific enthalpy of seawater using the computationally-
!  efficient expression for specific volume in terms of SA, CT and p
!  (Roquet et al., 2014).
!
!  SA  =  Absolute Salinity                                        [ g/kg ]
!  CT  =  Conservative Temperature (ITS-90)                       [ deg C ]
!  p   =  sea pressure                                             [ dbar ]
!         ( i.e. absolute pressure - 10.1325 dbar ) 
!
!  enthalpy  =  specific enthalpy                                  [ J/kg ]
!--------------------------------------------------------------------------

use gsw_mod_toolbox, only : gsw_dynamic_enthalpy

use gsw_mod_teos10_constants, only : gsw_cp0

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sa, ct, p

real (r8) :: gsw_enthalpy

gsw_enthalpy = gsw_cp0*ct + gsw_dynamic_enthalpy(sa,ct,p)

return
end function

!--------------------------------------------------------------------------
!==========================================================================
elemental function gsw_enthalpy_ct_exact (sa, ct, p)
!==========================================================================
!
!  Calculates specific enthalpy of seawater from Absolute Salinity and 
!  Conservative Temperature and pressure.  
!
!  Note that this function uses the full Gibbs function.
!    
!  SA  =  Absolute Salinity                                        [ g/kg ]
!  CT  =  Conservative Temperature (ITS-90)                       [ deg C ]
!  p   =  sea pressure                                             [ dbar ]
!         ( i.e. absolute pressure - 10.1325 dbar )
!
!  enthalpy_CT_exact  =  specific enthalpy                         [ J/kg ]
!--------------------------------------------------------------------------

use gsw_mod_toolbox, only : gsw_enthalpy_t_exact, gsw_t_from_ct

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sa, ct, p

real (r8) :: gsw_enthalpy_ct_exact

real (r8) :: t

t = gsw_t_from_ct(sa,ct,p) 
gsw_enthalpy_ct_exact = gsw_enthalpy_t_exact(sa,t,p)

return
end function

!--------------------------------------------------------------------------
!==========================================================================
elemental function gsw_enthalpy_diff (sa, ct, p_shallow, p_deep)
!==========================================================================
!
!  Calculates the difference of the specific enthalpy of seawater between 
!  two different pressures, p_deep (the deeper pressure) and p_shallow
!  (the shallower pressure), at the same values of SA and CT.  This 
!  function uses the computationally-efficient expression for specific 
!  volume in terms of SA, CT and p (Roquet et al., 2014).  The output
!  (enthalpy_diff_CT) is the specific enthalpy evaluated at (SA,CT,p_deep)
!  minus the specific enthalpy at (SA,CT,p_shallow). 
!
!  SA         =  Absolute Salinity                                 [ g/kg ]
!  CT         =  Conservative Temperature (ITS-90)                [ deg C ]
!  p_shallow  =  upper sea pressure                                [ dbar ]
!                ( i.e. shallower absolute pressure - 10.1325 dbar ) 
!  p_deep     =  lower sea pressure                                [ dbar ]
!                ( i.e. deeper absolute pressure - 10.1325 dbar )
!
!  enthalpy_diff_CT  =  difference of specific enthalpy            [ J/kg ]
!                       (deep minus shallow)
!--------------------------------------------------------------------------

use gsw_mod_teos10_constants, only : db2pa, gsw_sfac, offset

use gsw_mod_specvol_coefficients

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sa, ct, p_shallow, p_deep

real (r8) :: gsw_enthalpy_diff

real (r8) :: dynamic_enthalpy_shallow, dynamic_enthalpy_deep
real (r8) :: part_1, part_2, part_3, part_4, part_5, xs, ys, z_deep, z_shallow

xs = sqrt(gsw_sfac*sa + offset)
ys = ct*0.025_r8
z_shallow = p_shallow*1e-4_r8
z_deep = p_deep*1e-4_r8

part_1 = h001 + xs*(h101 + xs*(h201 + xs*(h301 + xs*(h401 &
    + xs*(h501 + h601*xs))))) + ys*(h011 + xs*(h111 + xs*(h211 + xs*(h311 &
    + xs*(h411 + h511*xs)))) + ys*(h021 + xs*(h121 + xs*(h221 + xs*(h321 &
    + h421*xs))) + ys*(h031 + xs*(h131 + xs*(h231 + h331*xs)) + ys*(h041 &
    + xs*(h141 + h241*xs) + ys*(h051 + h151*xs + h061*ys)))))

part_2 = h002 + xs*(h102 + xs*(h202 + xs*(h302 + xs*(h402 + h502*xs)))) &
    + ys*(h012 + xs*(h112 + xs*(h212 + xs*(h312 + h412*xs))) + ys*(h022 &
    + xs*(h122 + xs*(h222 + h322*xs)) + ys*(h032 + xs*(h132 + h232*xs) &
    + ys*(h042 + h142*xs + h052*ys))))

part_3 = h003 + xs*(h103 + xs*(h203 + xs*(h303 + h403*xs))) + ys*(h013 &
    + xs*(h113 + xs*(h213 + h313*xs)) + ys*(h023 + xs*(h123 + h223*xs) &
    + ys*(h033 + h133*xs + h043*ys)))

part_4 = h004 + xs*(h104 + h204*xs) + ys*(h014 + h114*xs + h024*ys)

part_5 = h005 + h105*xs + h015*ys

dynamic_enthalpy_shallow =  z_shallow*(part_1 + z_shallow*(part_2 &
    + z_shallow*(part_3 + z_shallow*(part_4 + z_shallow*(part_5 &
    + z_shallow*(h006 + h007*z_shallow))))))

dynamic_enthalpy_deep = z_deep*(part_1 + z_deep*(part_2 + z_deep*(part_3 &
    + z_deep*(part_4 + z_deep*(part_5 + z_deep*(h006 + h007*z_deep))))))

gsw_enthalpy_diff = (dynamic_enthalpy_deep &
                     - dynamic_enthalpy_shallow)*db2pa*1e4_r8

return
end function

!--------------------------------------------------------------------------
!==========================================================================
elemental subroutine gsw_enthalpy_first_derivatives (sa, ct, p, h_sa, h_ct)
!==========================================================================
!
!  Calculates the following two derivatives of specific enthalpy (h) of
!  seawater using the computationally-efficient expression for 
!  specific volume in terms of SA, CT and p (Roquet et al., 2014).  
!   (1) h_SA, the derivative with respect to Absolute Salinity at 
!       constant CT and p, and
!   (2) h_CT, derivative with respect to CT at constant SA and p. 
!  Note that h_P is specific volume (1/rho) it can be caclulated by calling
!  gsw_specvol(SA,CT,p).
!
!  SA  =  Absolute Salinity                                        [ g/kg ]
!  CT  =  Conservative Temperature (ITS-90)                       [ deg C ]
!  p   =  sea pressure                                             [ dbar ]
!         ( i.e. absolute pressure - 10.1325 dbar )
!
!  h_SA  =  The first derivative of specific enthalpy with respect to 
!           Absolute Salinity at constant CT and p.     
!                                            [ J/(kg (g/kg))]  i.e. [ J/g ]
!  h_CT  =  The first derivative of specific enthalpy with respect to 
!           CT at constant SA and p.                           [ J/(kg K) ]
!--------------------------------------------------------------------------

use gsw_mod_teos10_constants, only : gsw_cp0, gsw_sfac, offset

use gsw_mod_specvol_coefficients

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sa, ct, p
real (r8), intent(out), optional :: h_sa, h_ct

real (r8) :: dynamic_h_ct_part, dynamic_h_sa_part, xs, ys, z

xs = sqrt(gsw_sfac*sa + offset)
ys = ct*0.025_r8
z = p*1e-4_r8

if (present(h_sa)) then

    dynamic_h_sa_part =  z*(h101 + xs*(2.0_r8*h201 + xs*(3.0_r8*h301 &
        + xs*(4.0_r8*h401 + xs*(5.0_r8*h501 + 6.0_r8*h601*xs)))) + ys*(h111 &
        + xs*(2.0_r8*h211 + xs*(3.0_r8*h311 + xs*(4.0_r8*h411 &
        + 5.0_r8*h511*xs))) + ys*(h121 + xs*(2.0_r8*h221 + xs*(3.0_r8*h321 &
        + 4.0_r8*h421*xs)) + ys*(h131 + xs*(2.0_r8*h231 + 3.0_r8*h331*xs) &
        + ys*(h141 + 2.0_r8*h241*xs + h151*ys)))) + z*(h102 &
        + xs*(2.0_r8*h202 + xs*(3.0_r8*h302 + xs*(4.0_r8*h402 &
        + 5.0_r8*h502*xs))) + ys*(h112 + xs*(2.0_r8*h212 + xs*(3.0_r8*h312 &
        + 4.0_r8*h412*xs)) + ys*(h122 + xs*(2.0_r8*h222 + 3.0_r8*h322*xs) &
        + ys*(h132 + 2.0_r8*h232*xs + h142*ys ))) + z*(h103 + xs*(2.0_r8*h203 &
        + xs*(3.0_r8*h303 + 4.0_r8*h403*xs)) + ys*(h113 + xs*(2.0_r8*h213 &
        + 3.0_r8*h313*xs) + ys*(h123 + 2.0_r8*h223*xs + h133*ys)) &
        + z*(h104 + 2.0_r8*h204*xs + h114*ys + h105*z))))

    h_sa = 1e8_r8*0.5_r8*gsw_sfac*dynamic_h_sa_part/xs

end if

if (present(h_ct)) then

    dynamic_h_ct_part = z*(h011 + xs*(h111 + xs*(h211 + xs*(h311 + xs*(h411 &
        + h511*xs)))) + ys*(2.0_r8*(h021 + xs*(h121 + xs*(h221 + xs*(h321 &
        + h421*xs)))) + ys*(3.0_r8*(h031 + xs*(h131 + xs*(h231 + h331*xs))) &
        + ys*(4.0_r8*(h041 + xs*(h141 + h241*xs)) + ys*(5.0_r8*(h051 &
        + h151*xs) + 6.0_r8*h061*ys)))) + z*(h012 + xs*(h112 + xs*(h212 &
        + xs*(h312 + h412*xs))) + ys*(2.0_r8*(h022 + xs*(h122 + xs*(h222 &
        + h322*xs))) + ys*(3.0_r8*(h032 + xs*(h132 + h232*xs)) &
        + ys*(4.0_r8*(h042 + h142*xs) + 5.0_r8*h052*ys))) + z*(h013 &
        + xs*(h113 + xs*(h213 + h313*xs)) + ys*(2.0_r8*(h023 + xs*(h123 &
        + h223*xs)) + ys*(3.0_r8*(h033 + h133*xs) + 4.0_r8*h043*ys)) &
        + z*(h014 + h114*xs + 2.0_r8*h024*ys + h015*z ))))

    h_ct = gsw_cp0 + 1e8_r8*0.025_r8*dynamic_h_ct_part

end if

return
end subroutine

!--------------------------------------------------------------------------
!==========================================================================
elemental subroutine gsw_enthalpy_first_derivatives_ct_exact (sa, ct, p, &
                                                              h_sa, h_ct)
!==========================================================================
!
!  Calculates the following two derivatives of specific enthalpy (h)
!   (1) h_SA, the derivative with respect to Absolute Salinity at 
!       constant CT and p, and
!   (2) h_CT, derivative with respect to CT at constant SA and p. 
!  Note that h_P is specific volume (1/rho) it can be calulated by calling
!  gsw_specvol_CT_exact(SA,CT,p). This function uses the full Gibbs function.
!
!  SA  =  Absolute Salinity                                        [ g/kg ]
!  CT  =  Conservative Temperature (ITS-90)                       [ deg C ]
!  p   =  sea pressure                                             [ dbar ]
!         ( i.e. absolute pressure - 10.1325 dbar )
!
!  h_SA  =  The first derivative of specific enthalpy with respect to 
!           Absolute Salinity at constant CT and p.     
!                                            [ J/(kg (g/kg))]  i.e. [ J/g ]
!  h_CT  =  The first derivative of specific enthalpy with respect to 
!           CT at constant SA and p.                           [ J/(kg K) ]
!--------------------------------------------------------------------------

use gsw_mod_toolbox, only : gsw_pt_from_ct, gsw_t_from_ct

use gsw_mod_teos10_constants, only : gsw_cp0, gsw_sfac, gsw_t0, rec_db2pa

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sa, ct, p
real (r8), intent(out), optional :: h_sa, h_ct

real (r8) :: g_sa_mod_pt, g_sa_mod_t, pt0, t, temp_ratio
real (r8) :: x, y, y_pt, z

t = gsw_t_from_ct(sa,ct,p)
pt0 = gsw_pt_from_ct(sa,ct)  

temp_ratio = (gsw_t0 + t)/(gsw_t0 + pt0)

if (present(h_ct)) h_ct = gsw_cp0*temp_ratio

if (.not. present(h_sa)) return

x = sqrt(gsw_sfac*sa)
y = 0.025_r8*t
z = rec_db2pa*p !note.the input pressure (p) is sea pressure in units of dbar.

g_sa_mod_t = 8645.36753595126_r8 + z*(-6620.98308089678_r8 + &
       z*(769.588305957198_r8 + z*(-193.0648640214916_r8 + (31.6816345533648_r8 - 5.24960313181984_r8*z)*z))) + &
       x*(-7296.43987145382_r8 + x*(8103.20462414788_r8 + &
       y*(2175.341332000392_r8 + y*(-274.2290036817964_r8 + &
       y*(197.4670779425016_r8 + y*(-68.5590309679152_r8 + 9.98788038278032_r8*y))) - 90.6734234051316_r8*z) + &
       x*(-5458.34205214835_r8 - 980.14153344888_r8*y + &
       x*(2247.60742726704_r8 - 340.1237483177863_r8*x + 220.542973797483_r8*y) + 180.142097805543_r8*z) + &
       z*(-219.1676534131548_r8 + (-16.32775915649044_r8 - 120.7020447884644_r8*z)*z)) + &
       z*(598.378809221703_r8 + z*(-156.8822727844005_r8 + (204.1334828179377_r8 - 10.23755797323846_r8*z)*z)) + &
       y*(-1480.222530425046_r8 + z*(-525.876123559641_r8 + (249.57717834054571_r8 - 88.449193048287_r8*z)*z) + &
       y*(-129.1994027934126_r8 + z*(1149.174198007428_r8 + z*(-162.5751787551336_r8 + 76.9195462169742_r8*z)) + &
       y*(-30.0682112585625_r8 - 1380.9597954037708_r8*z + y*(2.626801985426835_r8 + 703.695562834065_r8*z))))) + &
       y*(1187.3715515697959_r8 + z*(1458.233059470092_r8 + &
       z*(-687.913805923122_r8 + z*(249.375342232496_r8 + z*(-63.313928772146_r8 + 14.09317606630898_r8*z)))) + &
       y*(1760.062705994408_r8 + y*(-450.535298526802_r8 + &
       y*(182.8520895502518_r8 + y*(-43.3206481750622_r8 + 4.26033941694366_r8*y) + &
       z*(-595.457483974374_r8 + (149.452282277512_r8 - 72.9745838003176_r8*z)*z)) + &
       z*(1388.489628266536_r8 + z*(-409.779283929806_r8 + (227.123395681188_r8 - 22.2565468652826_r8*z)*z))) + &
       z*(-1721.528607567954_r8 + z*(674.819060538734_r8 + &
       z*(-356.629112415276_r8 + (88.4080716616_r8 - 15.84003094423364_r8*z)*z)))))
  
g_sa_mod_t = 0.5_r8*gsw_sfac*g_sa_mod_t   
    
y_pt = 0.025_r8*pt0

g_sa_mod_pt = 8645.36753595126_r8 + &
        x*(-7296.43987145382_r8 + x*(8103.20462414788_r8 + &
        y_pt*(2175.341332000392_r8 + y_pt*(-274.2290036817964_r8 + &
        y_pt*(197.4670779425016_r8 + y_pt*(-68.5590309679152_r8 + 9.98788038278032_r8*y_pt)))) + &
        x*(-5458.34205214835_r8 - 980.14153344888_r8*y_pt + &
        x*(2247.60742726704_r8 - 340.1237483177863_r8*x + 220.542973797483_r8*y_pt))) + &
        y_pt*(-1480.222530425046_r8 + y_pt*(-129.1994027934126_r8 + &
        y_pt*(-30.0682112585625_r8 + y_pt*2.626801985426835_r8)))) + &
        y_pt*(1187.3715515697959_r8 + y_pt*(1760.062705994408_r8 + y_pt*(-450.535298526802_r8 + &
        y_pt*(182.8520895502518_r8 + y_pt*(-43.3206481750622_r8 + 4.26033941694366_r8*y_pt)))))
    
g_sa_mod_pt = 0.5_r8*gsw_sfac*g_sa_mod_pt   

h_sa = g_sa_mod_t - temp_ratio*g_sa_mod_pt

return
end subroutine

!--------------------------------------------------------------------------
!==========================================================================
elemental function gsw_enthalpy_ice (t, p)
!==========================================================================
!
! Calculates the specific enthalpy of ice (h_Ih). 
!
!  t  =  in-situ temperature (ITS-90)                             [ deg C ]
!  p  =  sea pressure                                              [ dbar ]
!        ( i.e. absolute pressure - 10.1325 dbar ) 
!
!  gsw_enthalpy_ice  :  specific enthalpy of ice                   [ J/kg ]
!--------------------------------------------------------------------------

use gsw_mod_teos10_constants, only : gsw_t0, db2pa

use gsw_mod_gibbs_ice_coefficients

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: t, p

real (r8) :: gsw_enthalpy_ice

real (r8) :: tau, dzi, g0
complex (r8) :: r2, sqtau_t1, sqtau_t2, g

tau = (t + gsw_t0)*rec_t3p

dzi = db2pa*p*rec_pt

g0 = g00 + dzi*(g01 + dzi*(g02 + dzi*(g03 + g04*dzi)))

r2 = r20 + dzi*(r21 + r22*dzi)

sqtau_t1 = (tau/t1)**2
sqtau_t2 = (tau/t2)**2

g = r1*t1*(log(1.0_r8 - sqtau_t1) + sqtau_t1) &
    + r2*t2*(log(1.0_r8 - sqtau_t2) + sqtau_t2)

gsw_enthalpy_ice = g0 + t3p*real(g,r8)

return
end function

!--------------------------------------------------------------------------
!==========================================================================
elemental subroutine gsw_enthalpy_second_derivatives (sa, ct, p, h_sa_sa, &
                                                      h_sa_ct, h_ct_ct)
! =========================================================================
!
!  Calculates the following three second-order derivatives of specific
!  enthalpy (h),using the computationally-efficient expression for 
!  specific volume in terms of SA, CT and p (Roquet et al., 2014).
!   (1) h_SA_SA, second-order derivative with respect to Absolute Salinity 
!       at constant CT & p.
!   (2) h_SA_CT, second-order derivative with respect to SA & CT at 
!       constant p. 
!   (3) h_CT_CT, second-order derivative with respect to CT at constant SA 
!       and p. 
!
!  SA  =  Absolute Salinity                                        [ g/kg ]
!  CT  =  Conservative Temperature (ITS-90)                       [ deg C ]
!  p   =  sea pressure                                             [ dbar ]
!         ( i.e. absolute pressure - 10.1325 dbar )
!
!  h_SA_SA  =  The second derivative of specific enthalpy with respect to 
!              Absolute Salinity at constant CT & p.    [ J/(kg (g/kg)^2) ]
!  h_SA_CT  =  The second derivative of specific enthalpy with respect to 
!              SA and CT at constant p.                  [ J/(kg K(g/kg)) ]
!  h_CT_CT  =  The second derivative of specific enthalpy with respect to 
!              CT at constant SA and p.                      [ J/(kg K^2) ]
!--------------------------------------------------------------------------

use gsw_mod_teos10_constants, only : gsw_sfac, offset

use gsw_mod_specvol_coefficients

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sa, ct, p
real (r8), intent(out), optional :: h_sa_sa, h_sa_ct, h_ct_ct

real (r8) :: dynamic_h_ct_ct_part, dynamic_h_sa_ct_part, dynamic_h_sa_sa_part
real (r8) :: xs, xs2, ys, z

xs = sqrt(gsw_sfac*sa + offset)
ys = ct*0.025_r8
z = p*1e-4_r8

if (present(h_sa_sa)) then

    xs2 = xs**2
    dynamic_h_sa_sa_part = z*(-h101 + xs2*(3.0_r8*h301 + xs*(8.0_r8*h401 &
        + xs*(15.0_r8*h501 + 24.0_r8*h601*xs))) + ys*(- h111 &
        + xs2*(3.0_r8*h311 + xs*(8.0_r8*h411 + 15.0_r8*h511*xs)) + ys*(-h121 &
        + xs2*(3.0_r8*h321 + 8.0_r8*h421*xs) + ys*(-h131 + 3.0_r8*h331*xs2 &
        + ys*(-h141 - h151*ys)))) + z*(-h102 + xs2*(3.0_r8*h302 &
        + xs*(8.0_r8*h402 + 15.0_r8*h502*xs)) + ys*(-h112 + xs2*(3.0_r8*h312 &
        + 8.0_r8*h412*xs) + ys*(-h122 + 3.0_r8*h322*xs2 + ys*(-h132 &
        - h142*ys ))) + z*(xs2*(8.0_r8*h403*xs + 3.0_r8*h313*ys) &
        + z*(-h103 + 3.0_r8*h303*xs2 + ys*(-h113 + ys*(-h123 - h133*ys)) &
        + z*(-h104 - h114*ys - h105*z)))))

    h_sa_sa = 1e8_r8*0.25_r8*gsw_sfac*gsw_sfac*dynamic_h_sa_sa_part/xs**3

end if

if (present(h_sa_ct)) then

    dynamic_h_sa_ct_part = z*(h111 + xs*(2.0_r8*h211 + xs*(3.0_r8*h311 &
        + xs*(4.0_r8*h411 + 5.0_r8*h511*xs))) + ys*(2.0_r8*h121 &
        + xs*(4.0_r8*h221 + xs*(6.0_r8*h321 + 8.0_r8*h421*xs)) &
        + ys*(3.0_r8*h131 + xs*(6.0_r8*h231 + 9.0_r8*h331*xs) &
        + ys*(4.0_r8*h141 + 8.0_r8*h241*xs + 5.0_r8*h151*ys ))) + z*(h112 &
        + xs*(2.0_r8*h212 + xs*(3.0_r8*h312 + 4.0_r8*h412*xs)) &
        + ys*(2.0_r8*h122 + xs*(4.0_r8*h222 + 6.0_r8*h322*xs) &
        + ys*(3.0_r8*h132 + 6.0_r8*h232*xs + 4.0_r8*h142*ys)) + z*(h113 &
        + xs*(2.0_r8*h213 + 3.0_r8*h313*xs) + ys*(2.0_r8*h123 &
        + 4.0_r8*h223*xs + 3.0_r8*h133*ys) + h114*z)))

    h_sa_ct = 1e8_r8*0.025_r8*0.5_r8*gsw_sfac*dynamic_h_sa_ct_part/xs

end if

if (present(h_ct_ct)) then

    dynamic_h_ct_ct_part = z*(2.0_r8*h021 + xs*(2.0_r8*h121 + xs*(2.0_r8*h221 &
        + xs*(2.0_r8*h321 + 2.0_r8*h421*xs))) + ys*(6.0_r8*h031 &
        + xs*(6.0_r8*h131 + xs*(6.0_r8*h231 + 6.0_r8*h331*xs)) &
        + ys*(12.0_r8*h041 + xs*(12.0_r8*h141 + 12.0_r8*h241*xs) &
        + ys*(20.0_r8*h051 + 20.0_r8*h151*xs + 30.0_r8*h061*ys))) &
        + z*(2.0_r8*h022 + xs*(2.0_r8*h122 + xs*(2.0_r8*h222 &
        + 2.0_r8*h322*xs)) + ys*(6.0_r8*h032 + xs*(6.0_r8*h132 &
        + 6.0_r8*h232*xs) + ys*(12.0_r8*h042 + 12.0_r8*h142*xs &
        + 20.0_r8*h052*ys)) + z*(2.0_r8*h023 + xs*(2.0_r8*h123 &
        + 2.0_r8*h223*xs) + ys*(6.0_r8*h133*xs + 6.0_r8*h033 &
        + 12.0_r8*h043*ys) + 2.0_r8*h024*z)))

    h_ct_ct = 1e8_r8*6.25e-4_r8*dynamic_h_ct_ct_part

end if

return
end subroutine

!--------------------------------------------------------------------------
!==========================================================================
elemental subroutine gsw_enthalpy_second_derivatives_ct_exact (sa, ct, p, &
                                                 h_sa_sa, h_sa_ct, h_ct_ct)
!==========================================================================
!
!  Calculates three second-order derivatives of specific enthalpy (h).
!  Note that this function uses the full Gibbs function.
!
!  sa  =  Absolute Salinity                                        [ g/kg ]
!  ct  =  Conservative Temperature (ITS-90)                       [ deg C ]
!  p   =  sea pressure                                             [ dbar ]
!         ( i.e. absolute pressure - 10.1325 dbar )
!
!  h_sa_sa  =  The second derivative of specific enthalpy with respect to 
!              Absolute Salinity at constant ct & p.    [ J/(kg (g/kg)^2) ]
!  h_sa_ct  =  The second derivative of specific enthalpy with respect to 
!              sa and ct at constant p.                  [ J/(kg K(g/kg)) ]
!  h_ct_ct  =  The second derivative of specific enthalpy with respect to 
!              ct at constant sa and p.                      [ J/(kg K^2) ]
!--------------------------------------------------------------------------

use gsw_mod_toolbox, only : gsw_gibbs, gsw_pt_from_ct, gsw_pt_from_t

use gsw_mod_teos10_constants, only : gsw_cp0, gsw_t0

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sa, ct, p
real (r8), intent(out), optional :: h_sa_sa, h_sa_ct, h_ct_ct

real (r8) :: factor, gsa_pt0, gsat_pt0, gsat, part_b, pt0, h_ct_ct_val
real (r8) :: rec_abs_pt0, rec_gtt_pt0, rec_gtt, t, temp_ratio
real (r8) :: gsasa, gsasa_pt0

integer, parameter :: n0=0, n1=1, n2=2
real (r8), parameter :: pr0 = 0.0_r8, sa_small = 1e-100_r8

pt0 = gsw_pt_from_ct(sa,ct)
rec_abs_pt0 = 1.0_r8/(gsw_t0 + pt0)
t = gsw_pt_from_t(sa,pt0,pr0,p)
temp_ratio = (gsw_t0 + t)*rec_abs_pt0

rec_gtt_pt0 = 1.0_r8/gsw_gibbs(n0,n2,n0,sa,pt0,pr0)
rec_gtt = 1.0_r8/gsw_gibbs(n0,n2,n0,sa,t,p)

! h_ct_ct is naturally well-behaved as sa approaches zero. 
h_ct_ct_val = gsw_cp0*gsw_cp0* &
    (temp_ratio*rec_gtt_pt0 - rec_gtt)*(rec_abs_pt0*rec_abs_pt0)

if (present(h_ct_ct)) h_ct_ct = h_ct_ct_val

if (.not. present(h_sa_sa) .and. .not. present(h_sa_ct)) return

gsat_pt0 = gsw_gibbs(n1,n1,n0,sa,pt0,pr0)
gsat = gsw_gibbs(n1,n1,n0,sa,t,p)
gsa_pt0 = gsw_gibbs(n1,n0,n0,sa,pt0,pr0)

part_b = (temp_ratio*gsat_pt0*rec_gtt_pt0 - gsat*rec_gtt)*rec_abs_pt0
factor = gsa_pt0/gsw_cp0

if (present(h_sa_sa)) then

    gsasa = gsw_gibbs(n2,n0,n0,sa,t,p)
    gsasa_pt0 = gsw_gibbs(n2,n0,n0,sa,pt0,pr0)

    ! h_sa_sa has a singularity at sa = 0, and blows up as sa approaches zero.  
    h_sa_sa = gsasa - temp_ratio*gsasa_pt0  &
        + temp_ratio*gsat_pt0*gsat_pt0*rec_gtt_pt0  &
        - gsat*gsat*rec_gtt  &
        - 2.0_r8*gsa_pt0*part_b + (factor*factor)*h_ct_ct_val

end if
if (.not. present(h_sa_ct)) return

! h_sa_ct should not blow up as sa approaches zero.  The following lines
! of code ensure that the h_sa_ct output of this function does not blow
! up in this limit.  That is, when sa < 1e-100 g/kg, we force the h_sa_ct 
! output to be the same as if sa = 1e-100 g/kg.  
if (sa .lt. sa_small) then
    rec_gtt_pt0 = 1.0_r8/gsw_gibbs(n0,n2,n0,sa_small,pt0,pr0)
    rec_gtt = 1.0_r8/gsw_gibbs(n0,n2,n0,sa_small,t,p)
    gsat_pt0 = gsw_gibbs(n1,n1,n0,sa_small,pt0,pr0)
    gsat = gsw_gibbs(n1,n1,n0,sa_small,t,p)
    gsa_pt0 = gsw_gibbs(n1,n0,n0,sa_small,pt0,pr0)
    part_b = (temp_ratio*gsat_pt0*rec_gtt_pt0 - gsat*rec_gtt)*rec_abs_pt0
    factor = gsa_pt0/gsw_cp0
end if

h_sa_ct  = gsw_cp0*part_b - factor*h_ct_ct_val

return
end subroutine

!--------------------------------------------------------------------------
!==========================================================================
elemental function gsw_enthalpy_sso_0 (p)
!==========================================================================
!  This function calculates enthalpy at the Standard Ocean Salinity, SSO, 
!  and at a Conservative Temperature of zero degrees C, as a function of
!  pressure, p, in dbar, using a streamlined version of the
!  computationally-efficient expression for specific volume, that is, a 
!  streamlined version of the code "gsw_enthalpy(SA,CT,p)".
!==========================================================================

use gsw_mod_teos10_constants, only : db2pa

use gsw_mod_specvol_coefficients

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: p

real (r8) :: gsw_enthalpy_sso_0

real (r8) :: dynamic_enthalpy_sso_0_p, z

z = p*1e-4_r8

dynamic_enthalpy_sso_0_p = &
      z*( 9.726613854843870e-4_r8 + z*(-2.252956605630465e-5_r8 &
    + z*( 2.376909655387404e-6_r8 + z*(-1.664294869986011e-7_r8 &
    + z*(-5.988108894465758e-9_r8 + z*(h006 + h007*z))))))

gsw_enthalpy_sso_0 = dynamic_enthalpy_sso_0_p*db2pa*1e4_r8

return
end function

!--------------------------------------------------------------------------
!==========================================================================
elemental function gsw_enthalpy_t_exact (sa, t, p) 
!==========================================================================
!
! Calculates the specific enthalpy of seawater
!
! sa     : Absolute Salinity                               [g/kg]
! t      : in-situ temperature                             [deg C]
! p      : sea pressure                                    [dbar]
! 
! gsw_enthalpy_t_exact : specific enthalpy                 [J/kg]
!--------------------------------------------------------------------------

use gsw_mod_toolbox, only : gsw_gibbs

use gsw_mod_teos10_constants, only : gsw_t0

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sa, t, p 

real (r8) :: gsw_enthalpy_t_exact

integer, parameter :: n0=0, n1=1

gsw_enthalpy_t_exact = gsw_gibbs(n0,n0,n0,sa,t,p) - &
                                (t+gsw_t0)*gsw_gibbs(n0,n1,n0,sa,t,p)

return
end function

!--------------------------------------------------------------------------
!==========================================================================
elemental subroutine gsw_entropy_first_derivatives (sa, ct, eta_sa, eta_ct)
! =========================================================================
!
!  Calculates the following two partial derivatives of specific entropy
!  (eta) 
!   (1) eta_SA, the derivative with respect to Absolute Salinity at 
!       constant Conservative Temperature, and
!   (2) eta_CT, the derivative with respect to Conservative Temperature at 
!       constant Absolute Salinity. 
!
!  SA  =  Absolute Salinity                                        [ g/kg ]
!  CT  =  Conservative Temperature (ITS-90)                       [ deg C ]
!
!  eta_SA =  The derivative of specific entropy with respect to 
!            Absolute Salinity (in units of g kg^-1) at constant  
!            Conservative Temperature.  
!            eta_SA has units of:         [ J/(kg K(g/kg))]  or [ J/(g K) ]
!  eta_CT =  The derivative of specific entropy with respect to 
!            Conservative Temperature at constant Absolute Salinity.
!            eta_CT has units of:                            [ J/(kg K^2) ]
!--------------------------------------------------------------------------

use gsw_mod_teos10_constants, only : gsw_cp0, gsw_t0

use gsw_mod_toolbox, only : gsw_gibbs, gsw_pt_from_ct

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sa, ct
real (r8), intent(out), optional :: eta_sa, eta_ct

real (r8) :: pt

integer, parameter :: n0=0, n1=1
real (r8), parameter :: pr0 = 0.0_r8

pt = gsw_pt_from_ct(sa,ct)

if (present(eta_sa)) eta_sa = -(gsw_gibbs(n1,n0,n0,sa,pt,pr0))/(gsw_t0 + pt)

if (present(eta_ct)) eta_ct = gsw_cp0/(gsw_t0 + pt)

return
end subroutine

!--------------------------------------------------------------------------
!==========================================================================
elemental function gsw_entropy_from_pt (sa, pt)
!==========================================================================
!
!  Calculates specific entropy of seawater. 
!
!  SA  =  Absolute Salinity                                        [ g/kg ]
!  pt  =  potential temperature (ITS-90)                          [ deg C ]
!
!  entropy  =  specific entropy                                [ J/(kg*K) ]
!--------------------------------------------------------------------------

use gsw_mod_toolbox, only : gsw_gibbs

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sa, pt

real (r8) :: gsw_entropy_from_pt

integer, parameter :: n0 = 0 
integer, parameter :: n1 = 1

real (r8), parameter :: pr0 = 0.0_r8

gsw_entropy_from_pt = -gsw_gibbs(n0,n1,n0,sa,pt,pr0)

return
end function

!--------------------------------------------------------------------------
!==========================================================================
elemental function gsw_entropy_from_t (sa, t, p) 
!==========================================================================
!
! Calculates the specific entropy of seawater
!
! sa     : Absolute Salinity                               [g/kg]
! t      : in-situ temperature                             [deg C]
! p      : sea pressure                                    [dbar]
! 
! gsw_entropy_from_t : specific entropy                    [J/(kg K)]
!--------------------------------------------------------------------------

use gsw_mod_toolbox, only : gsw_gibbs

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sa, t, p 

real (r8) :: gsw_entropy_from_t

integer, parameter :: n0=0, n1=1

gsw_entropy_from_t = -gsw_gibbs(n0,n1,n0,sa,t,p)

return
end function

!--------------------------------------------------------------------------
!==========================================================================
elemental function gsw_entropy_ice (t, p)
!==========================================================================
!
!  Calculates specific entropy of ice. 
!
!  t  =  in-situ temperature (ITS-90)                             [ deg C ]
!  p  =  sea pressure                                              [ dbar ]
!         ( i.e. absolute pressure - 10.1325 dbar ) 
!
!  ice_entropy  =  specific entropy of ice                 [ J kg^-1 K^-1 ]
!--------------------------------------------------------------------------

use gsw_mod_toolbox, only : gsw_gibbs_ice

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: t, p

real (r8) :: gsw_entropy_ice

gsw_entropy_ice = -gsw_gibbs_ice(1,0,t,p)

return
end function

!--------------------------------------------------------------------------
!==========================================================================
elemental function gsw_entropy_part (sa, t, p)
!==========================================================================
!
! entropy minus the terms that are a function of only SA
!
! sa     : Absolute Salinity                               [g/kg]
! t      : in-situ temperature                             [deg C]
! p      : sea pressure                                    [dbar]
! 
! gsw_entropy_part : entropy part
!--------------------------------------------------------------------------

use gsw_mod_teos10_constants, only : gsw_sfac

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sa, t, p

real (r8) :: gsw_entropy_part

real (r8) :: x2, x, y, z, g03, g08

x2 = gsw_sfac*sa
x = sqrt(x2)
y = t*0.025_r8
z = p*1e-4_r8

g03 = z*(-270.983805184062_r8 + &
    z*(776.153611613101_r8 + z*(-196.51255088122_r8 + (28.9796526294175_r8 - 2.13290083518327_r8*z)*z))) + &
    y*(-24715.571866078_r8 + z*(2910.0729080936_r8 + &
    z*(-1513.116771538718_r8 + z*(546.959324647056_r8 + z*(-111.1208127634436_r8 + 8.68841343834394_r8*z)))) + &
    y*(2210.2236124548363_r8 + z*(-2017.52334943521_r8 + &
    z*(1498.081172457456_r8 + z*(-718.6359919632359_r8 + (146.4037555781616_r8 - 4.9892131862671505_r8*z)*z))) + &
    y*(-592.743745734632_r8 + z*(1591.873781627888_r8 + &
    z*(-1207.261522487504_r8 + (608.785486935364_r8 - 105.4993508931208_r8*z)*z)) + &
    y*(290.12956292128547_r8 + z*(-973.091553087975_r8 + &
    z*(602.603274510125_r8 + z*(-276.361526170076_r8 + 32.40953340386105_r8*z))) + &
    y*(-113.90630790850321_r8 + y*(21.35571525415769_r8 - 67.41756835751434_r8*z) + &
    z*(381.06836198507096_r8 + z*(-133.7383902842754_r8 + 49.023632509086724_r8*z)))))))

g08 = x2*(z*(729.116529735046_r8 + &
    z*(-343.956902961561_r8 + z*(124.687671116248_r8 + z*(-31.656964386073_r8 + 7.04658803315449_r8*z)))) + &
    x*( x*(y*(-137.1145018408982_r8 + y*(148.10030845687618_r8 + y*(-68.5590309679152_r8 + 12.4848504784754_r8*y))) - &
    22.6683558512829_r8*z) + z*(-175.292041186547_r8 + (83.1923927801819_r8 - 29.483064349429_r8*z)*z) + &
    y*(-86.1329351956084_r8 + z*(766.116132004952_r8 + z*(-108.3834525034224_r8 + 51.2796974779828_r8*z)) + &
    y*(-30.0682112585625_r8 - 1380.9597954037708_r8*z + y*(3.50240264723578_r8 + 938.26075044542_r8*z)))) + &
    y*(1760.062705994408_r8 + y*(-675.802947790203_r8 + &
    y*(365.7041791005036_r8 + y*(-108.30162043765552_r8 + 12.78101825083098_r8*y) + &
    z*(-1190.914967948748_r8 + (298.904564555024_r8 - 145.9491676006352_r8*z)*z)) + &
    z*(2082.7344423998043_r8 + z*(-614.668925894709_r8 + (340.685093521782_r8 - 33.3848202979239_r8*z)*z))) + &
    z*(-1721.528607567954_r8 + z*(674.819060538734_r8 + &
    z*(-356.629112415276_r8 + (88.4080716616_r8 - 15.84003094423364_r8*z)*z)))))

gsw_entropy_part = -(g03 + g08)*0.025_r8

return
end function

!--------------------------------------------------------------------------
!==========================================================================
elemental function gsw_entropy_part_zerop (sa, pt0)
!==========================================================================
!
! entropy part evaluated at the sea surface
!
! sa     : Absolute Salinity                               [g/kg]
! pt0    : insitu temperature                              [deg C]
! 
! gsw_entropy_part_zerop : entropy part at the sea surface
!--------------------------------------------------------------------------

use gsw_mod_teos10_constants, only : gsw_sfac

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sa, pt0

real (r8) :: gsw_entropy_part_zerop

real (r8) :: x2, x, y, g03, g08

x2 = gsw_sfac*sa
x = sqrt(x2)
y = pt0*0.025_r8

g03 = y*(-24715.571866078_r8 + y*(2210.2236124548363_r8 + &
    y*(-592.743745734632_r8 + y*(290.12956292128547_r8 + &
    y*(-113.90630790850321_r8 + y*21.35571525415769_r8)))))

g08 = x2*(x*(x*(y*(-137.1145018408982_r8 + y*(148.10030845687618_r8 + &
    y*(-68.5590309679152_r8 + 12.4848504784754_r8*y)))) + &
    y*(-86.1329351956084_r8 + y*(-30.0682112585625_r8 + y*3.50240264723578_r8))) + &
    y*(1760.062705994408_r8 + y*(-675.802947790203_r8 + &
    y*(365.7041791005036_r8 + y*(-108.30162043765552_r8 + 12.78101825083098_r8*y)))))

gsw_entropy_part_zerop = -(g03 + g08)*0.025_r8

return
end function

!--------------------------------------------------------------------------
!==========================================================================
elemental subroutine gsw_entropy_second_derivatives (sa, ct, eta_sa_sa, &
                                                     eta_sa_ct, eta_ct_ct)
! =========================================================================
!
!  Calculates the following three second-order partial derivatives of 
!  specific entropy (eta) 
!   (1) eta_SA_SA, the second derivative with respect to Absolute  
!       Salinity at constant Conservative Temperature, and
!   (2) eta_SA_CT, the derivative with respect to Absolute Salinity and 
!       Conservative Temperature. 
!   (3) eta_CT_CT, the second derivative with respect to Conservative 
!       Temperature at constant Absolute Salinity. 
!
!  SA  =  Absolute Salinity                                        [ g/kg ]
!  CT  =  Conservative Temperature (ITS-90)                       [ deg C ]
!
!  eta_SA_SA =  The second derivative of specific entropy with respect 
!               to Absolute Salinity (in units of g kg^-1) at constant 
!               Conservative Temperature.    
!               eta_SA_SA has units of:                 [ J/(kg K(g/kg)^2)]
!  eta_SA_CT =  The second derivative of specific entropy with respect 
!               to Conservative Temperature at constant Absolute 
!               Salinity. eta_SA_CT has units of:     [ J/(kg (g/kg) K^2) ]
!  eta_CT_CT =  The second derivative of specific entropy with respect 
!               to Conservative Temperature at constant Absolute
!               Salinity.  eta_CT_CT has units of:           [ J/(kg K^3) ]
!--------------------------------------------------------------------------

use gsw_mod_teos10_constants, only : gsw_cp0, gsw_t0

use gsw_mod_toolbox, only : gsw_gibbs, gsw_pt_from_ct

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sa, ct
real (r8), intent(out), optional :: eta_sa_sa, eta_sa_ct, eta_ct_ct

real (r8) :: abs_pt, ct_pt, ct_sa, pt, ct_ct

integer, parameter :: n0=0, n1=1, n2=2
real (r8), parameter :: pr0 = 0.0_r8

pt = gsw_pt_from_ct(sa,ct)
abs_pt = gsw_t0 + pt

ct_pt = -(abs_pt*gsw_gibbs(n0,n2,n0,sa,pt,pr0))/gsw_cp0

ct_ct = -gsw_cp0/(ct_pt*abs_pt*abs_pt)

if (present(eta_sa_ct) .or. present(eta_sa_sa)) then

    ct_sa = (gsw_gibbs(n1,n0,n0,sa,pt,pr0) - &
               (abs_pt*gsw_gibbs(n1,n1,n0,sa,pt,pr0)))/gsw_cp0

    if (present(eta_sa_ct)) eta_sa_ct = -ct_sa*ct_ct

    if (present(eta_sa_sa)) eta_sa_sa = -gsw_gibbs(n2,n0,n0,sa,pt,pr0)/abs_pt +&
                                     ct_sa*ct_sa*ct_ct
end if

if (present(eta_ct_ct)) eta_ct_ct = ct_ct

return
end subroutine

!--------------------------------------------------------------------------
!==========================================================================
elemental function gsw_fdelta (p, long, lat)
!==========================================================================
!
! Calculates fdelta. 
!
! p      : sea pressure                                    [dbar]
! long   : longitude                                       [deg E]     
! lat    : latitude                                        [deg N]
!
! gsw_fdelta : Absolute Salinty Anomaly                    [unitless]
!--------------------------------------------------------------------------

use gsw_mod_toolbox, only : gsw_saar

use gsw_mod_error_functions, only : gsw_error_code, gsw_error_limit

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: p, long, lat

real (r8) :: gsw_fdelta

real (r8) ::  saar

character (*), parameter :: func_name = "gsw_fdelta"

saar = gsw_saar(p,long,lat)

if (saar.gt.gsw_error_limit) then
   gsw_fdelta = gsw_error_code(1,func_name)
else
   gsw_fdelta = ((1.0_r8 + 0.35_r8)*saar)/(1.0_r8 - 0.35_r8*saar)
end if

return
end function

!--------------------------------------------------------------------------
!==========================================================================
elemental subroutine gsw_frazil_properties (sa_bulk, h_bulk, p, &
                                            sa_final, ct_final, w_ih_final)
!==========================================================================
!
!  Calculates the mass fraction of ice (mass of ice divided by mass of ice
!  plus seawater), w_Ih_final, which results from given values of the bulk
!  Absolute Salinity, SA_bulk, bulk enthalpy, h_bulk, occuring at pressure
!  p.  The final values of Absolute Salinity, SA_final, and Conservative
!  Temperature, CT_final, of the interstitial seawater phase are also
!  returned.  This code assumes that there is no dissolved air in the
!  seawater (that is, saturation_fraction is assumed to be zero
!  throughout the code).
!
!  When the mass fraction w_Ih_final is calculated as being a positive
!  value, the seawater-ice mixture is at thermodynamic equlibrium.  
!
!  This code returns w_Ih_final = 0 when the input bulk enthalpy, h_bulk, 
!  is sufficiently large (i.e. sufficiently "warm") so that there is no ice 
!  present in the final state.  In this case the final state consists of 
!  only seawater rather than being an equlibrium mixture of seawater and 
!  ice which occurs when w_Ih_final is positive.  Note that when 
!  w_Ih_final = 0, the final seawater is not at the freezing temperature. 
!
!  SA_bulk =  bulk Absolute Salinity of the seawater and ice mixture
!                                                                  [ g/kg ]
!  h_bulk  =  bulk enthalpy of the seawater and ice mixture        [ J/kg ]
!  p       =  sea pressure                                         [ dbar ]
!             ( i.e. absolute pressure - 10.1325 dbar )
!
!  SA_final    =  Absolute Salinity of the seawater in the final state, 
!                 whether or not any ice is present.               [ g/kg ]
!  CT_final    =  Conservative Temperature of the seawater in the the final
!                 state, whether or not any ice is present.       [ deg C ]
!  w_Ih_final  =  mass fraction of ice in the final seawater-ice mixture.
!                 If this ice mass fraction is positive, the system is at 
!                 thermodynamic equilibrium.  If this ice mass fraction is 
!                 zero there is no ice in the final state which consists 
!                 only of seawater which is warmer than the freezing 
!                 temperature.                                   [unitless]
!--------------------------------------------------------------------------

use gsw_mod_toolbox, only : gsw_cp_ice, gsw_ct_freezing_exact
use gsw_mod_toolbox, only : gsw_ct_freezing_first_derivatives
use gsw_mod_toolbox, only : gsw_ct_from_enthalpy_exact, gsw_enthalpy_ct_exact
use gsw_mod_toolbox, only : gsw_enthalpy_first_derivatives_ct_exact
use gsw_mod_toolbox, only : gsw_enthalpy_ice, gsw_t_freezing_exact
use gsw_mod_toolbox, only : gsw_t_freezing_first_derivatives

use gsw_mod_error_functions, only : gsw_error_code, gsw_error_limit

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sa_bulk, h_bulk, p
real (r8), intent(out) :: sa_final, ct_final, w_ih_final

integer :: number_of_iterations
real (r8) :: cp_ih, ctf_sa, ctf, dfunc_dw_ih, dfunc_dw_ih_mean_poly
real (r8) :: func, func0, hf, h_hat_ct, h_hat_sa
real (r8) :: h_ihf, sa, tf_sa, tf, w_ih_mean, w_ih_old, w_ih

! Throughout this code seawater is taken to contain no dissolved air.
real (r8), parameter :: saturation_fraction = 0.0_r8

real (r8), parameter :: num_f = 5.0e-2_r8
real (r8), parameter :: num_f2 = 6.9e-7_r8
real (r8), parameter :: num_p = 2.21_r8

character (*), parameter :: func_name = "gsw_frazil_properties"

!--------------------------------------------------------------------------
! Finding func0
!--------------------------------------------------------------------------
ctf = gsw_ct_freezing_exact(sa_bulk,p,saturation_fraction)
func0 = h_bulk - gsw_enthalpy_ct_exact(sa_bulk,ctf,p)

!--------------------------------------------------------------------------
! When func0 is zero or positive we can immediately calculate the three
! outputs, as the bulk enthalpy, h_bulk, is too large to allow any ice
! at thermodynamic equilibrium. The result will be (warm) seawater with no
! frazil ice being present. The three outputs can be set and the rest of
! this code does not need to be performed.
!--------------------------------------------------------------------------
if (func0 .ge. 0.0_r8) then
    sa_final = sa_bulk
    ct_final = gsw_ct_from_enthalpy_exact(sa_bulk,h_bulk,p)
    w_ih_final = 0.0_r8
    return
endif

!--------------------------------------------------------------------------
! Begin to find the solution for those data points that have func0 < 0,
! implying that the output will be a positive ice mass fraction w_Ih_final.
!
! Do a quasi-Newton step with a separate polynomial estimate of the
! derivative of func with respect to the ice mass fraction.  This section
! of the code delivers initial values of both w_Ih and SA to the rest
! of the more formal modified Newtons Method approach of McDougall and
! Wotherspoon (2014).
!--------------------------------------------------------------------------
dfunc_dw_ih_mean_poly = 3.347814e+05_r8 &
                        - num_f*func0*(1.0_r8 + num_f2*func0) - num_p*p
w_ih = min(-func0/dfunc_dw_ih_mean_poly, 0.95_r8)
sa = sa_bulk/(1.0_r8 - w_ih)
if (sa .lt. 0.0_r8 .or. sa .gt. 120.0_r8) then
    sa_final = gsw_error_code(1,func_name)
    ct_final = sa_final
    w_ih_final = sa_final
    return
endif
    
!--------------------------------------------------------------------------
! Calculating the estimate of the derivative of func, dfunc_dw_Ih, to be
! fed into the iterative Newton's Method.
!--------------------------------------------------------------------------
ctf = gsw_ct_freezing_exact(sa,p,saturation_fraction)
hf = gsw_enthalpy_ct_exact(sa,ctf,p)
tf = gsw_t_freezing_exact(sa,p,saturation_fraction)
h_ihf = gsw_enthalpy_ice(tf,p)
cp_ih = gsw_cp_ice(tf,p)
call gsw_enthalpy_first_derivatives_ct_exact(sa,ctf,p,h_hat_sa,h_hat_ct)
call gsw_ct_freezing_first_derivatives(sa,p,saturation_fraction,ctf_sa)
call gsw_t_freezing_first_derivatives(sa,p,saturation_fraction,tf_sa)
    
dfunc_dw_ih = hf - h_ihf &
        - sa*(h_hat_sa + h_hat_ct*ctf_sa + w_ih*cp_ih*tf_sa/(1.0_r8 - w_ih))

!--------------------------------------------------------------------------
! Enter the main McDougall-Wotherspoon (2014) modified Newton-Raphson loop
!--------------------------------------------------------------------------
do number_of_iterations = 1, 3

    if (number_of_iterations .gt. 1) then
        ! on the first iteration these values are already known
        ctf = gsw_ct_freezing_exact(sa,p,saturation_fraction)
        hf = gsw_enthalpy_ct_exact(sa,ctf,p)
        tf = gsw_t_freezing_exact(sa,p,saturation_fraction)
        h_ihf = gsw_enthalpy_ice(tf,p)
    end if
    
    func = h_bulk - (1.0_r8 - w_ih)*hf - w_ih*h_ihf
    
    w_ih_old = w_ih
    w_ih = w_ih_old - func/dfunc_dw_ih
    w_ih_mean = 0.5_r8*(w_ih + w_ih_old)

    if (w_ih_mean .gt. 0.9_r8) then
        ! This ensures that the mass fraction of ice never exceeds 0.9
        sa_final = gsw_error_code(1+number_of_iterations,func_name)
        ct_final = sa_final
        w_ih_final = sa_final
        return
    endif

    sa = sa_bulk/(1.0_r8 - w_ih_mean)
    ctf = gsw_ct_freezing_exact(sa,p,saturation_fraction)
    hf = gsw_enthalpy_ct_exact(sa,ctf,p)
    tf = gsw_t_freezing_exact(sa,p,saturation_fraction)
    h_ihf = gsw_enthalpy_ice(tf,p)
    cp_ih = gsw_cp_ice(tf,p)
    call gsw_enthalpy_first_derivatives_ct_exact(sa,ctf,p,h_hat_sa,h_hat_ct)
    call gsw_ct_freezing_first_derivatives(sa,p,saturation_fraction,ctf_sa)
    call gsw_t_freezing_first_derivatives(sa,p,saturation_fraction,tf_sa)

    dfunc_dw_ih = hf - h_ihf - sa*(h_hat_sa + h_hat_ct*ctf_sa &
                                   + w_ih_mean*cp_ih*tf_sa/(1.0_r8 - w_ih_mean))

    w_ih = w_ih_old - func/dfunc_dw_ih

    if (w_ih .gt. 0.9_r8) then
        ! This ensures that the mass fraction of ice never exceeds 0.9
        sa_final = gsw_error_code(4+number_of_iterations,func_name)
        ct_final = sa_final
        w_ih_final = sa_final
        return
    endif

    sa = sa_bulk/(1.0_r8 - w_ih)
end do
    
sa_final = sa
ct_final = gsw_ct_freezing_exact(sa,p,saturation_fraction)
w_ih_final = w_ih
    
if (w_ih_final .lt. 0.0_r8) then
    ! This will only trap cases that are smaller than zero by just
    ! machine precision
    sa_final = sa_bulk
    ct_final = gsw_ct_from_enthalpy_exact(sa_final,h_bulk,p)
    w_ih_final = 0.0_r8
end if
    
return
end subroutine

!--------------------------------------------------------------------------
!==========================================================================
elemental subroutine gsw_frazil_properties_potential (sa_bulk, h_pot_bulk,&
                                         p, sa_final, ct_final, w_ih_final)
!==========================================================================
!
!  Calculates the mass fraction of ice (mass of ice divided by mass of ice
!  plus seawater), w_Ih_final, which results from given values of the bulk
!  Absolute Salinity, SA_bulk, bulk potential enthalpy, h_pot_bulk,
!  occuring at pressure p.  The final equilibrium values of Absolute
!  Salinity, SA_final, and Conservative Temperature, CT_final, of the
!  interstitial seawater phase are also returned.  This code assumes that
!  there is no dissolved air in the seawater (that is, saturation_fraction
!  is assumed to be zero thoughout the code).
!
!  When the mass fraction w_Ih_final is calculated as being a positive
!  value, the seawater-ice mixture is at thermodynamic equlibrium.  
!
!  This code returns w_Ih_final = 0 when the input bulk enthalpy, h_bulk, 
!  is sufficiently large (i.e. sufficiently "warm") so that there is no ice 
!  present in the final state.  In this case the final state consists of 
!  only seawater rather than being an equlibrium mixture of seawater and 
!  ice which occurs when w_Ih_final is positive.  Note that when 
!  w_Ih_final = 0, the final seawater is not at the freezing temperature. 
!
!  Note that this code uses the exact forms of CT_freezing and
!  pot_enthalpy_ice_freezing.
!
!  SA_bulk     =  bulk Absolute Salinity of the seawater and ice mixture
!                                                                  [ g/kg ]
!  h_pot_bulk  =  bulk potential enthalpy of the seawater and ice mixture
!                                                                  [ J/kg ]
!  p           =  sea pressure                                  [ dbar ]
!                  ( i.e. absolute pressure - 10.1325 dbar )
!
!  SA_final    =  Absolute Salinity of the seawater in the final state, 
!                 whether or not any ice is present.               [ g/kg ]
!  CT_final    =  Conservative Temperature of the seawater in the the final
!                 state, whether or not any ice is present.       [ deg C ]
!  w_Ih_final  =  mass fraction of ice in the final seawater-ice mixture.
!                 If this ice mass fraction is positive, the system is at 
!                 thermodynamic equilibrium.  If this ice mass fraction is 
!                 zero there is no ice in the final state which consists 
!                 only of seawater which is warmer than the freezing 
!                 temperature.                                   [unitless]
!--------------------------------------------------------------------------

use gsw_mod_teos10_constants, only : gsw_cp0

use gsw_mod_toolbox, only : gsw_ct_freezing_exact
use gsw_mod_toolbox, only : gsw_ct_freezing_first_derivatives
use gsw_mod_toolbox, only : gsw_pot_enthalpy_ice_freezing
use gsw_mod_toolbox, only : gsw_pot_enthalpy_ice_freezing_first_derivatives

use gsw_mod_error_functions, only : gsw_error_code, gsw_error_limit

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sa_bulk, h_pot_bulk, p
real (r8), intent(out) :: sa_final, ct_final, w_ih_final

integer :: iterations, max_iterations

real (r8) :: ctf_sa, ctf, dfunc_dw_ih, dfunc_dw_ih_mean_poly, dpot_h_ihf_dsa
real (r8) :: func, func0, h_pot_ihf, sa, w_ih_old, w_ih, x, xa, y, z

real (r8), parameter :: f01 = -9.041191886754806e-1_r8
real (r8), parameter :: f02 =  4.169608567309818e-2_r8
real (r8), parameter :: f03 = -9.325971761333677e-3_r8
real (r8), parameter :: f04 =  4.699055851002199e-2_r8
real (r8), parameter :: f05 = -3.086923404061666e-2_r8
real (r8), parameter :: f06 =  1.057761186019000e-2_r8
real (r8), parameter :: f07 = -7.349302346007727e-2_r8
real (r8), parameter :: f08 =  1.444842576424337e-1_r8
real (r8), parameter :: f09 = -1.408425967872030e-1_r8
real (r8), parameter :: f10 =  1.070981398567760e-1_r8
real (r8), parameter :: f11 = -1.768451760854797e-2_r8
real (r8), parameter :: f12 = -4.013688314067293e-1_r8
real (r8), parameter :: f13 =  7.209753205388577e-1_r8
real (r8), parameter :: f14 = -1.807444462285120e-1_r8
real (r8), parameter :: f15 =  1.362305015808993e-1_r8
real (r8), parameter :: f16 = -9.500974920072897e-1_r8
real (r8), parameter :: f17 =  1.192134856624248_r8
real (r8), parameter :: f18 = -9.191161283559850e-2_r8
real (r8), parameter :: f19 = -1.008594411490973_r8
real (r8), parameter :: f20 =  8.020279271484482e-1_r8
real (r8), parameter :: f21 = -3.930534388853466e-1_r8
real (r8), parameter :: f22 = -2.026853316399942e-2_r8
real (r8), parameter :: f23 = -2.722731069001690e-2_r8
real (r8), parameter :: f24 =  5.032098120548072e-2_r8
real (r8), parameter :: f25 = -2.354888890484222e-2_r8
real (r8), parameter :: f26 = -2.454090179215001e-2_r8
real (r8), parameter :: f27 =  4.125987229048937e-2_r8
real (r8), parameter :: f28 = -3.533404753585094e-2_r8
real (r8), parameter :: f29 =  3.766063025852511e-2_r8
real (r8), parameter :: f30 = -3.358409746243470e-2_r8
real (r8), parameter :: f31 = -2.242158862056258e-2_r8
real (r8), parameter :: f32 =  2.102254738058931e-2_r8
real (r8), parameter :: f33 = -3.048635435546108e-2_r8
real (r8), parameter :: f34 = -1.996293091714222e-2_r8
real (r8), parameter :: f35 =  2.577703068234217e-2_r8
real (r8), parameter :: f36 = -1.292053030649309e-2_r8
    
real (r8), parameter :: g01 =  3.332286683867741e5_r8
real (r8), parameter :: g02 =  1.416532517833479e4_r8
real (r8), parameter :: g03 = -1.021129089258645e4_r8
real (r8), parameter :: g04 =  2.356370992641009e4_r8
real (r8), parameter :: g05 = -8.483432350173174e3_r8
real (r8), parameter :: g06 =  2.279927781684362e4_r8
real (r8), parameter :: g07 =  1.506238790315354e4_r8
real (r8), parameter :: g08 =  4.194030718568807e3_r8
real (r8), parameter :: g09 = -3.146939594885272e5_r8
real (r8), parameter :: g10 = -7.549939721380912e4_r8
real (r8), parameter :: g11 =  2.790535212869292e6_r8
real (r8), parameter :: g12 =  1.078851928118102e5_r8
real (r8), parameter :: g13 = -1.062493860205067e7_r8
real (r8), parameter :: g14 =  2.082909703458225e7_r8
real (r8), parameter :: g15 = -2.046810820868635e7_r8
real (r8), parameter :: g16 =  8.039606992745191e6_r8
real (r8), parameter :: g17 = -2.023984705844567e4_r8
real (r8), parameter :: g18 =  2.871769638352535e4_r8
real (r8), parameter :: g19 = -1.444841553038544e4_r8
real (r8), parameter :: g20 =  2.261532522236573e4_r8
real (r8), parameter :: g21 = -2.090579366221046e4_r8
real (r8), parameter :: g22 = -1.128417003723530e4_r8
real (r8), parameter :: g23 =  3.222965226084112e3_r8
real (r8), parameter :: g24 = -1.226388046175992e4_r8
real (r8), parameter :: g25 =  1.506847628109789e4_r8
real (r8), parameter :: g26 = -4.584670946447444e4_r8
real (r8), parameter :: g27 =  1.596119496322347e4_r8
real (r8), parameter :: g28 = -6.338852410446789e4_r8
real (r8), parameter :: g29 =  8.951570926106525e4_r8
    
real (r8), parameter :: saturation_fraction = 0.0_r8

character (*), parameter :: func_name = "gsw_frazil_properties_potential"

!--------------------------------------------------------------------------
! Finding func0.  This is the value of the function, func, that would
! result in the output w_Ih_final being exactly zero.
!--------------------------------------------------------------------------
func0 = h_pot_bulk - &
               gsw_cp0*gsw_ct_freezing_exact(sa_bulk,p,saturation_fraction)

!--------------------------------------------------------------------------
! Setting the three outputs for data points that have func0 non-negative
!--------------------------------------------------------------------------
if (func0 >= 0.0_r8) then
    ! When func0 is zero or positive then the final answer will contain
    ! no frazil ice; that is, it will be pure seawater that is warmer
    ! han the freezing temperature.  If func0 >= 0 we do not need to go
    ! through the modified Newton-Raphson procedure and we can simply
    ! write down the answer, as in the following 4 lines of code.
    sa_final = sa_bulk
    ct_final = h_pot_bulk/gsw_cp0
    w_ih_final = 0.0_r8
    return
end if

!--------------------------------------------------------------------------
! Begin finding the solution for data points that have func0 < 0, so that
! the output will have a positive ice mass fraction w_Ih_final.
!--------------------------------------------------------------------------
   
! Evalaute a polynomial for w_Ih in terms of SA_bulk, func0 and p
x = sa_bulk*1e-2_r8  
y = func0/3e5_r8
z = p*1e-4_r8

w_ih = y*(f01 + x*(f02 + x*(f03 + x*(f04 + x*(f05 + f06*x)))) &
     + y*(f07 + x*(f08 + x*(f09 + x*(f10 + f11*x))) + y*(f12 + x*(f13 &
     + x*(f14 + f15*x)) + y*(f16 + x*(f17 + f18*x) + y*(f19 + f20*x &
     + f21*y)))) + z*(f22 + x*(f23 + x*(f24 + f25*x)) + y*(x*(f26 + f27*x) &
     + y*(f28 + f29*x + f30*y)) + z*(f31 + x*(f32 + f33*x) + y*(f34 &
     + f35*x + f36*y))))

if (w_ih .gt. 0.9_r8) then
    ! The ice mass fraction out of this code is restricted to be less than 0.9.
    sa_final = gsw_error_code(1,func_name)
    ct_final = sa_final
    w_ih_final = sa_final
    return
end if

! The initial guess at the absolute salinity of the interstitial seawater
sa = sa_bulk/(1.0_r8 - w_ih)

!--------------------------------------------------------------------------
! Doing a Newton step with a separate polynomial estimate of the mean
! derivative dfunc_dw_Ih_mean_poly.
!--------------------------------------------------------------------------
ctf = gsw_ct_freezing_exact(sa,p,saturation_fraction)
h_pot_ihf = gsw_pot_enthalpy_ice_freezing(sa,p)
func = h_pot_bulk - (1.0_r8 - w_ih)*gsw_cp0*ctf - w_ih*h_pot_ihf

xa = sa*1e-2_r8

dfunc_dw_ih_mean_poly = g01 + xa*(g02 + xa*(g03 + xa*(g04 + g05*xa))) &
    + w_ih*(xa*(g06 + xa*(g07 + g08*xa)) + w_ih*(xa*(g09 + g10*xa) &
    + w_ih*xa*(g11 + g12*xa + w_ih*(g13 + w_ih*(g14 + w_ih*(g15 &
    + g16*w_ih)))))) + z*(g17 + xa*(g18 + g19*xa) + w_ih*(g20 &
    + w_ih*(g21 + g22*w_ih) + xa*(g23 + g24*xa*w_ih)) &
    + z*(g25 + xa*(g26 + g27*xa) + w_ih*(g28 + g29*w_ih)))

w_ih_old = w_ih
w_ih = w_ih_old - func/dfunc_dw_ih_mean_poly

sa = sa_bulk/(1.0_r8 - w_ih)

!--------------------------------------------------------------------------
! Calculating the estimate of the derivative of func, dfunc_dw_Ih, to be
! fed into Newton's Method.
!--------------------------------------------------------------------------
ctf = gsw_ct_freezing_exact(sa,p,saturation_fraction)

h_pot_ihf = gsw_pot_enthalpy_ice_freezing(sa,p)

call gsw_ct_freezing_first_derivatives(sa,p,saturation_fraction,ctf_sa)
call gsw_pot_enthalpy_ice_freezing_first_derivatives(sa,p,dpot_h_ihf_dsa)

dfunc_dw_ih = gsw_cp0*ctf - h_pot_ihf - &
                 sa*(gsw_cp0*ctf_sa + w_ih*dpot_h_ihf_dsa/(1.0_r8 - w_ih))
    
if (w_ih .ge. 0.0_r8 .and. w_ih .le. 0.20_r8 .and. sa .gt. 15.0_r8 &
    .and. sa .lt. 60.0_r8 .and. p .le. 3000.0_r8) then
    max_iterations = 1
else if (w_ih .ge. 0.0_r8 .and. w_ih .le. 0.85_r8 .and. sa .gt. 0.0_r8 &
    .and. sa .lt. 120.0_r8 .and. p .le. 3500.0_r8) then
    max_iterations = 2
else
    max_iterations = 3
end if
    
do iterations = 1, max_iterations

    if (iterations .gt. 1) then
        ! On the first iteration ctf and h_pot_ihf are both known
        ctf = gsw_ct_freezing_exact(sa,p,saturation_fraction)
        h_pot_ihf = gsw_pot_enthalpy_ice_freezing(sa,p)
    end if

    ! This is the function, func, whose zero we seek ...
    func = h_pot_bulk - (1.0_r8 - w_ih)*gsw_cp0*ctf - w_ih*h_pot_ihf

    w_ih_old = w_ih
    w_ih = w_ih_old - func/dfunc_dw_ih

    if (w_ih .gt. 0.9_r8) then
        sa_final = gsw_error_code(1+iterations,func_name)
        ct_final = sa_final
        w_ih_final = sa_final
        return
    end if

    sa = sa_bulk/(1.0_r8 - w_ih)

end do

if (w_ih .lt. 0.0_r8) then
    sa_final = sa_bulk
    ct_final = h_pot_bulk/gsw_cp0
    w_ih_final = 0.0_r8
else
    sa_final = sa
    ct_final = gsw_ct_freezing_exact(sa,p,saturation_fraction)
    w_ih_final = w_ih
end if
    
return
end subroutine

!--------------------------------------------------------------------------
!==========================================================================
elemental subroutine gsw_frazil_properties_potential_poly (sa_bulk, &
                             h_pot_bulk, p, sa_final, ct_final, w_ih_final)
!==========================================================================
!
!  Calculates the mass fraction of ice (mass of ice divided by mass of ice
!  plus seawater), w_Ih_final, which results from given values of the bulk
!  Absolute Salinity, SA_bulk, bulk potential enthalpy, h_pot_bulk,
!  occuring at pressure p.  The final equilibrium values of Absolute
!  Salinity, SA_final, and Conservative Temperature, CT_final, of the
!  interstitial seawater phase are also returned.  This code assumes that
!  there is no dissolved air in the seawater (that is, saturation_fraction
!  is assumed to be zero thoughout the code).
!
!  When the mass fraction w_Ih_final is calculated as being a positive
!  value, the seawater-ice mixture is at thermodynamic equlibrium.  
!
!  This code returns w_Ih_final = 0 when the input bulk enthalpy, h_bulk, 
!  is sufficiently large (i.e. sufficiently "warm") so that there is no ice 
!  present in the final state.  In this case the final state consists of 
!  only seawater rather than being an equlibrium mixture of seawater and 
!  ice which occurs when w_Ih_final is positive.  Note that when 
!  w_Ih_final = 0, the final seawater is not at the freezing temperature. 
!
!  Note that this code uses the polynomial forms of CT_freezing and
!  pot_enthalpy_ice_freezing. This code is intended to be used in ocean
!  models where the model prognostic variables are SA_bulk and h_pot_bulk.
!
!  SA_bulk     =  bulk Absolute Salinity of the seawater and ice mixture
!                                                                  [ g/kg ]
!  h_pot_bulk  =  bulk potential enthalpy of the seawater and ice mixture
!                                                                  [ J/kg ]
!  p           =  sea pressure                                  [ dbar ]
!                  ( i.e. absolute pressure - 10.1325 dbar )
!
!  SA_final    =  Absolute Salinity of the seawater in the final state, 
!                 whether or not any ice is present.               [ g/kg ]
!  CT_final    =  Conservative Temperature of the seawater in the the final
!                 state, whether or not any ice is present.       [ deg C ]
!  w_Ih_final  =  mass fraction of ice in the final seawater-ice mixture.
!                 If this ice mass fraction is positive, the system is at 
!                 thermodynamic equilibrium.  If this ice mass fraction is 
!                 zero there is no ice in the final state which consists 
!                 only of seawater which is warmer than the freezing 
!                 temperature.                                   [unitless]
!--------------------------------------------------------------------------

use gsw_mod_teos10_constants, only : gsw_cp0

use gsw_mod_toolbox, only : gsw_ct_freezing_poly
use gsw_mod_toolbox, only : gsw_ct_freezing_first_derivatives_poly
use gsw_mod_toolbox, only : gsw_pot_enthalpy_ice_freezing_poly
use gsw_mod_toolbox, only : gsw_pot_enthalpy_ice_freezing_first_derivatives_poly

use gsw_mod_error_functions, only : gsw_error_code, gsw_error_limit

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sa_bulk, h_pot_bulk, p
real (r8), intent(out) :: sa_final, ct_final, w_ih_final

integer :: iterations, max_iterations

real (r8) :: ctf_sa, ctf, dfunc_dw_ih, dfunc_dw_ih_mean_poly, dpot_h_ihf_dsa
real (r8) :: func, func0, h_pot_ihf, sa, w_ih_old, w_ih, x, xa, y, z

real (r8), parameter :: f01 = -9.041191886754806e-1_r8
real (r8), parameter :: f02 =  4.169608567309818e-2_r8
real (r8), parameter :: f03 = -9.325971761333677e-3_r8
real (r8), parameter :: f04 =  4.699055851002199e-2_r8
real (r8), parameter :: f05 = -3.086923404061666e-2_r8
real (r8), parameter :: f06 =  1.057761186019000e-2_r8
real (r8), parameter :: f07 = -7.349302346007727e-2_r8
real (r8), parameter :: f08 =  1.444842576424337e-1_r8
real (r8), parameter :: f09 = -1.408425967872030e-1_r8
real (r8), parameter :: f10 =  1.070981398567760e-1_r8
real (r8), parameter :: f11 = -1.768451760854797e-2_r8
real (r8), parameter :: f12 = -4.013688314067293e-1_r8
real (r8), parameter :: f13 =  7.209753205388577e-1_r8
real (r8), parameter :: f14 = -1.807444462285120e-1_r8
real (r8), parameter :: f15 =  1.362305015808993e-1_r8
real (r8), parameter :: f16 = -9.500974920072897e-1_r8
real (r8), parameter :: f17 =  1.192134856624248_r8
real (r8), parameter :: f18 = -9.191161283559850e-2_r8
real (r8), parameter :: f19 = -1.008594411490973_r8
real (r8), parameter :: f20 =  8.020279271484482e-1_r8
real (r8), parameter :: f21 = -3.930534388853466e-1_r8
real (r8), parameter :: f22 = -2.026853316399942e-2_r8
real (r8), parameter :: f23 = -2.722731069001690e-2_r8
real (r8), parameter :: f24 =  5.032098120548072e-2_r8
real (r8), parameter :: f25 = -2.354888890484222e-2_r8
real (r8), parameter :: f26 = -2.454090179215001e-2_r8
real (r8), parameter :: f27 =  4.125987229048937e-2_r8
real (r8), parameter :: f28 = -3.533404753585094e-2_r8
real (r8), parameter :: f29 =  3.766063025852511e-2_r8
real (r8), parameter :: f30 = -3.358409746243470e-2_r8
real (r8), parameter :: f31 = -2.242158862056258e-2_r8
real (r8), parameter :: f32 =  2.102254738058931e-2_r8
real (r8), parameter :: f33 = -3.048635435546108e-2_r8
real (r8), parameter :: f34 = -1.996293091714222e-2_r8
real (r8), parameter :: f35 =  2.577703068234217e-2_r8
real (r8), parameter :: f36 = -1.292053030649309e-2_r8
    
real (r8), parameter :: g01 =  3.332286683867741e5_r8
real (r8), parameter :: g02 =  1.416532517833479e4_r8
real (r8), parameter :: g03 = -1.021129089258645e4_r8
real (r8), parameter :: g04 =  2.356370992641009e4_r8
real (r8), parameter :: g05 = -8.483432350173174e3_r8
real (r8), parameter :: g06 =  2.279927781684362e4_r8
real (r8), parameter :: g07 =  1.506238790315354e4_r8
real (r8), parameter :: g08 =  4.194030718568807e3_r8
real (r8), parameter :: g09 = -3.146939594885272e5_r8
real (r8), parameter :: g10 = -7.549939721380912e4_r8
real (r8), parameter :: g11 =  2.790535212869292e6_r8
real (r8), parameter :: g12 =  1.078851928118102e5_r8
real (r8), parameter :: g13 = -1.062493860205067e7_r8
real (r8), parameter :: g14 =  2.082909703458225e7_r8
real (r8), parameter :: g15 = -2.046810820868635e7_r8
real (r8), parameter :: g16 =  8.039606992745191e6_r8
real (r8), parameter :: g17 = -2.023984705844567e4_r8
real (r8), parameter :: g18 =  2.871769638352535e4_r8
real (r8), parameter :: g19 = -1.444841553038544e4_r8
real (r8), parameter :: g20 =  2.261532522236573e4_r8
real (r8), parameter :: g21 = -2.090579366221046e4_r8
real (r8), parameter :: g22 = -1.128417003723530e4_r8
real (r8), parameter :: g23 =  3.222965226084112e3_r8
real (r8), parameter :: g24 = -1.226388046175992e4_r8
real (r8), parameter :: g25 =  1.506847628109789e4_r8
real (r8), parameter :: g26 = -4.584670946447444e4_r8
real (r8), parameter :: g27 =  1.596119496322347e4_r8
real (r8), parameter :: g28 = -6.338852410446789e4_r8
real (r8), parameter :: g29 =  8.951570926106525e4_r8
    
real (r8), parameter :: saturation_fraction = 0.0_r8

character (*), parameter :: func_name = "gsw_frazil_properties_potential"

!--------------------------------------------------------------------------
! Finding func0.  This is the value of the function, func, that would
! result in the output w_Ih_final being exactly zero.
!--------------------------------------------------------------------------
func0 = h_pot_bulk - &
               gsw_cp0*gsw_ct_freezing_poly(sa_bulk,p,saturation_fraction)

!--------------------------------------------------------------------------
! Setting the three outputs for data points that have func0 non-negative
!--------------------------------------------------------------------------
if (func0 >= 0.0_r8) then
    ! When func0 is zero or positive then the final answer will contain
    ! no frazil ice; that is, it will be pure seawater that is warmer
    ! han the freezing temperature.  If func0 >= 0 we do not need to go
    ! through the modified Newton-Raphson procedure and we can simply
    ! write down the answer, as in the following 4 lines of code.
    sa_final = sa_bulk
    ct_final = h_pot_bulk/gsw_cp0
    w_ih_final = 0.0_r8
    return
end if

!--------------------------------------------------------------------------
! Begin finding the solution for data points that have func0 < 0, so that
! the output will have a positive ice mass fraction w_Ih_final.
!--------------------------------------------------------------------------
   
! Evalaute a polynomial for w_Ih in terms of SA_bulk, func0 and p
x = sa_bulk*1e-2_r8  
y = func0/3e5_r8
z = p*1e-4_r8

w_ih = y*(f01 + x*(f02 + x*(f03 + x*(f04 + x*(f05 + f06*x)))) &
     + y*(f07 + x*(f08 + x*(f09 + x*(f10 + f11*x))) + y*(f12 + x*(f13 &
     + x*(f14 + f15*x)) + y*(f16 + x*(f17 + f18*x) + y*(f19 + f20*x &
     + f21*y)))) + z*(f22 + x*(f23 + x*(f24 + f25*x)) + y*(x*(f26 + f27*x) &
     + y*(f28 + f29*x + f30*y)) + z*(f31 + x*(f32 + f33*x) + y*(f34 &
     + f35*x + f36*y))))

if (w_ih .gt. 0.9_r8) then
    ! The ice mass fraction out of this code is restricted to be less than 0.9.
    sa_final = gsw_error_code(1,func_name)
    ct_final = sa_final
    w_ih_final = sa_final
    return
end if

! The initial guess at the absolute salinity of the interstitial seawater
sa = sa_bulk/(1.0_r8 - w_ih)

!--------------------------------------------------------------------------
! Doing a Newton step with a separate polynomial estimate of the mean
! derivative dfunc_dw_Ih_mean_poly.
!--------------------------------------------------------------------------
ctf = gsw_ct_freezing_poly(sa,p,saturation_fraction)
h_pot_ihf = gsw_pot_enthalpy_ice_freezing_poly(sa,p)
func = h_pot_bulk - (1.0_r8 - w_ih)*gsw_cp0*ctf - w_ih*h_pot_ihf

xa = sa*1e-2_r8

dfunc_dw_ih_mean_poly = g01 + xa*(g02 + xa*(g03 + xa*(g04 + g05*xa))) &
    + w_ih*(xa*(g06 + xa*(g07 + g08*xa)) + w_ih*(xa*(g09 + g10*xa) &
    + w_ih*xa*(g11 + g12*xa + w_ih*(g13 + w_ih*(g14 + w_ih*(g15 &
    + g16*w_ih)))))) + z*(g17 + xa*(g18 + g19*xa) + w_ih*(g20 &
    + w_ih*(g21 + g22*w_ih) + xa*(g23 + g24*xa*w_ih)) &
    + z*(g25 + xa*(g26 + g27*xa) + w_ih*(g28 + g29*w_ih)))

w_ih_old = w_ih
w_ih = w_ih_old - func/dfunc_dw_ih_mean_poly

sa = sa_bulk/(1.0_r8 - w_ih)

!--------------------------------------------------------------------------
! Calculating the estimate of the derivative of func, dfunc_dw_Ih, to be
! fed into Newton's Method.
!--------------------------------------------------------------------------
ctf = gsw_ct_freezing_poly(sa,p,saturation_fraction)

h_pot_ihf = gsw_pot_enthalpy_ice_freezing_poly(sa,p)

call gsw_ct_freezing_first_derivatives_poly(sa,p,saturation_fraction,ctf_sa)
call gsw_pot_enthalpy_ice_freezing_first_derivatives_poly(sa,p,dpot_h_ihf_dsa)

dfunc_dw_ih = gsw_cp0*ctf - h_pot_ihf - &
                 sa*(gsw_cp0*ctf_sa + w_ih*dpot_h_ihf_dsa/(1.0_r8 - w_ih))
    
if (w_ih .ge. 0.0_r8 .and. w_ih .le. 0.20_r8 .and. sa .gt. 15.0_r8 &
    .and. sa .lt. 60.0_r8 .and. p .le. 3000.0_r8) then
    max_iterations = 1
else if (w_ih .ge. 0.0_r8 .and. w_ih .le. 0.85_r8 .and. sa .gt. 0.0_r8 &
    .and. sa .lt. 120.0_r8 .and. p .le. 3500.0_r8) then
    max_iterations = 2
else
    max_iterations = 3
end if
    
do iterations = 1, max_iterations

    if (iterations .gt. 1) then
        ! On the first iteration ctf and h_pot_ihf are both known
        ctf = gsw_ct_freezing_poly(sa,p,saturation_fraction)
        h_pot_ihf = gsw_pot_enthalpy_ice_freezing_poly(sa,p)
    end if

    ! This is the function, func, whose zero we seek ...
    func = h_pot_bulk - (1.0_r8 - w_ih)*gsw_cp0*ctf - w_ih*h_pot_ihf

    w_ih_old = w_ih
    w_ih = w_ih_old - func/dfunc_dw_ih

    if (w_ih .gt. 0.9_r8) then
        sa_final = gsw_error_code(1+iterations,func_name)
        ct_final = sa_final
        w_ih_final = sa_final
        return
    end if

    sa = sa_bulk/(1.0_r8 - w_ih)

end do

if (w_ih .lt. 0.0_r8) then
    sa_final = sa_bulk
    ct_final = h_pot_bulk/gsw_cp0
    w_ih_final = 0.0_r8
else
    sa_final = sa
    ct_final = gsw_ct_freezing_poly(sa,p,saturation_fraction)
    w_ih_final = w_ih
end if
    
return
end subroutine

!--------------------------------------------------------------------------
!==========================================================================
elemental subroutine gsw_frazil_ratios_adiabatic (sa, p, w_ih, &
                              dsa_dct_frazil, dsa_dp_frazil, dct_dp_frazil)
!==========================================================================
!
!  Calculates the ratios of SA, CT and P changes when frazil ice forms or
!  melts in response to an adiabatic change in pressure of a mixture of 
!  seawater and frazil ice crystals.  
!
!  Note that the first output, dSA_dCT_frazil, is dSA/dCT rather than 
!  dCT/dSA.  This is done so that when SA = 0, the output, dSA/dCT, is zero 
!  whereas dCT/dSA would then be infinite. 
!
!  Also note that both dSA_dP_frazil and dCT_dP_frazil are the pressure
!  derivatives with the pressure measured in Pa not dbar.  
!
!  SA  =  Absolute Salinity of seawater                            [ g/kg ]
!  p   =  sea pressure of seawater at which melting occurs         [ dbar ]
!         ( i.e. absolute pressure - 10.1325d0 dbar ) 
!  w_Ih  =  mass fraction of ice, that is the mass of ice divided by the 
!           sum of the masses of ice and seawater.  That is, the mass of
!           ice divided by the mass of the final mixed fluid.  
!           w_Ih must be between 0 and 1.                      [ unitless ]
!
!  dSA_dCT_frazil =  the ratio of the changes in Absolute Salinity 
!                    to that of Conservative Temperature       [ g/(kg K) ] 
!  dSA_dP_frazil  =  the ratio of the changes in Absolute Salinity 
!                    to that of pressure (in Pa)              [ g/(kg Pa) ] 
!  dCT_dP_frazil  =  the ratio of the changes in Conservative Temperature
!                    to that of pressure (in Pa)                   [ K/Pa ]
!--------------------------------------------------------------------------

use gsw_mod_toolbox, only : gsw_ct_freezing_exact, gsw_enthalpy_ct_exact
use gsw_mod_toolbox, only : gsw_t_freezing_exact, gsw_enthalpy_ice
use gsw_mod_toolbox, only : gsw_adiabatic_lapse_rate_ice, gsw_cp_ice
use gsw_mod_toolbox, only : gsw_enthalpy_first_derivatives_ct_exact
use gsw_mod_toolbox, only : gsw_ct_freezing_first_derivatives
use gsw_mod_toolbox, only : gsw_t_freezing_first_derivatives

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sa, p, w_ih
real (r8), intent(out) :: dsa_dct_frazil, dsa_dp_frazil, dct_dp_frazil

real (r8) :: bracket1, bracket2, cp_ih, gamma_ih, h, h_ih, part, rec_bracket3
real (r8) :: tf, wcp, h_hat_sa, h_hat_ct, tf_sa, tf_p, ctf, ctf_sa, ctf_p

real (r8), parameter :: saturation_fraction = 0.0_r8

ctf = gsw_ct_freezing_exact(sa,p,saturation_fraction)
tf = gsw_t_freezing_exact(sa,p,saturation_fraction)
h = gsw_enthalpy_ct_exact(sa,ctf,p)
h_ih = gsw_enthalpy_ice(tf,p)
cp_ih = gsw_cp_ice(tf,p)
gamma_ih = gsw_adiabatic_lapse_rate_ice(tf,p)
call gsw_enthalpy_first_derivatives_ct_exact(sa,ctf,p,h_hat_sa,h_hat_ct)
call gsw_t_freezing_first_derivatives(sa,p,saturation_fraction,tf_sa,tf_p)
call gsw_ct_freezing_first_derivatives(sa,p,saturation_fraction,ctf_sa,ctf_p)

wcp = cp_ih*w_ih/(1.0_r8 - w_ih)
part = (tf_p - gamma_ih)/ctf_p

bracket1 = h_hat_ct + wcp*part
bracket2 = h - h_ih - sa*(h_hat_sa + wcp*(tf_sa - part*ctf_sa))
rec_bracket3 = 1.0_r8/(h - h_ih - sa*(h_hat_sa + h_hat_ct*ctf_sa + wcp*tf_sa))

dsa_dct_frazil = sa*(bracket1/bracket2)
dsa_dp_frazil = sa*ctf_p*bracket1*rec_bracket3
dct_dp_frazil = ctf_p*bracket2*rec_bracket3

return
end subroutine

!--------------------------------------------------------------------------
!==========================================================================
elemental subroutine gsw_frazil_ratios_adiabatic_poly (sa, p, w_ih, &
                              dsa_dct_frazil, dsa_dp_frazil, dct_dp_frazil)
!==========================================================================
!
!  Calculates the ratios of SA, CT and P changes when frazil ice forms or
!  melts in response to an adiabatic change in pressure of a mixture of 
!  seawater and frazil ice crystals.  
!
!  Note that the first output, dSA_dCT_frazil, is dSA/dCT rather than 
!  dCT/dSA.  This is done so that when SA = 0, the output, dSA/dCT, is zero 
!  whereas dCT/dSA would then be infinite. 
!
!  Also note that both dSA_dP_frazil and dCT_dP_frazil are the pressure
!  derivatives with the pressure measured in Pa not dbar.  
!
!  SA  =  Absolute Salinity of seawater                            [ g/kg ]
!  p   =  sea pressure of seawater at which melting occurs         [ dbar ]
!         ( i.e. absolute pressure - 10.1325d0 dbar ) 
!  w_Ih  =  mass fraction of ice, that is the mass of ice divided by the 
!           sum of the masses of ice and seawater.  That is, the mass of
!           ice divided by the mass of the final mixed fluid.  
!           w_Ih must be between 0 and 1.                      [ unitless ]
!
!  dSA_dCT_frazil =  the ratio of the changes in Absolute Salinity 
!                    to that of Conservative Temperature       [ g/(kg K) ] 
!  dSA_dP_frazil  =  the ratio of the changes in Absolute Salinity 
!                    to that of pressure (in Pa)              [ g/(kg Pa) ] 
!  dCT_dP_frazil  =  the ratio of the changes in Conservative Temperature
!                    to that of pressure (in Pa)                   [ K/Pa ]
!--------------------------------------------------------------------------

use gsw_mod_toolbox, only : gsw_ct_freezing_poly, gsw_enthalpy
use gsw_mod_toolbox, only : gsw_t_freezing_poly, gsw_enthalpy_ice
use gsw_mod_toolbox, only : gsw_adiabatic_lapse_rate_ice, gsw_cp_ice
use gsw_mod_toolbox, only : gsw_enthalpy_first_derivatives
use gsw_mod_toolbox, only : gsw_ct_freezing_first_derivatives_poly
use gsw_mod_toolbox, only : gsw_t_freezing_first_derivatives_poly

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sa, p, w_ih
real (r8), intent(out) :: dsa_dct_frazil, dsa_dp_frazil, dct_dp_frazil

real (r8) :: bracket1, bracket2, cp_ih, gamma_ih, h, h_ih, part, rec_bracket3
real (r8) :: tf, wcp, h_hat_sa, h_hat_ct, tf_sa, tf_p, ctf, ctf_sa, ctf_p

real (r8), parameter :: saturation_fraction = 0.0_r8

tf = gsw_t_freezing_poly(sa,p,saturation_fraction)
ctf = gsw_ct_freezing_poly(sa,p,saturation_fraction)
h = gsw_enthalpy(sa,ctf,p)
h_ih = gsw_enthalpy_ice(tf,p)
cp_ih = gsw_cp_ice(tf,p)
gamma_ih = gsw_adiabatic_lapse_rate_ice(tf,p)
call gsw_enthalpy_first_derivatives(sa,ctf,p,h_hat_sa,h_hat_ct)
call gsw_t_freezing_first_derivatives_poly(sa,p,saturation_fraction,tf_sa,tf_p)
call gsw_ct_freezing_first_derivatives_poly(sa,p,saturation_fraction,ctf_sa,ctf_p)

wcp = cp_ih*w_ih/(1.0_r8 - w_ih)
part = (tf_p - gamma_ih)/ctf_p

bracket1 = h_hat_ct + wcp*part
bracket2 = h - h_ih - sa*(h_hat_sa + wcp*(tf_sa - part*ctf_sa))
rec_bracket3 = 1.0_r8/(h - h_ih - sa*(h_hat_sa + h_hat_ct*ctf_sa + wcp*tf_sa))

dsa_dct_frazil = sa*(bracket1/bracket2)
dsa_dp_frazil = sa*ctf_p*bracket1*rec_bracket3
dct_dp_frazil = ctf_p*bracket2*rec_bracket3

return
end subroutine

!--------------------------------------------------------------------------
!==========================================================================
pure function gsw_geo_strf_dyn_height (sa, ct, p, p_ref)
!==========================================================================
!
!  Calculates dynamic height anomaly as the integral of specific volume
!  anomaly from the pressure p of the bottle to the reference pressure
!  p_ref.
!
!  Hence, geo_strf_dyn_height is the dynamic height anomaly with respect
!  to a given reference pressure.  This is the geostrophic streamfunction 
!  for the difference between the horizontal velocity at the pressure 
!  concerned, p, and the horizontal velocity at p_ref.  Dynamic height 
!  anomaly is the geostrophic streamfunction in an isobaric surface.  The 
!  reference values used for the specific volume anomaly are 
!  SSO = 35.16504 g/kg and CT = 0 deg C.  This function calculates 
!  specific volume anomaly using the computationally efficient 
!  expression for specific volume of Roquet et al. (2015). 
!
!  This function evaluates the pressure integral of specific volume using 
!  SA and CT interpolated with respect to pressure using the method of 
!  Reiniger and Ross (1968).  It uses a weighted mean of (i) values 
!  obtained from linear interpolation of the two nearest data points, and 
!  (ii) a linear extrapolation of the pairs of data above and below.  This 
!  "curve fitting" method resembles the use of cubic splines.  
!
!  SA    =  Absolute Salinity                                      [ g/kg ]
!  CT    =  Conservative Temperature (ITS-90)                     [ deg C ]
!  p     =  sea pressure                                           [ dbar ]
!           ( i.e. absolute pressure - 10.1325 dbar )
!  p_ref =  reference pressure                                     [ dbar ]
!           ( i.e. reference absolute pressure - 10.1325 dbar )
!
!  geo_strf_dyn_height  =  dynamic height anomaly               [ m^2/s^2 ]
!   Note. If p_ref exceeds the pressure of the deepest bottle on a 
!     vertical profile, the dynamic height anomaly for each bottle 
!     on the whole vertical profile is returned as NaN.
!--------------------------------------------------------------------------

use gsw_mod_toolbox, only : gsw_linear_interp_sa_ct, gsw_rr68_interp_sa_ct
use gsw_mod_toolbox, only : gsw_specvol_anom_standard

use gsw_mod_error_functions, only : gsw_error_code

use gsw_mod_teos10_constants, only : db2pa

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sa(:), ct(:), p(:), p_ref

real (r8) :: gsw_geo_strf_dyn_height(size(sa))

integer, allocatable :: iidata(:)

real (r8), allocatable :: b(:), b_av(:), dp(:), dp_i(:)
real (r8), allocatable :: sa_i(:), ct_i(:), p_i(:)
real (r8), allocatable :: geo_strf_dyn_height0(:)

integer :: p_cnt, top_pad, i, nz, ibottle, ipref, np_max, np, ibpr

real (r8) :: dp_min, dp_max, p_min, p_max

! This max_dp_i is the limit we choose for the evaluation of specific
! volume in the pressure integration. That is, the vertical integration
! of specific volume with respect to pressure is perfomed with the pressure
! increment being no more than max_dp_i (the default value being 1 dbar).
real (r8), parameter :: max_dp_i = 1.0_r8

! This p_max_limit is the maximum pressure we're likely to encounter (at
! least in real-world situations - Mariana Trench is just over 11km depth).
! It is used to set memory limits for array allocation (via np_max_limit).
real (r8), parameter :: p_max_limit = 20000.0_r8

integer, parameter :: np_max_limit = nint(p_max_limit/max_dp_i)

character (*), parameter :: func_name = "gsw_geo_strf_dyn_height"

nz = size(sa)

allocate (dp(nz-1))

dp = p(2:nz) - p(1:nz-1)
dp_min = minval(dp)
dp_max = maxval(dp)

if (.not. (dp_min .gt. 0.0_r8)) then
    ! pressure must be monotonic (do negative test to trap NaNs)
    gsw_geo_strf_dyn_height = gsw_error_code(1,func_name)
    return
end if
p_min = p(1)
p_max = p(nz)

if (.not. (p_ref .le. p_max)) then
    ! the reference pressure p_ref is deeper than all bottles (do negative
    ! test to trap NaNs)
    gsw_geo_strf_dyn_height = gsw_error_code(2,func_name)
    return
end if

! Determine if there is a "bottle" at exactly p_ref
ipref = -1
do ibottle = 1, nz
    if (p(ibottle) .eq. p_ref) then
        ipref = ibottle
        exit
    end if
end do

if ((dp_max .le. max_dp_i) .and. (p(1) .eq. 0.0_r8) .and. (ipref .gt. 0)) then

    ! vertical resolution is good (bottle gap is no larger than max_dp_i) 
    ! & the vertical profile begins at the surface (i.e. at p = 0 dbar) 
    ! & the profile contains a "bottle" at exactly p_ref. 
    
    allocate (b(nz), b_av(nz-1))

    b = gsw_specvol_anom_standard(sa,ct,p)
    
    b_av = 0.5_r8*(b(1:nz-1) + b(2:nz))    
    
    ! "geo_strf_dyn_height0" is the dynamic height anomaly with respect
    ! to p_ref = 0 (the surface).  

    allocate (geo_strf_dyn_height0(nz))

    geo_strf_dyn_height0 = (/ 0.0_r8, b_av*dp*db2pa /)
    do i = 2, nz   ! cumulative sum
        geo_strf_dyn_height0(i) = geo_strf_dyn_height0(i-1) &
                                  - geo_strf_dyn_height0(i)
    end do
    gsw_geo_strf_dyn_height = geo_strf_dyn_height0 - geo_strf_dyn_height0(ipref)

else

    ! Test if there are vertical gaps between adjacent "bottles" which are
    ! greater than max_dp_i, and that there is a "bottle" exactly at the 
    ! reference pressure.

    allocate (iidata(nz))

    ibpr = 0
    if ((dp_max .le. max_dp_i) .and. (ipref .gt. 0)) then

        ! Vertical resolution is already good (no larger than max_dp_i), and
        ! there is a "bottle" at exactly p_ref. 
                
        if (p_min .gt. 0.0_r8) then
            ! resolution is fine and there is a bottle at p_ref, but
            ! there is not a bottle at p = 0. So add an extra bottle.
            allocate (sa_i(nz+1), ct_i(nz+1), p_i(nz+1))
            sa_i = (/ sa(1), sa /)
            ct_i = (/ ct(1), ct /)
            p_i = (/ 0.0_r8, p /)
            ibpr = ipref + 1
            iidata = (/ (i, i=2,nz+1) /)
        else
            ! resolution is fine, there is a bottle at p_ref, and
            ! there is a bottle at p = 0
            allocate (sa_i(nz), ct_i(nz), p_i(nz))
            sa_i = sa
            ct_i = ct
            p_i = p
            ibpr = ipref
            iidata = (/ (i, i=1,nz) /)
        end if
        p_cnt = size(p_i)

    else

        ! interpolation is needed.
        np_max = 2*nint(maxval(p/max_dp_i)+0.5_r8)
        if (np_max.gt.np_max_limit) then
            gsw_geo_strf_dyn_height = gsw_error_code(3,func_name)
            return
        end if
        allocate (p_i(np_max))

        if (p_min .gt. 0.0_r8) then
            ! there is not a bottle at p = 0.
            if (p_ref .lt. p_min) then
                ! p_ref is shallower than the minimum bottle pressure.
                p_i(1) = 0.0_r8
                call p_sequence(p_i(1),p_ref,p_i(2:),np)
                p_cnt = np + 1
                ibpr = p_cnt
                call p_sequence(p_ref,p_min,p_i(p_cnt+1:),np)
                p_cnt = p_cnt + np
                top_pad = p_cnt  
            else
                ! p_ref is deeper than the minimum bottle pressure. 
                p_i(1:2) = (/ 0.0_r8, p_min /)
                top_pad = 2
                p_cnt = 2
            end if
        else
            ! there is a bottle at p = 0.
            p_i(1) = p_min
            top_pad = 1
            p_cnt = 1
        end if

        do ibottle = 1, nz-1

            iidata(ibottle) = p_cnt
            if (p(ibottle) .eq. p_ref) ibpr = p_cnt

            if (p(ibottle) .lt. p_ref .and. p(ibottle+1) .gt. p_ref) then
                ! ... reference pressure is spanned by bottle pairs -
                ! need to include p_ref as an interpolated pressure.
                call p_sequence(p(ibottle),p_ref,p_i(p_cnt+1:),np)
                p_cnt = p_cnt + np
                ibpr = p_cnt
                call p_sequence(p_ref,p(ibottle+1),p_i(p_cnt+1:),np)
                p_cnt = p_cnt + np
            else
                ! ... reference pressure is not spanned by bottle pairs.
                call p_sequence(p(ibottle),p(ibottle+1),p_i(p_cnt+1:),np)
                p_cnt = p_cnt + np
            end if

        end do

        iidata(nz) = p_cnt
        if (p(nz) .eq. p_ref) ibpr = p_cnt

        allocate (sa_i(p_cnt), ct_i(p_cnt))

        if (top_pad .gt. 1) &
            call gsw_linear_interp_sa_ct(sa,ct,p,p_i(1:top_pad-1),sa_i,ct_i)
        call gsw_rr68_interp_sa_ct(sa,ct,p,p_i(top_pad:p_cnt), &
                                   sa_i(top_pad:),ct_i(top_pad:))
    end if

    allocate (b(p_cnt), b_av(p_cnt-1), dp_i(p_cnt-1))
    allocate (geo_strf_dyn_height0(p_cnt))

    b = gsw_specvol_anom_standard(sa_i(:p_cnt),ct_i(:p_cnt),p_i(:p_cnt))
    b_av = 0.5_r8*(b(1:p_cnt-1) + b(2:p_cnt))
    dp_i = p_i(2:p_cnt) - p_i(1:p_cnt-1)

    geo_strf_dyn_height0 = (/ 0.0_r8, b_av*dp_i /)
    do i = 2, p_cnt   ! cumulative sum
        geo_strf_dyn_height0(i) = geo_strf_dyn_height0(i-1) &
                                  - geo_strf_dyn_height0(i)
    end do
    gsw_geo_strf_dyn_height = (geo_strf_dyn_height0(iidata) &
                               - geo_strf_dyn_height0(ibpr))*db2pa

end if
return

contains

    pure subroutine p_sequence (p1, p2, pseq, nps)

    implicit none

    real (r8), intent(in) :: p1, p2
    real (r8), intent(inout) :: pseq(:)
    integer, intent(out), optional :: nps

    real (r8) :: dp, pstep

    integer :: n, i

    dp = p2 - p1
    n = ceiling(dp/max_dp_i)
    pstep = dp/n

    if (present(nps)) nps = n

    ! Generate the sequence ensuring that the value of p2 is exact to
    ! avoid round-off issues, ie. don't do "pseq = (p1+pstep*i, i=1,n)".

    pseq = (/ (p2-pstep*i, i=n-1,0,-1) /)

    return
    end subroutine p_sequence

end function

!--------------------------------------------------------------------------
!==========================================================================
pure subroutine gsw_geo_strf_dyn_height_pc (sa, ct, delta_p, &
                                            geo_strf_dyn_height_pc, p_mid)
!==========================================================================
!
!  Calculates dynamic height anomaly as the integral of specific volume 
!  anomaly from the the sea surface pressure (0 Pa) to the pressure p.
!  This function, gsw_geo_strf_dyn_height_pc, is to used when the 
!  Absolute Salinity and Conservative Temperature are piecewise constant in 
!  the vertical over sucessive pressure intervals of delta_p (such as in
!  a forward "z-coordinate" ocean model).  "geo_strf_dyn_height_pc" is
!  the dynamic height anomaly with respect to the sea surface.  That is, 
!  "geo_strf_dyn_height_pc" is the geostrophic streamfunction for the 
!  difference between the horizontal velocity at the pressure concerned, p,
!  and the horizontal velocity at the sea surface.  Dynamic height anomaly 
!  is the geostrophic streamfunction in an isobaric surface.  The reference
!  values used for the specific volume anomaly are SA = SSO = 35.16504 g/kg
!  and CT = 0 deg C.  The output values of geo_strf_dyn_height_pc are 
!  given at the mid-point pressures, p_mid, of each layer in which SA and 
!  CT are vertically piecewice constant (pc).  This function calculates 
!  enthalpy using the computationally-efficient 75-term expression for 
!  specific volume of Roquet et al., (2015). 
!
!  SA       =  Absolute Salinity                                   [ g/kg ]
!  CT       =  Conservative Temperature (ITS-90)                  [ deg C ]
!  delta_p  =  difference in sea pressure between the deep and     [ dbar ]
!              shallow extents of each layer in which SA and CT
!              are vertically constant. delta_p must be positive.
!              
!  Note. sea pressure is absolute pressure minus 10.1325 dbar.
!
!  geo_strf_dyn_height_pc =  dynamic height anomaly             [ m^2/s^2 ]
!  p_mid                  =  mid-point pressure in each layer      [ dbar ]
!--------------------------------------------------------------------------

use gsw_mod_toolbox, only : gsw_enthalpy_diff, gsw_enthalpy_sso_0

use gsw_mod_error_functions, only : gsw_error_code

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sa(:), ct(:), delta_p(:)
real (r8), intent(out) :: geo_strf_dyn_height_pc(:), p_mid(:)

integer :: i, np
real (r8), allocatable :: delta_h(:), delta_h_half(:), dyn_height_deep(:)
real (r8), allocatable :: p_deep(:), p_shallow(:)

character (*), parameter :: func_name = "gsw_geo_strf_dyn_height_pc"

if (any(delta_p .lt. 0.0_r8)) then
    geo_strf_dyn_height_pc = gsw_error_code(1,func_name)
    p_mid = geo_strf_dyn_height_pc
    return
end if

np = size(delta_p)
allocate (delta_h(np),delta_h_half(np),dyn_height_deep(np))
allocate (p_deep(np),p_shallow(np))

p_deep(1) = delta_p(1)
do i = 2, np
   p_deep(i) = p_deep(i-1) + delta_p(i)
end do
p_shallow = p_deep - delta_p

delta_h = gsw_enthalpy_diff(sa,ct,p_shallow,p_deep)

dyn_height_deep(1) = -delta_h(1)
do i = 2, np
   dyn_height_deep(i) = dyn_height_deep(i-1) - delta_h(i)
end do
!            This is Phi minus Phi_0 of Eqn. (3.32.2) of IOC et al. (2010).

p_mid = 0.5_r8*(p_shallow  + p_deep)
delta_h_half = gsw_enthalpy_diff(sa,ct,p_mid,p_deep)

geo_strf_dyn_height_pc = gsw_enthalpy_sso_0(p_mid) + &
                           dyn_height_deep + delta_h_half

return
end subroutine

!--------------------------------------------------------------------------
!==========================================================================
elemental function gsw_gibbs (ns, nt, np, sa, t, p)
!==========================================================================
!
! seawater specific Gibbs free energy and derivatives up to order 2
!
! ns     : order of s derivative
! nt     : order of t derivative
! np     : order of p derivative
! sa     : Absolute Salinity                               [g/kg]
! t      : temperature                                     [deg C]
! p      : sea pressure                                    [dbar]
! 
! gsw_gibbs  : specific Gibbs energy or its derivative
!--------------------------------------------------------------------------

use gsw_mod_teos10_constants, only : gsw_sfac

use gsw_mod_error_functions, only : gsw_error_code

use gsw_mod_kinds

implicit none

integer, intent(in) :: ns, nt, np
real (r8), intent(in) :: sa, t, p

real (r8) :: gsw_gibbs

real (r8) :: x2, x, y, z, g03, g08

character (*), parameter :: func_name = "gsw_gibbs"

x2 = gsw_sfac*sa
x = sqrt(x2)
y = t*0.025_r8
z = p*1e-4_r8

if (ns.eq.0 .and. nt.eq.0 .and. np.eq.0) then
          
  g03 = 101.342743139674_r8 + z*(100015.695367145_r8 + &
      z*(-2544.5765420363_r8 + z*(284.517778446287_r8 + &
      z*(-33.3146754253611_r8 + (4.20263108803084_r8 - 0.546428511471039_r8*z)*z)))) + &
      y*(5.90578347909402_r8 + z*(-270.983805184062_r8 + &
      z*(776.153611613101_r8 + z*(-196.51255088122_r8 + (28.9796526294175_r8 - 2.13290083518327_r8*z)*z))) + &
      y*(-12357.785933039_r8 + z*(1455.0364540468_r8 + &
      z*(-756.558385769359_r8 + z*(273.479662323528_r8 + z*(-55.5604063817218_r8 + 4.34420671917197_r8*z)))) + &
      y*(736.741204151612_r8 + z*(-672.50778314507_r8 + &
      z*(499.360390819152_r8 + z*(-239.545330654412_r8 + (48.8012518593872_r8 - 1.66307106208905_r8*z)*z))) + &
      y*(-148.185936433658_r8 + z*(397.968445406972_r8 + &
      z*(-301.815380621876_r8 + (152.196371733841_r8 - 26.3748377232802_r8*z)*z)) + &
      y*(58.0259125842571_r8 + z*(-194.618310617595_r8 + &
      z*(120.520654902025_r8 + z*(-55.2723052340152_r8 + 6.48190668077221_r8*z))) + &
      y*(-18.9843846514172_r8 + y*(3.05081646487967_r8 - 9.63108119393062_r8*z) + &
      z*(63.5113936641785_r8 + z*(-22.2897317140459_r8 + 8.17060541818112_r8*z))))))))
          
  g08 = x2*(1416.27648484197_r8 + z*(-3310.49154044839_r8 + &
        z*(384.794152978599_r8 + z*(-96.5324320107458_r8 + (15.8408172766824_r8 - 2.62480156590992_r8*z)*z))) + &
        x*(-2432.14662381794_r8 + x*(2025.80115603697_r8 + &
        y*(543.835333000098_r8 + y*(-68.5572509204491_r8 + &
        y*(49.3667694856254_r8 + y*(-17.1397577419788_r8 + 2.49697009569508_r8*y))) - 22.6683558512829_r8*z) + &
        x*(-1091.66841042967_r8 - 196.028306689776_r8*y + &
        x*(374.60123787784_r8 - 48.5891069025409_r8*x + 36.7571622995805_r8*y) + 36.0284195611086_r8*z) + &
        z*(-54.7919133532887_r8 + (-4.08193978912261_r8 - 30.1755111971161_r8*z)*z)) + &
        z*(199.459603073901_r8 + z*(-52.2940909281335_r8 + (68.0444942726459_r8 - 3.41251932441282_r8*z)*z)) + &
        y*(-493.407510141682_r8 + z*(-175.292041186547_r8 + (83.1923927801819_r8 - 29.483064349429_r8*z)*z) + &
        y*(-43.0664675978042_r8 + z*(383.058066002476_r8 + z*(-54.1917262517112_r8 + 25.6398487389914_r8*z)) + &
        y*(-10.0227370861875_r8 - 460.319931801257_r8*z + y*(0.875600661808945_r8 + 234.565187611355_r8*z))))) + &
        y*(168.072408311545_r8 + z*(729.116529735046_r8 + &
        z*(-343.956902961561_r8 + z*(124.687671116248_r8 + z*(-31.656964386073_r8 + 7.04658803315449_r8*z)))) + &
        y*(880.031352997204_r8 + y*(-225.267649263401_r8 + &
        y*(91.4260447751259_r8 + y*(-21.6603240875311_r8 + 2.13016970847183_r8*y) + &
        z*(-297.728741987187_r8 + (74.726141138756_r8 - 36.4872919001588_r8*z)*z)) + &
        z*(694.244814133268_r8 + z*(-204.889641964903_r8 + (113.561697840594_r8 - 11.1282734326413_r8*z)*z))) + &
        z*(-860.764303783977_r8 + z*(337.409530269367_r8 + &
        z*(-178.314556207638_r8 + (44.2040358308_r8 - 7.92001547211682_r8*z)*z))))))
        
  if (sa.gt.0.0_r8) &
        g08 = g08 + x2*(5812.81456626732_r8 + 851.226734946706_r8*y)*log(x)

  gsw_gibbs = g03 + g08
  
elseif (ns.eq.1 .and. nt.eq.0 .and. np.eq.0) then
        
  g08 = 8645.36753595126_r8 + z*(-6620.98308089678_r8 + &
       z*(769.588305957198_r8 + z*(-193.0648640214916_r8 + (31.6816345533648_r8 - 5.24960313181984_r8*z)*z))) + &
       x*(-7296.43987145382_r8 + x*(8103.20462414788_r8 + &
       y*(2175.341332000392_r8 + y*(-274.2290036817964_r8 + &
       y*(197.4670779425016_r8 + y*(-68.5590309679152_r8 + 9.98788038278032_r8*y))) - 90.6734234051316_r8*z) + &
       x*(-5458.34205214835_r8 - 980.14153344888_r8*y + &
       x*(2247.60742726704_r8 - 340.1237483177863_r8*x + 220.542973797483_r8*y) + 180.142097805543_r8*z) + &
       z*(-219.1676534131548_r8 + (-16.32775915649044_r8 - 120.7020447884644_r8*z)*z)) + &
       z*(598.378809221703_r8 + z*(-156.8822727844005_r8 + (204.1334828179377_r8 - 10.23755797323846_r8*z)*z)) + &
       y*(-1480.222530425046_r8 + z*(-525.876123559641_r8 + (249.57717834054571_r8 - 88.449193048287_r8*z)*z) + &
       y*(-129.1994027934126_r8 + z*(1149.174198007428_r8 + z*(-162.5751787551336_r8 + 76.9195462169742_r8*z)) + &
       y*(-30.0682112585625_r8 - 1380.9597954037708_r8*z + y*(2.626801985426835_r8 + 703.695562834065_r8*z))))) + &
       y*(1187.3715515697959_r8 + z*(1458.233059470092_r8 + &
       z*(-687.913805923122_r8 + z*(249.375342232496_r8 + z*(-63.313928772146_r8 + 14.09317606630898_r8*z)))) + &
       y*(1760.062705994408_r8 + y*(-450.535298526802_r8 + &
       y*(182.8520895502518_r8 + y*(-43.3206481750622_r8 + 4.26033941694366_r8*y) + &
       z*(-595.457483974374_r8 + (149.452282277512_r8 - 72.9745838003176_r8*z)*z)) + &
       z*(1388.489628266536_r8 + z*(-409.779283929806_r8 + (227.123395681188_r8 - 22.2565468652826_r8*z)*z))) + &
       z*(-1721.528607567954_r8 + z*(674.819060538734_r8 + &
       z*(-356.629112415276_r8 + (88.4080716616_r8 - 15.84003094423364_r8*z)*z)))))
  
  if (sa.gt.0_r8) then
    g08 = g08 + (11625.62913253464_r8 + 1702.453469893412_r8*y)*log(x)
  else
    g08 = 0.0_r8
  endif
  
  gsw_gibbs = 0.5*gsw_sfac*g08

elseif (ns.eq.0 .and. nt.eq.1 .and. np.eq.0) then
               
  g03 = 5.90578347909402_r8 + z*(-270.983805184062_r8 + &
       z*(776.153611613101_r8 + z*(-196.51255088122_r8 + (28.9796526294175_r8 - 2.13290083518327_r8*z)*z))) + &
       y*(-24715.571866078_r8 + z*(2910.0729080936_r8 + &
       z*(-1513.116771538718_r8 + z*(546.959324647056_r8 + z*(-111.1208127634436_r8 + 8.68841343834394_r8*z)))) + &
       y*(2210.2236124548363_r8 + z*(-2017.52334943521_r8 + &
       z*(1498.081172457456_r8 + z*(-718.6359919632359_r8 + (146.4037555781616_r8 - 4.9892131862671505_r8*z)*z))) + &
       y*(-592.743745734632_r8 + z*(1591.873781627888_r8 + &
       z*(-1207.261522487504_r8 + (608.785486935364_r8 - 105.4993508931208_r8*z)*z)) + &
       y*(290.12956292128547_r8 + z*(-973.091553087975_r8 + &
       z*(602.603274510125_r8 + z*(-276.361526170076_r8 + 32.40953340386105_r8*z))) + &
       y*(-113.90630790850321_r8 + y*(21.35571525415769_r8 - 67.41756835751434_r8*z) + &
       z*(381.06836198507096_r8 + z*(-133.7383902842754_r8 + 49.023632509086724_r8*z)))))))
              
  g08 = x2*(168.072408311545_r8 + z*(729.116529735046_r8 + &
        z*(-343.956902961561_r8 + z*(124.687671116248_r8 + z*(-31.656964386073_r8 + 7.04658803315449_r8*z)))) + &
        x*(-493.407510141682_r8 + x*(543.835333000098_r8 + x*(-196.028306689776_r8 + 36.7571622995805_r8*x) + &
        y*(-137.1145018408982_r8 + y*(148.10030845687618_r8 + y*(-68.5590309679152_r8 + 12.4848504784754_r8*y))) - &
        22.6683558512829_r8*z) + z*(-175.292041186547_r8 + (83.1923927801819_r8 - 29.483064349429_r8*z)*z) + &
        y*(-86.1329351956084_r8 + z*(766.116132004952_r8 + z*(-108.3834525034224_r8 + 51.2796974779828_r8*z)) + &
        y*(-30.0682112585625_r8 - 1380.9597954037708_r8*z + y*(3.50240264723578_r8 + 938.26075044542_r8*z)))) + &
        y*(1760.062705994408_r8 + y*(-675.802947790203_r8 + &
        y*(365.7041791005036_r8 + y*(-108.30162043765552_r8 + 12.78101825083098_r8*y) + &
        z*(-1190.914967948748_r8 + (298.904564555024_r8 - 145.9491676006352_r8*z)*z)) + &
        z*(2082.7344423998043_r8 + z*(-614.668925894709_r8 + (340.685093521782_r8 - 33.3848202979239_r8*z)*z))) + &
        z*(-1721.528607567954_r8 + z*(674.819060538734_r8 + &
        z*(-356.629112415276_r8 + (88.4080716616_r8 - 15.84003094423364_r8*z)*z)))))
      
  if (sa.gt.0_r8) g08 = g08 + 851.226734946706_r8*x2*log(x)
  
  gsw_gibbs = (g03 + g08)*0.025_r8

elseif (ns.eq.0 .and. nt.eq.0 .and. np.eq.1) then
    
  g03 = 100015.695367145_r8 + z*(-5089.1530840726_r8 + &
        z*(853.5533353388611_r8 + z*(-133.2587017014444_r8 + (21.0131554401542_r8 - 3.278571068826234_r8*z)*z))) + &
        y*(-270.983805184062_r8 + z*(1552.307223226202_r8 + &
        z*(-589.53765264366_r8 + (115.91861051767_r8 - 10.664504175916349_r8*z)*z)) + &
        y*(1455.0364540468_r8 + z*(-1513.116771538718_r8 + &
        z*(820.438986970584_r8 + z*(-222.2416255268872_r8 + 21.72103359585985_r8*z))) + &
        y*(-672.50778314507_r8 + z*(998.720781638304_r8 + &
        z*(-718.6359919632359_r8 + (195.2050074375488_r8 - 8.31535531044525_r8*z)*z)) + &
        y*(397.968445406972_r8 + z*(-603.630761243752_r8 + (456.589115201523_r8 - 105.4993508931208_r8*z)*z) + &
        y*(-194.618310617595_r8 + y*(63.5113936641785_r8 - 9.63108119393062_r8*y + &
        z*(-44.5794634280918_r8 + 24.511816254543362_r8*z)) + &
        z*(241.04130980405_r8 + z*(-165.8169157020456_r8 + &
        25.92762672308884_r8*z)))))))                                                           
  
  g08 = x2*(-3310.49154044839_r8 + z*(769.588305957198_r8 + &
        z*(-289.5972960322374_r8 + (63.3632691067296_r8 - 13.1240078295496_r8*z)*z)) + &
        x*(199.459603073901_r8 + x*(-54.7919133532887_r8 + 36.0284195611086_r8*x - 22.6683558512829_r8*y + &
        (-8.16387957824522_r8 - 90.52653359134831_r8*z)*z) + &
        z*(-104.588181856267_r8 + (204.1334828179377_r8 - 13.65007729765128_r8*z)*z) + &
        y*(-175.292041186547_r8 + (166.3847855603638_r8 - 88.449193048287_r8*z)*z + &
        y*(383.058066002476_r8 + y*(-460.319931801257_r8 + 234.565187611355_r8*y) + &
        z*(-108.3834525034224_r8 + 76.9195462169742_r8*z)))) + &
        y*(729.116529735046_r8 + z*(-687.913805923122_r8 + &
        z*(374.063013348744_r8 + z*(-126.627857544292_r8 + 35.23294016577245_r8*z))) + &
        y*(-860.764303783977_r8 + y*(694.244814133268_r8 + &
        y*(-297.728741987187_r8 + (149.452282277512_r8 - 109.46187570047641_r8*z)*z) + &
        z*(-409.779283929806_r8 + (340.685093521782_r8 - 44.5130937305652_r8*z)*z)) + &
        z*(674.819060538734_r8 + z*(-534.943668622914_r8 + (176.8161433232_r8 - 39.600077360584095_r8*z)*z)))))
     
  gsw_gibbs = (g03 + g08)*1e-8_r8

elseif (ns.eq.0 .and. nt.eq.2 .and. np.eq.0) then

  g03 = -24715.571866078_r8 + z*(2910.0729080936_r8 + z* &
       (-1513.116771538718_r8 + z*(546.959324647056_r8 + z*(-111.1208127634436_r8 + 8.68841343834394_r8*z)))) + &
       y*(4420.4472249096725_r8 + z*(-4035.04669887042_r8 + &
       z*(2996.162344914912_r8 + z*(-1437.2719839264719_r8 + (292.8075111563232_r8 - 9.978426372534301_r8*z)*z))) + &
       y*(-1778.231237203896_r8 + z*(4775.621344883664_r8 + &
       z*(-3621.784567462512_r8 + (1826.356460806092_r8 - 316.49805267936244_r8*z)*z)) + &
       y*(1160.5182516851419_r8 + z*(-3892.3662123519_r8 + &
       z*(2410.4130980405_r8 + z*(-1105.446104680304_r8 + 129.6381336154442_r8*z))) + &
       y*(-569.531539542516_r8 + y*(128.13429152494615_r8 - 404.50541014508605_r8*z) + &
       z*(1905.341809925355_r8 + z*(-668.691951421377_r8 + 245.11816254543362_r8*z))))))

  g08 = x2*(1760.062705994408_r8 + x*(-86.1329351956084_r8 + &
        x*(-137.1145018408982_r8 + y*(296.20061691375236_r8 + y*(-205.67709290374563_r8 + 49.9394019139016_r8*y))) + &
        z*(766.116132004952_r8 + z*(-108.3834525034224_r8 + 51.2796974779828_r8*z)) + &
        y*(-60.136422517125_r8 - 2761.9195908075417_r8*z + y*(10.50720794170734_r8 + 2814.78225133626_r8*z))) + &
        y*(-1351.605895580406_r8 + y*(1097.1125373015109_r8 + y*(-433.20648175062206_r8 + 63.905091254154904_r8*y) + &
        z*(-3572.7449038462437_r8 + (896.713693665072_r8 - 437.84750280190565_r8*z)*z)) + &
        z*(4165.4688847996085_r8 + z*(-1229.337851789418_r8 + (681.370187043564_r8 - 66.7696405958478_r8*z)*z))) + &
        z*(-1721.528607567954_r8 + z*(674.819060538734_r8 + &
        z*(-356.629112415276_r8 + (88.4080716616_r8 - 15.84003094423364_r8*z)*z))))
     
  gsw_gibbs = (g03 + g08)*0.000625_r8  

elseif (ns.eq.1 .and. nt.eq.0 .and. np.eq.1) then

  g08 = -6620.98308089678_r8 + z*(1539.176611914396_r8 + &
        z*(-579.1945920644748_r8 + (126.7265382134592_r8 - 26.2480156590992_r8*z)*z)) + &
        x*(598.378809221703_r8 + x*(-219.1676534131548_r8 + 180.142097805543_r8*x - 90.6734234051316_r8*y + &
        (-32.65551831298088_r8 - 362.10613436539325_r8*z)*z) + &
        z*(-313.764545568801_r8 + (612.4004484538132_r8 - 40.95023189295384_r8*z)*z) + &
        y*(-525.876123559641_r8 + (499.15435668109143_r8 - 265.347579144861_r8*z)*z + &
        y*(1149.174198007428_r8 + y*(-1380.9597954037708_r8 + 703.695562834065_r8*y) + &
        z*(-325.1503575102672_r8 + 230.7586386509226_r8*z)))) + &
        y*(1458.233059470092_r8 + z*(-1375.827611846244_r8 + &
        z*(748.126026697488_r8 + z*(-253.255715088584_r8 + 70.4658803315449_r8*z))) + &
        y*(-1721.528607567954_r8 + y*(1388.489628266536_r8 + &
        y*(-595.457483974374_r8 + (298.904564555024_r8 - 218.92375140095282_r8*z)*z) + &
        z*(-819.558567859612_r8 + (681.370187043564_r8 - 89.0261874611304_r8*z)*z)) + &
        z*(1349.638121077468_r8 + z*(-1069.887337245828_r8 + (353.6322866464_r8 - 79.20015472116819_r8*z)*z))))    
                                                          
  gsw_gibbs = g08*gsw_sfac*0.5e-8_r8
         
elseif (ns.eq.0 .and. nt.eq.1 .and. np.eq.1) then

  g03 = -270.983805184062_r8 + z*(1552.307223226202_r8 + z*(-589.53765264366_r8 + &
          (115.91861051767_r8 - 10.664504175916349_r8*z)*z)) + &
        y*(2910.0729080936_r8 + z*(-3026.233543077436_r8 + &
        z*(1640.877973941168_r8 + z*(-444.4832510537744_r8 + 43.4420671917197_r8*z))) + &
        y*(-2017.52334943521_r8 + z*(2996.162344914912_r8 + &
        z*(-2155.907975889708_r8 + (585.6150223126464_r8 - 24.946065931335752_r8*z)*z)) + &
        y*(1591.873781627888_r8 + z*(-2414.523044975008_r8 + (1826.356460806092_r8 - 421.9974035724832_r8*z)*z) + &
        y*(-973.091553087975_r8 + z*(1205.20654902025_r8 + z*(-829.084578510228_r8 + 129.6381336154442_r8*z)) + &
        y*(381.06836198507096_r8 - 67.41756835751434_r8*y + z*(-267.4767805685508_r8 + 147.07089752726017_r8*z))))))
    
  g08 = x2*(729.116529735046_r8 + z*(-687.913805923122_r8 + &
        z*(374.063013348744_r8 + z*(-126.627857544292_r8 + 35.23294016577245_r8*z))) + &
        x*(-175.292041186547_r8 - 22.6683558512829_r8*x + (166.3847855603638_r8 - 88.449193048287_r8*z)*z + &
        y*(766.116132004952_r8 + y*(-1380.9597954037708_r8 + 938.26075044542_r8*y) + &
        z*(-216.7669050068448_r8 + 153.8390924339484_r8*z))) + &
        y*(-1721.528607567954_r8 + y*(2082.7344423998043_r8 + &
        y*(-1190.914967948748_r8 + (597.809129110048_r8 - 437.84750280190565_r8*z)*z) + &
        z*(-1229.337851789418_r8 + (1022.055280565346_r8 - 133.5392811916956_r8*z)*z)) + &
        z*(1349.638121077468_r8 + z*(-1069.887337245828_r8 + (353.6322866464_r8 - 79.20015472116819_r8*z)*z))))
    
  gsw_gibbs = (g03 + g08)*2.5e-10_r8

elseif (ns.eq.1 .and. nt.eq.1 .and. np.eq.0) then

    g08 = 1187.3715515697959_r8 + z*(1458.233059470092_r8 + &
        z*(-687.913805923122_r8 + z*(249.375342232496_r8 + z*(-63.313928772146_r8 + 14.09317606630898_r8*z)))) + &
        x*(-1480.222530425046_r8 + x*(2175.341332000392_r8 + x*(-980.14153344888_r8 + 220.542973797483_r8*x) + &
        y*(-548.4580073635929_r8 + y*(592.4012338275047_r8 + y*(-274.2361238716608_r8 + 49.9394019139016_r8*y))) - &
        90.6734234051316_r8*z) + z*(-525.876123559641_r8 + (249.57717834054571_r8 - 88.449193048287_r8*z)*z) + &
        y*(-258.3988055868252_r8 + z*(2298.348396014856_r8 + z*(-325.1503575102672_r8 + 153.8390924339484_r8*z)) + &
        y*(-90.2046337756875_r8 - 4142.8793862113125_r8*z + y*(10.50720794170734_r8 + 2814.78225133626_r8*z)))) + &
        y*(3520.125411988816_r8 + y*(-1351.605895580406_r8 + &
        y*(731.4083582010072_r8 + y*(-216.60324087531103_r8 + 25.56203650166196_r8*y) + &
        z*(-2381.829935897496_r8 + (597.809129110048_r8 - 291.8983352012704_r8*z)*z)) + &
        z*(4165.4688847996085_r8 + z*(-1229.337851789418_r8 + (681.370187043564_r8 - 66.7696405958478_r8*z)*z))) + &
        z*(-3443.057215135908_r8 + z*(1349.638121077468_r8 + &
        z*(-713.258224830552_r8 + (176.8161433232_r8 - 31.68006188846728_r8*z)*z))))

    if (sa.gt.0_r8) g08 = g08 + 1702.453469893412_r8*log(x)

    gsw_gibbs = 0.5_r8*gsw_sfac*0.025_r8*g08

elseif (ns.eq.2 .and. nt.eq.0 .and. np.eq.0) then

    g08 = 2.0_r8*(8103.20462414788_r8 + &
        y*(2175.341332000392_r8 + y*(-274.2290036817964_r8 + &
        y*(197.4670779425016_r8 + y*(-68.5590309679152_r8 + 9.98788038278032_r8*y))) - 90.6734234051316_r8*z) + &
        1.5_r8*x*(-5458.34205214835_r8 - 980.14153344888_r8*y + &
        (4.0_r8/3.0_r8)*x*(2247.60742726704_r8 - 340.1237483177863_r8*1.25_r8*x + 220.542973797483_r8*y) + &
        180.142097805543_r8*z) + &
        z*(-219.1676534131548_r8 + (-16.32775915649044_r8 - 120.7020447884644_r8*z)*z))

    if (x.gt.0_r8) then
        g08 = g08 + (-7296.43987145382_r8 + z*(598.378809221703_r8 + &
            z*(-156.8822727844005_r8 + (204.1334828179377_r8 - 10.23755797323846_r8*z)*z)) + &
            y*(-1480.222530425046_r8 + z*(-525.876123559641_r8 + &
            (249.57717834054571_r8 - 88.449193048287_r8*z)*z) + &
            y*(-129.1994027934126_r8 + z*(1149.174198007428_r8 + &
            z*(-162.5751787551336_r8 + 76.9195462169742_r8*z)) + &
            y*(-30.0682112585625_r8 - 1380.9597954037708_r8*z + &
            y*(2.626801985426835_r8 + 703.695562834065_r8*z)))))/x + &
            (11625.62913253464_r8 + 1702.453469893412_r8*y)/x2
    else
        g08 = 0.0_r8
    end if

    gsw_gibbs = 0.25_r8*gsw_sfac*gsw_sfac*g08

elseif (ns.eq.0 .and. nt.eq.0 .and. np.eq.2) then
           
  g03 = -5089.1530840726_r8 + z*(1707.1066706777221_r8 + &
      z*(-399.7761051043332_r8 + (84.0526217606168_r8 - 16.39285534413117_r8*z)*z)) + &
     y*(1552.307223226202_r8 + z*(-1179.07530528732_r8 + (347.75583155301_r8 - 42.658016703665396_r8*z)*z) + &
      y*(-1513.116771538718_r8 + z*(1640.877973941168_r8 + z*(-666.7248765806615_r8 + 86.8841343834394_r8*z)) + &
        y*(998.720781638304_r8 + z*(-1437.2719839264719_r8 + (585.6150223126464_r8 - 33.261421241781_r8*z)*z) + &
         y*(-603.630761243752_r8 + (913.178230403046_r8 - 316.49805267936244_r8*z)*z + &
           y*(241.04130980405_r8 + y*(-44.5794634280918_r8 + 49.023632509086724_r8*z) + &
            z*(-331.6338314040912_r8 + 77.78288016926652_r8*z))))))
            
  g08 = x2*(769.588305957198_r8 + z*(-579.1945920644748_r8 + (190.08980732018878_r8 - 52.4960313181984_r8*z)*z) + &
      x*(-104.588181856267_r8 + x*(-8.16387957824522_r8 - 181.05306718269662_r8*z) + &
       (408.2669656358754_r8 - 40.95023189295384_r8*z)*z + &
       y*(166.3847855603638_r8 - 176.898386096574_r8*z + y*(-108.3834525034224_r8 + 153.8390924339484_r8*z))) + &
      y*(-687.913805923122_r8 + z*(748.126026697488_r8 + z*(-379.883572632876_r8 + 140.9317606630898_r8*z)) + &
       y*(674.819060538734_r8 + z*(-1069.887337245828_r8 + (530.4484299696_r8 - 158.40030944233638_r8*z)*z) + &
         y*(-409.779283929806_r8 + y*(149.452282277512_r8 - 218.92375140095282_r8*z) + &
          (681.370187043564_r8 - 133.5392811916956_r8*z)*z))))
    
  gsw_gibbs = (g03 + g08)*1e-16_r8 

else

  gsw_gibbs = gsw_error_code(1,func_name)

end if

return
end function

!--------------------------------------------------------------------------
! =========================================================================
elemental function gsw_gibbs_ice (nt, np, t, p)
! =========================================================================
!
!  Ice specific Gibbs energy and derivatives up to order 2.
!
!  nt  =  order of t derivative                      [ integers 0, 1 or 2 ]
!  np  =  order of p derivative                      [ integers 0, 1 or 2 ]
!  t   =  in-situ temperature (ITS-90)                            [ deg C ]
!  p   =  sea pressure                                             [ dbar ]
!   
!  gibbs_ice = Specific Gibbs energy of ice or its derivatives.
!            The Gibbs energy (when nt = np = 0) has units of:     [ J/kg ]
!            The temperature derivatives are output in units of: 
!                                                      [ (J/kg) (K)^(-nt) ]
!            The pressure derivatives are output in units of:
!                                                     [ (J/kg) (Pa)^(-np) ]
!            The mixed derivatives are output in units of:
!                                           [ (J/kg) (K)^(-nt) (Pa)^(-np) ]
!  Note. The derivatives are taken with respect to pressure in Pa, not
!    withstanding that the pressure input into this routine is in dbar.
!--------------------------------------------------------------------------

use gsw_mod_teos10_constants, only : gsw_t0, db2pa

use gsw_mod_error_functions, only : gsw_error_code

use gsw_mod_gibbs_ice_coefficients

use gsw_mod_kinds

implicit none

integer, intent(in) :: nt, np
real (r8), intent(in) :: t, p

real (r8) :: gsw_gibbs_ice

real (r8) :: dzi, g0, g0p, g0pp, sqrec_pt
complex (r8) :: r2, r2p, r2pp, g, sqtau_t1, sqtau_t2, tau, tau_t1, tau_t2

real (r8), parameter :: s0 = -3.32733756492168e3_r8

character (*), parameter :: func_name = "gsw_gibbs_ice"

tau = cmplx((t + gsw_t0)*rec_t3p,0.0_r8,r8)

dzi = db2pa*p*rec_pt

if (nt.eq.0 .and. np.eq.0) then
    
    tau_t1 = tau/t1
    sqtau_t1 = tau_t1*tau_t1
    tau_t2 = tau/t2
    sqtau_t2 = tau_t2*tau_t2
    
    g0 = g00 + dzi*(g01 + dzi*(g02 + dzi*(g03 + g04*dzi)))

    r2 = r20 + dzi*(r21 + r22*dzi)
         
    g = r1*(tau*log((1.0_r8 + tau_t1)/(1.0_r8 - tau_t1)) &
        + t1*(log(1.0_r8 - sqtau_t1) - sqtau_t1)) &
        + r2*(tau*log((1.0_r8 + tau_t2)/(1.0_r8 - tau_t2)) &
        + t2*(log(1.0_r8 - sqtau_t2) - sqtau_t2))
   
    gsw_gibbs_ice = real(g0 - t3p*(s0*tau - g),r8)
    
elseif (nt.eq.1 .and. np.eq.0) then
    
    tau_t1 = tau/t1
    tau_t2 = tau/t2
    
    r2 = r20 + dzi*(r21 + r22*dzi)
    
    g = r1*(log((1.0_r8 + tau_t1)/(1.0_r8 - tau_t1)) - 2.0_r8*tau_t1) &
        + r2*(log((1.0_r8 + tau_t2)/(1.0_r8 - tau_t2)) - 2.0_r8*tau_t2)
    
    gsw_gibbs_ice = -s0 + real(g,r8)
        
elseif (nt.eq.0 .and. np.eq.1) then
    
    tau_t2 = tau/t2
    sqtau_t2 = tau_t2*tau_t2
    
    g0p = rec_pt*(g01 + dzi*(2.0_r8*g02 + dzi*(3.0_r8*g03 + 4.0_r8*g04*dzi)))
    
    r2p = rec_pt*(r21 + 2.0_r8*r22*dzi)
        
    g = r2p*(tau*log((1.0_r8 + tau_t2)/(1.0_r8 - tau_t2)) &
        + t2*(log(1.0_r8 - sqtau_t2) - sqtau_t2))
    
    gsw_gibbs_ice = g0p + t3p*real(g,r8)

elseif (nt.eq.1 .and. np.eq.1) then
    
    tau_t2 = tau/t2

    r2p = rec_pt*(r21 + 2.0_r8*r22*dzi) 
    
    g = r2p*(log((1.0_r8 + tau_t2)/(1.0_r8 - tau_t2)) - 2.0_r8*tau_t2)
    
    gsw_gibbs_ice = real(g,r8)
    
elseif (nt.eq.2 .and. np.eq.0) then
    
    r2 = r20 + dzi*(r21 + r22*dzi)
    
    g = r1*(1.0_r8/(t1 - tau) + 1.0_r8/(t1 + tau) - 2.0_r8/t1) &
        + r2*(1.0_r8/(t2 - tau) + 1.0_r8/(t2 + tau) - 2.0_r8/t2)
    
    gsw_gibbs_ice = rec_t3p*real(g,r8)
    
elseif (nt.eq.0 .and. np.eq.2) then
   
    sqrec_pt = rec_pt*rec_pt
    
    tau_t2 = tau/t2
    sqtau_t2 = tau_t2*tau_t2
    
    g0pp = sqrec_pt*(2.0_r8*g02 + dzi*(6.0_r8*g03 + 12.0_r8*g04*dzi))
    
    r2pp = 2.0_r8*r22*sqrec_pt
       
    g = r2pp*(tau*log((1.0_r8 + tau_t2)/(1.0_r8 - tau_t2)) &
        + t2*(log(1.0_r8 - sqtau_t2) - sqtau_t2))

   gsw_gibbs_ice = g0pp + t3p*real(g,r8)
    
else

   gsw_gibbs_ice = gsw_error_code(1,func_name)

end if
    
return
end function

!--------------------------------------------------------------------------
! =========================================================================
elemental function gsw_gibbs_ice_part_t (t, p)
! =========================================================================
!
!  part of the the first temperature derivative of Gibbs energy of ice
!  that is the outout is gibbs_ice(1,0,t,p) + S0
!
!  t   =  in-situ temperature (ITS-90)                            [ deg C ]
!  p   =  sea pressure                                             [ dbar ]
!
!  gibbs_ice_part_t = part of temperature derivative       [ J kg^-1 K^-1 ]
!--------------------------------------------------------------------------

use gsw_mod_teos10_constants, only : gsw_t0, db2pa

use gsw_mod_gibbs_ice_coefficients

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: t, p

real (r8) :: gsw_gibbs_ice_part_t

real (r8) :: dzi, tau
complex (r8) :: g, tau_t1, tau_t2, r2

tau = (t + gsw_t0)*rec_t3p

dzi = db2pa*p*rec_pt

tau_t1 = tau/t1
tau_t2 = tau/t2

r2 = r20 + dzi*(r21 + r22*dzi)

g = r1*(log((1.0_r8 + tau_t1)/(1.0_r8 - tau_t1)) - 2.0_r8*tau_t1) &
    + r2*(log((1.0_r8 + tau_t2)/(1.0_r8 - tau_t2)) - 2.0_r8*tau_t2)

gsw_gibbs_ice_part_t = real(g,r8)

return
end function

!--------------------------------------------------------------------------
! =========================================================================
elemental function gsw_gibbs_ice_pt0 (pt0)
! =========================================================================
!
!  Part of the the first temperature derivative of Gibbs energy of ice
!  that is the outout is "gibbs_ice(1,0,pt0,0) + s0"
!
!  pt0  =  potential temperature with reference sea pressure of zero dbar
!                                                                 [ deg C ]
!
!  gsw_gibbs_ice_pt0 = part of temperature derivative     [ J kg^-1 K^-1 ]
!--------------------------------------------------------------------------

use gsw_mod_teos10_constants, only : gsw_t0

use gsw_mod_gibbs_ice_coefficients

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: pt0

real (r8) :: gsw_gibbs_ice_pt0

real (r8) :: tau
complex (r8) :: g, tau_t1, tau_t2

tau = (pt0 + gsw_t0)*rec_t3p

tau_t1 = tau/t1
tau_t2 = tau/t2

g = r1*(log((1.0_r8 + tau_t1)/(1.0_r8 - tau_t1)) - 2.0_r8*tau_t1) &
    + r20*(log((1.0_r8 + tau_t2)/(1.0_r8 - tau_t2)) - 2.0_r8*tau_t2)

gsw_gibbs_ice_pt0 = real(g,r8)

return
end function

!--------------------------------------------------------------------------
! =========================================================================
elemental function gsw_gibbs_ice_pt0_pt0 (pt0)
! =========================================================================
!
!  The second temperature derivative of Gibbs energy of ice at the 
!  potential temperature with reference sea pressure of zero dbar.  That is
!  the output is gibbs_ice(2,0,pt0,0). 
!
!  pt0  =  potential temperature with reference sea pressure of zero dbar
!                                                                 [ deg C ]
!
!  gsw_gibbs_ice_pt0_pt0 = temperature second derivative at pt0
!--------------------------------------------------------------------------

use gsw_mod_teos10_constants, only : gsw_t0

use gsw_mod_gibbs_ice_coefficients

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: pt0

real (r8) :: gsw_gibbs_ice_pt0_pt0

real (r8) :: tau
complex (r8) :: g

tau = (pt0 + gsw_t0)*rec_t3p

g = r1*(1.0_r8/(t1 - tau) + 1.0_r8/(t1 + tau) - 2.0_r8/t1) &
    + r20*(1.0_r8/(t2 - tau) + 1.0_r8/(t2 + tau) - 2.0_r8/t2)

gsw_gibbs_ice_pt0_pt0 = rec_t3p*real(g,r8)

return
end function

!--------------------------------------------------------------------------
!==========================================================================
elemental function gsw_gibbs_pt0_pt0 (sa, pt0)
!==========================================================================
!
! gibbs_tt at (sa,pt,0)
!
! sa     : Absolute Salinity                            [g/kg]
! pt0    : potential temperature                        [deg C]
! 
! gsw_gibbs_pt0_pt0 : gibbs_tt at (sa,pt,0)         
!--------------------------------------------------------------------------

use gsw_mod_teos10_constants, only : gsw_sfac

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sa, pt0

real (r8) :: gsw_gibbs_pt0_pt0

real (r8) :: x2, x, y, g03, g08

x2 = gsw_sfac*sa
x = sqrt(x2)
y = pt0*0.025_r8

g03 = -24715.571866078_r8 + &
    y*(4420.4472249096725_r8 + &
    y*(-1778.231237203896_r8 + &
    y*(1160.5182516851419_r8 + &
    y*(-569.531539542516_r8 + y*128.13429152494615_r8))))

g08 = x2*(1760.062705994408_r8 + x*(-86.1329351956084_r8 + &
    x*(-137.1145018408982_r8 + y*(296.20061691375236_r8 + &
    y*(-205.67709290374563_r8 + 49.9394019139016_r8*y))) + &
    y*(-60.136422517125_r8 + y*10.50720794170734_r8)) + &
    y*(-1351.605895580406_r8 + y*(1097.1125373015109_r8 + &
    y*(-433.20648175062206_r8 + 63.905091254154904_r8*y))))

gsw_gibbs_pt0_pt0 = (g03 + g08)*0.000625_r8

return
end function

!--------------------------------------------------------------------------
!==========================================================================
elemental function gsw_grav (lat, p)  
!==========================================================================
!
! Calculates acceleration due to gravity as a function of latitude and as
!  a function of pressure in the ocean.
!
! lat  =  latitude in decimal degress north                [ -90 ... +90 ]  
!  p  =  sea pressure                                              [ dbar ]
! 
! gsw_grav : grav  =  gravitational acceleration               [ m s^-2 ]
!--------------------------------------------------------------------------

use gsw_mod_toolbox, only : gsw_z_from_p

use gsw_mod_teos10_constants, only : deg2rad, gamma

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: lat, p  

real (r8) :: gsw_grav

real (r8) :: sin2, gs, z

sin2 = sin(lat*deg2rad)**2
gs = 9.780327_r8*(1.0_r8 + (5.2792e-3_r8 + (2.32e-5_r8*sin2))*sin2) 

z = gsw_z_from_p(p,lat)

gsw_grav = gs*(1.0_r8 - gamma*z)        ! z is the height corresponding to p.
                                        ! Note. In the ocean z is negative.
return
end function

!--------------------------------------------------------------------------
!==========================================================================
elemental function gsw_helmholtz_energy_ice (t, p)
!==========================================================================
!
!  Calculates the Helmholtz energy of ice. 
!
!  t  =  in-situ temperature (ITS-90)                             [ deg C ]
!  p  =  sea pressure                                              [ dbar ]
!         ( i.e. absolute pressure - 10.1325 dbar ) 
!
!  Helmholtz_energy_ice  =  Helmholtz energy of ice                [ J/kg ]
!--------------------------------------------------------------------------

use gsw_mod_toolbox, only : gsw_gibbs_ice

use gsw_mod_teos10_constants, only : gsw_p0, db2pa

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: t, p

real (r8) :: gsw_helmholtz_energy_ice

gsw_helmholtz_energy_ice = gsw_gibbs_ice(0,0,t,p) &
                           - (db2pa*p + gsw_p0)*gsw_gibbs_ice(0,1,t,p)

return
end function

!--------------------------------------------------------------------------
!==========================================================================
elemental function gsw_hill_ratio_at_sp2 (t)
!==========================================================================
!
!  Calculates the Hill ratio, which is the adjustment needed to apply for
!  Practical Salinities smaller than 2.  This ratio is defined at a 
!  Practical Salinity = 2 and in-situ temperature, t using PSS-78. The Hill
!  ratio is the ratio of 2 to the output of the Hill et al. (1986) formula
!  for Practical Salinity at the conductivity ratio, Rt, at which Practical
!  Salinity on the PSS-78 scale is exactly 2.
!--------------------------------------------------------------------------

use gsw_mod_sp_coefficients

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: t

real (r8) :: gsw_hill_ratio_at_sp2

real (r8), parameter :: g0 = 2.641463563366498e-1_r8
real (r8), parameter :: g1 = 2.007883247811176e-4_r8
real (r8), parameter :: g2 = -4.107694432853053e-6_r8
real (r8), parameter :: g3 = 8.401670882091225e-8_r8
real (r8), parameter :: g4 = -1.711392021989210e-9_r8
real (r8), parameter :: g5 = 3.374193893377380e-11_r8
real (r8), parameter :: g6 = -5.923731174730784e-13_r8
real (r8), parameter :: g7 = 8.057771569962299e-15_r8
real (r8), parameter :: g8 = -7.054313817447962e-17_r8
real (r8), parameter :: g9 = 2.859992717347235e-19_r8
real (r8), parameter :: sp2 = 2.0_r8

real (r8) :: t68, ft68, rtx0, dsp_drtx, sp_est, rtx, rtxm, x, part1, part2
real (r8) :: sqrty, sp_hill_raw_at_sp2

t68 = t*1.00024_r8
ft68 = (t68 - 15.0_r8)/(1.0_r8 + k*(t68 - 15.0_r8))

!--------------------------------------------------------------------------
! Find the initial estimates of Rtx (Rtx0) and of the derivative dSP_dRtx
! at SP = 2. 
!--------------------------------------------------------------------------
rtx0 = g0 + t68*(g1 + t68*(g2 + t68*(g3 + t68*(g4 + t68*(g5 &
          + t68*(g6 + t68*(g7 + t68*(g8 + t68*g9))))))))
     
dsp_drtx =  a1 + (2.0_r8*a2 + (3.0_r8*a3 + (4.0_r8*a4 + 5.0_r8*a5*rtx0)*rtx0)* &
                  rtx0)*rtx0 + &
      ft68*(b1 + (2.0_r8*b2 + (3.0_r8*b3 + (4.0_r8*b4 + 5.0_r8*b5*rtx0)*rtx0)* &
                  rtx0)*rtx0) 

!--------------------------------------------------------------------------
! Begin a single modified Newton-Raphson iteration to find Rt at SP = 2.
!--------------------------------------------------------------------------
sp_est = a0 + (a1 + (a2 + (a3 + (a4 + a5*rtx0)*rtx0)*rtx0)*rtx0)*rtx0 &
 + ft68*(b0 + (b1 + (b2 + (b3 + (b4 + b5*rtx0)*rtx0)*rtx0)*rtx0)*rtx0)
rtx = rtx0 - (sp_est - sp2)/dsp_drtx
rtxm = 0.5_r8*(rtx + rtx0)
dsp_drtx =  a1 + (2.0_r8*a2 + (3.0_r8*a3 + (4.0_r8*a4 + 5.0_r8*a5*rtxm)*rtxm)* &
                  rtxm)*rtxm &
    + ft68*(b1 + (2.0_r8*b2 + (3.0_r8*b3 + (4.0_r8*b4 + 5.0_r8*b5*rtxm)*rtxm)* &
                  rtxm)*rtxm)
rtx = rtx0 - (sp_est - sp2)/dsp_drtx

! This is the end of one full iteration of the modified Newton-Raphson 
! iterative equation solver.  The error in Rtx at this point is equivalent 
! to an error in SP of 9e-16 psu.  
                                
x = 400.0_r8*rtx*rtx
sqrty = 10.0_r8*rtx
part1 = 1.0_r8 + x*(1.5_r8 + x) 
part2 = 1.0_r8 + sqrty*(1.0_r8 + sqrty*(1.0_r8 + sqrty))
sp_hill_raw_at_sp2 = SP2 - a0/part1 - b0*ft68/part2

gsw_hill_ratio_at_sp2 = 2.0_r8/sp_hill_raw_at_sp2

return
end function

!--------------------------------------------------------------------------
!==========================================================================
elemental subroutine gsw_ice_fraction_to_freeze_seawater (sa, ct, p, &
                                          t_ih, sa_freeze, ct_freeze, w_ih)
!==========================================================================
!
!  Calculates the mass fraction of ice (mass of ice divided by mass of ice
!  plus seawater), which, when melted into seawater having (SA,CT,p) causes 
!  the final dilute seawater to be at the freezing temperature.  The other
!  outputs are the Absolute Salinity and Conservative Temperature of the
!  final diluted seawater.  
!
!  SA   =  Absolute Salinity of seawater                           [ g/kg ]
!  CT   =  Conservative Temperature of seawater (ITS-90)          [ deg C ]
!  p    =  sea pressure                                            [ dbar ]
!            ( i.e. absolute pressure - 10.1325d0 dbar )
!  t_Ih =  in-situ temperature of the ice at pressure p (ITS-90)  [ deg C ]
!
!  SA_freeze = Absolute Salinity of seawater after the mass fraction of 
!              ice, ice_fraction, at temperature t_Ih has melted into the
!              original seawater, and the final mixture is at the freezing
!              temperature of seawater.                            [ g/kg ]
!
!  CT_freeze = Conservative Temperature of seawater after the mass 
!              fraction, w_Ih, of ice at temperature t_Ih has melted into
!              the original seawater, and the final mixture is at the
!              freezing temperature of seawater.                  [ deg C ]
!
!  w_Ih      = mass fraction of ice, having in-situ temperature t_Ih, 
!              which, when melted into seawater at (SA,CT,p) leads to the
!              final diluted seawater being at the freezing temperature.
!              This output must be between 0 and 1.              [unitless]
!--------------------------------------------------------------------------

use gsw_mod_toolbox, only : gsw_enthalpy_ice, gsw_t_freezing_exact
use gsw_mod_toolbox, only : gsw_enthalpy_first_derivatives_ct_exact
use gsw_mod_toolbox, only : gsw_ct_freezing_exact, gsw_enthalpy_ct_exact
use gsw_mod_toolbox, only : gsw_ct_freezing_first_derivatives

use gsw_mod_error_functions, only : gsw_error_code, gsw_error_limit

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sa, ct, p, t_ih
real (r8), intent(out) :: sa_freeze, ct_freeze, w_ih

integer :: no_iter
real (r8) :: ctf, ctf_mean, ctf_old, ctf_plus1, ctf_zero
real (r8) :: dfunc_dsaf, func, func_plus1, func_zero, h, h_ih
real (r8) :: saf, saf_mean, saf_old, tf, h_hat_sa, h_hat_ct, ctf_sa

real (r8), parameter :: sa0 = 0.0_r8, saturation_fraction = 0.0_r8

character (*), parameter :: func_name = "gsw_ice_fraction_to_freeze_seawater"

ctf = gsw_ct_freezing_exact(sa,p,saturation_fraction)
if (ct .lt. ctf) then
    ! The seawater ct input is below the freezing temp
    sa_freeze = gsw_error_code(1,func_name)
    ct_freeze = sa_freeze
    w_ih = sa_freeze
    return
end if

tf = gsw_t_freezing_exact(sa0,p,saturation_fraction)
if (t_ih .gt. tf) then
    ! The input, t_Ih, exceeds the freezing temperature at sa = 0
    sa_freeze = gsw_error_code(2,func_name)
    ct_freeze = sa_freeze
    w_ih = sa_freeze
    return
end if

h = gsw_enthalpy_ct_exact(sa,ct,p)
h_ih = gsw_enthalpy_ice(t_ih,p)

ctf_zero = gsw_ct_freezing_exact(sa0,p,saturation_fraction)
func_zero = sa*(gsw_enthalpy_ct_exact(sa0,ctf_zero,p) - h_ih)

ctf_plus1 = gsw_ct_freezing_exact(sa+1.0_r8,p,saturation_fraction)
func_plus1 = sa*(gsw_enthalpy_ct_exact(sa+1.0_r8,ctf_plus1,p) - h) - (h - h_ih)

saf = -(sa+1.0_r8)*func_zero/(func_plus1 - func_zero)      ! initial guess
ctf = gsw_ct_freezing_exact(saf,p,saturation_fraction)
call gsw_enthalpy_first_derivatives_ct_exact(saf,ctf,p,h_hat_sa,h_hat_ct)
call gsw_ct_freezing_first_derivatives(saf,p,1.0_r8,ctfreezing_sa=ctf_sa)

dfunc_dsaf = sa*(h_hat_sa + h_hat_ct*ctf_sa) - (h - h_ih)

do no_iter = 1, 2
    saf_old = saf
    ctf_old = ctf
    func = sa*(gsw_enthalpy_ct_exact(saf_old,ctf_old,p) - h) &
           - (saf_old - sa)*(h - h_ih)
    saf = saf_old - func/dfunc_dsaf
    saf_mean = 0.5_r8*(saf + saf_old)
    ctf_mean = gsw_ct_freezing_exact(saf_mean,p,saturation_fraction)
    call gsw_enthalpy_first_derivatives_ct_exact(saf_mean,ctf_mean,p,h_hat_sa, &
                                                 h_hat_ct)
    call gsw_ct_freezing_first_derivatives(saf_mean,p,saturation_fraction, &
                                           ctfreezing_sa=ctf_sa)
    dfunc_dsaf = sa*(h_hat_sa + h_hat_ct*ctf_sa) - (h - h_ih)
    saf = saf_old - func/dfunc_dsaf 
    ctf = gsw_ct_freezing_exact(saf,p,saturation_fraction)
end do

! After these 2 iterations of this modified Newton-Raphson method, the
! error in SA_freeze is less than 1.3d0x10^-13 g/kg, in CT_freeze is less than
! 4x10^-13 deg C and in w_Ih is less than 3.8d0x10^-15 which represent machine 
! precision for these calculations. 

sa_freeze = saf
ct_freeze = ctf
w_ih = (h - gsw_enthalpy_ct_exact(sa_freeze,ct_freeze,p))/(h - h_ih)

return
end subroutine
!==========================================================================
elemental function gsw_internal_energy (sa, ct, p)  
!==========================================================================
!
!  Calculates specific internal energy of seawater.
!
!  sa     : Absolute Salinity                               [g/kg]
!  ct     : Conservative Temperature                       [deg C]
!  p      : sea pressure                                    [dbar]
! 
!  gsw_internal_energy  :  internal_energy of seawater    [ J/kg ]
!--------------------------------------------------------------------------

use gsw_mod_toolbox, only : gsw_enthalpy, gsw_specvol

use gsw_mod_teos10_constants, only : gsw_p0, db2pa

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sa, ct, p  

real (r8) :: gsw_internal_energy

gsw_internal_energy = gsw_enthalpy(sa,ct,p) - &
                                (gsw_p0 + db2pa*p)*gsw_specvol(sa,ct,p)
return
end function

!--------------------------------------------------------------------------


!==========================================================================
elemental function gsw_internal_energy_ice (t, p)
!==========================================================================
!
!  Calculates the specific internal energy of ice. 
!
!  t  =  in-situ temperature (ITS-90)                             [ deg C ]
!  p  =  sea pressure                                              [ dbar ]
!         ( i.e. absolute pressure - 10.1325 dbar ) 
!
!  internal_energy_ice  =  specific internal energy (u)              [J/kg]
!--------------------------------------------------------------------------

use gsw_mod_toolbox, only : gsw_gibbs_ice

use gsw_mod_teos10_constants, only : gsw_p0, gsw_t0, db2pa

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: t, p

real (r8) :: gsw_internal_energy_ice

gsw_internal_energy_ice = gsw_gibbs_ice(0,0,t,p) &
                          - (gsw_t0 + t)*gsw_gibbs_ice(1,0,t,p) &
                          - (db2pa*p + gsw_p0)*gsw_gibbs_ice(0,1,t,p)

return
end function

!--------------------------------------------------------------------------
!==========================================================================
pure subroutine gsw_ipv_vs_fnsquared_ratio (sa, ct, p, p_ref, &
                                            ipv_vs_fnsquared_ratio, p_mid)
!==========================================================================
!
!  Calculates the ratio of the vertical gradient of potential density to 
!  the vertical gradient of locally-referenced potential density.  This 
!  ratio is also the ratio of the planetary Isopycnal Potential Vorticity
!  (IPV) to f times N^2, hence the name for this variable,
!  IPV_vs_fNsquared_ratio (see Eqn. (3.20.5) of IOC et al. (2010)). 
!  The reference sea pressure, p_ref, of the potential density surface must
!  have a constant value.
!
!  IPV_vs_fNsquared_ratio is evaluated at the mid pressure between the 
!  individual data points in the vertical.

!  sa      : Absolute Salinity         (a profile (length nz))     [g/kg]
!  ct      : Conservative Temperature  (a profile (length nz))     [deg C]
!  p       : sea pressure              (a profile (length nz))     [dbar]
!  p_ref   : reference sea pressure of the potential density surface
!        ( i.e. absolute reference pressure - 10.1325 dbar )      [dbar]
!  IPV_vs_fNsquared_ratio
!          : The ratio of the vertical gradient of potential density
!            referenced to p_ref, to the vertical gradient of locally-
!            referenced potential density.  It is ouput on the same
!            vertical (M-1)xN grid as p_mid. 
!            IPV_vs_fNsquared_ratio is dimensionless.          [ unitless ]
!  p_mid   : Mid pressure between p grid  (length nz-1)           [dbar]
!--------------------------------------------------------------------------

use gsw_mod_toolbox, only : gsw_alpha, gsw_beta

use gsw_mod_error_functions, only : gsw_error_code

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sa(:), ct(:), p(:), p_ref
real (r8), intent(out) :: ipv_vs_fnsquared_ratio(:), p_mid(:)

integer :: nz, i
real (r8), dimension(:), allocatable :: dsa, sa_mid, dct, ct_mid, vp_ref
real (r8), dimension(:), allocatable :: alpha_mid, beta_mid, alpha_pref
real (r8), dimension(:), allocatable :: beta_pref, numerator, denominator

character (*), parameter :: func_name = "gsw_ipv_vs_fnsquared_ratio"

nz = size(sa)
if (size(ipv_vs_fnsquared_ratio).lt.nz-1 .or. size(p_mid).lt.nz-1) then
    ipv_vs_fnsquared_ratio = gsw_error_code(1,func_name)
    p_mid = ipv_vs_fnsquared_ratio(1)
    return
end if

allocate (dsa(nz-1), sa_mid(nz-1), dct(nz-1), ct_mid(nz-1))
allocate (vp_ref(nz-1), alpha_mid(nz-1), beta_mid(nz-1), alpha_pref(nz-1))
allocate (beta_pref(nz-1), numerator(nz-1), denominator(nz-1))

forall (i = 1: nz-1)
    dsa(i) = sa(i) - sa(i+1)
    dct(i) = ct(i) - ct(i+1)
    sa_mid(i) = 0.5_r8*(sa(i) + sa(i+1))
    ct_mid(i) = 0.5_r8*(ct(i) + ct(i+1))
    p_mid(i) = 0.5_r8*(p(i) + p(i+1))
end forall

vp_ref = p_ref
alpha_mid = gsw_alpha(sa_mid,ct_mid,p_mid(1:nz-1))
beta_mid = gsw_beta(sa_mid,ct_mid,p_mid(1:nz-1))
alpha_pref = gsw_alpha(sa_mid,ct_mid,vp_ref)
beta_pref = gsw_beta(sa_mid,ct_mid,vp_ref)

numerator = dct*alpha_pref - dsa*beta_pref
denominator = dct*alpha_mid - dsa*beta_mid

where (denominator /= 0.0_r8)
    ipv_vs_fnsquared_ratio = numerator/denominator
elsewhere
    ipv_vs_fnsquared_ratio = gsw_error_code(2,func_name)
end where

deallocate (dsa, sa_mid, dct, ct_mid)
deallocate (vp_ref, alpha_mid, beta_mid, alpha_pref)
deallocate (beta_pref, numerator, denominator)

return
end subroutine

!--------------------------------------------------------------------------
!==========================================================================
elemental function gsw_kappa (sa, ct, p)
!==========================================================================
!
!  Calculates the isentropic compressibility of seawater.  This function 
!  has inputs of Absolute Salinity and Conservative Temperature.  This 
!  function uses the computationally-efficient expression for 
!  specific volume in terms of SA, CT and p (Roquet et al., 2014).
!
!  SA  =  Absolute Salinity                                        [ g/kg ]
!  CT  =  Conservative Temperature (ITS-90)                       [ deg C ]
!  p   =  sea pressure                                             [ dbar ]
!         ( i.e. absolute pressure - 10.1325 dbar )
!
!  kappa  =  isentropic compressibility of seawater                [ 1/Pa ]
!--------------------------------------------------------------------------

use gsw_mod_teos10_constants, only : gsw_sfac, offset

use gsw_mod_specvol_coefficients

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sa, ct, p

real (r8) :: gsw_kappa

real (r8) :: v, v_p, xs, ys, z

xs = sqrt(gsw_sfac*sa + offset)
ys = ct*0.025_r8
z = p*1e-4_r8

v = v000 + xs*(v010 + xs*(v020 + xs*(v030 + xs*(v040 + xs*(v050 &
    + v060*xs))))) + ys*(v100 + xs*(v110 + xs*(v120 + xs*(v130 + xs*(v140 &
    + v150*xs)))) + ys*(v200 + xs*(v210 + xs*(v220 + xs*(v230 + v240*xs))) &
    + ys*(v300 + xs*(v310 + xs*(v320 + v330*xs)) + ys*(v400 + xs*(v410 &
    + v420*xs) + ys*(v500 + v510*xs + v600*ys))))) + z*(v001 + xs*(v011 &
    + xs*(v021 + xs*(v031 + xs*(v041 + v051*xs)))) + ys*(v101 + xs*(v111 &
    + xs*(v121 + xs*(v131 + v141*xs))) + ys*(v201 + xs*(v211 + xs*(v221 &
    + v231*xs)) + ys*(v301 + xs*(v311 + v321*xs) + ys*(v401 + v411*xs &
    + v501*ys)))) + z*(v002 + xs*(v012 + xs*(v022 + xs*(v032 + v042*xs))) &
    + ys*(v102 + xs*(v112 + xs*(v122 + v132*xs)) + ys*(v202 + xs*(v212 &
    + v222*xs) + ys*(v302 + v312*xs + v402*ys))) + z*(v003 + xs*(v013 &
    + v023*xs) + ys*(v103 + v113*xs + v203*ys) + z*(v004 + v014*xs + v104*ys &  
    + z*(v005 + v006*z)))))

v_p = c000 + xs*(c100 + xs*(c200 + xs*(c300 + xs*(c400 + c500*xs)))) & 
    + ys*(c010 + xs*(c110 + xs*(c210 + xs*(c310 + c410*xs))) + ys*(c020 &
    + xs*(c120 + xs*(c220 + c320*xs)) + ys*(c030 + xs*(c130 + c230*xs) &
    + ys*(c040 + c140*xs + c050*ys)))) + z*(c001 + xs*(c101 + xs*(c201 &
    + xs*(c301 + c401*xs))) + ys*(c011 + xs*(c111 + xs*(c211 + c311*xs)) &
    + ys*(c021 + xs*(c121 + c221*xs) + ys*(c031 + c131*xs + c041*ys))) &
    + z*( c002 + xs*(c102 + c202*xs) + ys*(c012 + c112*xs + c022*ys) &
    + z*(c003 + c103*xs + c013*ys + z*(c004 + c005*z))))

gsw_kappa = -1e-8_r8*v_p/v

return
end function

!--------------------------------------------------------------------------
!==========================================================================
elemental function gsw_kappa_const_t_ice (t, p)
!==========================================================================
!
!  Calculates isothermal compressibility of ice. 
!  Note. This is the compressibility of ice AT CONSTANT IN-SITU
!    TEMPERATURE
!
!  t  =  in-situ temperature (ITS-90)                             [ deg C ]
!  p  =  sea pressure                                              [ dbar ]
!         ( i.e. absolute pressure - 10.1325 dbar )
!
!  kappa_const_t_ice  =  isothermal compressibility                [ 1/Pa ]
!   Note. The output units are 1/Pa not 1/dbar.
!--------------------------------------------------------------------------

use gsw_mod_toolbox, only : gsw_gibbs_ice

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: t, p

real (r8) :: gsw_kappa_const_t_ice

gsw_kappa_const_t_ice = -gsw_gibbs_ice(0,2,t,p)/gsw_gibbs_ice(0,1,t,p)

return
end function

!--------------------------------------------------------------------------
!==========================================================================
elemental function gsw_kappa_ice (t, p)
!==========================================================================
!
!  Calculates the isentropic compressibility of ice. 
!  
!  t  =  in-situ temperature (ITS-90)                             [ deg C ]
!  p  =  sea pressure                                              [ dbar ]
!         ( i.e. absolute pressure - 10.1325 dbar )
!
!  kappa_ice  =  isentropic compressibility                        [ 1/Pa ]
!   Note. The output units are 1/Pa not 1/dbar.
!--------------------------------------------------------------------------

use gsw_mod_toolbox, only : gsw_gibbs_ice

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: t, p

real (r8) :: gsw_kappa_ice

real (r8) :: gi_tp, gi_tt

gi_tt = gsw_gibbs_ice(2,0,t,p) 
gi_tp = gsw_gibbs_ice(1,1,t,p)

gsw_kappa_ice = (gi_tp*gi_tp - gi_tt*gsw_gibbs_ice(0,2,t,p))/ &
                  (gsw_gibbs_ice(0,1,t,p)*gi_tt)

return
end function

!--------------------------------------------------------------------------
!==========================================================================
elemental function gsw_kappa_t_exact (sa, t, p)  
!==========================================================================
!
! isentropic compressibility of seawater
!
! sa     : Absolute Salinity                               [g/kg]
! t      : in-situ temperature                             [deg C]
! p      : sea pressure                                    [dbar]
!
! gsw_kappa_t_exact : isentropic compressibility           [1/Pa]
!--------------------------------------------------------------------------

use gsw_mod_toolbox, only : gsw_gibbs

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sa, t, p  

real (r8) :: gsw_kappa_t_exact

real (r8) :: g_tt, g_tp

integer, parameter :: n0=0, n1=1, n2=2

g_tt = gsw_gibbs(n0,n2,n0,sa,t,p)
g_tp = gsw_gibbs(n0,n1,n1,sa,t,p)

gsw_kappa_t_exact = (g_tp*g_tp - g_tt*gsw_gibbs(n0,n0,n2,sa,t,p)) / &
                         (gsw_gibbs(n0,n0,n1,sa,t,p)*g_tt)

return
end function

!--------------------------------------------------------------------------
!==========================================================================
elemental function gsw_latentheat_evap_ct (sa, ct) 
!==========================================================================
!
! Calculates latent heat, or enthalpy, of evaporation.
!
! sa     : Absolute Salinity                               [g/kg]
! ct     : Conservative Temperature                        [deg C]
! 
! gsw_latentheat_evaporation : latent heat of evaporation  [J/kg]
!--------------------------------------------------------------------------

use gsw_mod_teos10_constants, only : gsw_sfac

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sa, ct 

real (r8) :: gsw_latentheat_evap_ct

real (r8), parameter :: c0 =   2.499065844825125e6_r8
real (r8), parameter :: c1 =  -1.544590633515099e-1_r8
real (r8), parameter :: c2 =  -9.096800915831875e4_r8
real (r8), parameter :: c3 =   1.665513670736000e2_r8
real (r8), parameter :: c4 =   4.589984751248335e1_r8
real (r8), parameter :: c5 =   1.894281502222415e1_r8
real (r8), parameter :: c6 =   1.192559661490269e3_r8
real (r8), parameter :: c7 =  -6.631757848479068e3_r8
real (r8), parameter :: c8 =  -1.104989199195898e2_r8
real (r8), parameter :: c9 =  -1.207006482532330e3_r8
real (r8), parameter :: c10 = -3.148710097513822e3_r8
real (r8), parameter :: c11 =  7.437431482069087e2_r8
real (r8), parameter :: c12 =  2.519335841663499e3_r8
real (r8), parameter :: c13 =  1.186568375570869e1_r8
real (r8), parameter :: c14 =  5.731307337366114e2_r8
real (r8), parameter :: c15 =  1.213387273240204e3_r8
real (r8), parameter :: c16 =  1.062383995581363e3_r8
real (r8), parameter :: c17 = -6.399956483223386e2_r8
real (r8), parameter :: c18 = -1.541083032068263e3_r8
real (r8), parameter :: c19 =  8.460780175632090e1_r8
real (r8), parameter :: c20 = -3.233571307223379e2_r8
real (r8), parameter :: c21 = -2.031538422351553e2_r8
real (r8), parameter :: c22 =  4.351585544019463e1_r8
real (r8), parameter :: c23 = -8.062279018001309e2_r8
real (r8), parameter :: c24 =  7.510134932437941e2_r8
real (r8), parameter :: c25 =  1.797443329095446e2_r8
real (r8), parameter :: c26 = -2.389853928747630e1_r8
real (r8), parameter :: c27 =  1.021046205356775e2_r8

real (r8) :: x, y

x = sqrt(gsw_sfac*sa)
y = ct/40.0_r8

gsw_latentheat_evap_ct = c0 + x*(c1 + c4*y + x*(c3   &
    + y*(c7 + c12*y) + x*(c6 + y*(c11 + y*(c17 + c24*y)) &
    + x*(c10 + y*(c16 + c23*y) + x*(c15 + c22*y + c21*x)))))  &
    + y*(c2 + y*(c5 + c8*x + y*(c9 + x*(c13 + c18*x)  &
    + y*(c14 + x*(c19 + c25*x) + y*(c20 + c26*x + c27*y)))))

return
end function

!--------------------------------------------------------------------------
!==========================================================================
elemental function gsw_latentheat_evap_t (sa, t)  
!==========================================================================
!
! Calculates latent heat, or enthalpy, of evaporation.
!
! sa     : Absolute Salinity                               [g/kg]
! t      : in-situ temperature                             [deg C]
! 
! gsw_latentheat_evap_t : latent heat of evaporation       [J/kg]
!--------------------------------------------------------------------------

use gsw_mod_toolbox, only : gsw_ct_from_pt, gsw_latentheat_evap_ct

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sa, t  

real (r8) :: gsw_latentheat_evap_t

real (r8) :: ct

ct = gsw_ct_from_pt(sa,t)

gsw_latentheat_evap_t = gsw_latentheat_evap_ct(sa,ct)

return
end function

!--------------------------------------------------------------------------
!==========================================================================
elemental function gsw_latentheat_melting (sa, p)  
!==========================================================================
!
! Calculates latent heat, or enthalpy, of melting.
!
! sa     : Absolute Salinity                               [g/kg]
! p      : sea pressure                                    [dbar]
! 
! gsw_latentheat_melting : latent heat of melting          [kg/m^3]
!--------------------------------------------------------------------------

use gsw_mod_toolbox, only : gsw_enthalpy_ice, gsw_t_freezing_exact
use gsw_mod_toolbox, only : gsw_chem_potential_water_t_exact
use gsw_mod_toolbox, only : gsw_t_deriv_chem_potential_water_t_exact

use gsw_mod_teos10_constants, only : gsw_sfac, gsw_t0

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sa, p  

real (r8) :: gsw_latentheat_melting

real (r8) :: tf

tf = gsw_t_freezing_exact(sa,p,0.0_r8);

gsw_latentheat_melting = 1000.0_r8*(gsw_chem_potential_water_t_exact(sa,tf,p) &
           - (gsw_t0 + tf)*gsw_t_deriv_chem_potential_water_t_exact(sa,tf,p)) &
           - gsw_enthalpy_ice(tf,p);

return
end function

!--------------------------------------------------------------------------
!==========================================================================
pure subroutine gsw_linear_interp_sa_ct (sa, ct, p, p_i, sa_i, ct_i)
!==========================================================================
! This function interpolates the cast with respect to the interpolating 
! variable p. This function finds the values of SA, CT at p_i on this cast.
!
! VERSION NUMBER: 3.05 (27th January 2015)
!
! This fuction was adapted from Matlab's interp1q.
!==========================================================================

use gsw_mod_toolbox, only : gsw_util_sort_real

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sa(:), ct(:), p(:), p_i(:)
real (r8), intent(out) :: sa_i(:), ct_i(:)

logical, allocatable :: in_rng(:)
integer, allocatable :: j(:), k(:), r(:), jrev(:), ki(:)
real (r8), allocatable :: xi(:), xxi(:), u(:)

integer :: imax_p(1), imin_p(1), i, n, m, ii, np, npi
real (r8) :: max_p, min_p

np = size(p)
npi = size(p_i)

min_p = minval(p); imin_p = minloc(p)
max_p = maxval(p); imax_p = maxloc(p)

where (p_i .le. min_p)
    sa_i = sa(imin_p(1))  ! set equal to the shallowest bottle.
    ct_i = ct(imin_p(1))
end where
where (p_i .ge. max_p)
    sa_i = sa(imax_p(1))  ! set equal to the deepest bottle.
    ct_i = ct(imax_p(1))
end where

allocate (in_rng(npi))
in_rng = (p_i .gt. min_p .and. p_i .lt. max_p)

n = count(in_rng)
if (n .eq. 0) return

allocate (xi(n), k(n), ki(n), r(n), u(n))
m = np + n
allocate (xxi(m), j(m), jrev(m))

ii = 0
do i = 1, npi
    if (in_rng(i)) then
        ii = ii + 1
        xi(ii) = p_i(i)
        ki(ii) = i
    end if
end do

k = gsw_util_sort_real(xi)
xxi = (/ p, xi(k) /)
j = gsw_util_sort_real(xxi)

jrev(j) = (/ (i, i=1,size(j)) /)
r(k) = jrev(np+1:) - (/ (i, i=1,size(xi)) /)

u = (xi-p(r))/(p(r+1)-p(r))
sa_i(ki) = sa(r) + (sa(r+1)-sa(r))*u
ct_i(ki) = ct(r) + (ct(r+1)-ct(r))*u

return

end subroutine

!--------------------------------------------------------------------------
!==========================================================================
elemental function gsw_melting_ice_equilibrium_sa_ct_ratio (sa, p)
!==========================================================================
!
!  Calculates the ratio of SA to CT changes when ice melts into seawater
!  with both the seawater and the seaice temperatures being almost equal to
!  the equilibrium freezing temperature.  It is assumed that a small mass 
!  of ice melts into an infinite mass of seawater.  If indeed the 
!  temperature of the seawater and the ice were both equal to the freezing
!  temperature, then no melting or freezing would occur an imbalance 
!  between these three temperatures is needed for freezing or melting to
!  occur (the three temperatures being (1) the seawater temperature, 
!  (2) the ice temperature, and (3) the freezing temperature.  
!
!  The output, melting_ice_equilibrium_SA_CT_ratio, is dSA/dCT rather than 
!  dCT/dSA.  This is done so that when SA = 0, the output, dSA/dCT is zero
!  whereas dCT/dSA would be infinite. 
!
!  SA  =  Absolute Salinity of seawater                            [ g/kg ]
!  p   =  sea pressure at which the melting occurs                 [ dbar ]
!         ( i.e. absolute pressure - 10.1325d0 dbar ) 
!
!  melting_ice_equilibrium_SA_CT_ratio = the ratio dSA/dCT of SA to CT  
!                                changes when ice melts into seawater, with   
!                                the seawater and seaice being close to the  
!                                freezing temperature.         [ g/(kg K) ] 
!--------------------------------------------------------------------------

use gsw_mod_toolbox, only : gsw_ct_freezing_exact, gsw_enthalpy_ct_exact
use gsw_mod_toolbox, only : gsw_t_freezing_exact, gsw_enthalpy_ice
use gsw_mod_toolbox, only : gsw_enthalpy_first_derivatives_ct_exact

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sa, p

real (r8) :: gsw_melting_ice_equilibrium_sa_ct_ratio

real (r8) :: ctf, h, h_ih, t_seaice, h_hat_sa, h_hat_ct

real (r8), parameter :: saturation_fraction = 0.0_r8

ctf = gsw_ct_freezing_exact(sa,p,saturation_fraction)
t_seaice = gsw_t_freezing_exact(sa,p,saturation_fraction)

h = gsw_enthalpy_ct_exact(sa,ctf,p)
h_ih = gsw_enthalpy_ice(t_seaice,p)
call gsw_enthalpy_first_derivatives_ct_exact(sa,ctf,p,h_hat_sa,h_hat_ct)
          ! note that h_hat_ct is equal to cp0*(273.15 + t)/(273.15 + pt0)

gsw_melting_ice_equilibrium_sa_ct_ratio = sa*h_hat_ct/(h - h_ih - sa*h_hat_sa)

return
end function

!--------------------------------------------------------------------------
!==========================================================================
elemental function gsw_melting_ice_equilibrium_sa_ct_ratio_poly (sa, p)
!==========================================================================
!
!  Calculates the ratio of SA to CT changes when ice melts into seawater
!  with both the seawater and the seaice temperatures being almost equal to
!  the equilibrium freezing temperature.  It is assumed that a small mass 
!  of ice melts into an infinite mass of seawater.  If indeed the 
!  temperature of the seawater and the ice were both equal to the freezing
!  temperature, then no melting or freezing would occur an imbalance 
!  between these three temperatures is needed for freezing or melting to
!  occur (the three temperatures being (1) the seawater temperature, 
!  (2) the ice temperature, and (3) the freezing temperature.  
!
!  The output, melting_ice_equilibrium_SA_CT_ratio, is dSA/dCT rather than 
!  dCT/dSA.  This is done so that when SA = 0, the output, dSA/dCT is zero
!  whereas dCT/dSA would be infinite. 
!
!  SA  =  Absolute Salinity of seawater                            [ g/kg ]
!  p   =  sea pressure at which the melting occurs                 [ dbar ]
!         ( i.e. absolute pressure - 10.1325d0 dbar ) 
!
!  melting_ice_equilibrium_SA_CT_ratio = the ratio dSA/dCT of SA to CT  
!                                changes when ice melts into seawater, with   
!                                the seawater and seaice being close to the  
!                                freezing temperature.         [ g/(kg K) ] 
!--------------------------------------------------------------------------

use gsw_mod_toolbox, only : gsw_ct_freezing_poly, gsw_enthalpy
use gsw_mod_toolbox, only : gsw_t_freezing_poly, gsw_enthalpy_ice
use gsw_mod_toolbox, only : gsw_enthalpy_first_derivatives

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sa, p

real (r8) :: gsw_melting_ice_equilibrium_sa_ct_ratio_poly

real (r8) :: ctf, h, h_ih, t_seaice, h_hat_sa, h_hat_ct

real (r8), parameter :: saturation_fraction = 0.0_r8

ctf = gsw_ct_freezing_poly(sa,p,saturation_fraction)
t_seaice = gsw_t_freezing_poly(sa,p,saturation_fraction)

h = gsw_enthalpy(sa,ctf,p)
h_ih = gsw_enthalpy_ice(t_seaice,p)
call gsw_enthalpy_first_derivatives(sa,ctf,p,h_hat_sa,h_hat_ct)
          ! note that h_hat_ct is equal to cp0*(273.15 + t)/(273.15 + pt0)

gsw_melting_ice_equilibrium_sa_ct_ratio_poly = sa*h_hat_ct &
                                                    / (h - h_ih - sa*h_hat_sa)

return
end function

!--------------------------------------------------------------------------
!==========================================================================
elemental subroutine gsw_melting_ice_into_seawater (sa, ct, p, w_ih, t_ih,&
                                            sa_final, ct_final, w_ih_final)
!==========================================================================
!
!  Calculates the final Absolute Salinity, final Conservative Temperature 
!  and final ice mass fraction that results when a given mass fraction of 
!  ice melts and is mixed into seawater whose properties are (SA,CT,p).  
!  This code takes the seawater to contain no dissolved air.  
!
!  When the mass fraction w_Ih_final is calculated as being a positive
!  value, the seawater-ice mixture is at thermodynamic equlibrium.  
!
!  This code returns w_Ih_final = 0 when the input bulk enthalpy, h_bulk, 
!  is sufficiently large (i.e. sufficiently "warm") so that there is no ice 
!  present in the final state.  In this case the final state consists of 
!  only seawater rather than being an equlibrium mixture of seawater and 
!  ice which occurs when w_Ih_final is positive.  Note that when 
!  w_Ih_final = 0, the final seawater is not at the freezing temperature. 
!
!  SA   =  Absolute Salinity of seawater                           [ g/kg ]
!  CT   =  Conservative Temperature of seawater (ITS-90)          [ deg C ]
!  p    =  sea pressure at which the melting occurs                [ dbar ]
!         ( i.e. absolute pressure - 10.1325 dbar ) 
!  w_Ih =  mass fraction of ice, that is the mass of ice divided by the
!          sum of the masses of ice and seawater.  That is, the mass of 
!          ice divided by the mass of the final mixed fluid.  
!          w_Ih must be between 0 and 1.                       [ unitless ]
!  t_Ih =  the in-situ temperature of the ice (ITS-90)            [ deg C ]
!
!  SA_final    =  Absolute Salinity of the seawater in the final state, 
!                 whether or not any ice is present.               [ g/kg ]
!  CT_final    =  Conservative Temperature of the seawater in the the final
!                 state, whether or not any ice is present.       [ deg C ]
!  w_Ih_final  =  mass fraction of ice in the final seawater-ice mixture.
!                 If this ice mass fraction is positive, the system is at 
!                 thermodynamic equilibrium.  If this ice mass fraction is 
!                 zero there is no ice in the final state which consists 
!                 only of seawater which is warmer than the freezing 
!                 temperature.                                   [unitless]
!--------------------------------------------------------------------------

use gsw_mod_toolbox, only : gsw_ct_freezing_exact, gsw_enthalpy_ice
use gsw_mod_toolbox, only : gsw_frazil_properties, gsw_t_freezing_exact
use gsw_mod_toolbox, only : gsw_enthalpy_ct_exact

use gsw_mod_error_functions, only : gsw_error_code, gsw_error_limit

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sa, ct, p, w_ih, t_ih
real (r8), intent(out) :: sa_final, ct_final, w_ih_final

real (r8) :: ctf, h_bulk, sa_bulk, tf_ih

real (r8), parameter :: saturation_fraction = 0.0_r8

character (*), parameter :: func_name = "gsw_melting_ice_into_seawater"

ctf = gsw_ct_freezing_exact(sa,p,saturation_fraction)
if (ct .lt. ctf) then
    ! The seawater ct input is below the freezing temp
    sa_final = gsw_error_code(1,func_name)
    ct_final = sa_final
    w_ih_final = sa_final
    return
end if

tf_ih = gsw_t_freezing_exact(0.0_r8,p,saturation_fraction) - 1e-6_r8
if (t_ih .gt. tf_ih) then
    ! t_ih input exceeds the freezing temp. The 1e-6 C buffer in the allowable
    ! t_Ih is to ensure that there is some ice Ih in the sea ice.  
    sa_final = gsw_error_code(2,func_name)
    ct_final = sa_final
    w_ih_final = sa_final
    return
end if

sa_bulk = (1.0_r8 - w_ih)*sa
h_bulk = (1.0_r8 - w_ih)*gsw_enthalpy_ct_exact(sa,ct,p) &
                  + w_ih*gsw_enthalpy_ice(t_ih,p)
call gsw_frazil_properties(sa_bulk,h_bulk,p,sa_final,ct_final,w_ih_final)
if (sa_final .gt. gsw_error_limit) then
    sa_final = gsw_error_code(3,func_name,sa_final)
    ct_final = sa_final
    w_ih_final = sa_final
    return
endif
        
return
end subroutine

!--------------------------------------------------------------------------
!==========================================================================
elemental function gsw_melting_ice_sa_ct_ratio (sa, ct, p, t_ih)
!==========================================================================
!
!  Calculates the ratio of SA to CT changes when ice melts into seawater.
!  It is assumed that a small mass of ice melts into an infinite mass of
!  seawater.  Because of the infinite mass of seawater, the ice will always
!  melt.   
!
!  The output, melting_seaice_SA_CT_ratio, is dSA/dCT rather than dCT/dSA. 
!  This is done so that when SA = 0, the output, dSA/dCT is zero whereas 
!  dCT/dSA would be infinite. 
!
!  SA   =  Absolute Salinity of seawater                           [ g/kg ]
!  CT   =  Conservative Temperature of seawater (ITS-90)          [ deg C ]
!  p    =  sea pressure at which the melting occurs                [ dbar ]
!         ( i.e. absolute pressure - 10.1325d0 dbar ) 
!  t_Ih =  the in-situ temperature of the ice (ITS-90)            [ deg C ]
!
!  melting_ice_SA_CT_ratio = the ratio of SA to CT changes when ice melts
!                            into a large mass of seawater 
!                                                          [ g kg^-1 K^-1 ] 
!--------------------------------------------------------------------------

use gsw_mod_toolbox, only : gsw_ct_freezing_exact, gsw_enthalpy_ct_exact
use gsw_mod_toolbox, only : gsw_t_freezing_exact, gsw_enthalpy_ice
use gsw_mod_toolbox, only : gsw_enthalpy_first_derivatives_ct_exact

use gsw_mod_error_functions, only : gsw_error_code, gsw_error_limit

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sa, ct, p, t_ih

real (r8) :: gsw_melting_ice_sa_ct_ratio

real (r8) :: ctf, h, h_ih, tf, h_hat_sa, h_hat_ct

real (r8), parameter :: saturation_fraction = 0.0_r8

character (*), parameter :: func_name = "gsw_melting_ice_sa_ct_ratio"

ctf = gsw_ct_freezing_exact(sa,p,saturation_fraction)
if (ct .lt. ctf) then
    ! the seawater ct input is below the freezing temperature
    gsw_melting_ice_sa_ct_ratio = gsw_error_code(1,func_name)
    return
end if

tf = gsw_t_freezing_exact(0.0_r8,p,saturation_fraction)
if (t_ih .gt. tf) then
    ! t_ih exceeds the freezing temperature at sa = 0
    gsw_melting_ice_sa_ct_ratio = gsw_error_code(2,func_name)
    return
end if

h = gsw_enthalpy_ct_exact(sa,ct,p)
h_ih = gsw_enthalpy_ice(t_ih,p)
call gsw_enthalpy_first_derivatives_ct_exact(sa,ct,p,h_hat_sa,h_hat_ct)
	! Note that h_hat_CT is equal to cp0*(273.15 + t)/(273.15 + pt0)

gsw_melting_ice_sa_ct_ratio = sa*h_hat_ct/(h - h_ih - sa*h_hat_sa)

return
end function

!--------------------------------------------------------------------------
!==========================================================================
elemental function gsw_melting_ice_sa_ct_ratio_poly (sa, ct, p, t_ih)
!==========================================================================
!
!  Calculates the ratio of SA to CT changes when ice melts into seawater.
!  It is assumed that a small mass of ice melts into an infinite mass of
!  seawater.  Because of the infinite mass of seawater, the ice will always
!  melt.   
!
!  The output, melting_seaice_SA_CT_ratio, is dSA/dCT rather than dCT/dSA. 
!  This is done so that when SA = 0, the output, dSA/dCT is zero whereas 
!  dCT/dSA would be infinite. 
!
!  SA   =  Absolute Salinity of seawater                           [ g/kg ]
!  CT   =  Conservative Temperature of seawater (ITS-90)          [ deg C ]
!  p    =  sea pressure at which the melting occurs                [ dbar ]
!         ( i.e. absolute pressure - 10.1325d0 dbar ) 
!  t_Ih =  the in-situ temperature of the ice (ITS-90)            [ deg C ]
!
!  melting_ice_SA_CT_ratio = the ratio of SA to CT changes when ice melts
!                            into a large mass of seawater 
!                                                          [ g kg^-1 K^-1 ] 
!--------------------------------------------------------------------------

use gsw_mod_toolbox, only : gsw_ct_freezing_poly, gsw_enthalpy
use gsw_mod_toolbox, only : gsw_t_freezing_poly, gsw_enthalpy_ice
use gsw_mod_toolbox, only : gsw_enthalpy_first_derivatives

use gsw_mod_error_functions, only : gsw_error_code, gsw_error_limit

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sa, ct, p, t_ih

real (r8) :: gsw_melting_ice_sa_ct_ratio_poly

real (r8) :: ctf, h, h_ih, tf, h_hat_sa, h_hat_ct

real (r8), parameter :: saturation_fraction = 0.0_r8

character (*), parameter :: func_name = "gsw_melting_ice_sa_ct_ratio_poly"

ctf = gsw_ct_freezing_poly(sa,p,saturation_fraction)
if (ct .lt. ctf) then
    ! the seawater ct input is below the freezing temperature
    gsw_melting_ice_sa_ct_ratio_poly = gsw_error_code(1,func_name)
    return
end if

tf = gsw_t_freezing_poly(0.0_r8,p,saturation_fraction)
if (t_ih .gt. tf) then
    ! t_ih exceeds the freezing temperature at sa = 0
    gsw_melting_ice_sa_ct_ratio_poly = gsw_error_code(2,func_name)
    return
end if

h = gsw_enthalpy(sa,ct,p)
h_ih = gsw_enthalpy_ice(t_ih,p)
call gsw_enthalpy_first_derivatives(sa,ct,p,h_hat_sa,h_hat_ct)
	! Note that h_hat_CT is equal to cp0*(273.15 + t)/(273.15 + pt0)

gsw_melting_ice_sa_ct_ratio_poly = sa*h_hat_ct/(h - h_ih - sa*h_hat_sa)

return
end function

!--------------------------------------------------------------------------
!==========================================================================
elemental function gsw_melting_seaice_equilibrium_sa_ct_ratio (sa, p)
!==========================================================================
!
!  Calculates the ratio of SA to CT changes when sea ice melts into 
!  seawater with both the seawater and the sea ice temperatures being  
!  almost equal to the equilibrium freezing temperature.  It is assumed  
!  that a small mass of seaice melts into an infinite mass of seawater.  If 
!  indeed the temperature of the seawater and the sea ice were both equal  
!  to the freezing temperature, then no melting or freezing would occur; an  
!  imbalance between these three temperatures is needed for freezing or 
!  melting to occur (the three temperatures being (1) the seawater 
!  temperature, (2) the sea ice temperature, and (3) the freezing 
!  temperature.  
!
!  Note that the output of this function, dSA/dCT is independent of the 
!  sea ice salinity, SA_seaice.  That is, the output applies equally to  
!  pure ice Ih and to sea ice with seaice salinity, SA_seaice.  This result 
!  is proven in the manuscript, McDougall et al. (2013).  
!
!  The output, melting_seaice_equilibrium_SA_CT_ratio, is dSA/dCT rather  
!  than dCT/dSA.  This is done so that when SA = 0, the output, dSA/dCT is 
!  zero whereas dCT/dSA would be infinite. 
!
!  SA  =  Absolute Salinity of seawater                            [ g/kg ]
!  p   =  sea pressure at which the melting occurs                 [ dbar ]
!         ( i.e. absolute pressure - 10.1325 dbar ) 
!
!  melting_seaice_equilibrium_SA_CT_ratio = the ratio dSA/dCT of SA to CT  
!                            changes when sea ice melts into seawater, with   
!                            the seawater and sea ice being close to the  
!                            freezing temperature.             [ g/(kg K) ] 
!--------------------------------------------------------------------------

use gsw_mod_toolbox, only : gsw_ct_freezing_exact, gsw_enthalpy_ct_exact
use gsw_mod_toolbox, only : gsw_t_freezing_exact, gsw_enthalpy_ice
use gsw_mod_toolbox, only : gsw_enthalpy_first_derivatives_ct_exact

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sa, p

real (r8) :: gsw_melting_seaice_equilibrium_sa_ct_ratio

real (r8) :: ctf, h, h_ih, t_seaice, h_hat_sa, h_hat_ct

real (r8), parameter :: saturation_fraction = 0.0_r8

ctf = gsw_ct_freezing_exact(sa,p,saturation_fraction)
t_seaice = gsw_t_freezing_exact(sa,p,saturation_fraction)

h = gsw_enthalpy_ct_exact(sa,ctf,p)
h_ih = gsw_enthalpy_ice(t_seaice,p)
call gsw_enthalpy_first_derivatives_ct_exact(sa,ctf,p,h_hat_sa,h_hat_ct)

gsw_melting_seaice_equilibrium_sa_ct_ratio = sa*h_hat_ct / &
                                                 (h - h_ih - sa*h_hat_sa)
         
return
end function

!--------------------------------------------------------------------------
!==========================================================================
elemental function gsw_melting_seaice_equilibrium_sa_ct_ratio_poly (sa, p)
!==========================================================================
!
!  Calculates the ratio of SA to CT changes when sea ice melts into 
!  seawater with both the seawater and the sea ice temperatures being  
!  almost equal to the equilibrium freezing temperature.  It is assumed  
!  that a small mass of seaice melts into an infinite mass of seawater.  If 
!  indeed the temperature of the seawater and the sea ice were both equal  
!  to the freezing temperature, then no melting or freezing would occur; an  
!  imbalance between these three temperatures is needed for freezing or 
!  melting to occur (the three temperatures being (1) the seawater 
!  temperature, (2) the sea ice temperature, and (3) the freezing 
!  temperature.  
!
!  Note that the output of this function, dSA/dCT is independent of the 
!  sea ice salinity, SA_seaice.  That is, the output applies equally to  
!  pure ice Ih and to sea ice with seaice salinity, SA_seaice.  This result 
!  is proven in the manuscript, McDougall et al. (2013).  
!
!  The output, melting_seaice_equilibrium_SA_CT_ratio, is dSA/dCT rather  
!  than dCT/dSA.  This is done so that when SA = 0, the output, dSA/dCT is 
!  zero whereas dCT/dSA would be infinite. 
!
!  SA  =  Absolute Salinity of seawater                            [ g/kg ]
!  p   =  sea pressure at which the melting occurs                 [ dbar ]
!         ( i.e. absolute pressure - 10.1325 dbar ) 
!
!  melting_seaice_equilibrium_SA_CT_ratio = the ratio dSA/dCT of SA to CT  
!                            changes when sea ice melts into seawater, with   
!                            the seawater and sea ice being close to the  
!                            freezing temperature.             [ g/(kg K) ] 
!--------------------------------------------------------------------------

use gsw_mod_toolbox, only : gsw_ct_freezing_poly, gsw_enthalpy
use gsw_mod_toolbox, only : gsw_t_freezing_poly, gsw_enthalpy_ice
use gsw_mod_toolbox, only : gsw_enthalpy_first_derivatives

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sa, p

real (r8) :: gsw_melting_seaice_equilibrium_sa_ct_ratio_poly

real (r8) :: ctf, h, h_ih, t_seaice, h_hat_sa, h_hat_ct

real (r8), parameter :: saturation_fraction = 0.0_r8

ctf = gsw_ct_freezing_poly(sa,p,saturation_fraction)
t_seaice = gsw_t_freezing_poly(sa,p,saturation_fraction)

h = gsw_enthalpy(sa,ctf,p)
h_ih = gsw_enthalpy_ice(t_seaice,p)
call gsw_enthalpy_first_derivatives(sa,ctf,p,h_hat_sa,h_hat_ct)

gsw_melting_seaice_equilibrium_sa_ct_ratio_poly = sa*h_hat_ct / &
                                                     (h - h_ih - sa*h_hat_sa)
         
return
end function

!--------------------------------------------------------------------------
!==========================================================================
elemental subroutine gsw_melting_seaice_into_seawater (sa, ct, p, &
                         w_seaice, sa_seaice, t_seaice, sa_final, ct_final)
!==========================================================================
!
!  Calculates the Absolute Salinity and Conservative Temperature that 
!  results when a given mass of sea ice (or ice) melts and is mixed into a 
!  known mass of seawater (whose properties are (SA,CT,p)).  
!
!  If the ice contains no salt (e.g. if it is of glacial origin), then the 
!  input 'SA_seaice' should be set to zero.  
!
!  Ice formed at the sea surface (sea ice) typically contains between 2 g/kg
!  and 12 g/kg of salt (defined as the mass of salt divided by the mass of 
!  ice Ih plus brine) and this programme returns NaN's if the input  
!  SA_seaice is greater than 15 g/kg.  If the SA_seaice input is not zero,   
!  usually this would imply that the pressure p should be zero, as sea ice  
!  only occurs near the sea surface.  The code does not impose that p = 0 
!  if SA_seaice is non-zero.  Rather, this is left to the user.  
!
!  The Absolute Salinity, SA_brine, of the brine trapped in little pockets 
!  in the sea ice, is in thermodynamic equilibrium with the ice Ih that
!  surrounds these pockets.  As the sea ice temperature, t_seaice, may be 
!  less than the freezing temperature, SA_brine is usually greater than the
!  Absolute Salinity of the seawater at the time and place when and where 
!  the sea ice was formed.  So usually SA_brine will be larger than SA.  
!
!  SA  =  Absolute Salinity of seawater                            [ g/kg ]
!  CT  =  Conservative Temperature of seawater (ITS-90)           [ deg C ]
!  p   =  sea pressure at which the melting occurs                 [ dbar ]
!         ( i.e. absolute pressure - 10.1325 dbar ) 
!  w_seaice  =  mass fraction of sea ice, that is the mass of sea ice 
!               divided by the sum of the masses of sea ice and seawater. 
!               That is, the mass of sea ice divided by the mass of the 
!               final mixed fluid.  w_seaice must be between 0 and 1. 
!                                                              [ unitless ]
!  SA_seaice =  Absolute Salinity of sea ice, that is, the mass fraction of
!               salt in sea ice, expressed in g of salt per kg of sea ice.
!                                                                  [ g/kg ]
!  t_seaice  =  the in-situ temperature of the sea ice (or ice) (ITS-90)
!                                                                 [ deg C ]
!
!  SA_final  =  Absolute Salinity of the mixture of the melted sea ice 
!               (or ice) and the orignal seawater                  [ g/kg ]
!  CT_final  =  Conservative Temperature of the mixture of the melted 
!               sea ice (or ice) and the orignal seawater         [ deg C ]
!--------------------------------------------------------------------------

use gsw_mod_toolbox, only : gsw_ct_freezing_exact, gsw_enthalpy_ct_exact
use gsw_mod_toolbox, only : gsw_t_freezing_exact
use gsw_mod_toolbox, only : gsw_enthalpy_t_exact, gsw_enthalpy_ice
use gsw_mod_toolbox, only : gsw_sa_freezing_from_t, gsw_ct_from_enthalpy_exact

use gsw_mod_error_functions, only : gsw_error_code, gsw_error_limit

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sa, ct, p, w_seaice, sa_seaice, t_seaice
real (r8), intent(out) :: sa_final, ct_final

real (r8) :: ctf, h, h_brine, h_final, h_ih, sa_brine, tf_sa_seaice

real (r8), parameter :: saturation_fraction = 0.0_r8

character (*), parameter :: func_name = "gsw_melting_seaice_into_seawater"

ctf = gsw_ct_freezing_exact(sa,p,saturation_fraction)
if (ct .lt. ctf) then
    ! The seawater ct input is below the freezing temp
    sa_final = gsw_error_code(1,func_name)
    ct_final = sa_final
    return
end if

tf_sa_seaice = gsw_t_freezing_exact(sa_seaice,p,saturation_fraction) - 1e-6_r8
if (t_seaice .gt. tf_sa_seaice) then
    ! The 1e-6 C buffer in the allowable t_seaice is to ensure that there is
    ! some ice Ih in the sea ice. Without this buffer, that is if t_seaice
    ! is allowed to be exactly equal to tf_sa_seaice, the seaice is 
    ! actually 100% brine at Absolute Salinity of SA_seaice.
    sa_final = gsw_error_code(2,func_name)
    ct_final = sa_final
    return
end if

sa_brine = gsw_sa_freezing_from_t(t_seaice,p,saturation_fraction)
if (sa_brine .gt. gsw_error_limit) then
    sa_final = gsw_error_code(3,func_name,sa_brine)
    ct_final = sa_final
    return
end if
h_brine = gsw_enthalpy_t_exact(sa_brine,t_seaice,p)

h = gsw_enthalpy_ct_exact(sa,ct,p)
h_ih = gsw_enthalpy_ice(t_seaice,p)

h_final = h - w_seaice*(h - h_ih - (h_brine - h_ih)*sa_seaice/sa_brine)

sa_final = sa - w_seaice*(sa - sa_seaice)

!ctf = gsw_ct_freezing_exact(sa_final,p,saturation_fraction)
!
!if (h_final .lt. gsw_enthalpy_ct_exact(sa_final,ctf,p)) then
!    ! Melting this much seaice is not possible as it would result in
!    ! frozen seawater
!    sa_final = gsw_error_code(4,func_name)
!    ct_final = sa_final
!    return
!end if

ct_final = gsw_ct_from_enthalpy_exact(sa_final,h_final,p)
if (ct_final .gt. gsw_error_limit) then
    sa_final = ct_final
    return
endif

return
end subroutine

!--------------------------------------------------------------------------
!==========================================================================
elemental function gsw_melting_seaice_sa_ct_ratio (sa, ct, p, sa_seaice, &
                                                   t_seaice)
!==========================================================================
!
! Calculates the ratio of SA to CT changes when sea ice melts into seawater.
! It is assumed that a small mass of sea ice melts into an infinite mass of
! seawater.  Because of the infinite mass of seawater, the sea ice will 
! always melt.   
!
! Ice formed at the sea surface (sea ice) typically contains between 2 g/kg
! and 12 g/kg of salt (defined as the mass of salt divided by the mass of 
! ice Ih plus brine) and this programme returns NaN's if the input 
! SA_seaice is greater than 15 g/kg.  If the SA_seaice input is not zero, 
! usually this would imply that the pressure p should be zero, as sea ice 
! only occurs near the sea surface.  The code does not impose that p = 0 if 
! SA_seaice is non-zero.  Rather, this is left to the user.  
!
! The Absolute Salinity, SA_brine, of the brine trapped in little pockets 
! in the sea ice, is in thermodynamic equilibrium with the ice Ih that
! surrounds these pockets.  As the seaice temperature, t_seaice, may be 
! less than the freezing temperature, SA_brine is usually greater than the
! Absolute Salinity of the seawater at the time and place when and where 
! the sea ice was formed.  So usually SA_brine will be larger than SA.  
!
! The output, melting_seaice_SA_CT_ratio, is dSA/dCT rather than dCT/dSA. 
! This is done so that when (SA - seaice_SA) = 0, the output, dSA/dCT is 
! zero whereas dCT/dSA would be infinite. 
!
!  SA  =  Absolute Salinity of seawater                            [ g/kg ]
!  CT  =  Conservative Temperature of seawater (ITS-90)           [ deg C ]
!  p   =  sea pressure at which the melting occurs                 [ dbar ]
!         ( i.e. absolute pressure - 10.1325 dbar ) 
!  SA_seaice  =  Absolute Salinity of sea ice, that is, the mass fraction 
!                of salt in sea ice expressed in g of salt per kg of 
!                sea ice                                           [ g/kg ]
!  t_seaice = the in-situ temperature of the sea ice (ITS-90)     [ deg C ]
!
!  melting_seaice_SA_CT_ratio = the ratio dSA/dCT of SA to CT changes when
!                sea ice melts into a large mass of seawater   [ g/(kg K) ]
!--------------------------------------------------------------------------

use gsw_mod_toolbox, only : gsw_ct_freezing_exact, gsw_enthalpy_ct_exact
use gsw_mod_toolbox, only : gsw_sa_freezing_from_t, gsw_enthalpy_ice
use gsw_mod_toolbox, only : gsw_enthalpy_t_exact, gsw_t_freezing_exact
use gsw_mod_toolbox, only : gsw_enthalpy_first_derivatives_ct_exact

use gsw_mod_error_functions, only : gsw_error_code, gsw_error_limit

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sa, ct, p, sa_seaice, t_seaice

real (r8) :: gsw_melting_seaice_sa_ct_ratio

real (r8) :: ctf, delsa, h, h_brine, h_ih, sa_brine
real (r8) :: tf_sa_seaice, h_hat_sa, h_hat_ct

real (r8), parameter :: saturation_fraction = 0.0_r8

character (*), parameter :: func_name = "gsw_melting_seaice_sa_ct_ratio"

if (sa_seaice .lt. 0.0_r8 .or. sa_seaice .gt. 15.0_r8) then
    gsw_melting_seaice_sa_ct_ratio = gsw_error_code(1,func_name)
    return
end if

ctf = gsw_ct_freezing_exact(sa,p,saturation_fraction)
if (ct .lt. ctf) then    ! the seawater ct input is below the freezing temp
    gsw_melting_seaice_sa_ct_ratio = gsw_error_code(2,func_name)
    return
end if

!--------------------------------------------------------------------------
tf_sa_seaice = gsw_t_freezing_exact(sa_seaice,p,saturation_fraction) - 1e-6_r8
if (t_seaice .gt. tf_sa_seaice) then   ! t_seaice exceeds the freezing sa
    gsw_melting_seaice_sa_ct_ratio = gsw_error_code(3,func_name)
    return
end if
!
! The 1e-6 C buffer in the allowable t_seaice is to ensure that there is
! some ice Ih in the sea ice.  Without this buffer, that is if t_seaice
! is allowed to be exactly equal to tf_sa_seaice, the sea ice is actually
! 100% brine at Absolute Salinity of SA_seaice.
!--------------------------------------------------------------------------

h = gsw_enthalpy_ct_exact(sa,ct,p)
h_ih = gsw_enthalpy_ice(t_seaice,p)
call gsw_enthalpy_first_derivatives_ct_exact(sa,ct,p,h_hat_sa,h_hat_ct)

sa_brine = gsw_sa_freezing_from_t(t_seaice,p,saturation_fraction)
if (sa_brine .gt. gsw_error_limit) then
    gsw_melting_seaice_sa_ct_ratio = gsw_error_code(4,func_name,sa_brine)
    return
end if
h_brine = gsw_enthalpy_t_exact(sa_brine,t_seaice,p)
delsa = sa - sa_seaice

gsw_melting_seaice_sa_ct_ratio = h_hat_ct*delsa / &
              (h - h_ih - delsa*h_hat_sa - sa_seaice*(h_brine - h_ih)/sa_brine)

return
end function
!==========================================================================
elemental function gsw_melting_seaice_sa_ct_ratio_poly (sa, ct, p, &
                                                       sa_seaice, t_seaice)
!==========================================================================
!
! Calculates the ratio of SA to CT changes when sea ice melts into seawater.
! It is assumed that a small mass of sea ice melts into an infinite mass of
! seawater.  Because of the infinite mass of seawater, the sea ice will 
! always melt.   
!
! Ice formed at the sea surface (sea ice) typically contains between 2 g/kg
! and 12 g/kg of salt (defined as the mass of salt divided by the mass of 
! ice Ih plus brine) and this programme returns NaN's if the input 
! SA_seaice is greater than 15 g/kg.  If the SA_seaice input is not zero, 
! usually this would imply that the pressure p should be zero, as sea ice 
! only occurs near the sea surface.  The code does not impose that p = 0 if 
! SA_seaice is non-zero.  Rather, this is left to the user.  
!
! The Absolute Salinity, SA_brine, of the brine trapped in little pockets 
! in the sea ice, is in thermodynamic equilibrium with the ice Ih that
! surrounds these pockets.  As the seaice temperature, t_seaice, may be 
! less than the freezing temperature, SA_brine is usually greater than the
! Absolute Salinity of the seawater at the time and place when and where 
! the sea ice was formed.  So usually SA_brine will be larger than SA.  
!
! The output, melting_seaice_SA_CT_ratio, is dSA/dCT rather than dCT/dSA. 
! This is done so that when (SA - seaice_SA) = 0, the output, dSA/dCT is 
! zero whereas dCT/dSA would be infinite. 
!
!  SA  =  Absolute Salinity of seawater                            [ g/kg ]
!  CT  =  Conservative Temperature of seawater (ITS-90)           [ deg C ]
!  p   =  sea pressure at which the melting occurs                 [ dbar ]
!         ( i.e. absolute pressure - 10.1325 dbar ) 
!  SA_seaice  =  Absolute Salinity of sea ice, that is, the mass fraction 
!                of salt in sea ice expressed in g of salt per kg of 
!                sea ice                                           [ g/kg ]
!  t_seaice = the in-situ temperature of the sea ice (ITS-90)     [ deg C ]
!
!  melting_seaice_SA_CT_ratio = the ratio dSA/dCT of SA to CT changes when
!                sea ice melts into a large mass of seawater   [ g/(kg K) ]
!--------------------------------------------------------------------------

use gsw_mod_toolbox, only : gsw_ct_freezing_poly, gsw_enthalpy
use gsw_mod_toolbox, only : gsw_sa_freezing_from_t_poly, gsw_enthalpy_ice
use gsw_mod_toolbox, only : gsw_ct_from_t, gsw_t_freezing_poly
use gsw_mod_toolbox, only : gsw_enthalpy_first_derivatives

use gsw_mod_error_functions, only : gsw_error_code, gsw_error_limit

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sa, ct, p, sa_seaice, t_seaice

real (r8) :: gsw_melting_seaice_sa_ct_ratio_poly

real (r8) :: ctf, delsa, h, h_brine, h_ih, sa_brine, ct_brine
real (r8) :: tf_sa_seaice, h_hat_sa, h_hat_ct

real (r8), parameter :: saturation_fraction = 0.0_r8

character (*), parameter :: func_name = "gsw_melting_seaice_sa_ct_ratio_poly"

if (sa_seaice .lt. 0.0_r8 .or. sa_seaice .gt. 15.0_r8) then
    gsw_melting_seaice_sa_ct_ratio_poly = gsw_error_code(1,func_name)
    return
end if

ctf = gsw_ct_freezing_poly(sa,p,saturation_fraction)
if (ct .lt. ctf) then    ! the seawater ct input is below the freezing temp
    gsw_melting_seaice_sa_ct_ratio_poly = gsw_error_code(2,func_name)
    return
end if

!--------------------------------------------------------------------------
tf_sa_seaice = gsw_t_freezing_poly(sa_seaice,p,saturation_fraction) - 1e-6_r8
if (t_seaice .gt. tf_sa_seaice) then   ! t_seaice exceeds the freezing sa
    gsw_melting_seaice_sa_ct_ratio_poly = gsw_error_code(3,func_name)
    return
end if
!
! The 1e-6 C buffer in the allowable t_seaice is to ensure that there is
! some ice Ih in the sea ice.  Without this buffer, that is if t_seaice
! is allowed to be exactly equal to tf_sa_seaice, the sea ice is actually
! 100% brine at Absolute Salinity of SA_seaice.
!--------------------------------------------------------------------------

h = gsw_enthalpy(sa,ct,p)
h_ih = gsw_enthalpy_ice(t_seaice,p)
call gsw_enthalpy_first_derivatives(sa,ct,p,h_hat_sa,h_hat_ct)

sa_brine = gsw_sa_freezing_from_t_poly(t_seaice,p,saturation_fraction)
if (sa_brine .gt. gsw_error_limit) then
    gsw_melting_seaice_sa_ct_ratio_poly = gsw_error_code(4,func_name,sa_brine)
    return
end if
ct_brine = gsw_ct_from_t(sa_brine,t_seaice,p)
h_brine = gsw_enthalpy(sa_brine,ct_brine,p)
delsa = sa - sa_seaice

gsw_melting_seaice_sa_ct_ratio_poly = h_hat_ct*delsa / &
              (h - h_ih - delsa*h_hat_sa - sa_seaice*(h_brine - h_ih)/sa_brine)

return
end function
!==========================================================================
pure function gsw_mlp (sa, ct, p)
!==========================================================================
! 
!  Calculates the mixed-layer pressure as described in de Boyer Montgut 
!  et al. (2004).  The mlp is always deeper than 20 dbar, if the initial
!  estimate of the mlp is less than 20 dbar, the temperature and salinity  
!  of the bottles in the top 5 dbar are set to that of the bottle closest 
!  to 5 dbar.  This removes the effect if a thin layer of fresh water, 
!  such as that from a river outflow or from rain.
!
!  SA  =  Absolute Salinity                                        [ g/kg ]
!  CT  =  Conservative Temperature (ITS-90)                       [ deg C ]
!  p   =  sea pressure                                             [ dbar ]
!         ( i.e. absolute pressure - 10.1325 dbar )
!
!  mlp  =  mixed-layer pressure                                    [ dbar ]
!  
!==========================================================================

use gsw_mod_toolbox, only : gsw_rho

use gsw_mod_error_functions, only : gsw_error_code

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sa(:), ct(:), p(:)

real (r8) :: gsw_mlp

integer :: k(1), np

real (r8) :: min_p

real (r8), allocatable :: diff_rho0(:), dp(:), rho0(:)

character (*), parameter :: func_name = "gsw_mlp"

np = size(p)
allocate(dp(np-1))
dp = p(2:np) - p(1:np-1)
if (any(dp .le. 0.0_r8)) then  ! pressure must be monotonic and unique
    gsw_mlp = gsw_error_code(1,func_name)
    return
end if

min_p = minval(p)
if (min_p .gt. 20.0_r8) then  ! the profile starts at p greater than 20 dbar
    gsw_mlp = gsw_error_code(2,func_name)
    return
end if

allocate(rho0(np),diff_rho0(np))

rho0 = gsw_rho(sa,ct,0.0_r8)

diff_rho0 = (minval(rho0) + 0.3_r8) - rho0
k = minloc(diff_rho0, diff_rho0 .gt. 0.0_r8)
gsw_mlp = p(k(1))

if ((gsw_mlp-min_p) .lt. 20.0_r8) then

    ! If the mlp is less than 20 dbar it is possible that this density 
    ! difference is being effected by a thin fresh water layer at the surface,
    ! set the salinities and temperatures of the bottles in the top section of
    ! the cast to be equal to that of the bottle closest to 5 dbar.

    k = minloc(abs(p - 5.0_r8))
    rho0(1:k(1)-1) = rho0(k(1))

    diff_rho0 = (minval(rho0) + 0.3_r8) - rho0
    k = minloc(diff_rho0, diff_rho0 .gt. 0.0_r8)
    gsw_mlp = p(k(1))        

    if ((gsw_mlp-min_p) .lt. 20.0_r8) then
        gsw_mlp = gsw_error_code(3,func_name)
        return
    end if

end if

return
end function

!--------------------------------------------------------------------------
!==========================================================================
pure subroutine gsw_nsquared (sa, ct, p, lat, n2, p_mid)
!==========================================================================
!
!  Calculates the buoyancy frequency squared (N^2)(i.e. the Brunt-Vaisala 
!  frequency squared) at the mid pressure from the equation,
!
!
!           2      2             beta.d(SA) - alpha.d(CT)
!         N   =  g  .rho_local. -------------------------
!                                          dP
!
!  The pressure increment, dP, in the above formula is in Pa, so that it is
!  10^4 times the pressure increment dp in dbar. 
!
!  sa     : Absolute Salinity         (a profile (length nz))     [g/kg]
!  ct     : Conservative Temperature  (a profile (length nz))     [deg C]
!  p      : sea pressure              (a profile (length nz))     [dbar]
!  lat    : latitude                  (a profile (length nz))     [deg N]
!  n2     : Brunt-Vaisala Frequency squared  (length nz-1)        [s^-2]
!  p_mid  : Mid pressure between p grid      (length nz-1)        [dbar]
!--------------------------------------------------------------------------

use gsw_mod_toolbox, only : gsw_grav, gsw_rho_alpha_beta

use gsw_mod_teos10_constants, only : db2pa

use gsw_mod_error_functions, only : gsw_error_code

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sa(:), ct(:), p(:), lat(:)
real (r8), intent(out) :: n2(:), p_mid(:)

integer :: nz, k
real (r8), dimension(:), allocatable :: dsa, sa_mid, dct, ct_mid, dp, rho_mid
real (r8), dimension(:), allocatable :: alpha_mid, beta_mid, grav_local, grav

character (*), parameter :: func_name = "gsw_nsquared"

nz = size(sa)
if (size(n2).lt.nz-1 .or. size(p_mid).lt.nz-1) then
    n2 = gsw_error_code(1,func_name)
    p_mid = n2(1)
    return
end if

allocate (grav(nz), dsa(nz-1), sa_mid(nz-1), dct(nz-1), ct_mid(nz-1), dp(nz-1))
allocate (rho_mid(nz-1), alpha_mid(nz-1), beta_mid(nz-1), grav_local(nz-1))

grav = gsw_grav(lat(1:nz),p(1:nz))

forall (k = 1: nz-1)
    grav_local(k) = 0.5_r8*(grav(k) + grav(k+1))
    dsa(k) = (sa(k+1) - sa(k))
    sa_mid(k) = 0.5_r8*(sa(k) + sa(k+1))
    dct(k) = (ct(k+1) - ct(k))
    ct_mid(k) = 0.5_r8*(ct(k) + ct(k+1))
    dp(k) = (p(k+1) - p(k))
    p_mid(k) = 0.5_r8*(p(k) + p(k+1))
end forall

call gsw_rho_alpha_beta(sa_mid,ct_mid,p_mid(1:nz-1),rho_mid,alpha_mid,beta_mid)

n2(1:nz-1) = (grav_local**2) * (rho_mid/(db2pa*dp)) * &
             (beta_mid*dsa - alpha_mid*dct)

deallocate (grav, dsa, sa_mid, dct, ct_mid, dp)
deallocate (rho_mid, alpha_mid, beta_mid, grav_local)

return
end subroutine

!--------------------------------------------------------------------------
!==========================================================================
elemental function gsw_nsquared_lowerlimit (p, long, lat)
!==========================================================================
!
!  Calculates the minimum buoyancy (Brunt-Vaisala) frequency squared (N^2)
!  such that a cast is stable.
!
!  p  =  sea pressure                                              [ dbar ]
!         ( i.e. absolute pressure - 10.1325 dbar )
!  long =  longitude in decimal degrees                      [ 0 ... +360 ]
!                                                     or  [ -180 ... +180 ]
!  lat  =  latitude in decimal degrees north                [ -90 ... +90 ]
!
!  nsquared_lowerlimit  = Minimum Brunt-Vaisala Frequency squared [ 1/s^2 ]
!
!==========================================================================

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: p, long, lat

real (r8) :: gsw_nsquared_lowerlimit

real (r8) :: long360

if (long .lt. 0.0_r8) then
    long360 = long + 360.0_r8
else
    long360 = long
end if

gsw_nsquared_lowerlimit = (0.25_r8 + 0.75_r8*(exp(-p/1000.0_r8))) * 1e-7_r8

return
end function

!--------------------------------------------------------------------------
!==========================================================================
pure subroutine gsw_nsquared_min (sa, ct, p, lat, n2, n2_p, &
                               n2_specvol, n2_alpha, n2_beta, dsa, dct, dp)
!==========================================================================
! 
!  Calculates the minimum buoyancy frequency squared (N^2) (i.e. the 
!  Brunt-Vaisala frequency squared) between two bottles from the equation,
!
!           2      2     beta.dSA - alpha.dCT
!         N   =  g  . -------------------------
!                         specvol_local.dP
!
!  The pressure increment, dP, in the above formula is in Pa, so that it is
!  10^4 times the pressure increment dp in dbar. 
!
!  SA  =  Absolute Salinity                                        [ g/kg ]
!  CT  =  Conservative Temperature (ITS-90)                       [ deg C ]
!  p   =  sea pressure                                             [ dbar ]
!         ( i.e. absolute pressure - 10.1325 dbar )
!  lat =  latitude in decimal degrees north                 [ -90 ... +90 ]
!
!  N2         =  minimum Brunt-Vaisala Frequency squared          [ 1/s^2 ]
!  N2_p       =  pressure of minimum N2                            [ dbar ]
!  N2_specvol =  specific volume at the minimum N2                [ kg/m3 ]
!  N2_alpha   =  thermal expansion coefficient with respect         [ 1/K ]
!                to Conservative Temperature at the minimum N2
!  N2_beta    =  saline contraction coefficient at constant        [ kg/g ]
!                Conservative Temperature at the minimum N2
!  dSA        =  difference in salinity between bottles            [ g/kg ]
!  dCT        =  difference in Conservative Temperature between   [ deg C ]
!                bottles
!  dp         =  difference in pressure between bottles            [ dbar ]
!
!==========================================================================

use gsw_mod_toolbox, only : gsw_grav, gsw_specvol_alpha_beta

use gsw_mod_error_functions, only : gsw_error_code

use gsw_mod_kinds

use gsw_mod_teos10_constants, only : db2pa

implicit none

real (r8), intent(in) :: sa(:), ct(:), p(:), lat(:)
real (r8), intent(out) :: n2(:), n2_p(:), n2_specvol(:), n2_alpha(:)
real (r8), intent(out) :: n2_beta(:), dsa(:), dct(:), dp(:)

integer :: i, ideep, ishallow, mp

real (r8) :: n2_deep, n2_shallow
real (r8), allocatable :: alpha(:), beta(:), specvol(:), grav2(:)

character (*), parameter :: func_name = "gsw_nsquared_min"

mp = size(sa)
if (size(n2).lt.mp-1 .or. size(n2_p).lt.mp-1 .or. &
    size(n2_specvol).lt.mp-1 .or. size(n2_alpha).lt.mp-1 .or. &
    size(n2_beta).lt.mp-1 .or. size(dsa).lt.mp-1 .or. &
    size(dct).lt.mp-1 .or. size(dp).lt.mp-1) then
    n2 = gsw_error_code(1,func_name)
    n2_p = n2(1)
    n2_specvol = n2(1)
    n2_alpha = n2(1)
    n2_beta = n2(1)
    dsa = n2(1)
    dct = n2(1)
    dp = n2(1)
    return
end if

allocate(grav2(mp),specvol(mp),alpha(mp),beta(mp))

grav2 = gsw_grav(lat,p)**2

dp(1:mp-1)  =  p(2:mp) -  p(1:mp-1)
dsa(1:mp-1) = sa(2:mp) - sa(1:mp-1)
dct(1:mp-1) = ct(2:mp) - ct(1:mp-1)

call gsw_specvol_alpha_beta(sa,ct,p,specvol,alpha,beta)

ishallow = 1
ideep = 2
do i = 1, mp-1
    n2_shallow =  grav2(ishallow)/(specvol(ishallow)*db2pa*dp(i))* &
                 (beta(ishallow)*dsa(i) - alpha(ishallow)*dct(i))
    n2_deep =  grav2(ideep)/(specvol(ideep)*db2pa*dp(i))* &
              (beta(ideep)*dsa(i) - alpha(ideep)*dct(i))
    if (n2_shallow .lt. n2_deep) then
        n2(i) = n2_shallow
        n2_p(i) = p(ishallow)
        n2_specvol(i) = specvol(ishallow)
        n2_alpha(i) = alpha(ishallow)
        n2_beta(i) = beta(ishallow)
    else
        n2(i) = n2_deep
        n2_p(i) = p(ideep)
        n2_specvol(i) = specvol(ideep)
        n2_alpha(i) = alpha(ideep)
        n2_beta(i) = beta(ideep)
    end if
    ishallow = ishallow + 1
    ideep = ideep + 1
end do

return
end subroutine

!--------------------------------------------------------------------------
!==========================================================================
pure subroutine gsw_nsquared_min_const_t (sa, t, p, lat, n2, n2_p, &
                n2_specvol, n2_alpha, n2_beta, dsa, dct, dp, n2_beta_ratio)
!==========================================================================
! 
!  Calculates the minimum buoyancy frequency squared (N^2) (i.e. the 
!  Brunt-Vaisala frequency squared) between two bottles from the equation,
!
!           2      2     beta.dSA - alpha.dCT
!         N   =  g  . -------------------------
!                         specvol_local.dP
!
!  The pressure increment, dP, in the above formula is in Pa, so that it is
!  10^4 times the pressure increment dp in dbar. 
!
!  SA  =  Absolute Salinity                                        [ g/kg ]
!  t   =  In-situ temperature (ITS-90)                            [ deg C ]
!  p   =  Sea pressure                                             [ dbar ]
!         ( i.e. absolute pressure - 10.1325 dbar )
!  lat =  Latitude in decimal degrees north                 [ -90 ... +90 ]
!  Note: If lat is outside this range, a default gravitational
!        acceleration of 9.7963 m/s^2 (Griffies, 2004) will be applied.
!
!  N2         =  minimum Brunt-Vaisala Frequency squared          [ 1/s^2 ]
!  N2_p       =  pressure of minimum N2                            [ dbar ]
!  N2_specvol =  specific volume at the minimum N2                [ kg/m3 ]
!  N2_alpha   =  thermal expansion coefficient with respect         [ 1/K ]
!                to Conservative Temperature at the minimum N2
!  N2_beta    =  saline contraction coefficient at constant        [ kg/g ]
!                Conservative Temperature at the minimum N2
!  dSA        =  difference in salinity between bottles            [ g/kg ]
!  dCT        =  difference in Conservative Temperature between   [ deg C ]
!                bottles
!  dp         =  difference in pressure between bottles            [ dbar ]
!  N2_beta_ratio = ratio of the saline contraction             [ unitless ]
!                coefficient at constant Conservative Temperature to  
!                the saline contraction coefficient at constant in-situ  
!                temperature at the minimum N2
!
!==========================================================================

use gsw_mod_toolbox, only : gsw_grav, gsw_specvol_alpha_beta
use gsw_mod_toolbox, only : gsw_ct_from_t, gsw_beta_const_t_exact

use gsw_mod_kinds

use gsw_mod_teos10_constants, only : db2pa

implicit none

real (r8), intent(in) :: sa(:), t(:), p(:), lat
real (r8), intent(out) :: n2(:), n2_p(:), n2_specvol(:), n2_alpha(:)
real (r8), intent(out) :: n2_beta(:), dsa(:), dct(:), dp(:)
real (r8), intent(out) :: n2_beta_ratio(:)

integer :: i, ideep, ishallow, mp

real (r8) :: n2_deep, n2_shallow
real (r8), allocatable :: alpha(:), beta(:), specvol(:), grav2(:)
real (r8), allocatable :: ct(:), beta_ratio(:)

mp = size(sa)
allocate(grav2(mp),specvol(mp),alpha(mp),beta(mp),ct(mp),beta_ratio(mp))

if (lat .lt. -90.0_r8 .or. lat .gt. +90.0_r8) then
    grav2 = (/ (9.7963_r8**2, i=1,mp) /)
else
    grav2 = gsw_grav(lat,p)**2
end if

ct = gsw_ct_from_t(sa,t,p)

dp  =  p(2:mp) -  p(1:mp-1)
dsa = sa(2:mp) - sa(1:mp-1)
dct = ct(2:mp) - ct(1:mp-1)

call gsw_specvol_alpha_beta(sa,ct,p,specvol,alpha,beta)

beta_ratio = beta/gsw_beta_const_t_exact(sa,t,p)

ishallow = 1
ideep = 2
do i = 1, mp-1
    n2_shallow =  grav2(ishallow)/(specvol(ishallow)*db2pa*dp(i))* &
                 (beta(ishallow)*dsa(i) - alpha(ishallow)*dct(i))
    n2_deep =  grav2(ideep)/(specvol(ideep)*db2pa*dp(i))* &
              (beta(ideep)*dsa(i) - alpha(ideep)*dct(i))
    if (n2_shallow .lt. n2_deep) then
        n2(i) = n2_shallow
        n2_p(i) = p(ishallow)
        n2_specvol(i) = specvol(ishallow)
        n2_alpha(i) = alpha(ishallow)
        n2_beta(i) = beta(ishallow)
        n2_beta_ratio(i) = beta_ratio(ishallow)
    else
        n2(i) = n2_deep
        n2_p(i) = p(ideep)
        n2_specvol(i) = specvol(ideep)
        n2_alpha(i) = alpha(ideep)
        n2_beta(i) = beta(ideep)
        n2_beta_ratio(i) = beta_ratio(ideep)
    end if
    ishallow = ishallow + 1
    ideep = ideep + 1
end do

return
end subroutine

!--------------------------------------------------------------------------
!==========================================================================
elemental function gsw_p_from_z (z, lat, geo_strf_dyn_height, &
                                                   sea_surface_geopotental)
!==========================================================================
!
!  Calculates sea pressure from height using computationally-efficient 
!  75-term expression for density, in terms of SA, CT and p (Roquet et al.,
!  2015).  Dynamic height anomaly, geo_strf_dyn_height, if provided,
!  must be computed with its p_ref = 0 (the surface). Also if provided,
!  sea_surface_geopotental is the geopotential at zero sea pressure. This 
!  function solves Eqn.(3.32.3) of IOC et al. (2010) iteratively for p.  
!
!  Note. Height (z) is NEGATIVE in the ocean.  Depth is -z.  
!    Depth is not used in the GSW computer software library. 
!
!  z  =  height                                                       [ m ]
!  lat  =  latitude in decimal degrees north                [ -90 ... +90 ]
!   
! OPTIONAL:
!  geo_strf_dyn_height = dynamic height anomaly                 [ m^2/s^2 ]
!    Note that the reference pressure, p_ref, of geo_strf_dyn_height must
!     be zero (0) dbar.
!  sea_surface_geopotental = geopotential at zero sea pressure  [ m^2/s^2 ]
!
!  p  =  sea pressure                                              [ dbar ]
!        ( i.e. absolute pressure - 10.1325 dbar )
!==========================================================================

use gsw_mod_toolbox, only : gsw_enthalpy_sso_0, gsw_specvol_sso_0

use gsw_mod_teos10_constants, only : db2pa, deg2rad, gamma

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: z, lat
real (r8), intent(in), optional :: geo_strf_dyn_height
real (r8), intent(in), optional :: sea_surface_geopotental

real (r8) :: gsw_p_from_z

real (r8) :: c1, df_dp, f, g, gs, p, p_mid, p_old, sin2
real (r8) :: gsdh, ssg

if (present(geo_strf_dyn_height)) then
    gsdh = geo_strf_dyn_height
else
    gsdh = 0.0_r8
end if

if (present(sea_surface_geopotental)) then
    ssg = sea_surface_geopotental
else
    ssg = 0.0_r8
end if

g = gamma    ! If the graviational acceleration were to be regarded as 
             ! being depth-independent, which is often the case in 
             ! ocean models, then g would be set to be zero here,
             ! and the code below works perfectly well.  

sin2 = sin(lat*deg2rad)**2
gs = 9.780327_r8*(1.0_r8 + (5.2792e-3_r8 + (2.32e-5_r8*sin2))*sin2)

! Get the first estimate of p from Saunders (1981)
c1 =  5.25e-3_r8*sin2 + 5.92e-3_r8
p  = -2.0_r8*z/((1.0_r8-c1) + sqrt((1.0_r8-c1)*(1.0_r8-c1) + 8.84e-6_r8*z)) 

! Initial value of the derivative of f
df_dp = db2pa*gsw_specvol_sso_0(p)

f = gsw_enthalpy_sso_0(p) + gs*(z - 0.5_r8*g*(z*z)) - (gsdh + ssg)
p_old = p
p = p_old - f/df_dp
p_mid = 0.5_r8*(p + p_old)
df_dp = db2pa*gsw_specvol_sso_0(p_mid)
p = p_old - f/df_dp

! After this one iteration through this modified Newton-Raphson iterative
! procedure (McDougall and Wotherspoon, 2013), the remaining error in p is 
! at computer machine precision, being no more than 1.6e-10 dbar. 

gsw_p_from_z = p

return
end function

!--------------------------------------------------------------------------
!==========================================================================
elemental function gsw_pot_enthalpy_from_pt_ice (pt0_ice)
!==========================================================================
!
!  Calculates the potential enthalpy of ice from potential temperature of
!  ice (whose reference sea pressure is zero dbar).  
!
!  pt0_ice  =  potential temperature of ice (ITS-90)              [ deg C ]
!
!  gsw_pot_enthalpy_ice  =  potential enthalpy of ice              [ J/kg ]
!--------------------------------------------------------------------------

use gsw_mod_teos10_constants, only : gsw_t0

use gsw_mod_gibbs_ice_coefficients

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: pt0_ice

real (r8) :: gsw_pot_enthalpy_from_pt_ice

real (r8) :: tau
complex (r8) :: h0_part, sqtau_t1, sqtau_t2

tau = (pt0_ice + gsw_t0)*rec_t3p

sqtau_t1 = (tau/t1)**2
sqtau_t2 = (tau/t2)**2

h0_part = r1*t1*(log(1.0_r8 - sqtau_t1) + sqtau_t1) &
          + r20*t2*(log(1.0_r8 - sqtau_t2) + sqtau_t2)

gsw_pot_enthalpy_from_pt_ice = g00 + t3p*real(h0_part,r8) 
              
return
end function

!--------------------------------------------------------------------------
!==========================================================================
elemental function gsw_pot_enthalpy_from_pt_ice_poly (pt0_ice)
!==========================================================================
!
!  Calculates the potential enthalpy of ice from potential temperature of
!  ice (whose reference sea pressure is zero dbar).  This is a
!  compuationally efficient polynomial fit to the potential enthalpy of
!  ice.
!   
!  pt0_ice  =  potential temperature of ice (ITS-90)              [ deg C ]
!
!  pot_enthalpy_ice  =  potential enthalpy of ice                  [ J/kg ]
!--------------------------------------------------------------------------

use gsw_mod_toolbox, only : gsw_pt_from_pot_enthalpy_ice_poly
use gsw_mod_toolbox, only : gsw_pt_from_pot_enthalpy_ice_poly_dh

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: pt0_ice

real (r8) :: gsw_pot_enthalpy_from_pt_ice_poly

integer :: iteration
real (r8) :: df_dt, f, pot_enthalpy_ice
real (r8) :: pot_enthalpy_ice_mid, pot_enthalpy_ice_old

real (r8), parameter :: p0 = -3.333601570157700e5_r8
real (r8), parameter :: p1 =  2.096693916810367e3_r8
real (r8), parameter :: p2 =  3.687110754043292_r8
real (r8), parameter :: p3 =  4.559401565980682e-4_r8
real (r8), parameter :: p4 = -2.516011957758120e-6_r8
real (r8), parameter :: p5 = -1.040364574632784e-8_r8
real (r8), parameter :: p6 = -1.701786588412454e-10_r8
real (r8), parameter :: p7 = -7.667191301635057e-13_r8
    
! initial estimate of the potential enthalpy.
pot_enthalpy_ice = p0 + pt0_ice*(p1 + pt0_ice*(p2 + pt0_ice*(p3 &
                   + pt0_ice*(p4 + pt0_ice*(p5 + pt0_ice*(p6 &
                   + pt0_ice*p7))))))
               
df_dt = gsw_pt_from_pot_enthalpy_ice_poly_dh(pot_enthalpy_ice)

do iteration = 1, 5
    pot_enthalpy_ice_old = pot_enthalpy_ice
    f = gsw_pt_from_pot_enthalpy_ice_poly(pot_enthalpy_ice_old) - pt0_ice
    pot_enthalpy_ice = pot_enthalpy_ice_old - f/df_dt
    pot_enthalpy_ice_mid = 0.5_r8*(pot_enthalpy_ice + pot_enthalpy_ice_old)
    df_dt = gsw_pt_from_pot_enthalpy_ice_poly_dh(pot_enthalpy_ice_mid)
    pot_enthalpy_ice = pot_enthalpy_ice_old - f/df_dt
end do
               
! The error of this fit ranges between -6e-3 and 6e-3 J/kg over the potential 
! temperature range of -100 to 2 deg C, or the potential enthalpy range of 
! -5.7 x 10^5 to -3.3 x 10^5 J/kg. 
       
gsw_pot_enthalpy_from_pt_ice_poly = pot_enthalpy_ice

return
end function

!--------------------------------------------------------------------------
!==========================================================================
elemental function gsw_pot_enthalpy_ice_freezing (sa, p)
!==========================================================================
!
!  Calculates the potential enthalpy of ice at which seawater freezes.
!
!  SA  =  Absolute Salinity                                        [ g/kg ]
!  p   =  sea pressure                                             [ dbar ]
!         ( i.e. absolute pressure - 10.1325 dbar ) 
!
!  pot_enthalpy_ice_freezing = potential enthalpy of ice at freezing 
!                              of seawater                        [ deg C ]
!--------------------------------------------------------------------------

use gsw_mod_toolbox, only : gsw_pot_enthalpy_from_pt_ice, gsw_pt0_from_t_ice
use gsw_mod_toolbox, only : gsw_t_freezing_exact

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sa, p

real (r8) :: gsw_pot_enthalpy_ice_freezing

real (r8) :: pt0_ice, t_freezing

t_freezing = gsw_t_freezing_exact(sa,p,0.0_r8) 

pt0_ice = gsw_pt0_from_t_ice(t_freezing,p) 

gsw_pot_enthalpy_ice_freezing = gsw_pot_enthalpy_from_pt_ice(pt0_ice)

return
end function

!--------------------------------------------------------------------------
!==========================================================================
elemental subroutine gsw_pot_enthalpy_ice_freezing_first_derivatives (sa, &
              p, pot_enthalpy_ice_freezing_sa, pot_enthalpy_ice_freezing_p)
!==========================================================================
!
!  Calculates the first derivatives of the potential enthalpy of ice at
!  which seawater freezes, with respect to Absolute Salinity SA and
!  pressure P (in Pa).
!
!  SA  =  Absolute Salinity                                        [ g/kg ]
!  p   =  sea pressure                                             [ dbar ]
!         ( i.e. absolute pressure - 10.1325 dbar ) 
!
!  pot_enthalpy_ice_freezing_SA = the derivative of the potential enthalpy
!                  of ice at freezing (ITS-90) with respect to Absolute
!                  salinity at fixed pressure  [ K/(g/kg) ] i.e. [ K kg/g ]
!
!  pot_enthalpy_ice_freezing_P  = the derivative of the potential enthalpy
!                  of ice at freezing (ITS-90) with respect to pressure 
!                  (in Pa) at fixed Absolute Salinity              [ K/Pa ]
!--------------------------------------------------------------------------

use gsw_mod_toolbox, only : gsw_cp_ice, gsw_gibbs_ice, gsw_pt0_from_t_ice
use gsw_mod_toolbox, only : gsw_t_freezing_exact
use gsw_mod_toolbox, only : gsw_t_freezing_first_derivatives

use gsw_mod_teos10_constants, only : gsw_t0

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sa, p
real (r8), intent(out), optional :: pot_enthalpy_ice_freezing_sa
real (r8), intent(out), optional :: pot_enthalpy_ice_freezing_p

real (r8) :: cp_ihf, pt_icef, ratio_temp, tf, tf_p, tf_sa

real (r8), parameter :: saturation_fraction = 0.0_r8

tf = gsw_t_freezing_exact(sa,p,saturation_fraction)
pt_icef = gsw_pt0_from_t_ice(tf,p) 
ratio_temp = (gsw_t0 + pt_icef)/(gsw_t0 + tf)

cp_ihf = gsw_cp_ice(tf,p)

if (present(pot_enthalpy_ice_freezing_sa) .and. present(pot_enthalpy_ice_freezing_p)) then
    call gsw_t_freezing_first_derivatives(sa,p,saturation_fraction,tf_sa,tf_p)
else if (present(pot_enthalpy_ice_freezing_sa)) then
    call gsw_t_freezing_first_derivatives(sa,p,saturation_fraction, &
                                          tfreezing_sa=tf_sa)
else if (present(pot_enthalpy_ice_freezing_p)) then
    call gsw_t_freezing_first_derivatives(sa,p,saturation_fraction, &
                                          tfreezing_p=tf_p)
end if

if (present(pot_enthalpy_ice_freezing_sa)) &
                pot_enthalpy_ice_freezing_sa = ratio_temp*cp_ihf*tf_sa

if (present(pot_enthalpy_ice_freezing_p)) &
                pot_enthalpy_ice_freezing_p = ratio_temp*cp_ihf*tf_p &
                             - (gsw_t0 + pt_icef)*gsw_gibbs_ice(1,1,tf,p)

return
end subroutine

!--------------------------------------------------------------------------
!==========================================================================
elemental subroutine gsw_pot_enthalpy_ice_freezing_first_derivatives_poly(&
         sa, p, pot_enthalpy_ice_freezing_sa, pot_enthalpy_ice_freezing_p)
!==========================================================================
!
!  Calculates the first derivatives of the potential enthalpy of ice Ih at
!  which ice melts into seawater with Absolute Salinity SA and at pressure 
!  p.  This code uses the comptationally efficient polynomial fit of the
!  freezing potential enthalpy of ice Ih (McDougall et al., 2015).
!
!  SA  =  Absolute Salinity                                        [ g/kg ]
!  p   =  sea pressure                                             [ dbar ]
!         ( i.e. absolute pressure - 10.1325 dbar ) 
!
!  pot_enthalpy_ice_freezing_SA = the derivative of the potential enthalpy
!                of ice at freezing (ITS-90) with respect to Absolute
!                salinity at fixed pressure  [ (J/kg)/(g/kg) ] i.e. [ J/g ]
!                                            
!  pot_enthalpy_ice_freezing_P  = the derivative of the potential enthalpy
!                of ice at freezing (ITS-90) with respect to pressure 
!                (in Pa) at fixed Absolute Salinity           [ (J/kg)/Pa ]
!--------------------------------------------------------------------------

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sa, p
real (r8), intent(out), optional :: pot_enthalpy_ice_freezing_sa
real (r8), intent(out), optional :: pot_enthalpy_ice_freezing_p

real (r8) :: p_r, sa_r, x

real (r8), parameter :: d1 =  -1.249490228128056e4_r8
real (r8), parameter :: d2 =   1.336783910789822e4_r8
real (r8), parameter :: d3 =  -4.811989517774642e4_r8
real (r8), parameter :: d4 =   8.044864276240987e4_r8
real (r8), parameter :: d5 =  -7.124452125071862e4_r8
real (r8), parameter :: d6 =   2.280706828014839e4_r8
real (r8), parameter :: d7 =   0.315423710959628e3_r8
real (r8), parameter :: d8 =  -3.592775732074710e2_r8
real (r8), parameter :: d9 =   1.644828513129230e3_r8
real (r8), parameter :: d10 = -4.809640968940840e3_r8
real (r8), parameter :: d11 =  2.901071777977272e3_r8
real (r8), parameter :: d12 = -9.218459682855746e2_r8
real (r8), parameter :: d13 =  0.379377450285737e3_r8
real (r8), parameter :: d14 = -2.672164989849465e3_r8
real (r8), parameter :: d15 =  5.044317489422632e3_r8
real (r8), parameter :: d16 = -2.631711865886377e3_r8
real (r8), parameter :: d17 = -0.160245473297112e3_r8
real (r8), parameter :: d18 =  4.029061696035465e2_r8
real (r8), parameter :: d19 = -3.682950019675760e2_r8

real (r8), parameter :: f1 =  -2.034535061416256e4_r8
real (r8), parameter :: f2 =   0.315423710959628e3_r8
real (r8), parameter :: f3 =  -0.239518382138314e3_r8
real (r8), parameter :: f4 =   0.822414256564615e3_r8
real (r8), parameter :: f5 =  -1.923856387576336e3_r8
real (r8), parameter :: f6 =   0.967023925992424e3_r8
real (r8), parameter :: f7 =  -0.263384562367307e3_r8
real (r8), parameter :: f8 =  -5.051613740291480e3_r8
real (r8), parameter :: f9 =   7.587549005714740e2_r8
real (r8), parameter :: f10 = -3.562886653132620e3_r8
real (r8), parameter :: f11 =  5.044317489422632e3_r8
real (r8), parameter :: f12 = -2.105369492709102e3_r8
real (r8), parameter :: f13 =  6.387082316647800e2_r8
real (r8), parameter :: f14 = -4.807364198913360e2_r8
real (r8), parameter :: f15 =  8.058123392070929e2_r8
real (r8), parameter :: f16 = -5.524425029513641e2_r8

sa_r = sa*1e-2_r8
x = sqrt(sa_r)
p_r = p*1e-4_r8

if (present(pot_enthalpy_ice_freezing_sa)) pot_enthalpy_ice_freezing_sa = &
   (d1 + x*(d2  + x*(d3  + x*(d4  + x*(d5  + d6*x)))) &
       + p_r*(d7 + x*(d8 + x*(d9 + x*(d10 + x*(d11 + d12*x)))) & 
       + p_r*(d13 + x*(d14 + x*(d15 + d16*x)) &
       + p_r*(d17 + x*(d18 + d19*x)))))*1e-2_r8

if (present(pot_enthalpy_ice_freezing_p)) pot_enthalpy_ice_freezing_p = &
   (f1 + sa_r*(f2 + x*(f3 + x*(f4 + x*(f5 + x*(f6 + f7*x))))) &
       + p_r*(f8 + sa_r*(f9 + x*(f10 + x*(f11 + f12*x))) &
       + p_r*(f13 + sa_r*(f14 + x*(f15 + f16*x)))))*1e-8_r8

return
end subroutine

!--------------------------------------------------------------------------
!==========================================================================
elemental function gsw_pot_enthalpy_ice_freezing_poly (sa, p)
!==========================================================================
!
!  Calculates the potential enthalpy of ice at which seawater freezes.
!  The error of this fit ranges between -2.5 and 1 J/kg with an rms of 
!  1.07, between SA of 0 and 120 g/kg and p between 0 and 10,000 dbar (the
!  error in the fit is between -0.7 and 0.7 with an rms of
!  0.3, between SA of 0 and 120 g/kg and p between 0 and 5,000 dbar) when
!  compared with the potential enthalpy calculated from the exact in-situ 
!  freezing temperature which is found by a Newton-Raphson iteration of the 
!  equality of the chemical potentials of water in seawater and in ice.  
!
!  SA  =  Absolute Salinity                                        [ g/kg ]
!  p   =  sea pressure                                             [ dbar ]
!         ( i.e. absolute pressure - 10.1325 dbar ) 
!
!  pot_enthalpy_ice_freezing = potential enthalpy of ice at freezing 
!                              of seawater                         [ J/kg ]
!--------------------------------------------------------------------------

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sa, p

real (r8) :: gsw_pot_enthalpy_ice_freezing_poly

real (r8) :: p_r, sa_r, x

real (r8), parameter :: c0  = -3.333548730778702e5_r8
real (r8), parameter :: c1  = -1.249490228128056e4_r8
real (r8), parameter :: c2  =  0.891189273859881e4_r8
real (r8), parameter :: c3  = -2.405994758887321e4_r8
real (r8), parameter :: c4  =  3.217945710496395e4_r8
real (r8), parameter :: c5  = -2.374817375023954e4_r8
real (r8), parameter :: c6  =  0.651630522289954e4_r8
real (r8), parameter :: c7  = -2.034535061416256e4_r8
real (r8), parameter :: c8  = -0.252580687014574e4_r8
real (r8), parameter :: c9  =  0.021290274388826e4_r8
real (r8), parameter :: c10 =  0.315423710959628e3_r8
real (r8), parameter :: c11 = -0.239518382138314e3_r8
real (r8), parameter :: c12 =  0.379377450285737e3_r8
real (r8), parameter :: c13 =  0.822414256564615e3_r8
real (r8), parameter :: c14 = -1.781443326566310e3_r8
real (r8), parameter :: c15 = -0.160245473297112e3_r8
real (r8), parameter :: c16 = -1.923856387576336e3_r8
real (r8), parameter :: c17 =  2.522158744711316e3_r8
real (r8), parameter :: c18 =  0.268604113069031e3_r8
real (r8), parameter :: c19 =  0.967023925992424e3_r8
real (r8), parameter :: c20 = -1.052684746354551e3_r8
real (r8), parameter :: c21 = -0.184147500983788e3_r8
real (r8), parameter :: c22 = -0.263384562367307e3_r8

sa_r = sa*1e-2_r8
x = sqrt(sa_r)
p_r = p*1e-4_r8

gsw_pot_enthalpy_ice_freezing_poly = &
         c0 + sa_r*(c1 + x*(c2 + x*(c3 + x*(c4 + x*(c5 + c6*x))))) &
            + p_r*(c7 + p_r*(c8 + c9*p_r)) + sa_r*p_r*(c10 + p_r*(c12 &
            + p_r*(c15 + c21*sa_r)) + sa_r*(c13 + c17*p_r + c19*sa_r) &
            + x*(c11 + p_r*(c14 + c18*p_r) + sa_r*(c16 + c20*p_r + c22*sa_r)))

return
end function

!--------------------------------------------------------------------------
!==========================================================================
elemental function gsw_pot_rho_t_exact (sa, t, p, p_ref)  
!==========================================================================
!
! Calculates the potential density of seawater
!
! sa     : Absolute Salinity                               [g/kg]
! t      : in-situ temperature                             [deg C]
! p      : sea pressure                                    [dbar]
! p_ref  : reference sea pressure                          [dbar]
! 
! gsw_pot_rho_t_exact : potential density                  [kg/m^3]
!--------------------------------------------------------------------------

use gsw_mod_toolbox, only : gsw_pt_from_t, gsw_rho_t_exact

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sa, t, p, p_ref  

real (r8) :: gsw_pot_rho_t_exact

real (r8) :: pt

pt = gsw_pt_from_t(sa,t,p,p_ref)

gsw_pot_rho_t_exact = gsw_rho_t_exact(sa,pt,p_ref)

return
end function

!--------------------------------------------------------------------------
!==========================================================================
elemental function gsw_pressure_coefficient_ice (t, p)
!==========================================================================
!
!  Calculates pressure coefficient of ice. 
!
!  t  =  in-situ temperature (ITS-90)                             [ deg C ]
!  p  =  sea pressure                                              [ dbar ]
!         ( i.e. absolute pressure - 10.1325 dbar )
!
!  pressure_coefficient_ice  =  pressure coefficient of ice          [Pa/K]
!   Note. The output units are Pa/K NOT dbar/K.
!--------------------------------------------------------------------------

use gsw_mod_toolbox, only : gsw_gibbs_ice

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: t, p

real (r8) :: gsw_pressure_coefficient_ice

gsw_pressure_coefficient_ice = -gsw_gibbs_ice(1,1,t,p)/gsw_gibbs_ice(0,2,t,p)

return
end function

!--------------------------------------------------------------------------
!==========================================================================
elemental function gsw_pressure_freezing_ct (sa, ct, saturation_fraction)
!==========================================================================
!
!  Calculates the pressure (in dbar) of seawater at the freezing
!  temperature.  That is, the output is the pressure at which seawater,
!  with Absolute Salinity SA, Conservative Temperature CT, and with 
!  saturation_fraction of dissolved air, freezes.  If the input values are 
!  such that there is no value of pressure in the range between 0 dbar and 
!  10,000 dbar for which seawater is at the freezing temperature, the 
!  output, pressure_freezing, is put equal to NaN.
!
!  SA  =  Absolute Salinity of seawater                            [ g/kg ]
!  CT  =  Conservative Temperature (ITS-90)                       [ deg C ]
!  saturation_fraction = the saturation fraction of dissolved air in 
!                        seawater
!
!  pressure_freezing = sea pressure at which the seawater freezes  [ dbar ]
!        ( i.e. absolute pressure - 10.1325 dbar ) 
!--------------------------------------------------------------------------

use gsw_mod_toolbox, only : gsw_ct_freezing_first_derivatives, gsw_sa_p_inrange
use gsw_mod_toolbox, only : gsw_ct_freezing_exact

use gsw_mod_error_functions, only : gsw_error_code, gsw_error_limit

use gsw_mod_teos10_constants, only : rec_pa2db

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sa, ct, saturation_fraction

real (r8) :: gsw_pressure_freezing_ct

integer :: i_iter
real (r8) :: ct_freezing_p0, ct_freezing_p10000, dctf_dp, f
real (r8) :: pf, pfm, pf_old, ctfreezing_p

integer, parameter :: number_of_iterations = 3

! rec_Pa2dbar is to have dCTf_dp in units of K/dbar rather than K/Pa
 
character (*), parameter :: func_name = "gsw_pressure_freezing_ct"

! Find CT > CT_freezing_p0.  If this is the case, the input CT value
! represent seawater that will not be frozen at any positive p.  
ct_freezing_p0 = gsw_ct_freezing_exact(sa,0.0_r8,saturation_fraction)
if (ct .gt. ct_freezing_p0) then
    gsw_pressure_freezing_ct = gsw_error_code(1,func_name)
    return
end if
 
! Find CT < CT_freezing_p10000.  If this is the case, the input CT value
! represent seawater that is frozen even at p = 10,000 dbar.   
ct_freezing_p10000 = gsw_ct_freezing_exact(sa,1e4_r8,saturation_fraction)
if (ct .lt. ct_freezing_p10000) then
    gsw_pressure_freezing_ct = gsw_error_code(2,func_name)
    return
end if

! This is the initial (linear) guess of the freezing pressure, in dbar.  
pf = rec_pa2db*(ct_freezing_p0 - ct)/(ct_freezing_p0 - ct_freezing_p10000)

call gsw_ct_freezing_first_derivatives(sa,pf,saturation_fraction, &
                                       ctfreezing_p=ctfreezing_p)
dctf_dp = rec_pa2db*ctfreezing_p
    !  this dctf_dp is the initial value of the partial derivative of 
    !  ct_freezing with respect to pressure (in dbar) at fixed sa, 
    !  assuming that the saturation_fraction is zero. 

do i_iter = 1, number_of_iterations  
    pf_old = pf
    f = gsw_ct_freezing_exact(sa,pf_old,saturation_fraction) - ct
    pf = pf_old - f/dctf_dp
    pfm = 0.5_r8*(pf + pf_old)
    call gsw_ct_freezing_first_derivatives(sa,pfm,saturation_fraction, &
                                           ctfreezing_p=ctfreezing_p)
    dctf_dp = rec_pa2db*ctfreezing_p
    pf = pf_old - f/dctf_dp
end do

if (gsw_sa_p_inrange(sa,pf)) then
    gsw_pressure_freezing_ct = pf
else
    gsw_pressure_freezing_ct = gsw_error_code(3,func_name)
end if

return
end function

!--------------------------------------------------------------------------
!==========================================================================
elemental function gsw_pt0_cold_ice_poly (pot_enthalpy_ice)
!==========================================================================
!
!  Calculates an initial estimate of pt0_ice when it is less than about
!  -100 deg C. 
!
!  pot_enthalpy_ice  =  potential enthalpy of ice                  [ J/kg ]
!
!  pt0_cold_ice_poly  =  initial estimate of potential temperatur 
!                        of very cold ice in dgress C (not K)     [ deg C ] 
!--------------------------------------------------------------------------

use gsw_mod_teos10_constants, only : gsw_t0

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: pot_enthalpy_ice

real (r8) :: gsw_pt0_cold_ice_poly

real (r8) :: log_abs_theta0, log_h_diff

! h00 = gsw_enthalpy_ice(-gsw_t0,0)
real (r8), parameter :: h00 = -6.320202333358860e5_r8

real (r8), parameter :: s0 =  1.493103204647916_r8
real (r8), parameter :: s1 =  2.372788609320607e-1_r8
real (r8), parameter :: s2 = -2.014996002119374e-3_r8
real (r8), parameter :: s3 =  2.640600197732682e-6_r8
real (r8), parameter :: s4 =  3.134706016844293e-5_r8
real (r8), parameter :: s5 =  2.733592344937913e-6_r8
real (r8), parameter :: s6 =  4.726828010223258e-8_r8
real (r8), parameter :: s7 = -2.735193883189589e-9_r8
real (r8), parameter :: s8 = -8.547714991377670e-11_r8

log_h_diff = log(pot_enthalpy_ice - h00)

log_abs_theta0 = s0 + log_h_diff*(s1 + log_h_diff*(s2 + log_h_diff*(s3 &
                + log_h_diff*(s4 + log_h_diff*(s5 + log_h_diff*(s6 &
                + log_h_diff*(s7 + log_h_diff*s8)))))))

gsw_pt0_cold_ice_poly = exp(log_abs_theta0) - gsw_t0
    
return
end function

!--------------------------------------------------------------------------
!==========================================================================
elemental function gsw_pt0_from_t (sa, t, p) 
!==========================================================================
!   
! Calculates potential temperature with reference pressure, p_ref = 0 dbar. 
!
! sa     : Absolute Salinity                               [g/kg]
! t      : in-situ temperature                             [deg C]
! p      : sea pressure                                    [dbar]
!
! gsw_pt0_from_t : potential temperature, p_ref = 0        [deg C]
!--------------------------------------------------------------------------

use gsw_mod_toolbox, only : gsw_entropy_part, gsw_entropy_part_zerop
use gsw_mod_toolbox, only : gsw_gibbs_pt0_pt0

use gsw_mod_teos10_constants, only : gsw_cp0, gsw_sso, gsw_t0, gsw_ups

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sa, t, p 

real (r8) :: gsw_pt0_from_t

integer no_iter
real (r8) :: s1, true_entropy_part, pt0m
real (r8) :: pt0, pt0_old, dentropy, dentropy_dt

s1 = sa/gsw_ups

pt0 = t + p*( 8.65483913395442e-6_r8  - &
        s1 *  1.41636299744881e-6_r8  - &
         p *  7.38286467135737e-9_r8  + &
         t *(-8.38241357039698e-6_r8  + &
        s1 *  2.83933368585534e-8_r8  + &
         t *  1.77803965218656e-8_r8  + &
         p *  1.71155619208233e-10_r8))

dentropy_dt = gsw_cp0/((gsw_t0 + pt0)*(1.0_r8 - 0.05_r8*(1.0_r8 - sa/gsw_sso)))

true_entropy_part = gsw_entropy_part(sa,t,p)

do no_iter = 1, 2
    pt0_old = pt0
    dentropy = gsw_entropy_part_zerop(sa,pt0_old) - true_entropy_part
    pt0 = pt0_old - dentropy/dentropy_dt 
    pt0m = 0.5_r8*(pt0 + pt0_old)
    dentropy_dt = -gsw_gibbs_pt0_pt0(sa,pt0m)
    pt0 = pt0_old - dentropy/dentropy_dt
end do

gsw_pt0_from_t = pt0

return
end function

!--------------------------------------------------------------------------
! =========================================================================
elemental function gsw_pt0_from_t_ice (t, p)
! =========================================================================
!
!  Calculates potential temperature of ice Ih with a reference pressure of
!  0 dbar, from in-situ temperature, t.
!
!  t   =  in-situ temperature  (ITS-90)                           [ deg C ]
!  p   =  sea pressure                                             [ dbar ]
!         ( i.e. absolute pressure - 10.1325 dbar )
!
!  pt0_ice  =  potential temperature of ice Ih with reference pressure of
!              zero dbar (ITS-90)                                 [ deg C ]
!--------------------------------------------------------------------------

use gsw_mod_teos10_constants, only : gsw_t0

use gsw_mod_toolbox, only : gsw_gibbs_ice_part_t, gsw_gibbs_ice_pt0
use gsw_mod_toolbox, only : gsw_gibbs_ice_pt0_pt0

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: t, p

real (r8) :: gsw_pt0_from_t_ice

integer :: number_of_iterations
real (r8) :: dentropy, dentropy_dt, pt0_ice
real (r8) :: pt0_ice_old, ptm_ice, true_entropy

! This is the starting polynomial for pt0 of ice Ih.
real (r8), parameter :: s1 = -2.256611570832386e-4_r8
real (r8), parameter :: s2 = -6.045305921314694e-7_r8
real (r8), parameter :: s3 =  5.546699019612661e-9_r8
real (r8), parameter :: s4 =  1.795030639186685e-11_r8
real (r8), parameter :: s5 =  1.292346094030742e-9_r8

real (r8), parameter :: p1 = -2.259745637898635e-4_r8
real (r8), parameter :: p2 =  1.486236778150360e-9_r8
real (r8), parameter :: p3 =  6.257869607978536e-12_r8
real (r8), parameter :: p4 = -5.253795281359302e-7_r8
real (r8), parameter :: p5 =  6.752596995671330e-9_r8
real (r8), parameter :: p6 =  2.082992190070936e-11_r8
    
real (r8), parameter :: q1 = -5.849191185294459e-15_r8
real (r8), parameter :: q2 =  9.330347971181604e-11_r8
real (r8), parameter :: q3 =  3.415888886921213e-13_r8
real (r8), parameter :: q4 =  1.064901553161811e-12_r8
real (r8), parameter :: q5 = -1.454060359158787e-10_r8
real (r8), parameter :: q6 = -5.323461372791532e-13_r8

true_entropy = -gsw_gibbs_ice_part_t(t,p)

if (t.lt.-45.0_r8 .and. t.gt.-273.0_r8) then
    
    pt0_ice = t + p*(p1 + p*(p2 + p3*t) + t*(p4 + t*(p5 + p6*t)))
    
    if (pt0_ice.lt.-gsw_t0) pt0_ice = -gsw_t0
    
    ! we add 0.05d0 to the initial estimate of pt0_ice at
    ! temps less than -273 to ensure that it is never less than -273.15.
    if (pt0_ice.lt.-273.0_r8) pt0_ice = pt0_ice + 0.05_r8
    
    dentropy_dt = -gsw_gibbs_ice_pt0_pt0(pt0_ice)
        
    do number_of_iterations = 1, 3
        pt0_ice_old = pt0_ice
        dentropy = -gsw_gibbs_ice_pt0(pt0_ice_old) - true_entropy
        pt0_ice = pt0_ice_old - dentropy/dentropy_dt
        ptm_ice = 0.5_r8*(pt0_ice + pt0_ice_old)
        dentropy_dt = -gsw_gibbs_ice_pt0_pt0(ptm_ice)
        pt0_ice = pt0_ice_old - dentropy/dentropy_dt
    end do

else

    pt0_ice = t + p*(s1 + t*(s2 + t*(s3 + t*s4)) + s5*p)
    dentropy_dt = -gsw_gibbs_ice_pt0_pt0(pt0_ice)

    pt0_ice_old = pt0_ice
    dentropy = -gsw_gibbs_ice_pt0(pt0_ice_old) - true_entropy

    pt0_ice = pt0_ice_old - dentropy/dentropy_dt
    ptm_ice = 0.5_r8*(pt0_ice + pt0_ice_old)
    dentropy_dt = -gsw_gibbs_ice_pt0_pt0(ptm_ice)
    pt0_ice = pt0_ice_old - dentropy/dentropy_dt

end if
    
if (pt0_ice.lt.-273.0_r8) then
        
    pt0_ice = t + p*(q1 + p*(q2 + q3*t) + t*(q4 + t*(q5 + q6*t)))
        
    ! add 0.01d0 to the initial estimate of pt_ice used in the derivative to
    ! ensure that it is never less than -273.15d0 because the derivative
    ! approaches zero at absolute zero.
    dentropy_dt = -gsw_gibbs_ice_pt0_pt0(pt0_ice+0.01_r8)

    do number_of_iterations = 1, 3
        pt0_ice_old = pt0_ice
        dentropy = -gsw_gibbs_ice_pt0(pt0_ice_old) - true_entropy
        pt0_ice = pt0_ice_old - dentropy/dentropy_dt
        ptm_ice = 0.5_r8*(pt0_ice + pt0_ice_old)
        ! add 0.01d0 to the estimate of ptm_ice for temperatures less than
        ! -273 to ensure that they are never less than -273.15d0 because
        ! the derivative approaches zero at absolute zero and the addition
        ! of 0.01d0 degrees c ensures that when we divide by the derivatve
        ! in the modified newton routine the function does not blow up.
        ptm_ice = ptm_ice + 0.01_r8
        dentropy_dt = -gsw_gibbs_ice_pt0_pt0(ptm_ice)
        pt0_ice = pt0_ice_old - dentropy/dentropy_dt
    end do

end if
    
! For temperatures less than -273.1 degsC the maximum error is less than
! 2x10^-7 degsC. For temperatures between -273.1 and 273 the maximum error
! is less than 8x10^-8 degsC, and for temperatures greater than -273 degsC the
! maximum error is 1.5x10^-12 degsC.   These errors are over the whole
! ocean depths with p varying between 0 and 10,000 dbar, while the in-situ
! temperature varied independently between -273.15 and +2 degsC.

gsw_pt0_from_t_ice = pt0_ice

return
end function

!--------------------------------------------------------------------------
!==========================================================================
elemental subroutine gsw_pt_first_derivatives (sa, ct, pt_sa, pt_ct)
! =========================================================================
!
!  Calculates the following two partial derivatives of potential temperature 
!  (the regular potential temperature whose reference sea pressure is 0 dbar) 
!  (1) pt_SA, the derivative with respect to Absolute Salinity at 
!       constant Conservative Temperature, and
!  (2) pt_CT, the derivative with respect to Conservative Temperature at 
!       constant Absolute Salinity. 
!
!  SA  =  Absolute Salinity                                        [ g/kg ]
!  CT  =  Conservative Temperature (ITS-90)                       [ deg C ]
!
!  pt_SA =  The derivative of potential temperature with respect to 
!           Absolute Salinity at constant Conservative Temperature. 
!                                                               [ K/(g/kg)]
!  pt_CT =  The derivative of potential temperature with respect to 
!           Conservative Temperature at constant Absolute Salinity.
!           pt_CT is dimensionless.                            [ unitless ]
!--------------------------------------------------------------------------

use gsw_mod_teos10_constants, only : gsw_cp0, gsw_t0

use gsw_mod_toolbox, only : gsw_gibbs, gsw_pt_from_ct

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sa, ct
real (r8), intent(out), optional :: pt_sa, pt_ct

real (r8) :: abs_pt, ct_pt, ct_sa, pt

integer, parameter :: n0=0, n1=1, n2=2
real (r8), parameter :: pr0 = 0.0_r8

pt = gsw_pt_from_ct(sa,ct)
abs_pt = (gsw_t0 + pt)

ct_pt = -(abs_pt*gsw_gibbs(n0,n2,n0,sa,pt,pr0))/gsw_cp0

if (present(pt_sa)) then

    ct_sa = (gsw_gibbs(n1,n0,n0,sa,pt,pr0) -&
                abs_pt*gsw_gibbs(n1,n1,n0,sa,pt,pr0))/gsw_cp0

    pt_sa = -ct_sa/ct_pt

end if

if (present(pt_ct)) pt_ct = 1.0_r8/ct_pt

return
end subroutine

!--------------------------------------------------------------------------
!==========================================================================
elemental function gsw_pt_from_ct (sa, ct) 
!==========================================================================
!
! potential temperature of seawater from conservative temperature
!
! sa     : Absolute Salinity                               [g/kg]
! ct     : Conservative Temperature                        [deg C]
! p      : sea pressure                                    [dbar]
!
! gsw_pt_from_ct : potential temperature with              [deg C]
!                  reference pressure of  0 dbar
!--------------------------------------------------------------------------

use gsw_mod_toolbox, only : gsw_ct_from_pt, gsw_gibbs_pt0_pt0

use gsw_mod_teos10_constants, only : gsw_cp0, gsw_ups, gsw_t0

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sa, ct 

real (r8) :: gsw_pt_from_ct

real (r8) :: a5ct, b3ct, ct_factor, pt_num, pt_recden, ct_diff
real (r8) :: pt, pt_old, ptm, dpt_dct, s1

real (r8), parameter :: a0 = -1.446013646344788e-2_r8
real (r8), parameter :: a1 = -3.305308995852924e-3_r8
real (r8), parameter :: a2 =  1.062415929128982e-4_r8
real (r8), parameter :: a3 =  9.477566673794488e-1_r8
real (r8), parameter :: a4 =  2.166591947736613e-3_r8
real (r8), parameter :: a5 =  3.828842955039902e-3_r8

real (r8), parameter :: b0 =  1.0_r8
real (r8), parameter :: b1 =  6.506097115635800e-4_r8
real (r8), parameter :: b2 =  3.830289486850898e-3_r8
real (r8), parameter :: b3 =  1.247811760368034e-6_r8

s1 = sa/gsw_ups

a5ct = a5*ct
b3ct = b3*ct

ct_factor = (a3 + a4*s1 + a5ct)
pt_num = a0 + s1*(a1 + a2*s1) + ct*ct_factor
pt_recden = 1.0_r8/(b0 + b1*s1 + ct*(b2 + b3ct))
pt = pt_num*pt_recden

dpt_dct = (ct_factor + a5ct - (b2 + b3ct + b3ct)*pt)*pt_recden

! Start the 1.5 iterations through the modified Newton-Rapshon iterative,
! method, which is also known as the Newton-McDougall method. 

ct_diff = gsw_ct_from_pt(sa,pt) - ct
pt_old = pt
pt = pt_old - ct_diff*dpt_dct
ptm = 0.5_r8*(pt + pt_old)

dpt_dct = -gsw_cp0/((ptm + gsw_t0)*gsw_gibbs_pt0_pt0(sa,ptm))

pt = pt_old - ct_diff*dpt_dct
ct_diff = gsw_ct_from_pt(sa,pt) - ct
pt_old = pt
gsw_pt_from_ct = pt_old - ct_diff*dpt_dct

return 
end function

!--------------------------------------------------------------------------
! =========================================================================
elemental function gsw_pt_from_entropy (sa, entropy)
! =========================================================================
!
!  Calculates potential temperature with reference pressure p_ref = 0 dbar 
!  and with entropy as an input variable. 
!
!  SA       =  Absolute Salinity                                   [ g/kg ]
!  entropy  =  specific entropy                                   [ deg C ]
!
!  pt   =  potential temperature                                  [ deg C ]
!          with reference sea pressure (p_ref) = 0 dbar.
!  Note. The reference sea pressure of the output, pt, is zero dbar.
!--------------------------------------------------------------------------

use gsw_mod_teos10_constants, only : gsw_cp0, gsw_sso, gsw_t0

use gsw_mod_toolbox, only : gsw_entropy_from_pt, gsw_gibbs_pt0_pt0

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sa, entropy

real (r8) :: gsw_pt_from_entropy

integer :: number_of_iterations
real (r8) :: c, dentropy, dentropy_dt, ent_sa, part1, part2, pt, ptm
real (r8) :: pt_old

! Find the initial value of pt
part1 = 1.0_r8 - sa/gsw_sso
part2 = 1.0_r8 - 0.05_r8*part1
ent_sa = (gsw_cp0/gsw_t0)*part1*(1.0_r8 - 1.01_r8*part1)
c = (entropy - ent_sa)*(part2/gsw_cp0)
pt = gsw_t0*(exp(c) - 1.0_r8)
dentropy_dt = gsw_cp0/((gsw_t0 + pt)*part2)

do number_of_iterations = 1, 2
    pt_old = pt
    dentropy = gsw_entropy_from_pt(sa,pt_old) - entropy
    pt = pt_old - dentropy/dentropy_dt
    ptm = 0.5_r8*(pt + pt_old)
    dentropy_dt = -gsw_gibbs_pt0_pt0(sa,ptm)
    pt = pt_old - dentropy/dentropy_dt
end do
    
! Maximum error of 2.2x10^-6 degrees C for one iteration.
! Maximum error is 1.4x10^-14 degrees C for two iterations 
! (two iterations is the default, "for Number_of_iterations = 1:2"). 

gsw_pt_from_entropy = pt

return
end function

!--------------------------------------------------------------------------
!==========================================================================
elemental function gsw_pt_from_pot_enthalpy_ice (pot_enthalpy_ice)
!==========================================================================
!
!  Calculates the potential temperature of ice from the potential enthalpy 
!  of ice.  The reference sea pressure of both the potential temperature 
!  and the potential enthalpy is zero dbar. 
!
!  pot_enthalpy_ice  =  potential enthalpy of ice                  [ J/kg ]
!
!  pt0_ice  =  potential temperature of ice (ITS-90)              [ deg C ]
!--------------------------------------------------------------------------

use gsw_mod_toolbox, only : gsw_cp_ice, gsw_pot_enthalpy_from_pt_ice
use gsw_mod_toolbox, only : gsw_pt_from_pot_enthalpy_ice_poly_dh
use gsw_mod_toolbox, only : gsw_pt_from_pot_enthalpy_ice_poly
use gsw_mod_toolbox, only : gsw_pt0_cold_ice_poly

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: pot_enthalpy_ice

real (r8) :: gsw_pt_from_pot_enthalpy_ice

integer :: iteration
real (r8) :: df_dt, f, mod_pot_enthalpy_ice, pt0_cold_ice, recip_df_dt
real (r8) :: pt0_cold_ice_old, pt0_ice, pt0_ice_old, ptm_cold_ice, ptm_ice

real (r8), parameter :: h00 = -6.320202333358860e5_r8 !gsw_enthalpy_ice(-gsw_t0,0)
real (r8), parameter :: p0 = 0.0_r8

mod_pot_enthalpy_ice = max(pot_enthalpy_ice,h00)
   
if (mod_pot_enthalpy_ice.ge.-5.1e5_r8) then

    ! For input potential enthalpies greater than -5.1e-5, the above part of
    ! the code gives the output potential temperature of ice accurate to 1e-13
    ! degrees C.

    pt0_ice = gsw_pt_from_pot_enthalpy_ice_poly(mod_pot_enthalpy_ice)

    ! The variable "df_dt" below is the derivative of the above polynomial with
    ! respect to pot_enthalpy_ice.  This is the initial value of the derivative
    ! of the function f.
    recip_df_dt = gsw_pt_from_pot_enthalpy_ice_poly_dh(mod_pot_enthalpy_ice)

    pt0_ice_old = pt0_ice
    f = gsw_pot_enthalpy_from_pt_ice(pt0_ice_old) - mod_pot_enthalpy_ice
    pt0_ice = pt0_ice_old - f*recip_df_dt
    ptm_ice = 0.5_r8*(pt0_ice + pt0_ice_old)
    recip_df_dt = 1.0_r8/gsw_cp_ice(ptm_ice,p0)
    pt0_ice = pt0_ice_old - f*recip_df_dt

else

    ! For  pot_enthalpy_ice < -5.1e5 (or pt0_ice less than about -100 deg c)
    ! these temperatures are less than those found in nature on planet earth. 
    
    pt0_cold_ice = gsw_pt0_cold_ice_poly(mod_pot_enthalpy_ice) 
    
    df_dt = gsw_cp_ice(pt0_cold_ice+0.02_r8,p0)
    ! the heat capacity, cp, is
    ! evaluated at 0.02 c greater than usual in order to avoid stability
    ! issues and to ensure convergence near zero absolute temperature. 
    
    do iteration = 1, 6  
        pt0_cold_ice_old = pt0_cold_ice
        f = gsw_pot_enthalpy_from_pt_ice(pt0_cold_ice_old)-mod_pot_enthalpy_ice
        pt0_cold_ice = pt0_cold_ice_old - f/df_dt
        ptm_cold_ice = 0.5_r8*(pt0_cold_ice + pt0_cold_ice_old)        
        df_dt = gsw_cp_ice(ptm_cold_ice+0.02_r8,p0)
	! note the extra 0.02 c here as well    
        pt0_cold_ice = pt0_cold_ice_old - f/df_dt
    end do
    pt0_ice = pt0_cold_ice

end if

!The potential temerature has a maximum error as listed in the table below.
!
!  potential temerature error (deg C)  |  @ potential temerature (deg C)
!--------------------------------------|--------------------------------
!                0.012                 |     -273.15 to -273.12
!              4 x 10^-4               |     -232.12 to -273.0
!             2.5 x 10^-6              |          -273
!              7 x 10^-9               |          -272
!            3.7 x 10^-10              |          -270
!              6 x 10^-11              |          -268
!             2.5 x 10^11              |          -266
!             3 x 10^-12               |          -260
!             7 x 10^-13               |          -250
!            2.2 x 10^-13              |          -220
!            1.7 x 10^-13              |         >= -160
!  
! Note.  The above errors in each temperature range are machine precissions 
! for this calculation.

gsw_pt_from_pot_enthalpy_ice = pt0_ice

return
end function

!--------------------------------------------------------------------------
!==========================================================================
elemental function gsw_pt_from_pot_enthalpy_ice_poly (pot_enthalpy_ice)
!==========================================================================
!
!  Calculates the potential temperature of ice (whose reference sea 
!  pressure is zero dbar) from the potential enthalpy of ice.  This is a
!  compuationally efficient polynomial fit to the potential enthalpy of
!  ice.
!
!  pot_enthalpy_ice  =  potential enthalpy of ice                  [ J/kg ]
!
!  pt0_ice  =  potential temperature of ice (ITS-90)              [ deg C ]
!--------------------------------------------------------------------------

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: pot_enthalpy_ice

real (r8) :: gsw_pt_from_pot_enthalpy_ice_poly

real (r8), parameter :: q0 = 2.533588268773218e2_r8
real (r8), parameter :: q1 = 2.594351081876611e-3_r8
real (r8), parameter :: q2 = 1.765077810213815e-8_r8
real (r8), parameter :: q3 = 7.768070564290540e-14_r8
real (r8), parameter :: q4 = 2.034842254277530e-19_r8
real (r8), parameter :: q5 = 3.220014531712841e-25_r8
real (r8), parameter :: q6 = 2.845172809636068e-31_r8
real (r8), parameter :: q7 = 1.094005878892950e-37_r8
    
! The error of this fit ranges between -5e-5 and 2e-4 deg C over the potential 
! temperature range of -100 to 2 deg C, or the potential enthalpy range of 
! -5.7 x 10^5 to -3.3 x 10^5 J/kg. 
      
gsw_pt_from_pot_enthalpy_ice_poly = q0 &
         + pot_enthalpy_ice*(q1 + pot_enthalpy_ice*(q2 + pot_enthalpy_ice*(q3 &
         + pot_enthalpy_ice*(q4 + pot_enthalpy_ice*(q5 + pot_enthalpy_ice*(q6 &
         + pot_enthalpy_ice*q7))))))

return
end function

!--------------------------------------------------------------------------
!==========================================================================
elemental function gsw_pt_from_pot_enthalpy_ice_poly_dh (pot_enthalpy_ice)
!==========================================================================
!
!  Calculates the derivative of potential temperature of ice with respect 
!  to potential enthalpy.  This is based on the compuationally-efficient 
!  polynomial fit to the potential enthalpy of ice. 
!
!  pot_enthalpy_ice  =  potential enthalpy of ice                  [ J/kg ]
!
!  dpt0_ice_dh  =  derivative of potential temperature of ice 
!                  with respect to potential enthalpy             [ deg C ]
!--------------------------------------------------------------------------

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: pot_enthalpy_ice

real (r8) :: gsw_pt_from_pot_enthalpy_ice_poly_dh

real (r8), parameter :: q1 = 2.594351081876611e-3_r8
real (r8), parameter :: p2 = 3.530155620427630e-8_r8
real (r8), parameter :: p3 = 2.330421169287162e-13_r8
real (r8), parameter :: p4 = 8.139369017110120e-19_r8
real (r8), parameter :: p5 = 1.610007265856420e-24_r8
real (r8), parameter :: p6 = 1.707103685781641e-30_r8
real (r8), parameter :: p7 = 7.658041152250651e-37_r8

gsw_pt_from_pot_enthalpy_ice_poly_dh = q1 &
    + pot_enthalpy_ice*(p2 + pot_enthalpy_ice*(p3 &
    + pot_enthalpy_ice*(p4 + pot_enthalpy_ice*(p5 + pot_enthalpy_ice*(p6 &
    + pot_enthalpy_ice*p7)))))

return
end function

!--------------------------------------------------------------------------
!==========================================================================
elemental function gsw_pt_from_t (sa, t, p, p_ref) 
!==========================================================================
!   
! Calculates potential temperature of seawater from in-situ temperature 
!
! sa     : Absolute Salinity                               [g/kg]
! t      : in-situ temperature                             [deg C]
! p      : sea pressure                                    [dbar]
! p_ref  : reference sea pressure                          [dbar]
!
! gsw_pt_from_t : potential temperature                    [deg C]
!--------------------------------------------------------------------------

use gsw_mod_toolbox, only : gsw_entropy_part, gsw_gibbs

use gsw_mod_teos10_constants, only : gsw_cp0, gsw_sso, gsw_t0, gsw_ups

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sa, t, p, p_ref 

real (r8) :: gsw_pt_from_t

integer no_iter
real (r8) :: s1, pt, pt_old, dentropy, dentropy_dt
real (r8) :: true_entropy_part, ptm

integer, parameter :: n0=0, n2=2

s1 = sa/gsw_ups

pt = t + (p-p_ref)*( 8.65483913395442e-6_r8 - &
               s1 *  1.41636299744881e-6_r8 - &
         (p+p_ref)*  7.38286467135737e-9_r8 + &
               t  *(-8.38241357039698e-6_r8 + &
               s1 *  2.83933368585534e-8_r8 + &
               t  *  1.77803965218656e-8_r8 + &
         (p+p_ref)*  1.71155619208233e-10_r8))

dentropy_dt = gsw_cp0/((gsw_t0 + pt)*(1.0_r8 - 0.05_r8*(1.0_r8 - sa/gsw_sso)))

true_entropy_part = gsw_entropy_part(sa,t,p)

do no_iter = 1, 2
    pt_old = pt
    dentropy = gsw_entropy_part(sa,pt_old,p_ref) - true_entropy_part
    pt = pt_old - dentropy/dentropy_dt 
    ptm = 0.5_r8*(pt + pt_old)
    dentropy_dt = -gsw_gibbs(n0,n2,n0,sa,ptm,p_ref)
    pt = pt_old - dentropy/dentropy_dt
end do

gsw_pt_from_t = pt

return
end function

!--------------------------------------------------------------------------
! =========================================================================
elemental function gsw_pt_from_t_ice (t, p, p_ref)
! =========================================================================
!
!  Calculates potential temperature of ice Ih with the general reference
!  pressure, p_ref, from in-situ temperature, t.
!
!  A faster gsw routine exists if p_ref is indeed zero dbar.  This routine
!  is "gsw_pt0_from_t_ice(t,p)".
!
!  t  =  in-situ temperature (ITS-90)                             [ deg C ]
!  p  =  sea pressure                                              [ dbar ]
!         ( i.e. absolute pressure - 10.1325 dbar )
!  p_ref  =  reference pressure                                    [ dbar ]
!--------------------------------------------------------------------------

use gsw_mod_teos10_constants, only : gsw_t0

use gsw_mod_toolbox, only : gsw_gibbs_ice, gsw_gibbs_ice_part_t

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: t, p, p_ref

real (r8) :: gsw_pt_from_t_ice

integer :: number_of_iterations
real (r8) :: dentropy, dentropy_dt, dp
real (r8) :: pt_ice, pt_ice_old, ptm_ice, true_entropy

real (r8), parameter :: p1 = -2.259745637898635e-4_r8
real (r8), parameter :: p2 =  1.486236778150360e-9_r8
real (r8), parameter :: p3 =  6.257869607978536e-12_r8
real (r8), parameter :: p4 = -5.253795281359302e-7_r8
real (r8), parameter :: p5 =  6.752596995671330e-9_r8
real (r8), parameter :: p6 =  2.082992190070936e-11_r8

real (r8), parameter :: q1 = -5.849191185294459e-15_r8
real (r8), parameter :: q2 =  9.330347971181604e-11_r8
real (r8), parameter :: q3 =  3.415888886921213e-13_r8
real (r8), parameter :: q4 =  1.064901553161811e-12_r8
real (r8), parameter :: q5 = -1.454060359158787e-10_r8
real (r8), parameter :: q6 = -5.323461372791532e-13_r8

! This is the starting polynomial for pt of ice Ih.
dp = p - p_ref

pt_ice = t + dp*(p1 + (p + p_ref)*(p2 + p3*t) + t*(p4 + t*(p5 + p6*t)))

if (pt_ice.lt.-gsw_t0) pt_ice = -gsw_t0

if (pt_ice.lt.-273.0_r8) pt_ice = pt_ice + 0.05_r8
! we add 0.05 to the initial estimate of pt_ice at temps less than -273 to
! ensure that it is never less than -273.15.

dentropy_dt = -gsw_gibbs_ice(2,0,pt_ice,p_ref)

true_entropy = -gsw_gibbs_ice_part_t(t,p)

do number_of_iterations = 1, 3
    pt_ice_old = pt_ice
    dentropy = -gsw_gibbs_ice_part_t(pt_ice_old,p_ref) - true_entropy
    pt_ice = pt_ice_old - dentropy/dentropy_dt
    ptm_ice = 0.5_r8*(pt_ice + pt_ice_old)
    dentropy_dt = -gsw_gibbs_ice(2,0,ptm_ice,p_ref)
    pt_ice = pt_ice_old - dentropy/dentropy_dt
end do

if (pt_ice.lt.-273.0_r8) then
    
    pt_ice = t + (p - p_ref)*(q1 + (p + p_ref)*(q2 + q3*t) &
        + t*(q4 + t*(q5 + q6*t)))        

    dentropy_dt = -gsw_gibbs_ice(2,0,pt_ice+0.01_r8,p_ref)
    ! we add 0.01 to the initial estimate of pt_ice used in the derivative to
    ! ensure that it is never less than -273.15 because the derivative
    ! approaches zero at absolute zero.

    do number_of_iterations = 1, 3
        pt_ice_old = pt_ice
        dentropy = -gsw_gibbs_ice_part_t(pt_ice_old,p_ref) - true_entropy
        pt_ice = pt_ice_old - dentropy/dentropy_dt
        ptm_ice = 0.5_r8*(pt_ice + pt_ice_old)        
        ptm_ice = ptm_ice + 0.01_r8    
        dentropy_dt = -gsw_gibbs_ice(2,0,ptm_ice,p_ref)
        pt_ice = pt_ice_old - dentropy/dentropy_dt
    end do

end if

! For temperatures less than -273.1 degsC the maximum error is less than
! 2x10^-7 degsC. For temperatures between -273.1 and 273 the maximum error
! is less than 8x10^-8 degsC, and for temperatures greater than -273 degsC the
! maximum error is 1.5x10^-12 degsC.  These errors are over the whole
! ocean depths with both p and pref varying independently between 0 and
! 10,000 dbar, while the in-situ temperature varied independently between
! -273.15 and +2 degsC.

gsw_pt_from_t_ice = pt_ice

return
end function

!--------------------------------------------------------------------------
!==========================================================================
elemental subroutine gsw_pt_second_derivatives (sa, ct, pt_sa_sa, &
                                                pt_sa_ct, pt_ct_ct)
! =========================================================================
!
!  Calculates the following three second-order derivatives of potential 
!  temperature (the regular potential temperature which has a reference 
!  sea pressure of 0 dbar), 
!   (1) pt_SA_SA, the second derivative with respect to Absolute Salinity 
!       at constant Conservative Temperature,
!   (2) pt_SA_CT, the derivative with respect to Conservative Temperature
!       and Absolute Salinity, and
!   (3) pt_CT_CT, the second derivative with respect to Conservative 
!       Temperature at constant Absolute Salinity. 
!
!  SA  =  Absolute Salinity                                        [ g/kg ]
!  CT  =  Conservative Temperature (ITS-90)                       [ deg C ]
!
!  pt_SA_SA  =  The second derivative of potential temperature (the 
!               regular potential temperature which has reference sea 
!               pressure of 0 dbar) with respect to Absolute Salinity 
!               at constant Conservative Temperature.  
!               pt_SA_SA has units of:                     [ K/((g/kg)^2) ]
!  pt_SA_CT  =  The derivative of potential temperature with respect 
!               to Absolute Salinity and Conservative Temperature.   
!               pt_SA_CT has units of:                         [ 1/(g/kg) ]
!  pt_CT_CT  =  The second derivative of potential temperature (the 
!               regular one with p_ref = 0 dbar) with respect to 
!               Conservative Temperature at constant SA.  
!               pt_CT_CT has units of:                              [ 1/K ]
!--------------------------------------------------------------------------

use gsw_mod_toolbox, only : gsw_pt_first_derivatives

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sa, ct
real (r8), intent(out), optional :: pt_sa_sa, pt_sa_ct, pt_ct_ct

real (r8) :: ct_l, ct_u, pt_ct_l, pt_ct_u, pt_sa_l, pt_sa_u, sa_l, sa_u

real (r8), parameter :: dct = 1e-2_r8, dsa = 1e-3_r8

if (present(pt_sa_sa)) then

    sa_l = max(sa - dsa, 0.0_r8)
    sa_u = sa + dsa

    call gsw_pt_first_derivatives(sa_l,ct,pt_sa=pt_sa_l)
    call gsw_pt_first_derivatives(sa_u,ct,pt_sa=pt_sa_u)

    pt_sa_sa = (pt_sa_u - pt_sa_l)/(sa_u - sa_l)

end if

if (present(pt_sa_ct) .or. present(pt_ct_ct)) then

    ct_l = ct - dct
    ct_u = ct + dct

    if (present(pt_sa_ct) .and. present(pt_ct_ct)) then

        call gsw_pt_first_derivatives(sa,ct_l,pt_sa_l,pt_ct_l)
        call gsw_pt_first_derivatives(sa,ct_u,pt_sa_u,pt_ct_u)

        pt_sa_ct = (pt_sa_u - pt_sa_l)/(ct_u - ct_l)
        pt_ct_ct = (pt_ct_u - pt_ct_l)/(ct_u - ct_l)

    else if (present(pt_sa_ct) .and. .not. present(pt_ct_ct)) then

        call gsw_pt_first_derivatives(sa,ct_l,pt_sa=pt_sa_l)
        call gsw_pt_first_derivatives(sa,ct_u,pt_sa=pt_sa_u)

        pt_sa_ct = (pt_sa_u - pt_sa_l)/(ct_u - ct_l)

    else if (.not. present(pt_sa_ct) .and. present(pt_ct_ct)) then

        call gsw_pt_first_derivatives(sa,ct_l,pt_ct=pt_ct_l)
        call gsw_pt_first_derivatives(sa,ct_u,pt_ct=pt_ct_u)

        pt_ct_ct = (pt_ct_u - pt_ct_l)/(ct_u - ct_l)

    end if

end if

return
end subroutine

!--------------------------------------------------------------------------
!==========================================================================
elemental function gsw_rho (sa, ct, p)
!==========================================================================
! 
!  Calculates in-situ density from Absolute Salinity and Conservative 
!  Temperature, using the computationally-efficient expression for
!  specific volume in terms of SA, CT and p (Roquet et al., 2014).
!
!  Note that potential density with respect to reference pressure, pr, is
!  obtained by calling this function with the pressure argument being pr
!  (i.e. "gsw_rho(SA,CT,pr)").
!
!  SA  =  Absolute Salinity                                        [ g/kg ]
!  CT  =  Conservative Temperature (ITS-90)                       [ deg C ]
!  p   =  sea pressure                                             [ dbar ]
!         ( i.e. absolute pressure - 10.1325 dbar )
!
!  rho  =  in-situ density                                         [ kg/m ]
!--------------------------------------------------------------------------

use gsw_mod_toolbox, only : gsw_specvol

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sa, ct, p

real (r8) :: gsw_rho

gsw_rho = 1.0_r8/gsw_specvol(sa,ct,p)

return
end function

!--------------------------------------------------------------------------
!==========================================================================
elemental subroutine gsw_rho_alpha_beta (sa, ct, p, rho, alpha, beta)
!==========================================================================
! 
!  Calculates in-situ density, the appropiate thermal expansion coefficient
!  and the appropriate saline contraction coefficient of seawater from 
!  Absolute Salinity and Conservative Temperature.  This function uses the
!  computationally-efficient expression for specific volume in terms of 
!  SA, CT and p (Roquet et al., 2014).
!
!  Note that potential density (pot_rho) with respect to reference pressure
!  p_ref is obtained by calling this function with the pressure argument 
!  being p_ref as in [pot_rho, ~, ~] = gsw_rho_alpha_beta(SA,CT,p_ref).
!
!  SA  =  Absolute Salinity                                        [ g/kg ]
!  CT  =  Conservative Temperature (ITS-90)                       [ deg C ]
!  p   =  sea pressure                                             [ dbar ]
!         ( i.e. absolute pressure - 10.1325 dbar )
!
!  rho    =  in-situ density                                       [ kg/m ]
!  alpha  =  thermal expansion coefficient                          [ 1/K ]
!            with respect to Conservative Temperature
!  beta   =  saline (i.e. haline) contraction                      [ kg/g ]
!            coefficient at constant Conservative Temperature
!--------------------------------------------------------------------------

use gsw_mod_toolbox, only : gsw_rho

use gsw_mod_teos10_constants, only : gsw_sfac, offset

use gsw_mod_specvol_coefficients

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sa, ct, p
real (r8), intent(out), optional :: rho, alpha, beta

real (r8) :: v, v_ct_part, v_sa_part, xs, ys, z

xs = sqrt(gsw_sfac*sa + offset)
ys = ct*0.025_r8
z = p*1e-4_r8

v = v000 + xs*(v010 + xs*(v020 + xs*(v030 + xs*(v040 + xs*(v050 &
    + v060*xs))))) + ys*(v100 + xs*(v110 + xs*(v120 + xs*(v130 + xs*(v140 &
    + v150*xs)))) + ys*(v200 + xs*(v210 + xs*(v220 + xs*(v230 + v240*xs))) &
    + ys*(v300 + xs*(v310 + xs*(v320 + v330*xs)) + ys*(v400 + xs*(v410 &
    + v420*xs) + ys*(v500 + v510*xs + v600*ys))))) + z*(v001 + xs*(v011 &
    + xs*(v021 + xs*(v031 + xs*(v041 + v051*xs)))) + ys*(v101 + xs*(v111 &
    + xs*(v121 + xs*(v131 + v141*xs))) + ys*(v201 + xs*(v211 + xs*(v221 &
    + v231*xs)) + ys*(v301 + xs*(v311 + v321*xs) + ys*(v401 + v411*xs &
    + v501*ys)))) + z*(v002 + xs*(v012 + xs*(v022 + xs*(v032 + v042*xs))) &
    + ys*(v102 + xs*(v112 + xs*(v122 + v132*xs)) + ys*(v202 + xs*(v212 &
    + v222*xs) + ys*(v302 + v312*xs + v402*ys))) + z*(v003 + xs*(v013 &
    + v023*xs) + ys*(v103 + v113*xs + v203*ys) + z*(v004 + v014*xs + v104*ys &
    + z*(v005 + v006*z)))))

if (present(rho)) rho = 1.0_r8/v

if (present(alpha)) then

    v_ct_part = a000 + xs*(a100 + xs*(a200 + xs*(a300 + xs*(a400 + a500*xs)))) &
             + ys*(a010 + xs*(a110 + xs*(a210 + xs*(a310 + a410*xs))) &
             + ys*(a020 + xs*(a120 + xs*(a220 + a320*xs)) + ys*(a030 &
             + xs*(a130 + a230*xs) + ys*(a040 + a140*xs + a050*ys )))) &
             + z*(a001 + xs*(a101 + xs*(a201 + xs*(a301 + a401*xs))) &
             + ys*(a011 + xs*(a111 + xs*(a211 + a311*xs)) + ys*(a021 &
             + xs*(a121 + a221*xs) + ys*(a031 + a131*xs + a041*ys))) &
             + z*(a002 + xs*(a102 + xs*(a202 + a302*xs)) + ys*(a012 &
             + xs*(a112 + a212*xs) + ys*(a022 + a122*xs + a032*ys)) &
             + z*(a003 + a103*xs + a013*ys + a004*z))) 

    alpha = 0.025_r8*v_ct_part/v

end if

if (present(beta)) then

    v_sa_part = b000 + xs*(b100 + xs*(b200 + xs*(b300 + xs*(b400 + b500*xs)))) &
           + ys*(b010 + xs*(b110 + xs*(b210 + xs*(b310 + b410*xs))) &
           + ys*(b020 + xs*(b120 + xs*(b220 + b320*xs)) + ys*(b030 &
           + xs*(b130 + b230*xs) + ys*(b040 + b140*xs + b050*ys)))) &
           + z*(b001 + xs*(b101 + xs*(b201 + xs*(b301 + b401*xs))) &
           + ys*(b011 + xs*(b111 + xs*(b211 + b311*xs)) + ys*(b021 &
           + xs*(b121 + b221*xs) + ys*(b031 + b131*xs + b041*ys))) &
           + z*(b002 + xs*(b102 + xs*(b202 + b302*xs))+ ys*(b012 &
           + xs*(b112 + b212*xs) + ys*(b022 + b122*xs + b032*ys)) &
           + z*(b003 +  b103*xs + b013*ys + b004*z)))
 
    beta = -v_sa_part*0.5_r8*gsw_sfac/(v*xs)

end if

return
end subroutine

!--------------------------------------------------------------------------
!==========================================================================
elemental subroutine gsw_rho_alpha_beta_bsq (sa, ct, p, rho, alpha, beta, &
                                             stiffened)
!==========================================================================
!
!  Calculates in-situ density (or optionally stiffened density), and the
!  appropriate Boussinesq thermal expansion and Boussinesq saline contraction
!  coefficients of seawater from absolute salinity, conservative temperature
!  and pressure, using the computationally-efficient 55-term polynomial
!  expression for density (Roquet et al., 2014).
!
!  sa  =  Absolute Salinity                                        [ g/kg ]
!  ct  =  Conservative Temperature (ITS-90)                       [ deg C ]
!  p   =  sea pressure                                             [ dbar ]
!         ( i.e. absolute pressure - 10.1325 dbar )
!
!  rho   =  In-situ density (rho = r1 x rdot)                    [ kg/m^3 ]
!           (or optionally stiffened density, rdot)
!  alpha =  Boussinesq thermal expansion (=-drho/dCT)          [ kg/m^3/K ]
!  beta  =  Boussinesq haline contraction (=drho/dSA)     [ kg/m^3/(g/kg) ]
!
!  Check values (sa=30g/kg, ct=10degC, p=1e3dbar):
!  rho  = 1027.45140
!  a    = 0.179649406
!  b    = 0.765554495
!  r1   = 1.00447333
!  rdot = 1022.87574
!
!  Roquet, F., Madec, G., McDougall, T. J., Barker, P. M., 2014: Accurate 
!    polynomial expressions for the density and specific volume of 
!    seawater using the TEOS-10 standard. Ocean Modelling, 90:29-43.
!--------------------------------------------------------------------------

use gsw_mod_teos10_constants, only : gsw_ups

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sa, ct, p
logical, intent(in), optional :: stiffened
real (r8), intent(out) :: rho
real (r8), intent(out), optional :: alpha, beta

logical :: do_stiffened

real (r8) :: pp, rz0, rz1, rz2, rz3, ss, tt, r1

real (r8), parameter :: sau = 40.0_r8*gsw_ups
real (r8), parameter :: ctu = 40.0_r8
real (r8), parameter :: zu = 1e4_r8
real (r8), parameter :: deltas = 32.0_r8

real (r8), parameter :: r10 =  4.5238001132e-02_r8
real (r8), parameter :: r11 = -5.0691457704e-03_r8
real (r8), parameter :: r12 =  2.1990865986e-04_r8 
real (r8), parameter :: r13 =  6.2587720090e-05_r8
real (r8), parameter :: r14 =  1.5194795322e-05_r8
real (r8), parameter :: r15 = -1.6777531159e-06_r8

real (r8), parameter :: alp000 = -6.5047345980e-01_r8
real (r8), parameter :: alp100 =  1.6337444787e+00_r8
real (r8), parameter :: alp200 = -2.0484575392e+00_r8 
real (r8), parameter :: alp300 =  1.4268760685e+00_r8
real (r8), parameter :: alp400 = -4.4447427136e-01_r8
real (r8), parameter :: alp500 =  4.8463173700e-02_r8 
real (r8), parameter :: alp010 =  1.8523793418e+00_r8
real (r8), parameter :: alp110 = -3.0734838779e+00_r8
real (r8), parameter :: alp210 =  3.0136782240e+00_r8 
real (r8), parameter :: alp310 = -1.4543073694e+00_r8
real (r8), parameter :: alp410 =  2.7320572723e-01_r8
real (r8), parameter :: alp020 = -1.6234028145e+00_r8 
real (r8), parameter :: alp120 =  2.5061411737e+00_r8
real (r8), parameter :: alp220 = -1.4770589780e+00_r8
real (r8), parameter :: alp320 =  2.3782870611e-01_r8 
real (r8), parameter :: alp030 =  8.3587258634e-01_r8
real (r8), parameter :: alp130 = -1.1301873278e+00_r8
real (r8), parameter :: alp230 =  5.3494903247e-01_r8 
real (r8), parameter :: alp040 = -6.7823124325e-02_r8
real (r8), parameter :: alp140 = -5.9955123381e-02_r8
real (r8), parameter :: alp050 =  2.8648472338e-02_r8 
real (r8), parameter :: alp001 =  3.4328982507e-01_r8
real (r8), parameter :: alp101 =  1.0916815960e-01_r8
real (r8), parameter :: alp201 = -1.3224824721e-01_r8 
real (r8), parameter :: alp301 =  1.7830956551e-02_r8
real (r8), parameter :: alp011 = -3.7421662855e-01_r8
real (r8), parameter :: alp111 = -1.5721498096e-01_r8 
real (r8), parameter :: alp211 =  9.0708859933e-02_r8
real (r8), parameter :: alp021 =  1.9507636737e-01_r8
real (r8), parameter :: alp121 =  3.7400054411e-02_r8 
real (r8), parameter :: alp031 = -5.5882364387e-02_r8
real (r8), parameter :: alp002 = -7.1271116782e-02_r8
real (r8), parameter :: alp102 =  1.1117840325e-02_r8 
real (r8), parameter :: alp012 =  7.5296514078e-02_r8
real (r8), parameter :: alp003 = -6.3135413999e-03_r8 

real (r8), parameter :: bet000 =  1.0785939671e+01_r8
real (r8), parameter :: bet100 = -4.4465045269e+01_r8
real (r8), parameter :: bet200 =  7.6072094337e+01_r8 
real (r8), parameter :: bet300 = -6.3964420131e+01_r8
real (r8), parameter :: bet400 =  2.6898783594e+01_r8
real (r8), parameter :: bet500 = -4.5234968986e+00_r8 
real (r8), parameter :: bet010 = -8.1303841476e-01_r8
real (r8), parameter :: bet110 =  2.0388435182e+00_r8
real (r8), parameter :: bet210 = -2.1302689715e+00_r8 
real (r8), parameter :: bet310 =  8.8477644261e-01_r8
real (r8), parameter :: bet410 = -1.2058930400e-01_r8
real (r8), parameter :: bet020 =  7.6476477580e-01_r8 
real (r8), parameter :: bet120 = -1.4997670675e+00_r8
real (r8), parameter :: bet220 =  1.0856114040e+00_r8
real (r8), parameter :: bet320 = -2.7192349143e-01_r8 
real (r8), parameter :: bet030 = -4.1572985119e-01_r8
real (r8), parameter :: bet130 =  4.9004223351e-01_r8
real (r8), parameter :: bet230 = -1.1835625260e-01_r8 
real (r8), parameter :: bet040 =  1.4061037779e-01_r8
real (r8), parameter :: bet140 = -1.3310958936e-01_r8
real (r8), parameter :: bet050 =  5.9673736141e-03_r8 
real (r8), parameter :: bet001 = -5.2308076768e-01_r8
real (r8), parameter :: bet101 =  1.2084313165e+00_r8
real (r8), parameter :: bet201 = -1.1374069553e+00_r8 
real (r8), parameter :: bet301 =  3.2360305476e-01_r8
real (r8), parameter :: bet011 = -5.4327900468e-02_r8
real (r8), parameter :: bet111 =  1.3162756682e-01_r8 
real (r8), parameter :: bet211 = -2.6620905846e-02_r8
real (r8), parameter :: bet021 =  3.9119281064e-02_r8
real (r8), parameter :: bet121 = -4.5141568126e-02_r8 
real (r8), parameter :: bet031 = -6.2040874705e-03_r8
real (r8), parameter :: bet002 = -5.6500454603e-02_r8
real (r8), parameter :: bet102 =  6.7785665385e-02_r8 
real (r8), parameter :: bet012 = -5.5328304955e-03_r8
real (r8), parameter :: bet003 = -2.2276668383e-03_r8 

real (r8), parameter :: r000 =  8.0185969881e+02_r8
real (r8), parameter :: r100 =  8.6694399997e+02_r8
real (r8), parameter :: r200 = -1.7869886805e+03_r8 
real (r8), parameter :: r300 =  2.0381548497e+03_r8
real (r8), parameter :: r400 = -1.2853207957e+03_r8
real (r8), parameter :: r500 =  4.3240996619e+02_r8 
real (r8), parameter :: r600 = -6.0597695001e+01_r8
real (r8), parameter :: r010 =  2.6018938392e+01_r8
real (r8), parameter :: r110 = -6.5349779146e+01_r8 
real (r8), parameter :: r210 =  8.1938301569e+01_r8
real (r8), parameter :: r310 = -5.7075042739e+01_r8
real (r8), parameter :: r410 =  1.7778970855e+01_r8 
real (r8), parameter :: r510 = -1.9385269480e+00_r8
real (r8), parameter :: r020 = -3.7047586837e+01_r8
real (r8), parameter :: r120 =  6.1469677558e+01_r8 
real (r8), parameter :: r220 = -6.0273564480e+01_r8
real (r8), parameter :: r320 =  2.9086147388e+01_r8
real (r8), parameter :: r420 = -5.4641145446e+00_r8 
real (r8), parameter :: r030 =  2.1645370860e+01_r8
real (r8), parameter :: r130 = -3.3415215649e+01_r8
real (r8), parameter :: r230 =  1.9694119706e+01_r8 
real (r8), parameter :: r330 = -3.1710494147e+00_r8
real (r8), parameter :: r040 = -8.3587258634e+00_r8
real (r8), parameter :: r140 =  1.1301873278e+01_r8 
real (r8), parameter :: r240 = -5.3494903247e+00_r8
real (r8), parameter :: r050 =  5.4258499460e-01_r8
real (r8), parameter :: r150 =  4.7964098705e-01_r8 
real (r8), parameter :: r060 = -1.9098981559e-01_r8
real (r8), parameter :: r001 =  2.1989266031e+01_r8
real (r8), parameter :: r101 = -4.2043785414e+01_r8 
real (r8), parameter :: r201 =  4.8565183521e+01_r8
real (r8), parameter :: r301 = -3.0473875108e+01_r8
real (r8), parameter :: r401 =  6.5025796369e+00_r8 
real (r8), parameter :: r011 = -1.3731593003e+01_r8
real (r8), parameter :: r111 = -4.3667263842e+00_r8
real (r8), parameter :: r211 =  5.2899298884e+00_r8 
real (r8), parameter :: r311 = -7.1323826203e-01_r8
real (r8), parameter :: r021 =  7.4843325711e+00_r8
real (r8), parameter :: r121 =  3.1442996192e+00_r8 
real (r8), parameter :: r221 = -1.8141771987e+00_r8
real (r8), parameter :: r031 = -2.6010182316e+00_r8
real (r8), parameter :: r131 = -4.9866739215e-01_r8 
real (r8), parameter :: r041 =  5.5882364387e-01_r8
real (r8), parameter :: r002 =  1.1144125393e+00_r8
real (r8), parameter :: r102 = -4.5413502768e+00_r8 
real (r8), parameter :: r202 =  2.7242121539e+00_r8
real (r8), parameter :: r012 =  2.8508446713e+00_r8
real (r8), parameter :: r112 = -4.4471361300e-01_r8 
real (r8), parameter :: r022 = -1.5059302816e+00_r8
real (r8), parameter :: r003 =  1.9817079368e-01_r8
real (r8), parameter :: r103 = -1.7905369937e-01_r8 
real (r8), parameter :: r013 =  2.5254165600e-01_r8 

if (present(stiffened)) then
   do_stiffened = stiffened
else
   do_stiffened = .false.
end if

! reduced variables
ss = sqrt((sa+deltas)/sau)
tt = ct/ctu
pp =  p/zu

rz3 =    r013*tt+r103*ss+r003
rz2 =   (r022*tt+r112*ss+r012)*tt+(r202*ss+r102)*ss+r002
rz1 =      (((r041*tt+r131*ss+r031)*tt                                       &
        +    (r221*ss+r121)*ss+r021)*tt                                      &
        +   ((r311*ss+r211)*ss+r111)*ss+r011)*tt                             &
        +  (((r401*ss+r301)*ss+r201)*ss+r101)*ss+r001
rz0 =    (((((r060*tt+r150*ss+r050)*tt                                       &
        +    (r240*ss+r140)*ss+r040)*tt                                      &
        +   ((r330*ss+r230)*ss+r130)*ss+r030)*tt                             &
        +  (((r420*ss+r320)*ss+r220)*ss+r120)*ss+r020)*tt                    &
        + ((((r510*ss+r410)*ss+r310)*ss+r210)*ss+r110)*ss+r010)*tt           &
        +(((((r600*ss+r500)*ss+r400)*ss+r300)*ss+r200)*ss+r100)*ss+r000

! stiffened density (rdot)
rho = ((rz3*pp + rz2)*pp + rz1)*pp + rz0

if (present(alpha) .or. present(beta) .or. .not. do_stiffened) then

   ! vertical reference profile of density
   r1 = (((((r15*pp + r14)*pp + r13)*pp + r12)*pp + r11)*pp + r10)*pp + 1.0_r8

   ! in-situ density
   if (.not. do_stiffened) rho = r1 * rho 

   ! thermal expansion alpha
   if (present(alpha)) then
      alpha = ((alp003*pp                                                    &
                + alp012*tt + alp102*ss+alp002)*pp                           &
                + ((alp031*tt                                                &
                + alp121*ss+alp021)*tt                                       &
                + (alp211*ss+alp111)*ss+alp011)*tt                           &
                + ((alp301*ss+alp201)*ss+alp101)*ss+alp001)*pp               &
                + ((((alp050*tt                                              &
                + alp140*ss+alp040)*tt                                       &
                + (alp230*ss+alp130)*ss+alp030)*tt                           &
                + ((alp320*ss+alp220)*ss+alp120)*ss+alp020)*tt               &
                + (((alp410*ss+alp310)*ss+alp210)*ss+alp110)*ss+alp010)*tt   &
                + ((((alp500*ss+alp400)*ss+alp300)*ss+alp200)*ss+alp100)*ss  &
                + alp000
      alpha = alpha * r1
   end if

   ! haline contraction beta
   if (present(beta)) then
      beta = ((bet003*pp                                                     &
                + bet012*tt + bet102*ss+bet002)*pp                           &
                + ((bet031*tt                                                &
                + bet121*ss+bet021)*tt                                       &
                + (bet211*ss+bet111)*ss+bet011)*tt                           &
                + ((bet301*ss+bet201)*ss+bet101)*ss+bet001)*pp               &
                + ((((bet050*tt                                              &
                + bet140*ss+bet040)*tt                                       &
                + (bet230*ss+bet130)*ss+bet030)*tt                           &
                + ((bet320*ss+bet220)*ss+bet120)*ss+bet020)*tt               &
                + (((bet410*ss+bet310)*ss+bet210)*ss+bet110)*ss+bet010)*tt   &
                + ((((bet500*ss+bet400)*ss+bet300)*ss+bet200)*ss+bet100)*ss  &
                + bet000
      beta = beta * r1 / ss 
   end if
end if

return
end subroutine
!==========================================================================
elemental subroutine gsw_rho_first_derivatives (sa, ct, p, drho_dsa, &
                                                drho_dct, drho_dp)
!==========================================================================
! 
!  Calculates the three (3) partial derivatives of in-situ density with 
!  respect to Absolute Salinity, Conservative Temperature and pressure.  
!  Note that the pressure derivative is done with respect to pressure in 
!  Pa, not dbar. This function uses the computationally-efficient expression
!  for specific volume in terms of SA, CT and p (Roquet et al., 2014).  
!
!  SA  =  Absolute Salinity                                        [ g/kg ]
!  CT  =  Conservative Temperature (ITS-90)                       [ deg C ]
!  p   =  sea pressure                                             [ dbar ]
!         ( i.e. absolute pressure - 10.1325 dbar )
!
!  drho_dSA  =  partial derivatives of density             [ kg^2/(g m^3) ]
!                 with respect to Absolute Salinity
!  drho_dCT  =  partial derivatives of density               [ kg/(K m^3) ]
!                 with respect to Conservative Temperature
!  drho_dP   =  partial derivatives of density              [ kg/(Pa m^3) ]
!                 with respect to pressure in Pa
!--------------------------------------------------------------------------

use gsw_mod_teos10_constants, only : pa2db, gsw_sfac, offset

use gsw_mod_specvol_coefficients

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sa, ct, p
real (r8), intent(out), optional :: drho_dsa, drho_dct, drho_dp

real (r8) :: rho2, v_ct, v_p, v_sa, xs, ys, z, v

xs = sqrt(gsw_sfac*sa + offset)
ys = ct*0.025_r8
z = p*1e-4_r8

v = v000 + xs*(v010 + xs*(v020 + xs*(v030 + xs*(v040 + xs*(v050 &
    + v060*xs))))) + ys*(v100 + xs*(v110 + xs*(v120 + xs*(v130 + xs*(v140 &
    + v150*xs)))) + ys*(v200 + xs*(v210 + xs*(v220 + xs*(v230 + v240*xs))) &
    + ys*(v300 + xs*(v310 + xs*(v320 + v330*xs)) + ys*(v400 + xs*(v410 &
    + v420*xs) + ys*(v500 + v510*xs + v600*ys))))) + z*(v001 + xs*(v011 &
    + xs*(v021 + xs*(v031 + xs*(v041 + v051*xs)))) + ys*(v101 + xs*(v111 &
    + xs*(v121 + xs*(v131 + v141*xs))) + ys*(v201 + xs*(v211 + xs*(v221 &
    + v231*xs)) + ys*(v301 + xs*(v311 + v321*xs) + ys*(v401 + v411*xs &
    + v501*ys)))) + z*(v002 + xs*(v012 + xs*(v022 + xs*(v032 + v042*xs))) &
    + ys*(v102 + xs*(v112 + xs*(v122 + v132*xs)) + ys*(v202 + xs*(v212 &
    + v222*xs) + ys*(v302 + v312*xs + v402*ys))) + z*(v003 + xs*(v013 &
    + v023*xs) + ys*(v103 + v113*xs + v203*ys) + z*(v004 + v014*xs + v104*ys &
    + z*(v005 + v006*z)))))

rho2 = (1.0_r8/v)**2

if (present(drho_dsa)) then

    v_sa = b000 + xs*(b100 + xs*(b200 + xs*(b300 + xs*(b400 + b500*xs)))) &
           + ys*(b010 + xs*(b110 + xs*(b210 + xs*(b310 + b410*xs))) &
           + ys*(b020 + xs*(b120 + xs*(b220 + b320*xs)) + ys*(b030 &
           + xs*(b130 + b230*xs) + ys*(b040 + b140*xs + b050*ys)))) &
           + z*(b001 + xs*(b101 + xs*(b201 + xs*(b301 + b401*xs))) &
           + ys*(b011 + xs*(b111 + xs*(b211 + b311*xs)) + ys*(b021 &
           + xs*(b121 + b221*xs) + ys*(b031 + b131*xs + b041*ys))) &
           + z*(b002 + xs*(b102 + xs*(b202 + b302*xs))+ ys*(b012 &
           + xs*(b112 + b212*xs) + ys*(b022 + b122*xs + b032*ys)) &
           + z*(b003 +  b103*xs + b013*ys + b004*z)))

    drho_dsa = -rho2*0.5_r8*gsw_sfac*v_sa/xs

end if

if (present(drho_dct)) then

    v_ct = a000 + xs*(a100 + xs*(a200 + xs*(a300 + xs*(a400 + a500*xs)))) &
           + ys*(a010 + xs*(a110 + xs*(a210 + xs*(a310 + a410*xs))) &
           + ys*(a020 + xs*(a120 + xs*(a220 + a320*xs)) + ys*(a030 &
           + xs*(a130 + a230*xs) + ys*(a040 + a140*xs + a050*ys )))) &
           + z*(a001 + xs*(a101 + xs*(a201 + xs*(a301 + a401*xs))) &
           + ys*(a011 + xs*(a111 + xs*(a211 + a311*xs)) + ys*(a021 &
           + xs*(a121 + a221*xs) + ys*(a031 + a131*xs + a041*ys))) &
           + z*(a002 + xs*(a102 + xs*(a202 + a302*xs)) + ys*(a012 &
           + xs*(a112 + a212*xs) + ys*(a022 + a122*xs + a032*ys)) &
           + z*(a003 + a103*xs + a013*ys + a004*z))) 

    drho_dct = -rho2*0.025_r8*v_ct

end if

if (present(drho_dp)) then

    v_p = c000 + xs*(c100 + xs*(c200 + xs*(c300 + xs*(c400 + c500*xs)))) &
          + ys*(c010 + xs*(c110 + xs*(c210 + xs*(c310 + c410*xs))) + ys*(c020 &
          + xs*(c120 + xs*(c220 + c320*xs)) + ys*(c030 + xs*(c130 + c230*xs) &
          + ys*(c040 + c140*xs + c050*ys)))) + z*(c001 + xs*(c101 + xs*(c201 &
          + xs*(c301 + c401*xs))) + ys*(c011 + xs*(c111 + xs*(c211 + c311*xs)) &
          + ys*(c021 + xs*(c121 + c221*xs) + ys*(c031 + c131*xs + c041*ys))) &
          + z*(c002 + xs*(c102 + c202*xs) + ys*(c012 + c112*xs + c022*ys) &
          + z*(c003 + c103*xs + c013*ys + z*(c004 + c005*z))))
  
    drho_dp = -rho2*1e-4_r8*pa2db*v_p

end if

return
end subroutine

!--------------------------------------------------------------------------
!==========================================================================
elemental subroutine gsw_rho_first_derivatives_wrt_enthalpy (sa, ct, p, &
                                                             rho_sa, rho_h)
! =========================================================================
!
!  Calculates two first-order derivatives of specific volume (v).
!  Note that this function uses the using the computationally-efficient
!  expression for specific volume (Roquet et al., 2014).
!
!  SA  =  Absolute Salinity                                        [ g/kg ]
!  CT  =  Conservative Temperature (ITS-90)                       [ deg C ]
!  p   =  sea pressure                                             [ dbar ]
!         ( i.e. absolute pressure - 10.1325 dbar )
!
!  rho_SA =  The first derivative of rho with respect to 
!              Absolute Salinity at constant CT & p.    [ J/(kg (g/kg)^2) ]
!  rho_h  =  The first derivative of rho with respect to 
!              SA and CT at constant p.                  [ J/(kg K(g/kg)) ]
!--------------------------------------------------------------------------

use gsw_mod_toolbox, only : gsw_specvol
use gsw_mod_toolbox, only : gsw_specvol_first_derivatives_wrt_enthalpy

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sa, ct, p
real (r8), intent(out), optional :: rho_sa, rho_h

real (r8) :: rec_v2, v_h, v_sa

if (present(rho_sa) .and. present(rho_h)) then

    call gsw_specvol_first_derivatives_wrt_enthalpy(sa,ct,p,v_sa,v_h)

else if (present(rho_sa)) then

    call gsw_specvol_first_derivatives_wrt_enthalpy(sa,ct,p,v_sa=v_sa)

else if (present(rho_h)) then

    call gsw_specvol_first_derivatives_wrt_enthalpy(sa,ct,p,v_h=v_h)

end if

rec_v2 = (1.0_r8/gsw_specvol(sa,ct,p))**2

if (present(rho_sa)) rho_sa = -v_sa*rec_v2

if (present(rho_h)) rho_h = -v_h*rec_v2

return
end subroutine

!--------------------------------------------------------------------------
!==========================================================================
elemental function gsw_rho_ice (t, p)
!==========================================================================
! 
!  Calculates in-situ density of ice from in-situ temperature and pressure.
!  Note that the output, rho_ice, is density, not density anomaly;  that 
!  is, 1000 kg/m^3 is not subracted from it.  
!
!  t   =  in-situ temperature (ITS-90)                            [ deg C ]
!  p   =  sea pressure                                             [ dbar ]
!         ( i.e. absolute pressure - 10.1325 dbar )
!
!  rho_ice  =  in-situ density of ice (not density anomaly)      [ kg/m^3 ]
!--------------------------------------------------------------------------

use gsw_mod_toolbox, only : gsw_gibbs_ice

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: t, p

real (r8) :: gsw_rho_ice

gsw_rho_ice = 1.0_r8/gsw_gibbs_ice(0,1,t,p)

return
end function

!--------------------------------------------------------------------------
!==========================================================================
elemental subroutine gsw_rho_second_derivatives (sa, ct, p, rho_sa_sa, &
                                  rho_sa_ct, rho_ct_ct, rho_sa_p, rho_ct_p)
!==========================================================================
!
!  Calculates five second-order derivatives of rho. Note that this function
!  uses the using the computationally-efficient expression for specific
!  volume (Roquet et al., 2014).
!
!  SA  =  Absolute Salinity                                        [ g/kg ]
!  CT  =  Conservative Temperature (ITS-90)                       [ deg C ]
!  p   =  sea pressure                                             [ dbar ]
!         ( i.e. absolute pressure - 10.1325 dbar )
!
!  rho_SA_SA = The second-order derivative of rho with respect to 
!              Absolute Salinity at constant CT & p.    [ J/(kg (g/kg)^2) ]
!  rho_SA_CT = The second-order derivative of rho with respect to 
!              SA and CT at constant p.                  [ J/(kg K(g/kg)) ]
!  rho_CT_CT = The second-order derivative of rho with respect to CT at 
!              constant SA & p
!  rho_SA_P  = The second-order derivative with respect to SA & P at 
!              constant CT. 
!  rho_CT_P  = The second-order derivative with respect to CT & P at 
!              constant SA. 
!--------------------------------------------------------------------------

use gsw_mod_toolbox, only : gsw_specvol, gsw_specvol_first_derivatives
use gsw_mod_toolbox, only : gsw_specvol_second_derivatives

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sa, ct, p
real (r8), intent(out), optional :: rho_sa_sa, rho_sa_ct, rho_ct_ct
real (r8), intent(out), optional :: rho_sa_p, rho_ct_p

integer :: iflag1, iflag2
real (r8) :: rec_v, rec_v2, rec_v3, v_ct, v_ct_ct, v_ct_p, v_p, v_sa, v_sa_ct
real (r8) :: v_sa_p, v_sa_sa

iflag1 = 0
if (present(rho_sa_sa) .or. present(rho_sa_ct) &
                       .or. present(rho_sa_p)) iflag1 = ibset(iflag1,1)
if (present(rho_sa_ct) .or. present(rho_ct_ct) &
                       .or. present(rho_ct_p)) iflag1 = ibset(iflag1,2)
if (present(rho_sa_p) .or. present(rho_ct_p)) iflag1 = ibset(iflag1,3)

call gsw_specvol_first_derivatives(sa,ct,p,v_sa,v_ct,v_p,iflag=iflag1)

iflag2 = 0
if (present(rho_sa_sa)) iflag2 = ibset(iflag2,1)
if (present(rho_sa_ct)) iflag2 = ibset(iflag2,2)
if (present(rho_ct_ct)) iflag2 = ibset(iflag2,3)
if (present(rho_sa_p)) iflag2 = ibset(iflag2,4)
if (present(rho_ct_p)) iflag2 = ibset(iflag2,5)

call gsw_specvol_second_derivatives(sa,ct,p,v_sa_sa,v_sa_ct,v_ct_ct, &
                                    v_sa_p,v_ct_p,iflag=iflag2)

rec_v = 1.0_r8/gsw_specvol(sa,ct,p)
rec_v2 = rec_v**2
rec_v3 = rec_v2*rec_v

if (present(rho_sa_sa)) rho_sa_sa = -v_sa_sa*rec_v2 + 2.0_r8*v_sa*v_sa*rec_v3

if (present(rho_sa_ct)) rho_sa_ct = -v_sa_ct*rec_v2 + 2.0_r8*v_sa*v_ct*rec_v3

if (present(rho_ct_ct)) rho_ct_ct = -v_ct_ct*rec_v2 + 2.0_r8*v_ct*v_ct*rec_v3

if (present(rho_sa_p)) rho_sa_p = -v_sa_p*rec_v2 + 2.0_r8*v_sa*v_p*rec_v3

if (present(rho_ct_p)) rho_ct_p = -v_ct_p*rec_v2 + 2.0_r8*v_ct*v_p*rec_v3

return
end subroutine

!--------------------------------------------------------------------------
!==========================================================================
elemental subroutine gsw_rho_second_derivatives_wrt_enthalpy (sa, ct, p, &
                                              rho_sa_sa, rho_sa_h, rho_h_h)
! =========================================================================
!
!  Calculates three second-order derivatives of rho with respect to enthalpy.
!  Note that this function uses the using the computationally-efficient
!  expression for specific volume (Roquet et al., 2014).
!
!  SA  =  Absolute Salinity                                        [ g/kg ]
!  CT  =  Conservative Temperature (ITS-90)                       [ deg C ]
!  p   =  sea pressure                                             [ dbar ]
!         ( i.e. absolute pressure - 10.1325 dbar )
!
!  rho_SA_SA = The second-order derivative of rho with respect to 
!              Absolute Salinity at constant h & p.     [ J/(kg (g/kg)^2) ]
!  rho_SA_h  = The second-order derivative of rho with respect to 
!              SA and h at constant p.                   [ J/(kg K(g/kg)) ]
!  rho_h_h   = The second-order derivative of rho with respect to h at 
!              constant SA & p
!--------------------------------------------------------------------------

use gsw_mod_toolbox, only : gsw_specvol
use gsw_mod_toolbox, only : gsw_specvol_first_derivatives_wrt_enthalpy
use gsw_mod_toolbox, only : gsw_specvol_second_derivatives_wrt_enthalpy

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sa, ct, p
real (r8), intent(out), optional :: rho_sa_sa, rho_sa_h, rho_h_h

integer :: iflag1, iflag2
real (r8) :: rec_v, rec_v2, rec_v3, v_h, v_h_h, v_sa, v_sa_h, v_sa_sa

iflag1 = 0
if (present(rho_sa_sa) .or. present(rho_sa_h)) iflag1 = ibset(iflag1,1)
if (present(rho_sa_h) .or. present(rho_h_h)) iflag1 = ibset(iflag1,2)

call gsw_specvol_first_derivatives_wrt_enthalpy(sa,ct,p,v_sa,v_h,iflag=iflag1)

iflag2 = 0
if (present(rho_sa_sa)) iflag2 = ibset(iflag2,1)
if (present(rho_sa_h)) iflag2 = ibset(iflag2,2)
if (present(rho_h_h)) iflag2 = ibset(iflag2,3)

call gsw_specvol_second_derivatives_wrt_enthalpy(sa,ct,p,v_sa_sa,v_sa_h,v_h_h, &
                                                 iflag=iflag2)

rec_v = 1.0_r8/gsw_specvol(sa,ct,p)
rec_v2 = rec_v**2
rec_v3 = rec_v2*rec_v

if (present(rho_sa_sa)) rho_sa_sa = -v_sa_sa*rec_v2 + 2.0_r8*v_sa*v_sa*rec_v3

if (present(rho_sa_h)) rho_sa_h = -v_sa_h*rec_v2 + 2.0_r8*v_sa*v_h*rec_v3

if (present(rho_h_h)) rho_h_h = -v_h_h*rec_v2 + 2.0_r8*v_h*v_h*rec_v3

return
end subroutine

!--------------------------------------------------------------------------
!==========================================================================
elemental function gsw_rho_t_exact (sa, t, p) 
!==========================================================================
!
! Calculates in-situ density of seawater from Absolute Salinity and 
! in-situ temperature.
!
! sa     : Absolute Salinity                               [g/kg]
! t      : in-situ temperature                             [deg C]
! p      : sea pressure                                    [dbar]
! 
! gsw_rho_t_exact : in-situ density                        [kg/m^3]
!--------------------------------------------------------------------------

use gsw_mod_toolbox, only : gsw_gibbs

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sa, t, p 

real (r8) :: gsw_rho_t_exact

integer, parameter :: n0=0, n1=1

gsw_rho_t_exact = 1.0_r8/gsw_gibbs(n0,n0,n1,sa,t,p)

return
end function

!--------------------------------------------------------------------------
!==========================================================================
pure subroutine gsw_rr68_interp_sa_ct (sa, ct, p, p_i, sa_i, ct_i)
!==========================================================================
!
!  Interpolate Absolute Salinity and Conservative Temperature values to
!  arbitrary pressures using the Reiniger and Ross (1968) interpolation
!  scheme.
!  Note that this interpolation scheme requires at least four observed
!  bottles on the cast.
!
!  SA   =  Absolute Salinity                                  [ g/kg ]
!  CT   =  Conservative Temperature (ITS-90)                 [ deg C ]
!  p    =  sea pressure                                       [ dbar ]
!           ( i.e. absolute pressure - 10.1325 dbar )
!  p_i  =  pressures to interpolate to.
!
!  SA_i = interpolated SA values at pressures p_i.
!  CT_i = interpolated CT values at pressures p_i.
!--------------------------------------------------------------------------

use gsw_mod_toolbox, only : gsw_util_interp1q_int

use gsw_mod_error_functions, only : gsw_error_code

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sa(:), ct(:), p(:), p_i(:)
real (r8), intent(out) :: sa_i(:), ct_i(:)

integer :: i, j, mp, mp_i, nshallow, ncentral, ndeep

integer, allocatable :: ip(:), ip_i(:)
integer, allocatable :: ip_ishallow(:), ip_icentral(:), ip_ideep(:)
logical, allocatable :: shallow(:), central(:), deep(:)
real (r8), allocatable :: ip_shallow(:), ip_central(:), ip_deep(:)
real (r8), allocatable :: dp(:)

character (*), parameter :: func_name = "gsw_rr68_interp_sa_ct"

mp = size(p)
if (mp .lt. 4) then  ! need at least four bottles to perform this interpolation
    sa_i = gsw_error_code(1,func_name)
    ct_i = sa_i
    return
end if

allocate(dp(mp-1))
dp = p(2:mp) - p(1:mp-1)
if (any(dp .le. 0.0_r8)) then  ! pressure must be monotonic
    sa_i = gsw_error_code(2,func_name)
    ct_i = sa_i
    return
end if

mp_i = size(p_i)

allocate(shallow(mp_i),central(mp_i),deep(mp_i))
shallow = (p_i >= p(1)    .and. p_i <= p(2)   ); nshallow=count(shallow)
central = (p_i >= p(2)    .and. p_i <= p(mp-1)); ncentral=count(central)
deep    = (p_i >= p(mp-1) .and. p_i <= p(mp)  ); ndeep=count(deep)

if ((nshallow .eq. 0) .or. (ncentral .eq. 0) .or. (ndeep .eq. 0)) then
    sa_i = gsw_error_code(3,func_name)
    ct_i = sa_i
    return
end if

allocate(ip(mp),ip_i(mp_i))
ip = (/ (i, i=1,mp) /)
ip_i = (/ (i, i=1,mp_i) /)

allocate(ip_shallow(nshallow),ip_central(ncentral),ip_deep(ndeep))
allocate(ip_ishallow(nshallow),ip_icentral(ncentral),ip_ideep(ndeep))

! Calculate the 2 outer extrapolated values and the inner interpolated values

ip_icentral = pack(ip_i,central)
ip_central = gsw_util_interp1q_int(p,ip,p_i(ip_icentral))
call rr68_interp_section(0,ip_central,ip_icentral,sa_i,ct_i)

ip_ishallow = pack(ip_i,shallow)
ip_shallow = gsw_util_interp1q_int(p,ip,p_i(ip_ishallow))
call rr68_interp_section(-1,ip_shallow,ip_ishallow,sa_i,ct_i)

ip_ideep = pack(ip_i,deep)
ip_deep = gsw_util_interp1q_int(p,ip,p_i(ip_ideep))
call rr68_interp_section(+1,ip_deep,ip_ideep,sa_i,ct_i)

! Insert any observed bottles that are at the required interpolated pressures
do i = 1, mp_i
   do j = 1, mp
      if (p(j) .eq. p_i(i)) then
         sa_i(i) = sa(j)
         ct_i(i) = ct(j)
      end if
   end do
end do

return

contains

    pure subroutine rr68_interp_section (sectnum, ip_sect, ip_isect, sa_i, ct_i)

    use gsw_mod_toolbox, only : gsw_linear_interp_sa_ct

    implicit none

    integer, intent(in) :: sectnum, ip_isect(:)
    real (r8), intent(in) :: ip_sect(:)
    real (r8), intent(inout) :: sa_i(:), ct_i(:)

    integer nsect
    real (r8) :: m

    integer, allocatable :: ip_1(:), ip_2(:), ip_3(:), ip_4(:)
    real (r8), allocatable :: ct_12(:), ct_13(:), ct_23(:), ct_34(:), ctp1(:)
    real (r8), allocatable :: ctp2(:), ct_ref(:), ctref_denom(:)
    real (r8), allocatable :: ct_ref_minus_ctp1(:), ct_ref_minus_ctp2(:)
    real (r8), allocatable :: ctref_num(:)
    real (r8), allocatable :: gamma1_23(:), gamma1_24(:), gamma2_31(:)
    real (r8), allocatable :: gamma2_34(:), gamma2_41(:), gamma3_12(:)
    real (r8), allocatable :: gamma3_42(:), gamma4_12(:), gamma4_23(:)
    real (r8), allocatable :: sa_12(:), sa_13(:), sa_23(:), sa_34(:), sap1(:)
    real (r8), allocatable :: sap2(:), sa_ref(:), saref_denom(:)
    real (r8), allocatable :: sa_ref_minus_sap1(:), sa_ref_minus_sap2(:)
    real (r8), allocatable :: saref_num(:)

    nsect = size(ip_sect)
    allocate(ip_1(nsect),ip_2(nsect),ip_3(nsect),ip_4(nsect))

    allocate(sa_12(nsect),sa_13(nsect),sa_23(nsect),sa_34(nsect))
    allocate(sa_ref(nsect),saref_num(nsect),saref_denom(nsect))
    allocate(sap1(nsect),sap2(nsect),sa_ref_minus_sap1(nsect))
    allocate(sa_ref_minus_sap2(nsect))

    allocate(ct_12(nsect),ct_13(nsect),ct_23(nsect),ct_34(nsect))
    allocate(ct_ref(nsect),ctref_num(nsect),ctref_denom(nsect))
    allocate(ctp1(nsect),ctp2(nsect),ct_ref_minus_ctp1(nsect))
    allocate(ct_ref_minus_ctp2(nsect))

    allocate(gamma1_23(nsect),gamma2_31(nsect),gamma3_12(nsect))
    if (sectnum .eq. 0) then
        allocate(gamma2_34(nsect),gamma3_42(nsect),gamma4_23(nsect))
    else
        allocate(gamma1_24(nsect),gamma2_41(nsect),gamma4_12(nsect))
    end if

    if (sectnum .lt. 0) then       !  shallow
        ip_1 = floor(ip_sect)
        ip_2 = ceiling(ip_sect)
!        where (ip_1 .eq. ip_2) ip_2 = ip_1 + 1
        ip_3 = ip_2 + 1
        ip_4 = ip_3 + 1
    else if (sectnum .eq. 0) then  !  central
        ip_2 = floor(ip_sect)
        ip_3 = ceiling(ip_sect)
!        where (ip_2 .eq. ip_3) ip_2 = ip_3 - 1
        ip_1 = ip_2 - 1
!        where (ip_1 .lt. 1)
!            ip_1 = 1
!            ip_2 = 2
!            ip_3 = 3
!        end where
        ip_4 = ip_3 + 1
    else if (sectnum .gt. 0) then  !  deep
        ip_1 = ceiling(ip_sect)
        ip_2 = floor(ip_sect)
!        where (ip_1 .eq. ip_2) ip_2 = ip_1 - 1
        ip_3 = ip_2 - 1
        ip_4 = ip_3 - 1
    end if

    !eqn (3d)
    sa_34 = sa(ip_3) + ((sa(ip_4) - sa(ip_3))*(p_i(ip_isect) - p(ip_3))/ &
        (p(ip_4) - p(ip_3)))
    ct_34 = ct(ip_3) + ((ct(ip_4) - ct(ip_3))*(p_i(ip_isect) - p(ip_3))/ &
        (p(ip_4) - p(ip_3)))

    ! Construct the Reiniger & Ross reference curve equation.
    ! m = the power variable
    m = 1.7_r8

    if (sectnum .eq. 0) then

        sa_12 = sa(ip_1) + ((sa(ip_2) - sa(ip_1))*(p_i(ip_isect) - p(ip_1))/ &
            (p(ip_2) - p(ip_1)))
        ct_12 = ct(ip_1) + ((ct(ip_2) - ct(ip_1))*(p_i(ip_isect) - p(ip_1))/ &
            (p(ip_2) - p(ip_1)))

        call gsw_linear_interp_sa_ct(sa,ct,p,p_i(ip_isect),sa_23,ct_23)

        ! eqn (3a)
        saref_num = (abs(sa_23-sa_34)**m)*sa_12 + (abs(sa_12-sa_23)**m)*sa_34 
        ctref_num = (abs(ct_23-ct_34)**m)*ct_12 + (abs(ct_12-ct_23)**m)*ct_34

        saref_denom = abs(sa_23-sa_34)**m + abs(sa_12-sa_23)**m 
        ctref_denom = abs(ct_23-ct_34)**m + abs(ct_12-ct_23)**m 

        where (saref_denom .eq. 0.0_r8)
            sa_23 = sa_23 + 1.0e-6_r8
            saref_num = (abs(sa_23-sa_34)**m)*sa_12+(abs(sa_12-sa_23)**m)*sa_34
            saref_denom = abs(sa_23-sa_34)**m + abs(sa_12-sa_23)**m 
        end where
        where (ctref_denom .eq. 0.0_r8)
            ct_23 = ct_23 + 1.0e-6_r8
            ctref_num = (abs(ct_23-ct_34)**m)*ct_12+(abs(ct_12-ct_23)**m)*ct_34
            ctref_denom = abs(ct_23-ct_34)**m + abs(ct_12-ct_23)**m 
        end where

        sa_ref = 0.5_r8*(sa_23 + (saref_num/saref_denom))
        ct_ref = 0.5_r8*(ct_23 + (ctref_num/ctref_denom))

    else

        call gsw_linear_interp_sa_ct(sa,ct,p,p_i(ip_isect),sa_12,ct_12)

        sa_13 = sa(ip_1) + ((sa(ip_3) - sa(ip_1))*(p_i(ip_isect) - p(ip_1))/&
            (p(ip_3) - p(ip_1)))
        ct_13 = ct(ip_1) + ((ct(ip_3) - ct(ip_1))*(p_i(ip_isect) - p(ip_1))/&
            (p(ip_3) - p(ip_1)))

        sa_23 = sa(ip_2) + ((sa(ip_3) - sa(ip_2))*(p_i(ip_isect) - p(ip_2))/ &
            (p(ip_3) - p(ip_2)))
        ct_23 = ct(ip_2) + ((ct(ip_3) - ct(ip_2))*(p_i(ip_isect) - p(ip_2))/ &
            (p(ip_3) - p(ip_2)))

        !eqn (3a')
        saref_num = (abs(sa_12-sa_23)**m)*sa_34 + (abs(sa_12-sa_13)**m)*sa_23 
        ctref_num = (abs(ct_12-ct_23)**m)*ct_34 + (abs(ct_12-ct_13)**m)*ct_23

        saref_denom = abs(sa_12-sa_23)**m + abs(sa_12-sa_13)**m 
        ctref_denom = abs(ct_12-ct_23)**m + abs(ct_12-ct_13)**m 

        where (saref_denom .eq. 0.0_r8)
            sa_23 = sa_23 + 1.0e-6_r8
            saref_num = (abs(sa_12-sa_23)**m)*sa_34+(abs(sa_12-sa_13)**m)*sa_23
            saref_denom = abs(sa_12-sa_23)**m + abs(sa_12-sa_13)**m 
        end where
        where (ctref_denom .eq. 0.0_r8)
            ct_23 = ct_23 + 1.0e-6_r8
            ctref_num = (abs(ct_12-ct_23)**m)*ct_34+(abs(ct_12-ct_13)**m)*ct_23
            ctref_denom = abs(ct_12-ct_23)**m + abs(ct_12-ct_13)**m 
        end where

        sa_ref = 0.5_r8*(sa_12 + (saref_num/saref_denom))
        ct_ref = 0.5_r8*(ct_12 + (ctref_num/ctref_denom))

    end if

    !eqn (3c)
    gamma1_23 = ((p_i(ip_isect) - p(ip_2))*(p_i(ip_isect) - p(ip_3)))/ &
        ((p(ip_1) - p(ip_2))*(p(ip_1) - p(ip_3)))
    gamma2_31 = ((p_i(ip_isect) - p(ip_3))*(p_i(ip_isect) - p(ip_1)))/ &
        ((p(ip_2) - p(ip_3))*(p(ip_2) - p(ip_1)))
    gamma3_12 = ((p_i(ip_isect) - p(ip_1))*(p_i(ip_isect) - p(ip_2)))/ &
        ((p(ip_3) - p(ip_1))*(p(ip_3) - p(ip_2)))

    if (sectnum .eq. 0) then 
        gamma2_34 = ((p_i(ip_isect) - p(ip_3))*(p_i(ip_isect) - p(ip_4)))/ &
            ((p(ip_2) - p(ip_3))*(p(ip_2) - p(ip_4)))
        gamma3_42 = ((p_i(ip_isect) - p(ip_4))*(p_i(ip_isect) - p(ip_2)))/ &
            ((p(ip_3) - p(ip_4))*(p(ip_3) - p(ip_2)))
        gamma4_23 = ((p_i(ip_isect) - p(ip_2))*(p_i(ip_isect) - p(ip_3)))/ &
            ((p(ip_4) - p(ip_2))*(p(ip_4) - p(ip_3)))
    else
        gamma1_24 = ((p_i(ip_isect) - p(ip_2))*(p_i(ip_isect) - p(ip_4)))/ &
            ((p(ip_1) - p(ip_2))*(p(ip_1) - p(ip_4)))
        gamma2_41 = ((p_i(ip_isect) - p(ip_4))*(p_i(ip_isect) - p(ip_1)))/ &
            ((p(ip_2) - p(ip_4))*(p(ip_2) - p(ip_1)))
        gamma4_12 = ((p_i(ip_isect) - p(ip_1))*(p_i(ip_isect) - p(ip_2)))/ &
            ((p(ip_4) - p(ip_1))*(p(ip_4) - p(ip_2)))
    end if

    !eqn (3b/3b')
    sap1 = gamma1_23*sa(ip_1) + gamma2_31*sa(ip_2) + gamma3_12*sa(ip_3)
    ctp1 = gamma1_23*ct(ip_1) + gamma2_31*ct(ip_2) + gamma3_12*ct(ip_3)
    if (sectnum .eq. 0) then
        sap2 = gamma2_34*sa(ip_2) + gamma3_42*sa(ip_3) + gamma4_23*sa(ip_4)
        ctp2 = gamma2_34*ct(ip_2) + gamma3_42*ct(ip_3) + gamma4_23*ct(ip_4)
    else
        sap2 = gamma1_24*sa(ip_1) + gamma2_41*sa(ip_2) + gamma4_12*sa(ip_4)
        ctp2 = gamma1_24*ct(ip_1) + gamma2_41*ct(ip_2) + gamma4_12*ct(ip_4)
    end if

    !eqn (3)
    sa_ref_minus_sap1 = abs(sa_ref - sap1)
    sa_ref_minus_sap2 = abs(sa_ref - sap2)
    where (sa_ref_minus_sap1 .eq. 0.0_r8 .and. sa_ref_minus_sap2 .eq. 0.0_r8)
        sa_ref = sa_ref + 1.0e-6_r8
        sa_ref_minus_sap1 = abs(sa_ref - sap1)
        sa_ref_minus_sap2 = abs(sa_ref - sap2)
    end where

    ct_ref_minus_ctp1 = abs(ct_ref - ctp1)
    ct_ref_minus_ctp2 = abs(ct_ref - ctp2)
    where (ct_ref_minus_ctp1 .eq. 0.0_r8 .and. ct_ref_minus_ctp2 .eq. 0.0_r8)
        ct_ref = ct_ref + 1.0e-6_r8
        ct_ref_minus_ctp1 = abs(ct_ref - ctp1)
        ct_ref_minus_ctp2 = abs(ct_ref - ctp2)
    end where

    sa_i(ip_isect) = (sa_ref_minus_sap1*sap2 + sa_ref_minus_sap2*sap1) / &
                        (sa_ref_minus_sap1 + sa_ref_minus_sap2)
    ct_i(ip_isect) = (ct_ref_minus_ctp1*ctp2 + ct_ref_minus_ctp2*ctp1) / &
                        (ct_ref_minus_ctp1 + ct_ref_minus_ctp2)
    return
    end subroutine rr68_interp_section

end subroutine

!--------------------------------------------------------------------------
!==========================================================================
elemental function gsw_sa_freezing_estimate (p, saturation_fraction, ct, t)
!==========================================================================
!
! Form an estimate of SA from a polynomial in CT and p 
!
!--------------------------------------------------------------------------

use gsw_mod_teos10_constants, only : gsw_sso

use gsw_mod_toolbox, only : gsw_ct_from_t

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: p, saturation_fraction
real (r8), intent(in), optional :: ct, t

real (r8) :: gsw_sa_freezing_estimate

real (r8) :: ctx, ctsat, sa

! note that aa = 0.502500117621d0/35.16504
real (r8), parameter :: aa = 0.014289763856964_r8
real (r8), parameter :: bb = 0.057000649899720_r8

real (r8), parameter :: p0  =  2.570124672768757e-1_r8
real (r8), parameter :: p1  = -1.917742353032266e1_r8
real (r8), parameter :: p2  = -1.413382858617969e-2_r8
real (r8), parameter :: p3  = -5.427484830917552e-1_r8
real (r8), parameter :: p4  = -4.126621135193472e-4_r8
real (r8), parameter :: p5  = -4.176407833276121e-7_r8
real (r8), parameter :: p6  =  4.688217641883641e-5_r8
real (r8), parameter :: p7  = -3.039808885885726e-8_r8
real (r8), parameter :: p8  = -4.990118091261456e-11_r8
real (r8), parameter :: p9  = -9.733920711119464e-9_r8
real (r8), parameter :: p10 = -7.723324202726337e-12_r8
real (r8), parameter :: p11 =  7.121854166249257e-16_r8
real (r8), parameter :: p12 =  1.256474634100811e-12_r8
real (r8), parameter :: p13 =  2.105103897918125e-15_r8
real (r8), parameter :: p14 =  8.663811778227171e-19_r8

! A very rough estimate of sa to get the saturated ct
if (present(ct)) then
    sa = max(-(ct + 9e-4_r8*p)/0.06_r8, 0.0_r8)
    ctx = ct
else if (present(t)) then
    sa = max(-(t + 9e-4_r8*p)/0.06_r8, 0.0_r8)
    ctx = gsw_ct_from_t(sa,t,p)
else
    gsw_sa_freezing_estimate = 0.0_r8
    return
end if

! CTsat is the estimated value of CT if the seawater were saturated with
! dissolved air, recognizing that it actually has the air fraction
! saturation_fraction; see McDougall, Barker and Feistel, 2014).  

ctsat = ctx - (1.0_r8-saturation_fraction)* &
        (1e-3_r8)*(2.4_r8-aa*sa)*(1.0_r8+bb*(1.0_r8-sa/gsw_sso))

gsw_sa_freezing_estimate = p0 &
        + p*(p2 + p4*ctsat + p*(p5 + ctsat*(p7 + p9*ctsat) &
        + p*(p8  + ctsat*(p10 + p12*ctsat) + p*(p11 + p13*ctsat + p14*p)))) &
        + ctsat*(p1 + ctsat*(p3 + p6*p))

return
end function

!--------------------------------------------------------------------------
!==========================================================================
elemental function gsw_sa_freezing_from_ct (ct, p, saturation_fraction)
!==========================================================================
!
!  Calculates the Absolute Salinity of seawater at the freezing temperature.  
!  That is, the output is the Absolute Salinity of seawater, with 
!  Conservative Temperature CT, pressure p and the fraction 
!  saturation_fraction of dissolved air, that is in equilibrium 
!  with ice at the same in situ temperature and pressure.  If the input 
!  values are such that there is no positive value of Absolute Salinity for
!  which seawater is frozen, the output is made a NaN.
!
!  CT  =  Conservative Temperature of seawater (ITS-90)           [ deg C ]
!  p   =  sea pressure                                             [ dbar ]
!         ( i.e. absolute pressure - 10.1325 dbar ) 
!  saturation_fraction  =  the saturation fraction of dissolved air in 
!                          seawater
!
!  sa_freezing_from_ct  =  Absolute Salinity of seawater when it freezes,
!                 for given input values of its Conservative Temperature,
!                 pressure and air saturation fraction.            [ g/kg ]
!--------------------------------------------------------------------------

use gsw_mod_toolbox, only : gsw_ct_freezing_first_derivatives, gsw_sa_p_inrange
use gsw_mod_toolbox, only : gsw_ct_freezing_exact, gsw_sa_freezing_estimate

use gsw_mod_error_functions, only : gsw_error_code, gsw_error_limit

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: ct, p, saturation_fraction

real (r8) :: gsw_sa_freezing_from_ct

integer :: i_iter
real (r8) :: ct_freezing_zero_sa, f, ctfreezing_sa
real (r8) :: sa, sa_mean, sa_old

integer, parameter :: number_of_iterations = 3

! This is the band of sa within +- 2.5 g/kg of sa = 0, which we treat
! differently in calculating the initial values of both SA and dCT_dSA. 
real (r8), parameter :: sa_cut_off = 2.5_r8

character (*), parameter :: func_name = "gsw_sa_freezing_from_ct"

! Find CT > CT_freezing_zero_SA.  If this is the case, the input values
! represent seawater that is not frozen (for any positive SA). 
ct_freezing_zero_sa = gsw_ct_freezing_exact(0.0_r8,p,saturation_fraction)
if (ct .gt. ct_freezing_zero_sa) then
    gsw_sa_freezing_from_ct = gsw_error_code(1,func_name)
    return
end if

! Form the first estimate of SA from a polynomial in CT and p 
sa = gsw_sa_freezing_estimate(p,saturation_fraction,ct=ct)
if (sa .lt. -sa_cut_off) then
    gsw_sa_freezing_from_ct = gsw_error_code(2,func_name)
    return
end if

! Form the first estimate of CTfreezing_SA, the derivative of CT_freezing 
! with respect to SA at fixed p.  
sa = max(sa,0.0_r8)
call gsw_ct_freezing_first_derivatives(sa,p,saturation_fraction, &
                                       ctfreezing_sa=ctfreezing_sa)

! For -SA_cut_off < SA < SA_cut_off, replace the above estimate of SA  
! with one based on (CT_freezing_zero_SA - CT).
if (abs(sa) .lt. sa_cut_off) sa = (ct - ct_freezing_zero_sa)/ctfreezing_sa

!--------------------------------------------------------------------------
! Begin the modified Newton-Raphson method to solve  
! f = (CT_freezing - CT) = 0 for SA. 
!--------------------------------------------------------------------------
do i_iter = 1, number_of_iterations
    sa_old = sa
    f = gsw_ct_freezing_exact(sa,p,saturation_fraction) - ct
    sa = sa_old - f/ctfreezing_sa
    sa_mean = 0.5_r8*(sa + sa_old) 
    call gsw_ct_freezing_first_derivatives(sa_mean,p,saturation_fraction, &
                                           ctfreezing_sa=ctfreezing_sa)
    sa = sa_old - f/ctfreezing_sa
end do

if (gsw_sa_p_inrange(sa,p)) then
    gsw_sa_freezing_from_ct = sa
else
    gsw_sa_freezing_from_ct = gsw_error_code(3,func_name)
end if

return
end function

!--------------------------------------------------------------------------
!==========================================================================
elemental function gsw_sa_freezing_from_ct_poly (ct, p, saturation_fraction)
!==========================================================================
!
!  Calculates the Absolute Salinity of seawater at the freezing temperature.  
!  That is, the output is the Absolute Salinity of seawater, with the 
!  fraction saturation_fraction of dissolved air, that is in equilibrium 
!  with ice at Conservative Temperature CT and pressure p.  If the input 
!  values are such that there is no positive value of Absolute Salinity for
!  which seawater is frozen, the output is put equal to Nan.
!
!  CT  =  Conservative Temperature (ITS-90)                       [ deg C ]
!  p   =  sea pressure                                             [ dbar ]
!         ( i.e. absolute pressure - 10.1325 dbar ) 
!  saturation_fraction  =  the saturation fraction of dissolved air in 
!                          seawater
!
!  sa_freezing_from_ct  =  Absolute Salinity of seawater when it freezes,
!                 for given input values of Conservative Temperature
!                 pressure and air saturation fraction.            [ g/kg ]
!--------------------------------------------------------------------------

use gsw_mod_toolbox, only : gsw_ct_freezing_poly, gsw_sa_p_inrange
use gsw_mod_toolbox, only : gsw_ct_freezing_first_derivatives_poly
use gsw_mod_toolbox, only : gsw_sa_freezing_estimate

use gsw_mod_error_functions, only : gsw_error_code, gsw_error_limit

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: ct, p, saturation_fraction

real (r8) :: gsw_sa_freezing_from_ct_poly

integer :: i_iter
real (r8) :: ct_freezing, ct_freezing_zero_sa, dct_dsa
real (r8) :: sa, sa_old, sa_mean

integer, parameter :: number_of_iterations = 2

! This is the band of sa within +- 2.5 g/kg of sa = 0, which we treat
! differently in calculating the initial values of both SA and dCT_dSA. 
real (r8), parameter :: sa_cut_off = 2.5_r8

character (*), parameter :: func_name = "gsw_sa_freezing_from_ct_poly"

! Find CT > CT_freezing_zero_SA.  If this is the case, the input values
! represent seawater that is not frozen (at any positive SA). 
ct_freezing_zero_sa = gsw_ct_freezing_poly(0.0_r8,p,saturation_fraction)
if (ct .gt. ct_freezing_zero_sa) then
    gsw_sa_freezing_from_ct_poly = gsw_error_code(1,func_name)
    return
endif

! Form the first estimate of SA from a polynomial in CT and p 
sa = gsw_sa_freezing_estimate(p,saturation_fraction,ct=ct)
if (sa .lt. -sa_cut_off) then
    gsw_sa_freezing_from_ct_poly = gsw_error_code(2,func_name)
    return
endif

! Form the first estimate of dCT_dSA, the derivative of CT with respect 
! to SA at fixed p.  
sa = max(sa,0.0_r8)
call gsw_ct_freezing_first_derivatives_poly(sa,p,saturation_fraction, &
                                            ctfreezing_sa=dct_dsa)

! For -SA_cut_off < SA < SA_cut_off, replace the above estimate of SA  
! with one based on (CT_freezing_zero_SA - CT).
if (abs(sa) .lt. sa_cut_off) sa = (ct - ct_freezing_zero_sa)/dct_dsa

!--------------------------------------------------------------------------
! Begin the modified Newton-Raphson method to solve the root of 
! CT_freezing = CT for SA. 
!--------------------------------------------------------------------------
do i_iter = 1, number_of_iterations
    sa_old = sa
    ct_freezing = gsw_ct_freezing_poly(sa_old,p,saturation_fraction)
    sa = sa_old - (ct_freezing - ct)/dct_dsa
    sa_mean = 0.5_r8*(sa + sa_old)
    call gsw_ct_freezing_first_derivatives_poly(sa_mean,p,saturation_fraction, &
                                                ctfreezing_sa=dct_dsa)
    sa = sa_old - (ct_freezing - ct)/dct_dsa
enddo

if (gsw_sa_p_inrange(sa,p)) then
    gsw_sa_freezing_from_ct_poly = sa
else
    gsw_sa_freezing_from_ct_poly = gsw_error_code(3,func_name)
endif

return
end function

!--------------------------------------------------------------------------
!==========================================================================
elemental function gsw_sa_freezing_from_t (t, p, saturation_fraction)
!==========================================================================
!
!  Calculates the Absolute Salinity of seawater at the freezing temperature.  
!  That is, the output is the Absolute Salinity of seawater, with the 
!  fraction saturation_fraction of dissolved air, that is in equilibrium 
!  with ice at in-situ temperature t and pressure p.  If the input values 
!  are such that there is no positive value of Absolute Salinity for which 
!  seawater is frozen, the output is set to NaN.  
!
!  t  =  in-situ Temperature (ITS-90)                             [ deg C ]
!  p  =  sea pressure                                              [ dbar ]
!        ( i.e. absolute pressure - 10.1325 dbar ) 
!  saturation_fraction = the saturation fraction of dissolved air in 
!                        seawater
!  (i.e., saturation_fraction must be between 0 and 1, and the default 
!    is 1, completely saturated) 
!
!  sa_freezing_from_t  =  Absolute Salinity of seawater when it freezes, for 
!                given input values of in situ temperature, pressure and 
!                air saturation fraction.                          [ g/kg ] 
!---------------------------------------------------------------------------

use gsw_mod_toolbox, only : gsw_t_freezing_exact, gsw_sa_p_inrange
use gsw_mod_toolbox, only : gsw_t_freezing_first_derivatives
use gsw_mod_toolbox, only : gsw_sa_freezing_estimate

use gsw_mod_error_functions, only : gsw_error_code, gsw_error_limit

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: t, p, saturation_fraction

real (r8) :: gsw_sa_freezing_from_t

integer :: i_iter
real (r8) :: f, sa, sa_mean, sa_old, t_freezing_zero_sa, tfreezing_sa

integer, parameter :: number_of_iterations = 2

! This is the band of sa within +- 2.5 g/kg of sa = 0, which we treat
! differently in calculating the initial values of both SA and dCT_dSA. 
real (r8), parameter :: sa_cut_off = 2.5_r8

character (*), parameter :: func_name = "gsw_sa_freezing_from_t"

! Find t > t_freezing_zero_SA.  If this is the case, the input values
! represent seawater that is not frozen (at any positive SA). 
t_freezing_zero_sa = gsw_t_freezing_exact(0.0_r8,p,saturation_fraction)
if (t .gt. t_freezing_zero_sa) then
    gsw_sa_freezing_from_t = gsw_error_code(1,func_name)
    return
end if
 
! This is the inital guess of SA using a purpose-built polynomial in CT and p  
sa = gsw_sa_freezing_estimate(p,saturation_fraction,t=t)
if (sa .lt. -sa_cut_off) then
    gsw_sa_freezing_from_t = gsw_error_code(2,func_name)
    return
end if

! Form the first estimate of tfreezing_SA, the derivative of CT_freezing  
! with respect to SA at fixed p.    
sa = max(sa,0.0_r8)
call gsw_t_freezing_first_derivatives(sa,p,saturation_fraction, &
                                      tfreezing_sa=tfreezing_sa)

! For -SA_cut_off < SA < SA_cut_off, replace the above estimate of SA  
! with one based on (t_freezing_zero_SA - t).
if (abs(sa) .lt. sa_cut_off) sa = (t - t_freezing_zero_sa)/tfreezing_sa

!---------------------------------------------------------------------------
! Begin the modified Newton-Raphson method to find the root of 
! f = (t_freezing - t) = 0 for SA. 
!---------------------------------------------------------------------------
do i_iter = 1, number_of_iterations
    sa_old = sa
    f = gsw_t_freezing_exact(sa_old,p,saturation_fraction) - t
    sa = sa_old - f/tfreezing_sa
    sa_mean = 0.5_r8*(sa + sa_old)
    call gsw_t_freezing_first_derivatives(sa_mean,p,saturation_fraction, &
                                          tfreezing_sa=tfreezing_sa)
    sa = sa_old - f/tfreezing_sa   
end do

if (gsw_sa_p_inrange(sa,p)) then
    gsw_sa_freezing_from_t = sa
else
    gsw_sa_freezing_from_t = gsw_error_code(3,func_name)
end if

return
end function

!---------------------------------------------------------------------------
!==========================================================================
elemental function gsw_sa_freezing_from_t_poly (t, p, saturation_fraction)
!==========================================================================
!
!  Calculates the Absolute Salinity of seawater at the freezing temperature.  
!  That is, the output is the Absolute Salinity of seawater, with the 
!  fraction saturation_fraction of dissolved air, that is in equilibrium 
!  with ice at in-situ temperature t and pressure p.  If the input values 
!  are such that there is no positive value of Absolute Salinity for which 
!  seawater is frozen, the output is put equal to Nan.  
!
!  t  =  in-situ Temperature (ITS-90)                             [ deg C ]
!  p  =  sea pressure                                              [ dbar ]
!        ( i.e. absolute pressure - 10.1325 dbar ) 
!  saturation_fraction = the saturation fraction of dissolved air in 
!                        seawater
!
!  sa_freezing_from_t_poly  =  Absolute Salinity of seawater when it freezes,
!                for given input values of in situ temperature, pressure and 
!                air saturation fraction.                          [ g/kg ]
!--------------------------------------------------------------------------

use gsw_mod_toolbox, only : gsw_sa_p_inrange, gsw_sa_freezing_estimate
use gsw_mod_toolbox, only : gsw_t_freezing_first_derivatives_poly
use gsw_mod_toolbox, only : gsw_t_freezing_poly

use gsw_mod_error_functions, only : gsw_error_code, gsw_error_limit

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: t, p, saturation_fraction

real (r8) :: gsw_sa_freezing_from_t_poly

integer :: i_iter
real (r8) :: dt_dsa, sa, sa_old, sa_mean, t_freezing, t_freezing_zero_sa

integer, parameter :: number_of_iterations = 5

! This is the band of sa within +- 2.5 g/kg of sa = 0, which we treat
! differently in calculating the initial values of both SA and dCT_dSA. 
real (r8), parameter :: sa_cut_off = 2.5_r8

character (*), parameter :: func_name = "gsw_sa_freezing_from_t_poly"

! Find t > t_freezing_zero_SA.  If this is the case, the input values
! represent seawater that is not frozen (at any positive SA). 
t_freezing_zero_sa = gsw_t_freezing_poly(0.0_r8,p,saturation_fraction)
if (t .gt. t_freezing_zero_sa) then
    gsw_sa_freezing_from_t_poly = gsw_error_code(1,func_name)
    return
end if

! This is the inital guess of SA using a purpose-built polynomial in CT and p  
sa = gsw_sa_freezing_estimate(p,saturation_fraction,t=t)
if (sa .lt. -sa_cut_off) then
    gsw_sa_freezing_from_t_poly = gsw_error_code(2,func_name)
    return
end if

! Form the first estimate of dt_dSA, the derivative of t with respect 
! to SA at fixed p.
sa = max(sa,0.0_r8)
call gsw_t_freezing_first_derivatives_poly(sa,p,saturation_fraction, &
                                           tfreezing_sa=dt_dsa)

! For -SA_cut_off < SA < SA_cut_off, replace the above estimate of SA  
! with one based on (t_freezing_zero_SA - t).
if (abs(sa) .lt. sa_cut_off) sa = (t - t_freezing_zero_sa)/dt_dsa

!---------------------------------------------------------------------------
! Begin the modified Newton-Raphson method to find the root of 
! t_freezing = t for SA. 
!---------------------------------------------------------------------------
do i_iter = 1, number_of_iterations
    sa_old = sa
    t_freezing = gsw_t_freezing_poly(sa_old,p,saturation_fraction)
    sa = sa_old - (t_freezing - t)/dt_dsa
    sa_mean = 0.5_r8*(sa + sa_old)
    call gsw_t_freezing_first_derivatives_poly(sa_mean,p,saturation_fraction, &
                                               tfreezing_sa=dt_dsa)
    sa = sa_old - (t_freezing - t)/dt_dsa
end do

if (gsw_sa_p_inrange(sa,p)) then
    gsw_sa_freezing_from_t_poly = sa
else
    gsw_sa_freezing_from_t_poly = gsw_error_code(3,func_name)
end if

return
end function

!--------------------------------------------------------------------------
!==========================================================================
elemental function gsw_sa_from_rho (rho, ct, p)
!==========================================================================
!
!  Calculates the Absolute Salinity of a seawater sample, for given values
!  of its density, Conservative Temperature and sea pressure (in dbar). 

!  rho =  density of a seawater sample (e.g. 1026 kg/m^3).       [ kg/m^3 ]
!   Note. This input has not had 1000 kg/m^3 subtracted from it. 
!     That is, it is 'density', not 'density anomaly'.
!  ct  =  Conservative Temperature (ITS-90)                       [ deg C ]
!  p   =  sea pressure                                             [ dbar ]
!
!  sa  =  Absolute Salinity                                          [g/kg]
!--------------------------------------------------------------------------

use gsw_mod_toolbox, only : gsw_specvol, gsw_specvol_first_derivatives

use gsw_mod_error_functions, only : gsw_error_code, gsw_error_limit

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: rho, ct, p

real (r8) :: gsw_sa_from_rho

integer no_iter

real (r8) :: sa, v_lab, v_0, v_50, v_sa, sa_old, delta_v, sa_mean

character (*), parameter :: func_name = "gsw_sa_from_rho"

v_lab = 1.0_r8/rho
v_0 = gsw_specvol(0.0_r8,ct,p)
v_50 = gsw_specvol(50.0_r8,ct,p)

sa = 50.0_r8*(v_lab - v_0)/(v_50 - v_0)

if (sa.lt.0_r8.or.sa.gt.50_r8) then
    gsw_sa_from_rho = gsw_error_code(1,func_name)
    return
end if

v_sa = (v_50 - v_0)/50.0_r8

do no_iter = 1, 2 
    sa_old = sa
    delta_v = gsw_specvol(sa_old,ct,p) - v_lab
    sa = sa_old - delta_v/v_sa 
    sa_mean = 0.5_r8*(sa + sa_old)
    call gsw_specvol_first_derivatives(sa_mean,ct,p,v_sa)
    sa = sa_old - delta_v/v_sa
    if (sa.lt.0_r8.or.sa.gt.50_r8) then
        gsw_sa_from_rho = gsw_error_code(no_iter+1,func_name)
        return
    end if
end do

gsw_sa_from_rho = sa

return
end function

!--------------------------------------------------------------------------
!==========================================================================
elemental function gsw_sa_from_sp (sp, p, long, lat)       
!==========================================================================
!
! Calculates Absolute Salinity, SA, from Practical Salinity, SP
!
! sp     : Practical Salinity                              [unitless]
! p      : sea pressure                                    [dbar]
! long   : longitude                                       [DEG E]     
! lat    : latitude                                        [DEG N]
!
! gsw_sa_from_sp   : Absolute Salinity                     [g/kg]
!--------------------------------------------------------------------------

use gsw_mod_toolbox, only : gsw_saar, gsw_sa_from_sp_baltic

use gsw_mod_error_functions, only : gsw_error_code, gsw_error_limit

use gsw_mod_teos10_constants, only : gsw_ups

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sp, p, long, lat       

real (r8) :: gsw_sa_from_sp

real (r8) :: saar, sa_baltic

character (*), parameter :: func_name = "gsw_sa_from_sp"

sa_baltic = gsw_sa_from_sp_baltic(sp,long,lat)

if (sa_baltic .lt. 1e10_r8) then

   gsw_sa_from_sp = sa_baltic

else

   saar = gsw_saar(p,long,lat)
   if (saar .gt. gsw_error_limit) then
      gsw_sa_from_sp = gsw_error_code(1,func_name,saar)
   else
      gsw_sa_from_sp = gsw_ups*sp*(1.0_r8 + saar)
   end if

end if

return
end function

!--------------------------------------------------------------------------
!==========================================================================
elemental function gsw_sa_from_sp_baltic (sp, long, lat)
!==========================================================================
!
! For the Baltic Sea, calculates Absolute Salinity with a value
! computed analytically from Practical Salinity
!
! sp     : Practical Salinity                              [unitless]
! long   : longitude                                       [deg E]     
! lat    : latitude                                        [deg N]
!
! gsw_sa_from_sp_baltic : Absolute Salinity                [g/kg] 
!--------------------------------------------------------------------------

use gsw_mod_toolbox, only : gsw_util_xinterp1

use gsw_mod_baltic_data

use gsw_mod_teos10_constants, only : gsw_sso

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sp, long, lat

real (r8) :: gsw_sa_from_sp_baltic

real (r8) :: xx_left, xx_right

if (xb_left(2).lt.long .and. long.lt.xb_right(1) .and. &
    yb_left(1).lt.lat  .and.  lat.lt.yb_left(3)) then
  
    xx_left = gsw_util_xinterp1(yb_left, xb_left, 3, lat)
    
    xx_right = gsw_util_xinterp1(yb_right, xb_right, 2, lat)
    
    if(xx_left.le.long .and. long.le.xx_right) then
        gsw_sa_from_sp_baltic =((gsw_sso - 0.087_r8)/35.0_r8)*sp + 0.087_r8
    else
        gsw_sa_from_sp_baltic = 9e15_r8
    end if

else
    gsw_sa_from_sp_baltic = 9e15_r8
end if

return
end function

!--------------------------------------------------------------------------
!==========================================================================
elemental function gsw_sa_from_sstar (sstar, p, long, lat)  
!==========================================================================
!
! Calculates Absolute Salinity, SA, from Preformed Salinity, Sstar.
!
! Sstar  : Preformed Salinity                              [g/kg]
! p      : sea pressure                                    [dbar]
! long   : longitude                                       [deg E]     
! lat    : latitude                                        [deg N]
!
! gsw_sa_from_sstar   : Absolute Salinity                  [g/kg]
!--------------------------------------------------------------------------

use gsw_mod_toolbox, only : gsw_saar

use gsw_mod_error_functions, only : gsw_error_code, gsw_error_limit

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sstar, p, long, lat  

real (r8) :: gsw_sa_from_sstar

real (r8) :: saar

character (*), parameter :: func_name = "gsw_sa_from_sstar"

saar = gsw_saar(p,long,lat)

if (saar.gt.gsw_error_limit) then

   gsw_sa_from_sstar = gsw_error_code(1,func_name,saar)

else

   ! In the Baltic Sea, Sstar = SA, and note that gsw_saar returns zero
   ! for SAAR in the Baltic.

   gsw_sa_from_sstar = sstar*(1.0_r8 + saar)/(1.0_r8 - 0.35_r8*saar)

end if

return
end function

!--------------------------------------------------------------------------
!==========================================================================
elemental function gsw_sa_p_inrange (sa, p)
!==========================================================================
!
!  Check for any values that are out of the TEOS-10 range ...
!
!  SA  =  Absolute Salinity                                        [ g/kg ]
!  p   =  sea pressure                                             [ dbar ]
!         ( i.e. absolute pressure - 10.1325 dbar ) 
!---------------------------------------------------------------------------

use gsw_mod_kinds

implicit none


real (r8), intent(in) :: sa, p

logical :: gsw_sa_p_inrange

gsw_sa_p_inrange = .true.

if (p.gt.10000.0_r8 .or. sa.gt.120.0_r8 .or. &
    p + sa*71.428571428571402_r8.gt.13571.42857142857_r8) &
    gsw_sa_p_inrange = .false.

return
end function

!---------------------------------------------------------------------------
!==========================================================================
elemental function gsw_saar (p, long, lat)
!==========================================================================
!
! Calculates the Absolute Salinity Anomaly Ratio, SAAR at a geographic point.
!
! p      : sea pressure                                    [dbar]
! long   : longitude                                       [deg E]     
! lat    : latitude                                        [deg N]
!
! gsw_saar : Absolute Salinity Anomaly Ratio               [unitless]
!--------------------------------------------------------------------------

use gsw_mod_toolbox, only : gsw_add_barrier, gsw_add_mean, gsw_util_indx

use gsw_mod_error_functions, only : gsw_error_code

use gsw_mod_saar_data

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: p, long, lat

real (r8) :: gsw_saar

integer :: indx0, indy0, indz0, k, ndepth_max

real (r8), dimension(4) :: saar, saar_old
real (r8) :: p0_original, sa_upper, sa_lower, dlong, dlat
real (r8) :: r1, s1, t1, p_tmp, long360

character (*), parameter :: func_name = "gsw_saar"

long360 = long
if (long360.lt.0.0_r8) long360 = long360 + 360.0_r8

indx0 = floor(1.0_r8 + (nx-1)*(long360-longs_ref(1)) / &
              (longs_ref(nx)-longs_ref(1)))

indy0 = floor(1.0_r8 + (ny-1)*(lat-lats_ref(1)) / (lats_ref(ny)-lats_ref(1)))

if ((indx0.ge.1 .and. indx0.le.nx) .and. (indy0.ge.1 .and. indy0.le.ny)) then
   if (indx0.eq.nx) indx0 = nx-1
   if (indy0.eq.ny) indy0 = ny-1
else
   ! This is to catch any out-of-range or nonsense lat/long input (including
   ! NaN, +Inf, -Inf etc). Note: NaNs will not satisfy any "if" conditional
   ! so will be trapped by the "else".
   gsw_saar = gsw_error_code(2,func_name)
   return
end if

! Look for the maximum valid "ndepth_ref" value around our point.
! Note: invalid "ndepth_ref" values are NaNs (a hangover from the codes
! Matlab origins), but we have replaced the NaNs with a value of 999,
! hence we need an additional upper-limit check in the code below so they
! will not be recognised as valid values.
ndepth_max = -1
do k = 1,4
   if ((ndepth_ref(indy0+delj(k),indx0+deli(k)).gt.0) .and. &
       (ndepth_ref(indy0+delj(k),indx0+deli(k)).lt.99)) &
      ndepth_max = max(ndepth_max,ndepth_ref(indy0+delj(k),indx0+deli(k)))
end do

! If we are a long way from the ocean then there will be no valid "ndepth_ref"
! values near the point (ie. surrounded by NaNs) - so just return SAAR = 0.0
if (ndepth_max.eq.-1) then
   gsw_saar = 0.0_r8 
   return
end if 

p0_original = p
p_tmp = p
if (p_tmp.gt.p_ref(ndepth_max)) p_tmp = p_ref(ndepth_max)
call gsw_util_indx(p_ref,nz,p_tmp,indz0)

dlong = longs_ref(indx0+1) - longs_ref(indx0)
dlat = lats_ref(indy0+1) - lats_ref(indy0)

r1 = (long360-longs_ref(indx0)) / dlong
s1 = (lat-lats_ref(indy0)) / dlat
t1 = (p_tmp-p_ref(indz0))/(p_ref(indz0+1)-p_ref(indz0))

do k = 1,4
   saar(k) = saar_ref(indz0,indy0+delj(k),indx0+deli(k))
end do

if (  longs_pan(1).le.long360 .and. long360.le.longs_pan(npan)-0.001_r8 .and. &
    lats_pan(npan).le.lat     .and.     lat.le.lats_pan(1)) then
  saar_old = saar
  call gsw_add_barrier(saar_old,long360,lat,longs_ref(indx0), &
                       lats_ref(indy0),dlong,dlat,saar)
else if (abs(sum(saar)).ge.1e10_r8) then
  saar_old = saar
  call gsw_add_mean(saar_old,saar)
end if

sa_upper = (1.0_r8-s1)*(saar(1) + r1*(saar(2)-saar(1))) + s1*(saar(4) + &
              r1*(saar(3)-saar(4)))

do k = 1,4
   saar(k) = saar_ref(indz0+1,indy0+delj(k),indx0+deli(k))
end do

if (  longs_pan(1).le.long360 .and. long360.le.longs_pan(npan)-0.001_r8 .and. &
    lats_pan(npan).le.lat     .and.     lat.le.lats_pan(1)) then
   saar_old = saar
   call gsw_add_barrier(saar_old,long360,lat,longs_ref(indx0), &
                        lats_ref(indy0),dlong,dlat,saar)
else if (abs(sum(saar)).ge.1e10_r8) then 
   saar_old = saar
   call gsw_add_mean(saar_old,saar)
end if

sa_lower = (1.0_r8-s1)*(saar(1) + r1*(saar(2)-saar(1))) + s1*(saar(4) + &
              r1*(saar(3)-saar(4)))
if (abs(sa_lower).ge.1e10_r8) sa_lower = sa_upper
gsw_saar = sa_upper + t1*(sa_lower-sa_upper)

if (abs(gsw_saar).ge.1e10_r8) gsw_saar = gsw_error_code(3,func_name)

return
end function

!--------------------------------------------------------------------------
!==========================================================================
elemental subroutine gsw_seaice_fraction_to_freeze_seawater (sa, ct, p, &
                       sa_seaice, t_seaice, sa_freeze, ct_freeze, w_seaice)
!==========================================================================
!
!  Calculates the mass fraction of sea ice (mass of sea ice divided by mass 
!  of sea ice plus seawater), which, when melted into seawater having the
!  properties (SA,CT,p) causes the final seawater to be at the freezing 
!  temperature.  The other outputs are the Absolute Salinity and 
!  Conservative Temperature of the final seawater.  
!
!  SA        =  Absolute Salinity of seawater                      [ g/kg ]
!  CT        =  Conservative Temperature of seawater (ITS-90)     [ deg C ]
!  p         =  sea pressure                                       [ dbar ]
!            ( i.e. absolute pressure - 10.1325 dbar )
!  SA_seaice =  Absolute Salinity of sea ice, that is, the mass fraction of
!               salt in sea ice, expressed in g of salt per kg of sea ice.
!                                                                  [ g/kg ]
!  t_seaice  =  in-situ temperature of the sea ice at pressure p (ITS-90)
!                                                                 [ deg C ]
!
!  SA_freeze  =  Absolute Salinity of seawater after the mass fraction of
!                sea ice, w_seaice, at temperature t_seaice has melted into
!                the original seawater, and the final mixture is at the 
!                freezing temperature of seawater.                 [ g/kg ]
!
!  CT_freeze  =  Conservative Temperature of seawater after the mass 
!                fraction, w_seaice, of sea ice at temperature t_seaice has
!                melted into the original seawater, and the final mixture 
!                is at the freezing temperature of seawater.      [ deg C ]
!
!  w_seaice   =  mass fraction of sea ice, at SA_seaice and t_seaice, 
!                which, when melted into seawater at (SA,CT,p) leads to the
!                final mixed seawater being at the freezing temperature.  
!                This output is between 0 and 1.                 [unitless]
!--------------------------------------------------------------------------

use gsw_mod_toolbox, only : gsw_ct_freezing_exact, gsw_enthalpy_ct_exact
use gsw_mod_toolbox, only : gsw_sa_freezing_from_t, gsw_enthalpy_ice
use gsw_mod_toolbox, only : gsw_enthalpy_t_exact, gsw_t_freezing_exact
use gsw_mod_toolbox, only : gsw_enthalpy_first_derivatives_ct_exact
use gsw_mod_toolbox, only : gsw_ct_freezing_first_derivatives

use gsw_mod_error_functions, only : gsw_error_code, gsw_error_limit

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sa, ct, p, sa_seaice, t_seaice
real (r8), intent(out) :: sa_freeze, ct_freeze, w_seaice

integer :: number_of_iterations
real (r8) :: ctf, ctf_mean, ctf_old, ctf_plus1, ctf_zero
real (r8) :: dfunc_dsaf, func, func_plus1, func_zero, h, h_brine
real (r8) :: h_ih, sa_freezing, saf, saf_mean, saf_old
real (r8) :: salt_ratio, tf_sa_seaice, h_hat_sa, h_hat_ct, ctf_sa

real (r8), parameter :: sa0 = 0.0_r8, saturation_fraction = 0.0_r8

character (*), parameter :: func_name = "gsw_seaice_fraction_to_freeze_seawater"

ctf = gsw_ct_freezing_exact(sa,p,saturation_fraction)
if (ct .lt. ctf) then
    ! The seawater ct input is below the freezing temp
    sa_freeze = gsw_error_code(1,func_name)
    ct_freeze = sa_freeze
    w_seaice = sa_freeze
    return
end if

tf_sa_seaice = gsw_t_freezing_exact(sa_seaice,p,saturation_fraction) - 1e-6_r8
if (t_seaice .gt. tf_sa_seaice) then
    ! The 1e-6 C buffer in the allowable t_seaice is to ensure that there is
    ! some ice Ih in the sea ice.   Without this buffer, that is if t_seaice
    ! is allowed to be exactly equal to tf_sa_seaice, the sea ice is 
    ! actually 100% brine at Absolute Salinity of SA_seaice.
    sa_freeze = gsw_error_code(2,func_name)
    ct_freeze = sa_freeze
    w_seaice = sa_freeze
    return
end if

sa_freezing = gsw_sa_freezing_from_t(t_seaice,p,saturation_fraction)
if (sa_freezing .gt. gsw_error_limit) then
    sa_freeze = gsw_error_code(3,func_name,sa_freezing)
    ct_freeze = sa_freeze
    w_seaice = sa_freeze
    return
end if
h_brine = gsw_enthalpy_t_exact(sa_freezing,t_seaice,p)
salt_ratio = sa_seaice/sa_freezing

h = gsw_enthalpy_ct_exact(sa,ct,p)
h_ih = gsw_enthalpy_ice(t_seaice,p)

ctf_plus1 = gsw_ct_freezing_exact(sa+1.0_r8,p,saturation_fraction)
func_plus1 = (sa - sa_seaice)*(gsw_enthalpy_ct_exact(sa+1.0_r8,ctf_plus1,p) &
                - h) - (h - h_ih) + salt_ratio*(h_brine - h_ih)

ctf_zero = gsw_ct_freezing_exact(sa0,p,saturation_fraction)
func_zero = (sa - sa_seaice)*(gsw_enthalpy_ct_exact(sa0,ctf_zero,p) - h) &
             + sa*((h - h_ih) - salt_ratio*(h_brine - h_ih))

saf = -(sa+1.0_r8)*func_zero/(func_plus1 - func_zero) ! initial guess of saf
ctf = gsw_ct_freezing_exact(saf,p,saturation_fraction)
call gsw_enthalpy_first_derivatives_ct_exact(saf,ctf,p,h_hat_sa,h_hat_ct)
call gsw_ct_freezing_first_derivatives(saf,p,saturation_fraction, &
                                       ctfreezing_sa=ctf_sa)

dfunc_dsaf = (sa - sa_seaice)*(h_hat_sa + h_hat_ct*ctf_sa) &
              - (h - h_ih) + salt_ratio*(h_brine - h_ih)

do number_of_iterations = 1, 4
    saf_old = saf   
    ctf_old = ctf
    func = (sa - sa_seaice)*(gsw_enthalpy_ct_exact(saf_old,ctf_old,p) - h) &
         - (saf_old - sa)*((h - h_ih) - salt_ratio*(h_brine - h_ih))
    saf = saf_old - func/dfunc_dsaf
    saf_mean = 0.5_r8*(saf + saf_old)
    ctf_mean = gsw_ct_freezing_exact(saf_mean,p,saturation_fraction)
    call gsw_enthalpy_first_derivatives_ct_exact(saf_mean,ctf_mean,p, &
                                                 h_hat_sa,h_hat_ct)
    call gsw_ct_freezing_first_derivatives(saf_mean,p,saturation_fraction, &
                                           ctfreezing_sa=ctf_sa)
    dfunc_dsaf = (sa - sa_seaice)*(h_hat_sa + h_hat_ct*ctf_sa) &
                  - (h - h_ih) + salt_ratio*(h_brine - h_ih)
    saf = saf_old - func/dfunc_dsaf
    ctf = gsw_ct_freezing_exact(saf,p,saturation_fraction)
end do

! After these 4 iterations of this modified Newton-Raphson method, the
! errors in SA_freeze is less than 1.5x10^-12 g/kg, in CT_freeze is less than
! 2x10^-13 deg C and in w_seaice is less than 2.8x10^-13 which represent machine
! precision for these calculations. 

sa_freeze = saf
ct_freeze = ctf
w_seaice = (h - gsw_enthalpy_ct_exact(sa_freeze,ct_freeze,p)) / &
                           (h - h_ih - salt_ratio*(h_brine - h_ih))
return
end subroutine

!--------------------------------------------------------------------------
!==========================================================================
elemental function gsw_sigma0 (sa, ct)
!==========================================================================
! 
!  Calculates potential density anomaly with reference pressure of 0 dbar,
!  this being this particular potential density minus 1000 kg/m^3.  This
!  function has inputs of Absolute Salinity and Conservative Temperature.
!  This function uses the computationally-efficient expression for 
!  specific volume in terms of SA, CT and p (Roquet et al., 2014).
!
!  SA  =  Absolute Salinity                                        [ g/kg ]
!  CT  =  Conservative Temperature (ITS-90)                       [ deg C ]
!
!  sigma0  =  potential density anomaly with                     [ kg/m^3 ]
!             respect to a reference pressure of 0 dbar,   
!             that is, this potential density - 1000 kg/m^3.
!--------------------------------------------------------------------------

use gsw_mod_teos10_constants, only : gsw_sfac, offset

use gsw_mod_specvol_coefficients

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sa, ct

real (r8) :: gsw_sigma0

real (r8) :: vp0, xs, ys

xs = sqrt(gsw_sfac*sa + offset)
ys = ct*0.025_r8

vp0 = v000 + xs*(v010 + xs*(v020 + xs*(v030 + xs*(v040 + xs*(v050 &
    + v060*xs))))) + ys*(v100 + xs*(v110 + xs*(v120 + xs*(v130 + xs*(v140 &
    + v150*xs)))) + ys*(v200 + xs*(v210 + xs*(v220 + xs*(v230 + v240*xs))) &
    + ys*(v300 + xs*(v310 + xs*(v320 + v330*xs)) + ys*(v400 + xs*(v410 &
    + v420*xs) + ys*(v500 + v510*xs + v600*ys)))))

gsw_sigma0 = 1.0_r8/vp0 - 1000.0_r8

return
end function

!--------------------------------------------------------------------------
!==========================================================================
elemental function gsw_sigma1 (sa, ct) 
!==========================================================================
!
!  Calculates potential density anomaly with reference pressure of 1000 dbar,
!  this being this particular potential density minus 1000 kg/m^3.  This
!  function has inputs of Absolute Salinity and Conservative Temperature.
!
! sa     : Absolute Salinity                               [g/kg]
! ct     : Conservative Temperature                        [deg C]
! 
! gsw_sigma1  : potential density anomaly with reference pressure of 1000
!--------------------------------------------------------------------------

use gsw_mod_toolbox, only : gsw_rho

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sa, ct 

real (r8) :: gsw_sigma1

gsw_sigma1 = gsw_rho(sa,ct,1e3_r8) - 1e3_r8

return
end function

!--------------------------------------------------------------------------
!==========================================================================
elemental function gsw_sigma2 (sa, ct) 
!==========================================================================
!
!  Calculates potential density anomaly with reference pressure of 2000 dbar,
!  this being this particular potential density minus 1000 kg/m^3.  This
!  function has inputs of Absolute Salinity and Conservative Temperature.
!
! sa     : Absolute Salinity                               [g/kg]
! ct     : Conservative Temperature                        [deg C]
! 
! gsw_sigma2  : potential density anomaly with reference pressure of 2000
!--------------------------------------------------------------------------

use gsw_mod_toolbox, only : gsw_rho

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sa, ct 

real (r8) :: gsw_sigma2

gsw_sigma2 = gsw_rho(sa,ct,2e3_r8) - 1e3_r8

return
end function

!--------------------------------------------------------------------------
!==========================================================================
elemental function gsw_sigma3 (sa, ct) 
!==========================================================================
!
!  Calculates potential density anomaly with reference pressure of 3000 dbar,
!  this being this particular potential density minus 1000 kg/m^3.  This
!  function has inputs of Absolute Salinity and Conservative Temperature.
!
! sa     : Absolute Salinity                               [g/kg]
! ct     : Conservative Temperature                        [deg C]
! 
! gsw_sigma3  : potential density anomaly with reference pressure of 3000
!--------------------------------------------------------------------------

use gsw_mod_toolbox, only : gsw_rho

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sa, ct 

real (r8) :: gsw_sigma3

gsw_sigma3 = gsw_rho(sa,ct,3e3_r8) - 1e3_r8

return
end function

!--------------------------------------------------------------------------
!==========================================================================
elemental function gsw_sigma4 (sa, ct) 
!==========================================================================
!
!  Calculates potential density anomaly with reference pressure of 4000 dbar,
!  this being this particular potential density minus 1000 kg/m^3.  This
!  function has inputs of Absolute Salinity and Conservative Temperature.
!
! sa     : Absolute Salinity                               [g/kg]
! ct     : Conservative Temperature                        [deg C]
! 
! gsw_sigma4  : potential density anomaly with reference pressure of 4000
!--------------------------------------------------------------------------

use gsw_mod_toolbox, only : gsw_rho

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sa, ct 

real (r8) :: gsw_sigma4

gsw_sigma4 = gsw_rho(sa,ct,4e3_r8) - 1e3_r8

return
end function

!--------------------------------------------------------------------------
!==========================================================================
elemental function gsw_sound_speed (sa, ct, p)
!==========================================================================
!
!  Calculates the speed of sound in seawater.  This function has inputs of 
!  Absolute Salinity and Conservative Temperature.  This function uses the 
!  computationally-efficient expression for specific volume in terms of SA,
!  CT and p (Roquet et al., 2014).
!
!  SA  =  Absolute Salinity                                        [ g/kg ]
!  CT  =  Conservative Temperature (ITS-90)                       [ deg C ]
!  p   =  sea pressure                                             [ dbar ]
!         ( i.e. absolute pressure - 10.1325 dbar )
!
!  sound_speed  =  speed of sound in seawater                       [ m/s ]
!--------------------------------------------------------------------------

use gsw_mod_teos10_constants, only : gsw_sfac, offset

use gsw_mod_specvol_coefficients

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sa, ct, p

real (r8) :: gsw_sound_speed

real (r8) :: v, v_p, xs, ys, z

xs = sqrt(gsw_sfac*sa + offset)
ys = ct*0.025_r8
z = p*1e-4_r8

v = v000 + xs*(v010 + xs*(v020 + xs*(v030 + xs*(v040 + xs*(v050 &
    + v060*xs))))) + ys*(v100 + xs*(v110 + xs*(v120 + xs*(v130 + xs*(v140 &
    + v150*xs)))) + ys*(v200 + xs*(v210 + xs*(v220 + xs*(v230 + v240*xs))) &
    + ys*(v300 + xs*(v310 + xs*(v320 + v330*xs)) + ys*(v400 + xs*(v410 &
    + v420*xs) + ys*(v500 + v510*xs + v600*ys))))) + z*(v001 + xs*(v011 &
    + xs*(v021 + xs*(v031 + xs*(v041 + v051*xs)))) + ys*(v101 + xs*(v111 &
    + xs*(v121 + xs*(v131 + v141*xs))) + ys*(v201 + xs*(v211 + xs*(v221 &
    + v231*xs)) + ys*(v301 + xs*(v311 + v321*xs) + ys*(v401 + v411*xs &
    + v501*ys)))) + z*(v002 + xs*(v012 + xs*(v022 + xs*(v032 + v042*xs))) &
    + ys*(v102 + xs*(v112 + xs*(v122 + v132*xs)) + ys*(v202 + xs*(v212 &
    + v222*xs) + ys*(v302 + v312*xs + v402*ys))) + z*(v003 + xs*(v013 &
    + v023*xs) + ys*(v103 + v113*xs + v203*ys) + z*(v004 + v014*xs + v104*ys &  
    + z*(v005 + v006*z)))))

v_p = c000 + xs*(c100 + xs*(c200 + xs*(c300 + xs*(c400 + c500*xs)))) & 
    + ys*(c010 + xs*(c110 + xs*(c210 + xs*(c310 + c410*xs))) + ys*(c020 &
    + xs*(c120 + xs*(c220 + c320*xs)) + ys*(c030 + xs*(c130 + c230*xs) &
    + ys*(c040 + c140*xs + c050*ys)))) + z*(c001 + xs*(c101 + xs*(c201 &
    + xs*(c301 + c401*xs))) + ys*(c011 + xs*(c111 + xs*(c211 + c311*xs)) &
    + ys*(c021 + xs*(c121 + c221*xs) + ys*(c031 + c131*xs + c041*ys))) &
    + z*( c002 + xs*(c102 + c202*xs) + ys*(c012 + c112*xs + c022*ys) &
    + z*(c003 + c103*xs + c013*ys + z*(c004 + c005*z))))

gsw_sound_speed = 10000.0_r8*sqrt(-v**2/v_p)

return
end function

!--------------------------------------------------------------------------
!==========================================================================
elemental function gsw_sound_speed_ice (t, p)
!==========================================================================
!
!  Calculates the compression speed of sound in ice. 
!  
!  t   =  in-situ temperature (ITS-90)                            [ deg C ]
!  p   =  sea pressure                                             [ dbar ]
!         ( i.e. absolute pressure - 10.1325 dbar )
!
!  sound_speed_ice  =  compression speed of sound in ice            [ m/s ]
!--------------------------------------------------------------------------

use gsw_mod_toolbox, only : gsw_gibbs_ice

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: t, p

real (r8) :: gsw_sound_speed_ice

real (r8) :: gi_tp, gi_tt

gi_tt = gsw_gibbs_ice(2,0,t,p) 
gi_tp = gsw_gibbs_ice(1,1,t,p)

gsw_sound_speed_ice = gsw_gibbs_ice(0,1,t,p) * &
               sqrt(gi_tt/(gi_tp*gi_tp - gi_tt*gsw_gibbs_ice(0,2,t,p)))

return
end function

!--------------------------------------------------------------------------
!==========================================================================
elemental function gsw_sound_speed_t_exact (sa, t, p)  
!==========================================================================
!
! Calculates the sound speed of seawater
!
! sa     : Absolute Salinity                               [g/kg]
! t      : in-situ temperature                             [deg C]
! p      : sea pressure                                    [dbar]
! 
! gsw_sound_speed_t_exact : sound speed                    [m/s]
!--------------------------------------------------------------------------

use gsw_mod_toolbox, only : gsw_gibbs

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sa, t, p  

real (r8) :: gsw_sound_speed_t_exact

real (r8) :: g_tt, g_tp

integer, parameter :: n0=0, n1=1, n2=2

g_tt = gsw_gibbs(n0,n2,n0,sa,t,p)
g_tp = gsw_gibbs(n0,n1,n1,sa,t,p)

gsw_sound_speed_t_exact = gsw_gibbs(n0,n0,n1,sa,t,p) * &
               sqrt(g_tt/(g_tp*g_tp - g_tt*gsw_gibbs(n0,n0,n2,sa,t,p)))
                                    
return
end function

!--------------------------------------------------------------------------
!==========================================================================
elemental function gsw_sp_from_c (c, t, p)       
!==========================================================================
!
!  Calculates Practical Salinity, SP, from conductivity, C, primarily using
!  the PSS-78 algorithm.  Note that the PSS-78 algorithm for Practical 
!  Salinity is only valid in the range 2 < SP < 42.  If the PSS-78 
!  algorithm produces a Practical Salinity that is less than 2 then the 
!  Practical Salinity is recalculated with a modified form of the Hill et 
!  al. (1986) formula.  The modification of the Hill et al. (1986)
!  expression is to ensure that it is exactly consistent with PSS-78 
!  at SP = 2.  Note that the input values of conductivity need to be in 
!  units of mS/cm (not S/m). 
!
! c      : conductivity                                     [ mS/cm ]
! t      : in-situ temperature [ITS-90]                     [deg C]
! p      : sea pressure                                     [dbar]
!
! sp     : Practical Salinity                               [unitless]
!--------------------------------------------------------------------------

use gsw_mod_toolbox, only : gsw_hill_ratio_at_sp2

use gsw_mod_error_functions, only : gsw_error_code, gsw_error_limit

use gsw_mod_teos10_constants, only : gsw_c3515

use gsw_mod_sp_coefficients

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: c, t, p       

real (r8) :: gsw_sp_from_c

real (r8) :: sp, t68, ft68, r, rt_lc, rp, rt, rtx
real (r8) :: hill_ratio, x, sqrty, part1, part2, sp_hill_raw

character (*), parameter :: func_name = "gsw_sp_from_c"

t68 = t*1.00024_r8
ft68 = (t68 - 15.0_r8)/(1.0_r8 + k*(t68 - 15.0_r8))

! The dimensionless conductivity ratio, R, is the conductivity input, C,
! divided by the present estimate of C(SP=35, t_68=15, p=0) which is 
! 42.9140 mS/cm (=4.29140 S/m), (Culkin and Smith, 1980). 

r = c/gsw_c3515

! rt_lc corresponds to rt as defined in the UNESCO 44 (1983) routines.  
rt_lc = c0 + (c1 + (c2 + (c3 + c4*t68)*t68)*t68)*t68
rp = 1.0_r8 + (p*(e1 + e2*p + e3*p*p))/(1.0_r8 + d1*t68 + d2*t68*t68 + (d3 + d4*t68)*r)
rt = r/(rp*rt_lc)  

if (rt .lt. 0.0_r8) then
    gsw_sp_from_c = gsw_error_code(1,func_name)
    return
endif

rtx = sqrt(rt)

sp = a0 + (a1 + (a2 + (a3 + (a4 + a5*rtx)*rtx)*rtx)*rtx)*rtx + &
    ft68*(b0 + (b1 + (b2 + (b3 + (b4 + b5*rtx)*rtx)*rtx)*rtx)*rtx)

! The following section of the code is designed for SP < 2 based on the
! Hill et al. (1986) algorithm.  This algorithm is adjusted so that it is
! exactly equal to the PSS-78 algorithm at SP = 2.

if (sp .lt. 2.0_r8) then
    hill_ratio = gsw_hill_ratio_at_sp2(t)
    x = 400.0_r8*rt
    sqrty = 10.0_r8*rtx
    part1 = 1.0_r8 + x*(1.5_r8 + x)
    part2 = 1.0_r8 + sqrty*(1.0_r8 + sqrty*(1.0_r8 + sqrty))
    sp_hill_raw = sp - a0/part1 - b0*ft68/part2
    sp = hill_ratio*sp_hill_raw
endif

if (sp .lt. 0.0_r8) then
    gsw_sp_from_c = gsw_error_code(2,func_name)
else
    gsw_sp_from_c = sp
end if

return
end function

!--------------------------------------------------------------------------
!==========================================================================
elemental function gsw_sp_from_sa (sa, p, long, lat) 
!==========================================================================
!
! Calculates Practical salinity, sp, from Absolute salinity, sa  
!
! sa     : Absolute Salinity                               [g/kg]
! p      : sea pressure                                    [dbar]
! long   : longitude                                       [DEG E]     
! lat    : latitude                                        [DEG N]
!
! gsw_sp_from_sa      : Practical Salinity                 [unitless]
!--------------------------------------------------------------------------

use gsw_mod_toolbox, only : gsw_saar, gsw_sp_from_sa_baltic

use gsw_mod_error_functions, only : gsw_error_code, gsw_error_limit

use gsw_mod_teos10_constants, only : gsw_ups

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sa, p, long, lat 

real (r8) :: gsw_sp_from_sa

real (r8) :: saar, sp_baltic

character (*), parameter :: func_name = "gsw_sp_from_sa"

sp_baltic = gsw_sp_from_sa_baltic(sa,long,lat)

if (sp_baltic .lt. 1e10_r8) then

   gsw_sp_from_sa = sp_baltic

else

   saar = gsw_saar(p,long,lat)
   if (saar .gt. gsw_error_limit) then
      gsw_sp_from_sa = gsw_error_code(1,func_name,saar)
   else
      gsw_sp_from_sa = (sa/gsw_ups)/(1.0_r8 + saar)
   end if

end if

return
end function

!--------------------------------------------------------------------------
!==========================================================================
elemental function gsw_sp_from_sa_baltic (sa, long, lat)
!==========================================================================
!
! For the Baltic Sea, calculates Practical Salinity with a value
! computed analytically from Absolute Salinity
!
! sa     : Absolute Salinity                               [g/kg]
! long   : longitude                                       [deg E]     
! lat    : latitude                                        [deg N]
!
! gsw_sp_from_sa_baltic  : Practical Salinity              [unitless]
!--------------------------------------------------------------------------

use gsw_mod_toolbox, only : gsw_util_xinterp1

use gsw_mod_baltic_data

use gsw_mod_teos10_constants, only : gsw_sso

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sa, long, lat

real (r8) :: gsw_sp_from_sa_baltic

real (r8) :: xx_left, xx_right

if (xb_left(2).lt.long .and. long.lt.xb_right(1) .and. &
    yb_left(1).lt.lat  .and.  lat.lt.yb_left(3)) then
  
    xx_left = gsw_util_xinterp1(yb_left, xb_left, 3, lat)
    
    xx_right = gsw_util_xinterp1(yb_right, xb_right, 2, lat)
    
    if(xx_left.le.long .and. long.le.xx_right) then
        gsw_sp_from_sa_baltic = (35.0_r8/(gsw_sso - 0.087_r8))*(sa - 0.087_r8)
    else
        gsw_sp_from_sa_baltic = 9e15_r8
    end if
     
else
    gsw_sp_from_sa_baltic = 9e15_r8
end if

return
end function

!--------------------------------------------------------------------------
!==========================================================================
elemental function gsw_sp_from_sk (sk)       
!==========================================================================
!
! Calculates Practical Salinity, SP, from SK
!
!  SK    : Knudsen Salinity                        [parts per thousand, ppt]
!
! gsw_sp_from_sk  : Practical Salinity                              [unitless]
!--------------------------------------------------------------------------

use gsw_mod_teos10_constants, only : gsw_soncl

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sk       

real (r8) :: gsw_sp_from_sk

gsw_sp_from_sk = (sk - 0.03_r8)*(gsw_soncl/1.805_r8) 

return
end function

!--------------------------------------------------------------------------
!==========================================================================
elemental function gsw_sp_from_sr (sr)  
!==========================================================================
!
! Calculates Practical Salinity, sp, from Reference Salinity, sr. 
!
! sr     : Reference Salinity                              [g/kg]
!
! gsw_sp_from_sr  : Practical Salinity                     [unitless]
!--------------------------------------------------------------------------

use gsw_mod_teos10_constants, only : gsw_ups

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sr  

real (r8) :: gsw_sp_from_sr

gsw_sp_from_sr = sr/gsw_ups

return
end function

!--------------------------------------------------------------------------



!==========================================================================
elemental function gsw_sp_from_sstar (sstar, p, long, lat)  
!==========================================================================
!
! Calculates Practical Salinity, SP, from Preformed Salinity, Sstar. 
!
! sstar  : Preformed Salinity                              [g/kg]
! p      : sea pressure                                    [dbar]
! long   : longitude                                       [deg E]     
! lat    : latitude                                        [deg N]
!
! gsw_sp_from_Sstar : Preformed Salinity                   [g/kg]
!--------------------------------------------------------------------------

use gsw_mod_toolbox, only : gsw_saar, gsw_sp_from_sa_baltic

use gsw_mod_error_functions, only : gsw_error_code, gsw_error_limit

use gsw_mod_teos10_constants, only : gsw_ups

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sstar, p, long, lat  

real (r8) :: gsw_sp_from_sstar

real (r8) :: saar, sp_baltic

character (*), parameter :: func_name = "gsw_sp_from_sstar"

! In the Baltic Sea, SA = Sstar.
sp_baltic = gsw_sp_from_sa_baltic(sstar,long,lat)

if (sp_baltic .lt. 1e10_r8) then

   gsw_sp_from_sstar = sp_baltic

else

   saar = gsw_saar(p,long,lat)
   if (saar .gt. gsw_error_limit) then
      gsw_sp_from_sstar = gsw_error_code(1,func_name,saar)
   else
      gsw_sp_from_sstar = (sstar/gsw_ups)/(1.0_r8 - 0.35_r8*saar)
   end if

end if

return
end function

!--------------------------------------------------------------------------
!==========================================================================
elemental function gsw_specvol (sa, ct, p)
!==========================================================================
! 
!  Calculates specific volume from Absolute Salinity, Conservative 
!  Temperature and pressure, using the computationally-efficient
!  polynomial expression for specific volume (Roquet et al., 2014).
!
!  SA  =  Absolute Salinity                                        [ g/kg ]
!  CT  =  Conservative Temperature (ITS-90)                       [ deg C ]
!  p   =  sea pressure                                             [ dbar ]
!         ( i.e. absolute pressure - 10.1325 dbar )
!
!  specvol  =  specific volume                                   [ m^3/kg ]
!--------------------------------------------------------------------------

use gsw_mod_teos10_constants, only : gsw_sfac, offset

use gsw_mod_specvol_coefficients

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sa, ct, p

real (r8) :: gsw_specvol

real (r8) :: xs, ys, z

xs = sqrt(gsw_sfac*sa + offset)
ys = ct*0.025_r8
z = p*1e-4_r8

gsw_specvol = v000 + xs*(v010 + xs*(v020 + xs*(v030 + xs*(v040 + xs*(v050 &
    + v060*xs))))) + ys*(v100 + xs*(v110 + xs*(v120 + xs*(v130 + xs*(v140 &
    + v150*xs)))) + ys*(v200 + xs*(v210 + xs*(v220 + xs*(v230 + v240*xs))) &
    + ys*(v300 + xs*(v310 + xs*(v320 + v330*xs)) + ys*(v400 + xs*(v410 &
    + v420*xs) + ys*(v500 + v510*xs + v600*ys))))) + z*(v001 + xs*(v011 &
    + xs*(v021 + xs*(v031 + xs*(v041 + v051*xs)))) + ys*(v101 + xs*(v111 &
    + xs*(v121 + xs*(v131 + v141*xs))) + ys*(v201 + xs*(v211 + xs*(v221 &
    + v231*xs)) + ys*(v301 + xs*(v311 + v321*xs) + ys*(v401 + v411*xs &
    + v501*ys)))) + z*(v002 + xs*(v012 + xs*(v022 + xs*(v032 + v042*xs))) &
    + ys*(v102 + xs*(v112 + xs*(v122 + v132*xs)) + ys*(v202 + xs*(v212 &
    + v222*xs) + ys*(v302 + v312*xs + v402*ys))) + z*(v003 + xs*(v013 &
    + v023*xs) + ys*(v103 + v113*xs + v203*ys) + z*(v004 + v014*xs + v104*ys &  
    + z*(v005 + v006*z)))))

return
end function

!--------------------------------------------------------------------------
!==========================================================================
elemental subroutine gsw_specvol_alpha_beta (sa, ct, p, specvol, alpha, &
                                             beta)
!==========================================================================
! 
!  Calculates specific volume, the appropiate thermal expansion coefficient
!  and the appropriate saline contraction coefficient of seawater from 
!  Absolute Salinity and Conservative Temperature.  This function uses the
!  computationally-efficient expression for specific volume in terms of 
!  SA, CT and p (Roquet et al., 2014).
!
!  SA  =  Absolute Salinity                                        [ g/kg ]
!  CT  =  Conservative Temperature (ITS-90)                       [ deg C ]
!  p   =  sea pressure                                             [ dbar ]
!         ( i.e. absolute pressure - 10.1325 dbar )
!
!  specvol =  specific volume                                      [ m/kg ]
!  alpha   =  thermal expansion coefficient                         [ 1/K ]
!             with respect to Conservative Temperature
!  beta    =  saline (i.e. haline) contraction                     [ kg/g ]
!             coefficient at constant Conservative Temperature
!--------------------------------------------------------------------------

use gsw_mod_teos10_constants, only : gsw_sfac, offset

use gsw_mod_specvol_coefficients

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sa, ct, p
real (r8), intent(out), optional :: specvol, alpha, beta

real (r8) :: v, v_ct, v_sa_part, xs, ys, z

xs = sqrt(gsw_sfac*sa + offset)
ys = ct*0.025_r8
z = p*1e-4_r8

v = v000 + xs*(v010 + xs*(v020 + xs*(v030 + xs*(v040 + xs*(v050 &
    + v060*xs))))) + ys*(v100 + xs*(v110 + xs*(v120 + xs*(v130 + xs*(v140 &
    + v150*xs)))) + ys*(v200 + xs*(v210 + xs*(v220 + xs*(v230 + v240*xs))) &
    + ys*(v300 + xs*(v310 + xs*(v320 + v330*xs)) + ys*(v400 + xs*(v410 &
    + v420*xs) + ys*(v500 + v510*xs + v600*ys))))) + z*(v001 + xs*(v011 &
    + xs*(v021 + xs*(v031 + xs*(v041 + v051*xs)))) + ys*(v101 + xs*(v111 &
    + xs*(v121 + xs*(v131 + v141*xs))) + ys*(v201 + xs*(v211 + xs*(v221 &
    + v231*xs)) + ys*(v301 + xs*(v311 + v321*xs) + ys*(v401 + v411*xs &
    + v501*ys)))) + z*(v002 + xs*(v012 + xs*(v022 + xs*(v032 + v042*xs))) &
    + ys*(v102 + xs*(v112 + xs*(v122 + v132*xs)) + ys*(v202 + xs*(v212 &
    + v222*xs) + ys*(v302 + v312*xs + v402*ys))) + z*(v003 + xs*(v013 &
    + v023*xs) + ys*(v103 + v113*xs + v203*ys) + z*(v004 + v014*xs + v104*ys &  
    + z*(v005 + v006*z)))))

if (present(specvol)) specvol = v

if (present(alpha)) then

    v_ct = a000 + xs*(a100 + xs*(a200 + xs*(a300 + xs*(a400 + a500*xs)))) &
             + ys*(a010 + xs*(a110 + xs*(a210 + xs*(a310 + a410*xs))) &
             + ys*(a020 + xs*(a120 + xs*(a220 + a320*xs)) + ys*(a030 &
             + xs*(a130 + a230*xs) + ys*(a040 + a140*xs + a050*ys )))) &
             + z*(a001 + xs*(a101 + xs*(a201 + xs*(a301 + a401*xs))) &
             + ys*(a011 + xs*(a111 + xs*(a211 + a311*xs)) + ys*(a021 &
             + xs*(a121 + a221*xs) + ys*(a031 + a131*xs + a041*ys))) &
             + z*(a002 + xs*(a102 + xs*(a202 + a302*xs)) + ys*(a012 &
             + xs*(a112 + a212*xs) + ys*(a022 + a122*xs + a032*ys)) &
             + z*(a003 + a103*xs + a013*ys + a004*z))) 

    alpha = 0.025_r8*v_ct/v

end if

if (present(beta)) then

    v_sa_part = b000 + xs*(b100 + xs*(b200 + xs*(b300 + xs*(b400 + b500*xs)))) &
           + ys*(b010 + xs*(b110 + xs*(b210 + xs*(b310 + b410*xs))) &
           + ys*(b020 + xs*(b120 + xs*(b220 + b320*xs)) + ys*(b030 &
           + xs*(b130 + b230*xs) + ys*(b040 + b140*xs + b050*ys)))) &
           + z*(b001 + xs*(b101 + xs*(b201 + xs*(b301 + b401*xs))) &
           + ys*(b011 + xs*(b111 + xs*(b211 + b311*xs)) + ys*(b021 &
           + xs*(b121 + b221*xs) + ys*(b031 + b131*xs + b041*ys))) &
           + z*(b002 + xs*(b102 + xs*(b202 + b302*xs))+ ys*(b012 &
           + xs*(b112 + b212*xs) + ys*(b022 + b122*xs + b032*ys)) &
           + z*(b003 +  b103*xs + b013*ys + b004*z)))
 
    beta = -v_sa_part*0.5_r8*gsw_sfac/(v*xs)

end if

return
end subroutine

!--------------------------------------------------------------------------
!==========================================================================
elemental function gsw_specvol_anom_standard (sa, ct, p)  
!==========================================================================
!
!  Calculates specific volume anomaly of seawater.
!
!  sa     : Absolute Salinity                               [g/kg]
!  ct     : Conservative Temperature                        [deg C]
!  p      : sea pressure                                    [dbar]
! 
!  specvol_anom  :  specific volume anomaly of seawater
!--------------------------------------------------------------------------

use gsw_mod_toolbox, only : gsw_specvol, gsw_specvol_sso_0

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sa, ct, p  

real (r8) :: gsw_specvol_anom_standard

gsw_specvol_anom_standard = gsw_specvol(sa,ct,p) - gsw_specvol_sso_0(p)

return
end function

!--------------------------------------------------------------------------
!==========================================================================
elemental subroutine gsw_specvol_first_derivatives (sa, ct, p, v_sa, v_ct, &
                                                    v_p, iflag)
! =========================================================================
!
!  Calculates three first-order derivatives of specific volume (v).
!  Note that this function uses the computationally-efficient
!  expression for specific volume (Roquet et al., 2014).
!
!  SA  =  Absolute Salinity                                        [ g/kg ]
!  CT  =  Conservative Temperature (ITS-90)                       [ deg C ]
!  p   =  sea pressure                                             [ dbar ]
!         ( i.e. absolute pressure - 10.1325 dbar )
!
!  v_SA  =  The first derivative of specific volume with respect to 
!           Absolute Salinity at constant CT & p.       [ J/(kg (g/kg)^2) ]
!  v_CT  =  The first derivative of specific volume with respect to 
!           CT at constant SA and p.                     [ J/(kg K(g/kg)) ]
!  v_P   =  The first derivative of specific volume with respect to 
!           P at constant SA and CT.                         [ J/(kg K^2) ]
!--------------------------------------------------------------------------

use gsw_mod_teos10_constants, only : gsw_sfac, offset

use gsw_mod_specvol_coefficients

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sa, ct, p
integer, intent(in), optional :: iflag
real (r8), intent(out), optional :: v_sa, v_ct, v_p

integer :: i
logical :: flags(3)
real (r8) :: v_ct_part, v_p_part, v_sa_part, xs, ys, z

xs = sqrt(gsw_sfac*sa + offset)
ys = ct*0.025_r8
z = p*1e-4_r8

if (present(iflag)) then
    do i = 1, 3
        flags(i) = btest(iflag,i)
    end do
else
    flags = .true.
end if

if (present(v_sa) .and. flags(1)) then

    v_sa_part = b000 + xs*(b100 + xs*(b200 + xs*(b300 + xs*(b400 + b500*xs)))) &
           + ys*(b010 + xs*(b110 + xs*(b210 + xs*(b310 + b410*xs))) &
           + ys*(b020 + xs*(b120 + xs*(b220 + b320*xs)) + ys*(b030 &
           + xs*(b130 + b230*xs) + ys*(b040 + b140*xs + b050*ys)))) &
           + z*(b001 + xs*(b101 + xs*(b201 + xs*(b301 + b401*xs))) &
           + ys*(b011 + xs*(b111 + xs*(b211 + b311*xs)) + ys*(b021 &
           + xs*(b121 + b221*xs) + ys*(b031 + b131*xs + b041*ys))) &
           + z*(b002 + xs*(b102 + xs*(b202 + b302*xs))+ ys*(b012 &
           + xs*(b112 + b212*xs) + ys*(b022 + b122*xs + b032*ys)) &
           + z*(b003 +  b103*xs + b013*ys + b004*z)))
 
    v_sa = 0.5_r8*gsw_sfac*v_sa_part/xs

end if


if (present(v_ct) .and. flags(2)) then

    v_ct_part = a000 + xs*(a100 + xs*(a200 + xs*(a300 + xs*(a400 + a500*xs)))) &
             + ys*(a010 + xs*(a110 + xs*(a210 + xs*(a310 + a410*xs))) &
             + ys*(a020 + xs*(a120 + xs*(a220 + a320*xs)) + ys*(a030 &
             + xs*(a130 + a230*xs) + ys*(a040 + a140*xs + a050*ys )))) &
             + z*(a001 + xs*(a101 + xs*(a201 + xs*(a301 + a401*xs))) &
             + ys*(a011 + xs*(a111 + xs*(a211 + a311*xs)) + ys*(a021 &
             + xs*(a121 + a221*xs) + ys*(a031 + a131*xs + a041*ys))) &
             + z*(a002 + xs*(a102 + xs*(a202 + a302*xs)) + ys*(a012 &
             + xs*(a112 + a212*xs) + ys*(a022 + a122*xs + a032*ys)) &
             + z*(a003 + a103*xs + a013*ys + a004*z))) 

    v_ct = 0.025_r8*v_ct_part

end if

if (present(v_p) .and. flags(3)) then

    v_p_part = c000 + xs*(c100 + xs*(c200 + xs*(c300 + xs*(c400 + c500*xs)))) & 
        + ys*(c010 + xs*(c110 + xs*(c210 + xs*(c310 + c410*xs))) + ys*(c020 &
        + xs*(c120 + xs*(c220 + c320*xs)) + ys*(c030 + xs*(c130 + c230*xs) &
        + ys*(c040 + c140*xs + c050*ys)))) + z*(c001 + xs*(c101 + xs*(c201 &
        + xs*(c301 + c401*xs))) + ys*(c011 + xs*(c111 + xs*(c211 + c311*xs)) &
        + ys*(c021 + xs*(c121 + c221*xs) + ys*(c031 + c131*xs + c041*ys))) &
        + z*( c002 + xs*(c102 + c202*xs) + ys*(c012 + c112*xs + c022*ys) &
        + z*(c003 + c103*xs + c013*ys + z*(c004 + c005*z))))

    v_p = 1e-8_r8*v_p_part

end if

return
end subroutine

!--------------------------------------------------------------------------
!==========================================================================
elemental subroutine gsw_specvol_first_derivatives_wrt_enthalpy (sa, ct, &
                                                       p, v_sa, v_h, iflag)
! =========================================================================
!
!  Calculates two first-order derivatives of specific volume (v).
!  Note that this function uses the using the computationally-efficient
!  expression for specific volume (Roquet et al., 2014).
!
!  SA  =  Absolute Salinity                                        [ g/kg ]
!  CT  =  Conservative Temperature (ITS-90)                       [ deg C ]
!  p   =  sea pressure                                             [ dbar ]
!         ( i.e. absolute pressure - 10.1325 dbar )
!
!  v_SA  =  The first derivative of specific volume with respect to 
!              Absolute Salinity at constant CT & p.    [ J/(kg (g/kg)^2) ]
!  v_h  =  The first derivative of specific volume with respect to 
!              SA and CT at constant p.                  [ J/(kg K(g/kg)) ]
!--------------------------------------------------------------------------

use gsw_mod_toolbox, only : gsw_enthalpy_first_derivatives
use gsw_mod_toolbox, only : gsw_specvol_first_derivatives

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sa, ct, p
integer, intent(in), optional :: iflag
real (r8), intent(out), optional :: v_sa, v_h

integer :: i
logical :: flags(2)
real (r8) :: h_ct, h_sa, rec_h_ct, vct_ct, vct_sa

if (present(iflag)) then
    do i = 1, 2
        flags(i) = btest(iflag,i)
    end do
else
    flags = .true.
end if

if (present(v_sa) .and. flags(1)) then

    call gsw_specvol_first_derivatives(sa,ct,p,vct_sa,vct_ct)
    call gsw_enthalpy_first_derivatives(sa,ct,p,h_sa,h_ct)

else if (present(v_h) .and. flags(2)) then

    call gsw_specvol_first_derivatives(sa,ct,p,v_ct=vct_ct)
    call gsw_enthalpy_first_derivatives(sa,ct,p,h_ct=h_ct)

end if

rec_h_ct = 1.0_r8/h_ct

if (present(v_sa) .and. flags(1)) v_sa = vct_sa - (vct_ct*h_sa)*rec_h_ct

if (present(v_h) .and. flags(2)) v_h = vct_ct*rec_h_ct

return
end subroutine

!--------------------------------------------------------------------------
!==========================================================================
elemental function gsw_specvol_ice (t, p)
!==========================================================================
!
!  Calculates the specific volume of ice. 
! 
!  t  =  in-situ temperature (ITS-90)                             [ deg C ]
!  p  =  sea pressure                                              [ dbar ]
!         ( i.e. absolute pressure - 10.1325 dbar )
!
!  specvol_ice  =  specific volume                               [ m^3/kg ]
!--------------------------------------------------------------------------

use gsw_mod_toolbox, only : gsw_gibbs_ice

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: t, p

real (r8) :: gsw_specvol_ice

gsw_specvol_ice = gsw_gibbs_ice(0,1,t,p)

return
end function

!--------------------------------------------------------------------------
!==========================================================================
elemental subroutine gsw_specvol_second_derivatives (sa, ct, p, v_sa_sa, &
                                   v_sa_ct, v_ct_ct, v_sa_p, v_ct_p, iflag)
! =========================================================================
!
!  Calculates five second-order derivatives of specific volume (v).
!  Note that this function uses the computationally-efficient
!  expression for specific volume (Roquet et al., 2014).
!
!  SA  =  Absolute Salinity                                        [ g/kg ]
!  CT  =  Conservative Temperature (ITS-90)                       [ deg C ]
!  p   =  sea pressure                                             [ dbar ]
!         ( i.e. absolute pressure - 10.1325 dbar )
!
!  v_SA_SA  =  The second derivative of specific volume with respect to 
!              Absolute Salinity at constant CT & p.    [ J/(kg (g/kg)^2) ]
!  v_SA_CT  =  The second derivative of specific volume with respect to 
!              SA and CT at constant p.                  [ J/(kg K(g/kg)) ]
!  v_CT_CT  =  The second derivative of specific volume with respect to 
!              CT at constant SA and p.                      [ J/(kg K^2) ]
!  v_SA_P  =  The second derivative of specific volume with respect to 
!              SA and P at constant CT.                  [ J/(kg K(g/kg)) ]
!  v_CT_P  =  The second derivative of specific volume with respect to 
!              CT and P at constant SA.                  [ J/(kg K(g/kg)) ]
!--------------------------------------------------------------------------

use gsw_mod_teos10_constants, only : gsw_sfac, offset

use gsw_mod_specvol_coefficients

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sa, ct, p
integer, intent(in), optional :: iflag
real (r8), intent(out), optional :: v_sa_sa, v_sa_ct, v_ct_ct, v_sa_p, v_ct_p

integer :: i
logical :: flags(5)
real (r8) :: v_ct_ct_part, v_ct_p_part, v_sa_ct_part, v_sa_p_part
real (r8) :: v_sa_sa_part, xs, xs2, ys, z

xs2 = gsw_sfac*sa + offset
xs = sqrt(xs2)
ys = ct*0.025_r8
z = p*1e-4_r8

if (present(iflag)) then
    do i = 1, 5
        flags(i) = btest(iflag,i)
    end do
else
    flags = .true.
end if

if (present(v_sa_sa) .and. flags(1)) then

   v_sa_sa_part = (-b000 + xs2*(b200 + xs*(2.0_r8*b300 + xs*(3.0_r8*b400 &
       + 4.0_r8*b500*xs))) + ys*(-b010 + xs2*(b210 + xs*(2.0_r8*b310 &
       + 3.0_r8*b410*xs)) + ys*(-b020 + xs2*(b220 + 2.0_r8*b320*xs) &
       + ys*(-b030 + b230*xs2 + ys*(-b040 - b050*ys)))) + z*(-b001 &
       + xs2*(b201 + xs*(2.0_r8*b301 + 3.0_r8*b401*xs)) + ys*(-b011 &
       + xs2*(b211 + 2.0_r8*b311*xs) + ys*(-b021 + b221*xs2 &
       + ys*(-b031 - b041*ys))) + z*(-b002 + xs2*(b202 + 2.0_r8*b302*xs) &
       + ys*(-b012 + b212*xs2 + ys*(-b022 - b032*ys)) + z*(-b003 &
       - b013*ys - b004*z))))/xs2

   v_sa_sa = 0.25_r8*gsw_sfac*gsw_sfac*v_sa_sa_part/xs

end if

if (present(v_sa_ct) .and. flags(2)) then

   v_sa_ct_part = (b010 + xs*(b110 + xs*(b210 + xs*(b310 + b410*xs))) &
       + ys*(2.0_r8*(b020 + xs*(b120 + xs*(b220 + b320*xs))) &
       + ys*(3.0_r8*(b030 + xs*(b130 + b230*xs)) + ys*(4.0_r8*(b040 + b140*xs) &
       + 5.0_r8*b050*ys))) + z*(b011 + xs*(b111 + xs*(b211 + b311*xs)) &
       + ys*(2.0_r8*(b021 + xs*(b121 + b221*xs)) + ys*(3.0_r8*(b031 + b131*xs) &
       + 4.0_r8*b041*ys)) + z*(b012 + xs*(b112 + b212*xs) + ys*(2.0_r8*(b022 &
       + b122*xs) + 3.0_r8*b032*ys) + b013*z)))/xs

   v_sa_ct = 0.025_r8*0.5_r8*gsw_sfac*v_sa_ct_part

end if

if (present(v_ct_ct) .and. flags(3)) then

   v_ct_ct_part = a010 + xs*(a110 + xs*(a210 + xs*(a310 + a410*xs))) &
       + ys*(2.0_r8*(a020 + xs*(a120 + xs*(a220 + a320*xs))) &
       + ys*(3.0_r8*(a030 + xs*(a130 + a230*xs)) + ys*(4.0_r8*(a040 &
       + a140*xs) + 5.0_r8*a050*ys))) + z*( a011 + xs*(a111 + xs*(a211 &
       + a311*xs)) + ys*(2.0_r8*(a021 + xs*(a121 + a221*xs)) &
       + ys*(3.0_r8*(a031 + a131*xs) + 4.0_r8*a041*ys)) + z*(a012 &
       + xs*(a112 + a212*xs) + ys*(2.0_r8*(a022 + a122*xs) &
       + 3.0_r8*a032*ys) + a013*z))

   v_ct_ct = 0.025_r8*0.025_r8*v_ct_ct_part

end if

if (present(v_sa_p) .and. flags(4)) then

   v_sa_p_part = b001 + xs*(b101 + xs*(b201 + xs*(b301 + b401*xs))) + ys*(b011 &
       + xs*(b111 + xs*(b211 + b311*xs)) + ys*(b021 + xs*(b121 + b221*xs) &
       + ys*(b031 + b131*xs + b041*ys))) + z*(2.0_r8*(b002 + xs*(b102 &
       + xs*(b202 + b302*xs)) + ys*(b012 + xs*(b112 + b212*xs) + ys*(b022 &
       + b122*xs + b032*ys))) + z*(3.0_r8*(b003 + b103*xs + b013*ys) &
       + 4.0_r8*b004*z))

   v_sa_p = 1e-8_r8*0.5_r8*gsw_sfac*v_sa_p_part

end if

if (present(v_ct_p) .and. flags(5)) then

   v_ct_p_part = a001 + xs*(a101 + xs*(a201 + xs*(a301 + a401*xs))) + ys*(a011 &
        + xs*(a111 + xs*(a211 + a311*xs)) + ys*(a021 + xs*(a121 + a221*xs) &
        + ys*(a031 + a131*xs + a041*ys))) + z*(2.0_r8*(a002 + xs*(a102 &
        + xs*(a202 + a302*xs)) + ys*(a012 + xs*(a112 + a212*xs) + ys*(a022 &
        + a122*xs + a032*ys))) + z*(3.0_r8*(a003 + a103*xs + a013*ys) &
        + 4.0_r8*a004*z))

   v_ct_p = 1e-8_r8*0.025_r8*v_ct_p_part

end if

return
end subroutine

!--------------------------------------------------------------------------
!==========================================================================
elemental subroutine gsw_specvol_second_derivatives_wrt_enthalpy (sa, ct, &
                                          p, v_sa_sa, v_sa_h, v_h_h, iflag)
! =========================================================================
!
!  Calculates three first-order derivatives of specific volume (v) with
!  respect to enthalpy. Note that this function uses the using the
!  computationally-efficient expression for specific volume
!  (Roquet et al., 2014).
!
!  SA  =  Absolute Salinity                                        [ g/kg ]
!  CT  =  Conservative Temperature (ITS-90)                       [ deg C ]
!  p   =  sea pressure                                             [ dbar ]
!         ( i.e. absolute pressure - 10.1325 dbar )
!
!  v_SA_SA = The second-order derivative of specific volume with respect to 
!            Absolute Salinity at constant h & p.       [ J/(kg (g/kg)^2) ]
!  v_SA_h  = The second-order derivative of specific volume with respect to 
!            SA and h at constant p.                     [ J/(kg K(g/kg)) ]
!  v_h_h   = The second-order derivative with respect to h at 
!            constant SA & p.
!--------------------------------------------------------------------------

use gsw_mod_toolbox, only : gsw_enthalpy_first_derivatives
use gsw_mod_toolbox, only : gsw_enthalpy_second_derivatives
use gsw_mod_toolbox, only : gsw_specvol_first_derivatives
use gsw_mod_toolbox, only : gsw_specvol_second_derivatives

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sa, ct, p
integer, intent(in), optional :: iflag
real (r8), intent(out), optional :: v_sa_sa, v_sa_h, v_h_h

logical :: flags(3)
real (r8) :: h_ct, h_ct_ct, h_sa, h_sa_ct, h_sa_sa, rec_h_ct, v_h_h_part
real (r8) :: rec_h_ct2, v_ct, vct_ct_ct, vct_sa_ct, vct_sa_sa, v_sa_h_part

if (present(iflag)) then
    flags(1) = present(v_sa_sa) .and. btest(iflag,1)
    flags(2) = present(v_sa_h) .and. btest(iflag,2)
    flags(3) = present(v_h_h) .and. btest(iflag,3)
else
    flags(1) = present(v_sa_sa)
    flags(2) = present(v_sa_h)
    flags(3) = present(v_h_h)
end if

call gsw_specvol_first_derivatives(sa,ct,p,v_ct=v_ct)

if (flags(1) .or. flags(2)) then
   call gsw_enthalpy_first_derivatives(sa,ct,p,h_sa,h_ct)
else
   call gsw_enthalpy_first_derivatives(sa,ct,p,h_ct=h_ct)
end if

if (flags(1)) then
   call gsw_specvol_second_derivatives(sa,ct,p,vct_sa_sa,vct_sa_ct,vct_ct_ct)
else if (flags(2)) then
   call gsw_specvol_second_derivatives(sa,ct,p,v_sa_ct=vct_sa_ct, &
                                                v_ct_ct=vct_ct_ct)
else
   call gsw_specvol_second_derivatives(sa,ct,p,v_ct_ct=vct_ct_ct)
end if

if (flags(1)) then
   call gsw_enthalpy_second_derivatives(sa,ct,p,h_sa_sa,h_sa_ct,h_ct_ct)
else if (flags(2)) then
   call gsw_enthalpy_second_derivatives(sa,ct,p,h_sa_ct=h_sa_ct,h_ct_ct=h_ct_ct)
else
   call gsw_enthalpy_second_derivatives(sa,ct,p,h_ct_ct=h_ct_ct)
end if

rec_h_ct = 1.0_r8/h_ct
rec_h_ct2 = rec_h_ct**2.0_r8

v_h_h_part = (vct_ct_ct*h_ct - h_ct_ct*v_ct)*(rec_h_ct2*rec_h_ct)

if (flags(3)) v_h_h = v_h_h_part

if (flags(1) .or. flags(2)) then

    v_sa_h_part = (vct_sa_ct*h_ct - v_ct*h_sa_ct)*rec_h_ct2 - h_sa*v_h_h_part

    if (flags(2)) v_sa_h = v_sa_h_part

    if (flags(1)) v_sa_sa = vct_sa_sa - (h_ct*(vct_sa_ct*h_sa &
            - v_ct*h_sa_sa) + v_ct*h_sa*h_sa_ct)*rec_h_ct2 - h_sa*v_sa_h_part
end if

return
end subroutine

!--------------------------------------------------------------------------
!==========================================================================
elemental function gsw_specvol_sso_0 (p)
!==========================================================================
!  This function calculates specifc volume at the Standard Ocean Salinity,
!  SSO, and at a Conservative Temperature of zero degrees C, as a function 
!  of pressure, p, in dbar, using a streamlined version of the CT version
!  of specific volume, that is, a streamlined version of the code
!  "gsw_specvol(SA,CT,p)".
!==========================================================================

use gsw_mod_specvol_coefficients

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: p

real (r8) :: gsw_specvol_sso_0

real (r8) :: z
 
z = p*1e-4_r8

gsw_specvol_sso_0 = 9.726613854843870e-04_r8 + z*(-4.505913211160929e-05_r8 &
    + z*(7.130728965927127e-06_r8 + z*(-6.657179479768312e-07_r8 &
    + z*(-2.994054447232880e-08_r8 + z*(v005 + v006*z)))))

return
end function

!--------------------------------------------------------------------------
!==========================================================================
elemental function gsw_specvol_t_exact (sa, t, p)  
!==========================================================================
!
! Calulates the specific volume of seawater
!
! sa     : Absolute Salinity                               [g/kg]
! t      : in-situ temperature                             [deg C]
! p      : sea pressure                                    [dbar]
! 
! gsw_specvol_t_exact : specific volume                    [kg/m^3]
!--------------------------------------------------------------------------

use gsw_mod_toolbox, only : gsw_gibbs

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sa, t, p  

real (r8) :: gsw_specvol_t_exact

integer, parameter :: n0=0, n1=1

gsw_specvol_t_exact = gsw_gibbs(n0,n0,n1,sa,t,p)

return
end function

!--------------------------------------------------------------------------
!==========================================================================
elemental function gsw_spiciness0 (sa, ct)
!==========================================================================
! 
!  Calculates spiciness from Absolute Salinity and Conservative 
!  Temperature at a pressure of 0 dbar, as described by McDougall and 
!  Krzysik (2015).  This routine is based on the computationally-efficient 
!  expression for specific volume in terms of SA, CT and p (Roquet et al., 
!  2015).
!
!  SA  =  Absolute Salinity                                        [ g/kg ]
!  CT  =  Conservative Temperature (ITS-90)                       [ deg C ]
!
!  spiciness0  =  spiciness referenced to a pressure of 0 dbar, 
!                 i.e. the surface                             [ kg/m^3 ]
!--------------------------------------------------------------------------

use gsw_mod_teos10_constants, only : gsw_sfac, offset

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sa, ct

real (r8) :: gsw_spiciness0

real (r8) :: xs, ys

real (r8), parameter :: s01 = -9.22982898371678e1_r8
real (r8), parameter :: s02 = -1.35727873628866e1_r8
real (r8), parameter :: s03 =  1.87353650994010e1_r8
real (r8), parameter :: s04 = -1.61360047373455e1_r8
real (r8), parameter :: s05 =  3.76112762286425e1_r8
real (r8), parameter :: s06 = -4.27086671461257e1_r8
real (r8), parameter :: s07 =  2.00820111041594e1_r8
real (r8), parameter :: s08 =  2.87969717584045e2_r8
real (r8), parameter :: s09 =  1.13747111959674e1_r8
real (r8), parameter :: s10 =  6.07377192990680e1_r8
real (r8), parameter :: s11 = -7.37514033570187e1_r8
real (r8), parameter :: s12 = -7.51171878953574e1_r8
real (r8), parameter :: s13 =  1.63310989721504e2_r8
real (r8), parameter :: s14 = -8.83222751638095e1_r8
real (r8), parameter :: s15 = -6.41725302237048e2_r8
real (r8), parameter :: s16 =  2.79732530789261e1_r8
real (r8), parameter :: s17 = -2.49466901993728e2_r8
real (r8), parameter :: s18 =  3.26691295035416e2_r8
real (r8), parameter :: s19 =  2.66389243708181e1_r8
real (r8), parameter :: s20 = -2.93170905757579e2_r8
real (r8), parameter :: s21 =  1.76053907144524e2_r8
real (r8), parameter :: s22 =  8.27634318120224e2_r8
real (r8), parameter :: s23 = -7.02156220126926e1_r8
real (r8), parameter :: s24 =  3.82973336590803e2_r8
real (r8), parameter :: s25 = -5.06206828083959e2_r8
real (r8), parameter :: s26 =  6.69626565169529e1_r8
real (r8), parameter :: s27 =  3.02851235050766e2_r8
real (r8), parameter :: s28 = -1.96345285604621e2_r8
real (r8), parameter :: s29 = -5.74040806713526e2_r8
real (r8), parameter :: s30 =  7.03285905478333e1_r8
real (r8), parameter :: s31 = -2.97870298879716e2_r8
real (r8), parameter :: s32 =  3.88340373735118e2_r8
real (r8), parameter :: s33 = -8.29188936089122e1_r8
real (r8), parameter :: s34 = -1.87602137195354e2_r8
real (r8), parameter :: s35 =  1.27096944425793e2_r8
real (r8), parameter :: s36 =  2.11671167892147e2_r8
real (r8), parameter :: s37 = -3.15140919876285e1_r8
real (r8), parameter :: s38 =  1.16458864953602e2_r8
real (r8), parameter :: s39 = -1.50029730802344e2_r8
real (r8), parameter :: s40 =  3.76293848660589e1_r8
real (r8), parameter :: s41 =  6.47247424373200e1_r8
real (r8), parameter :: s42 = -4.47159994408867e1_r8
real (r8), parameter :: s43 = -3.23533339449055e1_r8
real (r8), parameter :: s44 =  5.30648562097667_r8
real (r8), parameter :: s45 = -1.82051249177948e1_r8
real (r8), parameter :: s46 =  2.33184351090495e1_r8
real (r8), parameter :: s47 = -6.22909903460368_r8
real (r8), parameter :: s48 = -9.55975464301446_r8
real (r8), parameter :: s49 =  6.61877073960113_r8
 
xs = sqrt(gsw_sfac*sa + offset)
ys = ct*0.025

gsw_spiciness0 = &
          s01 + ys*(s02 + ys*(s03 + ys*(s04 + ys*(s05 + ys*(s06 + s07*ys))))) &
    + xs*(s08 + ys*(s09 + ys*(s10 + ys*(s11 + ys*(s12 + ys*(s13 + s14*ys))))) &
    + xs*(s15 + ys*(s16 + ys*(s17 + ys*(s18 + ys*(s19 + ys*(s20 + s21*ys))))) &
    + xs*(s22 + ys*(s23 + ys*(s24 + ys*(s25 + ys*(s26 + ys*(s27 + s28*ys))))) &
    + xs*(s29 + ys*(s30 + ys*(s31 + ys*(s32 + ys*(s33 + ys*(s34 + s35*ys))))) &
    + xs*(s36 + ys*(s37 + ys*(s38 + ys*(s39 + ys*(s40 + ys*(s41 + s42*ys))))) &
    + xs*(s43 + ys*(s44 + ys*(s45 + ys*(s46 + ys*(s47 + ys*(s48 + s49*ys)))))))))))

return
end function

!--------------------------------------------------------------------------
!==========================================================================
elemental function gsw_spiciness1 (sa, ct)
!==========================================================================
! 
!  Calculates spiciness from Absolute Salinity and Conservative 
!  Temperature at a pressure of 1000 dbar, as described by McDougall and 
!  Krzysik (2015).  This routine is based on the computationally-efficient 
!  expression for specific volume in terms of SA, CT and p (Roquet et al., 
!  2015).
!
!  SA  =  Absolute Salinity                                        [ g/kg ]
!  CT  =  Conservative Temperature (ITS-90)                       [ deg C ]
!
!  spiciness1  =  spiciness referenced to a pressure of 1000 dbar 
!                                                                [ kg/m^3 ]
!--------------------------------------------------------------------------

use gsw_mod_teos10_constants, only : gsw_sfac, offset

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sa, ct

real (r8) :: gsw_spiciness1

real (r8) :: xs, ys

real (r8), parameter :: s01 = -9.19874584868912e1_r8
real (r8), parameter :: s02 = -1.33517268529408e1_r8
real (r8), parameter :: s03 =  2.18352211648107e1_r8
real (r8), parameter :: s04 = -2.01491744114173e1_r8
real (r8), parameter :: s05 =  3.70004204355132e1_r8
real (r8), parameter :: s06 = -3.78831543226261e1_r8
real (r8), parameter :: s07 =  1.76337834294554e1_r8
real (r8), parameter :: s08 =  2.87838842773396e2_r8
real (r8), parameter :: s09 =  2.14531420554522e1_r8
real (r8), parameter :: s10 =  3.14679705198796e1_r8
real (r8), parameter :: s11 = -4.04398864750692e1_r8
real (r8), parameter :: s12 = -7.70796428950487e1_r8
real (r8), parameter :: s13 =  1.36783833820955e2_r8
real (r8), parameter :: s14 = -7.36834317044850e1_r8
real (r8), parameter :: s15 = -6.41753415180701e2_r8
real (r8), parameter :: s16 =  1.33701981685590_r8 
real (r8), parameter :: s17 = -1.75289327948412e2_r8
real (r8), parameter :: s18 =  2.42666160657536e2_r8
real (r8), parameter :: s19 =  3.17062400799114e1_r8
real (r8), parameter :: s20 = -2.28131490440865e2_r8
real (r8), parameter :: s21 =  1.39564245068468e2_r8
real (r8), parameter :: s22 =  8.27747934506435e2_r8
real (r8), parameter :: s23 = -3.50901590694775e1_r8
real (r8), parameter :: s24 =  2.87473907262029e2_r8
real (r8), parameter :: s25 = -4.00227341144928e2_r8
real (r8), parameter :: s26 =  6.48307189919433e1_r8
real (r8), parameter :: s27 =  2.16433334701578e2_r8
real (r8), parameter :: s28 = -1.48273032774305e2_r8
real (r8), parameter :: s29 = -5.74545648799754e2_r8
real (r8), parameter :: s30 =  4.50446431127421e1_r8
real (r8), parameter :: s31 = -2.30714981343772e2_r8
real (r8), parameter :: s32 =  3.15958389253065e2_r8
real (r8), parameter :: s33 = -8.60635313930106e1_r8
real (r8), parameter :: s34 = -1.22978455069097e2_r8
real (r8), parameter :: s35 =  9.18287282626261e1_r8
real (r8), parameter :: s36 =  2.12120473062203e2_r8
real (r8), parameter :: s37 = -2.21528216973820e1_r8
real (r8), parameter :: s38 =  9.19013417923270e1_r8
real (r8), parameter :: s39 = -1.24400776026014e2_r8
real (r8), parameter :: s40 =  4.08512871163839e1_r8
real (r8), parameter :: s41 =  3.91127352213516e1_r8
real (r8), parameter :: s42 = -3.10508021853093e1_r8
real (r8), parameter :: s43 = -3.24790035899152e1_r8
real (r8), parameter :: s44 =  3.91029016556786_r8
real (r8), parameter :: s45 = -1.45362719385412e1_r8
real (r8), parameter :: s46 =  1.96136194246355e1_r8
real (r8), parameter :: s47 = -7.06035474689088_r8
real (r8), parameter :: s48 = -5.36884688614009_r8
real (r8), parameter :: s49 =  4.43247303092448_r8
 
xs = sqrt(gsw_sfac*sa + offset)
ys = ct*0.025

gsw_spiciness1 = &
          s01 + ys*(s02 + ys*(s03 + ys*(s04 + ys*(s05 + ys*(s06 + s07*ys))))) &
    + xs*(s08 + ys*(s09 + ys*(s10 + ys*(s11 + ys*(s12 + ys*(s13 + s14*ys))))) &
    + xs*(s15 + ys*(s16 + ys*(s17 + ys*(s18 + ys*(s19 + ys*(s20 + s21*ys))))) &
    + xs*(s22 + ys*(s23 + ys*(s24 + ys*(s25 + ys*(s26 + ys*(s27 + s28*ys))))) &
    + xs*(s29 + ys*(s30 + ys*(s31 + ys*(s32 + ys*(s33 + ys*(s34 + s35*ys))))) &
    + xs*(s36 + ys*(s37 + ys*(s38 + ys*(s39 + ys*(s40 + ys*(s41 + s42*ys))))) &
    + xs*(s43 + ys*(s44 + ys*(s45 + ys*(s46 + ys*(s47 + ys*(s48 + s49*ys)))))))))))

return
end function

!--------------------------------------------------------------------------
!==========================================================================
elemental function gsw_spiciness2 (sa, ct)
!==========================================================================
! 
!  Calculates spiciness from Absolute Salinity and Conservative 
!  Temperature at a pressure of 2000 dbar, as described by McDougall and 
!  Krzysik (2015).  This routine is based on the computationally-efficient 
!  expression for specific volume in terms of SA, CT and p (Roquet et al., 
!  2015).
!
!  SA  =  Absolute Salinity                                        [ g/kg ]
!  CT  =  Conservative Temperature (ITS-90)                       [ deg C ]
!
!  spiciness2  =  spiciness referenced to a pressure of 2000 dbar 
!                                                                [ kg/m^3 ]
!--------------------------------------------------------------------------

use gsw_mod_teos10_constants, only : gsw_sfac, offset

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sa, ct

real (r8) :: gsw_spiciness2

real (r8) :: xs, ys

real (r8), parameter :: s01 = -9.17327320732265e1_r8
real (r8), parameter :: s02 = -1.31200235147912e1_r8
real (r8), parameter :: s03 =  2.49574345782503e1_r8
real (r8), parameter :: s04 = -2.41678075247398e1_r8
real (r8), parameter :: s05 =  3.61654631402053e1_r8
real (r8), parameter :: s06 = -3.22582164667710e1_r8
real (r8), parameter :: s07 =  1.45092623982509e1_r8
real (r8), parameter :: s08 =  2.87776645983195e2_r8
real (r8), parameter :: s09 =  3.13902307672447e1_r8
real (r8), parameter :: s10 =  1.69777467534459_r8
real (r8), parameter :: s11 = -5.69630115740438_r8
real (r8), parameter :: s12 = -7.97586359017987e1_r8
real (r8), parameter :: s13 =  1.07507460387751e2_r8
real (r8), parameter :: s14 = -5.58234404964787e1_r8
real (r8), parameter :: s15 = -6.41708068766557e2_r8
real (r8), parameter :: s16 = -2.53494801286161e1_r8
real (r8), parameter :: s17 = -9.86755437385364e1_r8
real (r8), parameter :: s18 =  1.52406930795842e2_r8
real (r8), parameter :: s19 =  4.23888258264105e1_r8
real (r8), parameter :: s20 = -1.60118811141438e2_r8
real (r8), parameter :: s21 =  9.67497898053989e1_r8
real (r8), parameter :: s22 =  8.27674355478637e2_r8
real (r8), parameter :: s23 =  5.27561234412133e-1_r8
real (r8), parameter :: s24 =  1.87440206992396e2_r8
real (r8), parameter :: s25 = -2.83295392345171e2_r8
real (r8), parameter :: s26 =  5.14485994597635e1_r8
real (r8), parameter :: s27 =  1.29975755062696e2_r8
real (r8), parameter :: s28 = -9.36526588377456e1_r8
real (r8), parameter :: s29 = -5.74911728972948e2_r8
real (r8), parameter :: s30 =  1.91175851862772e1_r8
real (r8), parameter :: s31 = -1.59347231968841e2_r8
real (r8), parameter :: s32 =  2.33884725744938e2_r8
real (r8), parameter :: s33 = -7.87744010546157e1_r8
real (r8), parameter :: s34 = -6.04757235443685e1_r8
real (r8), parameter :: s35 =  5.27869695599657e1_r8
real (r8), parameter :: s36 =  2.12517758478878e2_r8
real (r8), parameter :: s37 = -1.24351794740528e1_r8
real (r8), parameter :: s38 =  6.53904308937490e1_r8
real (r8), parameter :: s39 = -9.44804080763788e1_r8
real (r8), parameter :: s40 =  3.93874257887364e1_r8
real (r8), parameter :: s41 =  1.49425448888996e1_r8
real (r8), parameter :: s42 = -1.62350721656367e1_r8
real (r8), parameter :: s43 = -3.25936844276669e1_r8
real (r8), parameter :: s44 =  2.44035700301595_r8
real (r8), parameter :: s45 = -1.05079633683795e1_r8
real (r8), parameter :: s46 =  1.51515796259082e1_r8
real (r8), parameter :: s47 = -7.06609886460683_r8
real (r8), parameter :: s48 = -1.48043337052968_r8
real (r8), parameter :: s49 =  2.10066653978515_r8
 
xs = sqrt(gsw_sfac*sa + offset)
ys = ct*0.025

gsw_spiciness2 = &
          s01 + ys*(s02 + ys*(s03 + ys*(s04 + ys*(s05 + ys*(s06 + s07*ys))))) &
    + xs*(s08 + ys*(s09 + ys*(s10 + ys*(s11 + ys*(s12 + ys*(s13 + s14*ys))))) &
    + xs*(s15 + ys*(s16 + ys*(s17 + ys*(s18 + ys*(s19 + ys*(s20 + s21*ys))))) &
    + xs*(s22 + ys*(s23 + ys*(s24 + ys*(s25 + ys*(s26 + ys*(s27 + s28*ys))))) &
    + xs*(s29 + ys*(s30 + ys*(s31 + ys*(s32 + ys*(s33 + ys*(s34 + s35*ys))))) &
    + xs*(s36 + ys*(s37 + ys*(s38 + ys*(s39 + ys*(s40 + ys*(s41 + s42*ys))))) &
    + xs*(s43 + ys*(s44 + ys*(s45 + ys*(s46 + ys*(s47 + ys*(s48 + s49*ys)))))))))))

return
end function

!--------------------------------------------------------------------------
!==========================================================================
elemental function gsw_sr_from_sp (sp) 
!==========================================================================
!
! Calculates Reference Salinity, SR, from Practical Salinity, SP. 
!
! sp     : Practical Salinity                              [unitless]
!
! gsw_sr_from_sp : Reference Salinity                      [g/kg]
!--------------------------------------------------------------------------

use gsw_mod_teos10_constants, only : gsw_ups

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sp 

real (r8) :: gsw_sr_from_sp

gsw_sr_from_sp = sp*gsw_ups

return
end function

!--------------------------------------------------------------------------



!==========================================================================
elemental function gsw_sstar_from_sa (sa, p, long, lat) 
!==========================================================================
!
! Calculates Preformed Salinity, Sstar, from Absolute Salinity, SA. 
!
! sa     : Absolute Salinity                               [g/kg]
! p      : sea pressure                                    [dbar]
! long   : longitude                                       [deg E]     
! lat    : latitude                                        [deg N]
!
! gsw_sstar_from_sa : Preformed Salinity                   [g/kg]
!--------------------------------------------------------------------------

use gsw_mod_toolbox, only : gsw_saar

use gsw_mod_error_functions, only : gsw_error_code, gsw_error_limit

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sa, p, long, lat 

real (r8) :: gsw_sstar_from_sa

real (r8) :: saar

character (*), parameter :: func_name = "gsw_sstar_from_sa"

! In the Baltic Sea, Sstar = sa, and note that gsw_saar returns zero
! for saar in the Baltic.

saar = gsw_saar(p,long,lat)

if (saar.gt.gsw_error_limit) then
   gsw_sstar_from_sa = gsw_error_code(1,func_name,saar)
else
   gsw_sstar_from_sa = sa*(1.0_r8 - 0.35_r8*saar)/(1.0_r8 + saar)
end if

return
end function

!--------------------------------------------------------------------------
!==========================================================================
elemental function gsw_sstar_from_sp (sp, p, long, lat) 
!==========================================================================
!
! Calculates Preformed Salinity, Sstar, from Practical Salinity, SP. 
!
! sp     : Practical Salinity                              [unitless]
! p      : sea pressure                                    [dbar]
! long   : longitude                                       [deg E]     
! lat    : latitude                                        [deg N]
!
! gsw_sstar_from_sp  : Preformed Salinity                  [g/kg]
!--------------------------------------------------------------------------

use gsw_mod_toolbox, only : gsw_saar, gsw_sa_from_sp_baltic

use gsw_mod_error_functions, only : gsw_error_code, gsw_error_limit

use gsw_mod_teos10_constants, only : gsw_ups

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sp, p, long, lat 

real (r8) :: gsw_sstar_from_sp

real (r8) :: saar, sstar_baltic

character (*), parameter :: func_name = "gsw_sstar_from_sp"

!In the Baltic Sea, Sstar = SA.
sstar_baltic = gsw_sa_from_sp_baltic(sp,long,lat)

if (sstar_baltic .lt. 1e10_r8) then

   gsw_sstar_from_sp = sstar_baltic

else

   saar = gsw_saar(p,long,lat)
   if (saar .gt. gsw_error_limit) then
      gsw_sstar_from_sp = gsw_error_code(1,func_name,saar)
   else
      gsw_sstar_from_sp = gsw_ups*sp*(1 - 0.35_r8*saar)
   end if

end if

return
end function

!--------------------------------------------------------------------------
!==========================================================================
elemental function gsw_t_deriv_chem_potential_water_t_exact (sa, t, p)
!==========================================================================
!
!  Calculates the temperature derivative of the chemical potential of water
!  in seawater so that it is valid at exactly SA = 0.
!
!  SA  =  Absolute Salinity                                        [ g/kg ]
!  t   =  in-situ temperature (ITS-90)                            [ deg C ]
!  p   =  sea pressure                                             [ dbar ]
!         ( i.e. absolute pressure - 10.1325 dbar ) 
!
!  chem_potential_water_dt  =  temperature derivative of the chemical 
!                           potential of water in seawater  [ J g^-1 K^-1 ]
!--------------------------------------------------------------------------

use gsw_mod_teos10_constants, only : gsw_sfac, rec_db2pa

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sa, t, p

real (r8) :: gsw_t_deriv_chem_potential_water_t_exact

real (r8) :: g03_t, g08_sa_t, x, x2, y, z, g08_t

real (r8), parameter :: kg2g = 1e-3_r8

! Note. The kg2g, a factor of 1e-3, is needed to convert the output of this
! function into units of J/g. See section (2.9) of the TEOS-10 Manual.

x2 = gsw_sfac*sa
x = sqrt(x2)
y = t*0.025_r8
z = p*rec_db2pa ! the input pressure (p) is sea pressure in units of dbar.

g03_t = 5.90578347909402_r8 + z*(-270.983805184062_r8 + &
        z*(776.153611613101_r8 + z*(-196.51255088122_r8 + (28.9796526294175_r8 - &
        2.13290083518327_r8*z)*z))) + &
        y*(-24715.571866078_r8 + z*(2910.0729080936_r8 + &
        z*(-1513.116771538718_r8 + z*(546.959324647056_r8 + &
        z*(-111.1208127634436_r8 + 8.68841343834394_r8*z)))) + &
        y*(2210.2236124548363_r8 + z*(-2017.52334943521_r8 + &
        z*(1498.081172457456_r8 + z*(-718.6359919632359_r8 + &
        (146.4037555781616_r8 - 4.9892131862671505_r8*z)*z))) + &
        y*(-592.743745734632_r8 + z*(1591.873781627888_r8 + &
        z*(-1207.261522487504_r8 + (608.785486935364_r8 - &
        105.4993508931208_r8*z)*z)) + &
        y*(290.12956292128547_r8 + z*(-973.091553087975_r8 + &
        z*(602.603274510125_r8 + z*(-276.361526170076_r8 + &
        32.40953340386105_r8*z))) + &
        y*(-113.90630790850321_r8 + y*(21.35571525415769_r8 - &
        67.41756835751434_r8*z) + &
        z*(381.06836198507096_r8 + z*(-133.7383902842754_r8 + &
        49.023632509086724_r8*z)))))))
    
g08_t = x2*(168.072408311545_r8 + &
        x*(-493.407510141682_r8 + x*(543.835333000098_r8 + &
        x*(-196.028306689776_r8 + 36.7571622995805_r8*x) + &
        y*(-137.1145018408982_r8 + y*(148.10030845687618_r8 + &
        y*(-68.5590309679152_r8 + 12.4848504784754_r8*y))) - &
        22.6683558512829_r8*z) + z*(-175.292041186547_r8 + &
        (83.1923927801819_r8 - 29.483064349429_r8*z)*z) + &
        y*(-86.1329351956084_r8 + z*(766.116132004952_r8 + &
        z*(-108.3834525034224_r8 + 51.2796974779828_r8*z)) + &
        y*(-30.0682112585625_r8 - 1380.9597954037708_r8*z + &
        y*(3.50240264723578_r8 + 938.26075044542_r8*z)))))
    
g08_sa_t = 1187.3715515697959_r8 + &
        x*(-1480.222530425046_r8 + x*(2175.341332000392_r8 + &
        x*(-980.14153344888_r8 + 220.542973797483_r8*x) + &
        y*(-548.4580073635929_r8 + y*(592.4012338275047_r8 + &
        y*(-274.2361238716608_r8 + 49.9394019139016_r8*y))) - &
        90.6734234051316_r8*z) + z*(-525.876123559641_r8 + &
        (249.57717834054571_r8 - 88.449193048287_r8*z)*z) + &
        y*(-258.3988055868252_r8 + z*(2298.348396014856_r8 + &
        z*(-325.1503575102672_r8 + 153.8390924339484_r8*z)) + &
        y*(-90.2046337756875_r8 - 4142.8793862113125_r8*z + &
        y*(10.50720794170734_r8 + 2814.78225133626_r8*z))))
    
gsw_t_deriv_chem_potential_water_t_exact = kg2g*((g03_t + g08_t)*0.025_r8 - &
                                           0.5_r8*gsw_sfac*0.025_r8*sa*g08_sa_t)
return
end function

!--------------------------------------------------------------------------
!==========================================================================
elemental function gsw_t_freezing (sa, p, saturation_fraction, poly)
!==========================================================================
!
!  Calculates the in-situ temperature at which seawater freezes.
!
!  sa  =  absolute Salinity                                        [ g/kg ]
!  p   =  sea pressure                                             [ dbar ]
!         ( i.e. absolute pressure - 10.1325 dbar ) 
!  saturation_fraction = the saturation fraction of dissolved air in 
!                        seawater
!
!  t_freezing = in-situ temperature at which seawater freezes.    [ deg C ]
!               (ITS-90)                
!--------------------------------------------------------------------------

use gsw_mod_toolbox, only : gsw_t_freezing_exact, gsw_t_freezing_poly

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sa, p, saturation_fraction
logical, intent(in), optional :: poly

real (r8) :: gsw_t_freezing

logical :: do_poly

if (present(poly)) then
    do_poly = poly
else
    do_poly = .false.
end if

if (do_poly) then
    gsw_t_freezing = gsw_t_freezing_poly(sa,p,saturation_fraction)
else
    gsw_t_freezing = gsw_t_freezing_exact(sa,p,saturation_fraction)
end if

return
end function

!--------------------------------------------------------------------------
!==========================================================================
elemental function gsw_t_freezing_exact (sa, p, saturation_fraction)
!==========================================================================
!
!  Calculates the in-situ temperature at which seawater freezes. The 
!  in-situ temperature freezing point is calculated from the exact 
!  in-situ freezing temperature which is found by a modified Newton-Raphson
!  iteration (McDougall and Wotherspoon, 2013) of the equality of the 
!  chemical potentials of water in seawater and in ice.
!
!  An alternative GSW function, gsw_t_freezing_poly, it is based on a 
!  computationally-efficient polynomial, and is accurate to within -5e-4 K 
!  and 6e-4 K, when compared with this function.
!
!  SA  =  Absolute Salinity                                        [ g/kg ]
!  p   =  sea pressure                                             [ dbar ]
!         ( i.e. absolute pressure - 10.1325 dbar ) 
!  saturation_fraction = the saturation fraction of dissolved air in 
!                        seawater
!  (i.e., saturation_fraction must be between 0 and 1, and the default 
!    is 1, completely saturated) 
!
!  t_freezing = in-situ temperature at which seawater freezes.    [ deg C ]
!--------------------------------------------------------------------------

use gsw_mod_teos10_constants, only : gsw_sso

use gsw_mod_toolbox, only : gsw_gibbs_ice, gsw_chem_potential_water_t_exact
use gsw_mod_toolbox, only : gsw_t_deriv_chem_potential_water_t_exact
use gsw_mod_toolbox, only : gsw_t_freezing_poly

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sa, p, saturation_fraction

real (r8) :: gsw_t_freezing_exact

real (r8) :: df_dt, tf, tfm, tf_old, f

! The initial value of t_freezing_exact (for air-free seawater)
tf = gsw_t_freezing_poly(sa,p,polynomial=.true.)

df_dt = 1e3_r8*gsw_t_deriv_chem_potential_water_t_exact(sa,tf,p) - &
                gsw_gibbs_ice(1,0,tf,p)
! df_dt here is the initial value of the derivative of the function f whose
! zero (f = 0) we are finding (see Eqn. (3.33.2) of IOC et al (2010)).  

tf_old = tf
f = 1e3_r8*gsw_chem_potential_water_t_exact(sa,tf_old,p) - &
                gsw_gibbs_ice(0,0,tf_old,p)
tf = tf_old - f/df_dt
tfm = 0.5_r8*(tf + tf_old)
df_dt = 1e3_r8*gsw_t_deriv_chem_potential_water_t_exact(sa,tfm,p) - &
                gsw_gibbs_ice(1,0,tfm,p)
tf = tf_old - f/df_dt

tf_old = tf
f = 1e3_r8*gsw_chem_potential_water_t_exact(sa,tf_old,p) - &
                gsw_gibbs_ice(0,0,tf_old,p)
tf = tf_old - f/df_dt

! Adjust for the effects of dissolved air
gsw_t_freezing_exact = tf - &
                saturation_fraction*(1e-3_r8)*(2.4_r8 - sa/(2.0_r8*gsw_sso)) 

return
end function

!--------------------------------------------------------------------------
!==========================================================================
elemental subroutine gsw_t_freezing_first_derivatives (sa, p, &
                            saturation_fraction, tfreezing_sa, tfreezing_p)
!==========================================================================
!
!  Calculates the frist derivatives of the in-situ temperature at which 
!  seawater freezes with respect to Absolute Salinity SA and pressure P (in
!  Pa).  These expressions come from differentiating the expression that
!  defines the freezing temperature, namely the equality between the 
!  chemical potentials of water in seawater and in ice.  
!
!  SA  =  Absolute Salinity                                        [ g/kg ]
!  p   =  sea pressure                                             [ dbar ]
!         ( i.e. absolute pressure - 10.1325 dbar ) 
!  saturation_fraction = the saturation fraction of dissolved air in 
!                        seawater
!
!  tfreezing_SA = the derivative of the in-situ freezing temperature 
!                 (ITS-90) with respect to Absolute Salinity at fixed    
!                 pressure                     [ K/(g/kg) ] i.e. [ K kg/g ] 
!
!  tfreezing_P  = the derivative of the in-situ freezing temperature  
!                 (ITS-90) with respect to pressure (in Pa) at fixed  
!                 Absolute Salinity                                [ K/Pa ]
!--------------------------------------------------------------------------

use gsw_mod_toolbox, only : gsw_t_deriv_chem_potential_water_t_exact
use gsw_mod_toolbox, only : gsw_specvol_ice, gsw_dilution_coefficient_t_exact
use gsw_mod_toolbox, only : gsw_entropy_ice, gsw_t_freezing_exact, gsw_gibbs
use gsw_mod_toolbox, only : gsw_specvol_t_exact

use gsw_mod_teos10_constants, only : gsw_sso

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sa, p, saturation_fraction
real (r8), intent(out), optional :: tfreezing_sa, tfreezing_p

real (r8) :: rec_denom, tf

real (r8), parameter :: g_per_kg = 1000.0_r8

tf = gsw_t_freezing_exact(sa,p,saturation_fraction) 
rec_denom = 1.0_r8/(g_per_kg*gsw_t_deriv_chem_potential_water_t_exact(sa,tf,p) &
                  + gsw_entropy_ice(tf,p))

if (present(tfreezing_sa)) tfreezing_sa = &
               gsw_dilution_coefficient_t_exact(sa,tf,p)*rec_denom &
               + saturation_fraction*(1e-3_r8)/(2.0_r8*gsw_sso)

if (present(tfreezing_p)) tfreezing_p = &
               -(gsw_specvol_t_exact(sa,tf,p) - sa*gsw_gibbs(1,0,1,sa,tf,p) &
               - gsw_specvol_ice(tf,p))*rec_denom

return
end subroutine

!--------------------------------------------------------------------------
!==========================================================================
elemental subroutine gsw_t_freezing_first_derivatives_poly (sa, p, &
                            saturation_fraction, tfreezing_sa, tfreezing_p)
!==========================================================================
!
!  Calculates the frist derivatives of the in-situ temperature at which 
!  seawater freezes with respect to Absolute Salinity SA and pressure P (in
!  Pa).  These expressions come from differentiating the expression that
!  defines the freezing temperature, namely the equality between the 
!  chemical potentials of water in seawater and in ice.  
!
!  SA  =  Absolute Salinity                                        [ g/kg ]
!  p   =  sea pressure                                             [ dbar ]
!         ( i.e. absolute pressure - 10.1325 dbar ) 
!  saturation_fraction = the saturation fraction of dissolved air in 
!                        seawater
!
!  tfreezing_SA = the derivative of the in-situ freezing temperature 
!                 (ITS-90) with respect to Absolute Salinity at fixed    
!                 pressure                     [ K/(g/kg) ] i.e. [ K kg/g ]
!
!  tfreezing_P  = the derivative of the in-situ freezing temperature  
!                 (ITS-90) with respect to pressure (in Pa) at fixed  
!                 Absolute Salinity                                [ K/Pa ]
!--------------------------------------------------------------------------

use gsw_mod_teos10_constants, only : gsw_sso

use gsw_mod_freezing_poly_coefficients

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sa, p, saturation_fraction
real (r8), intent(out), optional :: tfreezing_sa, tfreezing_p

real (r8) :: p_r, sa_r, x

real (r8), parameter :: c = 1e-3_r8/(2.0_r8*gsw_sso)

sa_r = sa*1e-2_r8
x = sqrt(sa_r)
p_r = p*1e-4_r8

if (present(tfreezing_sa)) tfreezing_sa = &
    (t1 + x*(1.5_r8*t2 + x*(2.0_r8*t3 + x*(2.5_r8*t4 + x*(3.0_r8*t5 &
        + 3.5_r8*t6*x)))) + p_r*(t10 + x*(1.5_r8*t11 + x*(2.0_r8*t13 &
        + x*(2.5_r8*t16 + x*(3.0_r8*t19 + 3.5_r8*t22*x)))) &
        + p_r*(t12 + x*(1.5_r8*t14 + x*(2.0_r8*t17 + 2.5_r8*t20*x)) &
        + p_r*(t15 + x*(1.5_r8*t18 + 2.0_r8*t21*x)))))*1e-2_r8 &
        + saturation_fraction*c

if (present(tfreezing_p)) tfreezing_p = &
    (t7 + sa_r*(t10 + x*(t11 + x*(t13 + x*(t16 + x*(t19 + t22*x))))) &
        + p_r*(2.0_r8*t8 + sa_r*(2.0_r8*t12 + x*(2.0_r8*t14 + x*(2.0_r8*t17 &
        + 2.0_r8*t20*x))) + p_r*(3.0_r8*t9 + sa_r*(3.0_r8*t15 + x*(3.0_r8*t18 &
        + 3.0_r8*t21*x)))))*1e-8_r8

return
end subroutine

!--------------------------------------------------------------------------
!==========================================================================
elemental function gsw_t_freezing_poly (sa, p, saturation_fraction, polynomial)
!==========================================================================
!
!  Calculates the in-situ temperature at which seawater freezes from a 
!  computationally efficient polynomial.
!
!  SA  =  Absolute Salinity                                        [ g/kg ]
!  p   =  sea pressure                                             [ dbar ]
!         ( i.e. absolute pressure - 10.1325 dbar ) 
!  saturation_fraction = the saturation fraction of dissolved air in 
!                        seawater
!
!  t_freezing = in-situ temperature at which seawater freezes.    [ deg C ]
!               (ITS-90)                
!--------------------------------------------------------------------------

use gsw_mod_teos10_constants, only : gsw_sso

use gsw_mod_freezing_poly_coefficients

use gsw_mod_toolbox, only : gsw_ct_freezing_poly, gsw_t_from_ct

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sa, p
real (r8), intent(in), optional :: saturation_fraction
logical, intent(in), optional :: polynomial

real (r8) :: gsw_t_freezing_poly

real (r8) :: p_r, sa_r, x, ctf, sfrac
logical :: direct_poly

if (present(polynomial)) then
   direct_poly = polynomial
else
   direct_poly = .false.
end if

if (.not. direct_poly) then

   if (present(saturation_fraction)) then
      sfrac = saturation_fraction
   else
      sfrac = 1.0_r8
   end if

   ctf = gsw_ct_freezing_poly(sa,p,sfrac)
   gsw_t_freezing_poly = gsw_t_from_ct(sa,ctf,p)

else

   ! Alternative calculation ...
   sa_r = sa*1e-2_r8
   x = sqrt(sa_r)
   p_r = p*1e-4_r8

   gsw_t_freezing_poly = t0 &
       + sa_r*(t1 + x*(t2 + x*(t3 + x*(t4 + x*(t5 + t6*x))))) &
       + p_r*(t7 + p_r*(t8 + t9*p_r)) &
       + sa_r*p_r*(t10 + p_r*(t12 + p_r*(t15 + t21*sa_r)) &
       + sa_r*(t13 + t17*p_r + t19*sa_r) &
       + x*(t11 + p_r*(t14 + t18*p_r) + sa_r*(t16 + t20*p_r + t22*sa_r)))

   if (.not. present(saturation_fraction)) return

   ! Adjust for the effects of dissolved air
   gsw_t_freezing_poly = gsw_t_freezing_poly - &
                  saturation_fraction*(1e-3_r8)*(2.4_r8 - sa/(2.0_r8*gsw_sso))
end if

return
end function

!--------------------------------------------------------------------------
!==========================================================================
elemental function gsw_t_from_ct (sa, ct, p)  
!==========================================================================
!
! Calculates in-situ temperature from Conservative Temperature of seawater  
!
! sa      : Absolute Salinity                              [g/kg]
! ct      : Conservative Temperature                       [deg C]
!
! gsw_t_from_ct : in-situ temperature                      [deg C]
!--------------------------------------------------------------------------

use gsw_mod_toolbox, only : gsw_pt_from_ct, gsw_pt_from_t

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sa, ct, p  

real (r8) :: gsw_t_from_ct

real (r8) :: pt0

real (r8), parameter :: p0 = 0.0_r8

pt0 = gsw_pt_from_ct(sa,ct)
gsw_t_from_ct = gsw_pt_from_t(sa,pt0,p0,p)

return
end function

!--------------------------------------------------------------------------
! =========================================================================
elemental function gsw_t_from_pt0_ice (pt0_ice, p)
! =========================================================================
!
!  Calculates in-situ temperature from the potential temperature of ice Ih 
!  with reference pressure, p_ref, of 0 dbar (the surface), and the 
!  in-situ pressure.
!
!  pt0_ice  =  potential temperature of ice Ih with reference pressure of 
!              zero dbar (ITS-90)                                 [ deg C ]
!  p        =  sea pressure                                        [ dbar ]
!         ( i.e. absolute pressure - 10.1325 dbar )
!--------------------------------------------------------------------------

use gsw_mod_toolbox, only : gsw_pt_from_t_ice

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: pt0_ice, p

real (r8) :: gsw_t_from_pt0_ice

real (r8), parameter :: p0 = 0.0_r8

gsw_t_from_pt0_ice = gsw_pt_from_t_ice(pt0_ice,p0,p)

return
end function

!--------------------------------------------------------------------------
!==========================================================================
elemental function gsw_thermobaric (sa, ct, p)
!==========================================================================
!
!  Calculates the thermobaric coefficient of seawater with respect to
!  Conservative Temperature.  This routine is based on the 
!  computationally-efficient expression for specific volume in terms of
!  SA, CT and p (Roquet et al., 2014).
!
!  SA  =  Absolute Salinity                                        [ g/kg ]
!  CT  =  Conservative Temperature (ITS-90)                       [ deg C ]
!  p   =  sea pressure                                             [ dbar ]
!         ( i.e. absolute pressure - 10.1325 dbar )
!
!  thermobaric  =  thermobaric coefficient with                [ 1/(K Pa) ] 
!                  respect to Conservative Temperature.           
!  Note. The pressure derivative is taken with respect to
!    pressure in Pa not dbar.
!--------------------------------------------------------------------------

use gsw_mod_toolbox, only : gsw_rho, gsw_specvol_first_derivatives
use gsw_mod_toolbox, only : gsw_specvol_second_derivatives

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sa, ct, p

real (r8) :: gsw_thermobaric

real (r8) :: v_ct, v_ct_p, v_sa, v_sa_p

call gsw_specvol_first_derivatives(sa,ct,p,v_sa=v_sa,v_ct=v_ct)

call gsw_specvol_second_derivatives(sa,ct,p,v_sa_p=v_sa_p,v_ct_p=v_ct_p)

gsw_thermobaric = gsw_rho(sa,ct,p)*(v_ct_p - (v_ct/v_sa)*v_sa_p)
 
return
end function

!--------------------------------------------------------------------------
!==========================================================================
pure subroutine gsw_turner_rsubrho (sa, ct, p, tu, rsubrho, p_mid)
!==========================================================================
!
!  Calculates the Turner angle and the Rsubrho as a function of pressure 
!  down a vertical water column.  These quantities express the relative 
!  contributions of the vertical gradients of Conservative Temperature 
!  and Absolute Salinity to the vertical stability (the square of the 
!  Brunt-Vaisala Frequency squared, N^2).  Tu and Rsubrho are evaluated at 
!  the mid pressure between the individual data points in the vertical.  
!  density in terms of SA, CT and p (IOC et al., 2010).
!
!  Note that in the double-diffusive literature, papers concerned with
!  the "diffusive" form of double-diffusive convection often define the 
!  stability ratio as the reciprocal of what is defined here as the 
!  stability ratio.  
!
! sa      : Absolute Salinity         (a profile (length nz))     [g/kg]
! ct      : Conservative Temperature  (a profile (length nz))     [deg C]
! p       : sea pressure              (a profile (length nz))     [dbar]
! tu      : Turner angle, on the same (nz-1) grid as p_mid.
!           Turner angle has units of:           [ degrees of rotation ]
! rsubrho : Stability Ratio, on the same (nz-1) grid as p_mid.
!           rsubrho is dimensionless.                       [ unitless ]
! p_mid   : Mid pressure between p grid  (length nz-1)           [dbar]
!--------------------------------------------------------------------------

use gsw_mod_toolbox, only : gsw_specvol_alpha_beta

use gsw_mod_teos10_constants, only : rad2deg

use gsw_mod_error_functions, only : gsw_error_code

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: sa(:), ct(:), p(:)
real (r8), intent(out) :: tu(:), rsubrho(:), p_mid(:)

integer :: nz, k
real (r8), dimension(:), allocatable :: dsa, sa_mid, dct, ct_mid, dp
real (r8), dimension(:), allocatable :: alpha_mid, beta_mid

character (*), parameter :: func_name = "gsw_turner_rsubrho"

nz = size(sa)
if (size(tu).lt.nz-1 .or. size(rsubrho).lt.nz-1 .or. size(p_mid).lt.nz-1) then
    tu = gsw_error_code(1,func_name)
    rsubrho = tu(1)
    p_mid = tu(1)
    return
end if

allocate (dsa(nz-1), sa_mid(nz-1), dct(nz-1), ct_mid(nz-1), dp(nz-1))
allocate (alpha_mid(nz-1), beta_mid(nz-1))

forall (k = 1: nz-1)
    dsa(k) = (sa(k) - sa(k+1))
    sa_mid(k) = 0.5_r8*(sa(k) + sa(k+1))
    dct(k) = (ct(k) - ct(k+1))
    ct_mid(k) = 0.5_r8*(ct(k) + ct(k+1))
    dp(k) = (p(k) - p(k+1))
    p_mid(k) = 0.5_r8*(p(k) + p(k+1))
end forall

call gsw_specvol_alpha_beta(sa_mid,ct_mid,p_mid(1:nz-1), &
                            alpha=alpha_mid,beta=beta_mid)

tu(1:nz-1) = rad2deg*atan2((alpha_mid*dct + beta_mid*dsa), &
                              (alpha_mid*dct - beta_mid*dsa))

where (dsa .ne. 0.0_r8)
   rsubrho = (alpha_mid*dct)/(beta_mid*dsa)
elsewhere
   rsubrho = gsw_error_code(2,func_name)
end where

deallocate (dsa, sa_mid, dct, ct_mid, dp)
deallocate (alpha_mid, beta_mid)

return
end subroutine

!--------------------------------------------------------------------------
!==========================================================================
pure subroutine gsw_util_indx (x, n, z, k)
!==========================================================================
!
!  Finds the index of the value in a monotonically increasing array
!
!  x     :  array of monotonically increasing values
!  n     :  length of the array
!  z     :  value to be indexed
!
!  k      : index K :- if x(k) <= z < x(k+1), or
!               n-1 :- if z = x(n)
!--------------------------------------------------------------------------

use gsw_mod_kinds

integer, intent(in) :: n
integer, intent(out) :: k
real (r8), intent(in), dimension(n) :: x
real (r8), intent(in) :: z

integer :: ku, kl, km

if(z.gt.x(1).and.z.lt.x(n)) then
   kl=1
   ku=n
   do while (ku-kl.gt.1)
      km=(ku+kl)/2
      if(z.gt.x(km)) then
         kl=km
      else
         ku=km
      endif
   end do
   k=kl
   if(z.eq.x(k+1)) then 
     k = k+1
   end if
elseif (z.le.x(1)) then
      k = 1
else    !if (z.ge.x(n)) then - removed (GBH 3/6/2015) so z=NaN has somewhere to go (otherwise k is undefined and gives segmentation fault)
      k = n-1
end if

return
end subroutine

!--------------------------------------------------------------------------

!==========================================================================
pure function gsw_util_interp1q_int (x, iy, x_i) result(y_i)
!==========================================================================
! Returns the value of the 1-D function iy (integer) at the points of column
! vector x_i using linear interpolation. The vector x specifies the
! coordinates of the underlying interval.
!==========================================================================

use gsw_mod_toolbox, only : gsw_util_sort_real

use gsw_mod_kinds

implicit none

integer, intent(in) :: iy(:)
real (r8), intent(in) :: x(:), x_i(:)

real (r8) :: y_i(size(x_i))

logical, allocatable :: in_rng(:)
integer, allocatable :: j(:), k(:), r(:), jrev(:), ki(:)
real (r8), allocatable :: xi(:), xxi(:), u(:)

integer :: imax_x(1), imin_x(1), i, n, m, ii, nx, nxi
real (r8) :: max_x, min_x

min_x = minval(x); imin_x = minloc(x)
max_x = maxval(x); imax_x = maxloc(x)

where (x_i .le. min_x) y_i = real(iy(imin_x(1)),r8)
where (x_i .ge. max_x) y_i = real(iy(imax_x(1)),r8)

nx = size(x)
nxi = size(x_i)

allocate (in_rng(nxi))
in_rng = (x_i .gt. min_x .and. x_i .lt. max_x)

n = count(in_rng)
if (n .eq. 0) return

allocate (xi(n), k(n), ki(n), r(n), u(n))
m = nx + n
allocate (xxi(m), j(m), jrev(m))

ii = 0
do i = 1, nxi
    if (in_rng(i)) then
        ii = ii + 1
        xi(ii) = x_i(i)
        ki(ii) = i
    end if
end do

k = gsw_util_sort_real(xi)
xxi = (/ x, xi(k) /)
j = gsw_util_sort_real(xxi)

jrev(j) = (/ (i, i=1,size(j)) /)
r(k) = jrev(nx+1:) - (/ (i, i=1,size(xi)) /)

u = (xi-x(r))/(x(r+1)-x(r))
y_i(ki) = iy(r) + (iy(r+1)-iy(r))*u

return

end function

!--------------------------------------------------------------------------
pure function gsw_util_sort_real (rarray) result(iarray)

! Sorts a real array and returns a list of sorted indices. The sort
! algorithm needs to be stable (should preserve the order of items of
! equal value).

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: rarray(:)    ! Values to be sorted

integer :: iarray(size(rarray))       ! Sorted ids

integer :: i, nx, depth_limit

nx = size(rarray)
depth_limit = 2*floor(log(nx*1.0_r8)/log(2.0_r8))

iarray = (/ (i, i=1,nx) /)
call introsort_index(iarray,1,nx,depth_limit)
return

contains

    !--------------------------------------------------------------------------

    pure recursive subroutine introsort_index (indx, lo, hi, depth_limit)

    ! An implementation of Introsort (also called introspective sort) - a
    ! sorting algorithm designed by David Musser in 1997. It is based on
    ! quicksort but introduces two optimizations to speed up the typical
    ! quicksort algorithm:
    ! - if the quicksort algorithm on any sub-array runs longer than expected
    !   (recursion is deeper than log(array size)) then switch to heapsort
    !   which is guaranteed to finish in O(n*log(n)) time;
    ! - if the number of elements is small enough then switch to an insertion
    !   sort which is faster on small data sets.
    ! These two optimizations allow introsort to outperform the “dumb”
    ! quicksort implementation on most arrays (even ~200 times on artificial
    ! “evil” arrays consisting of 100,000 elements).
    !
    ! Note 1: this version returns an array of sorted indices. So the sorted
    ! data is real_array(indx) - Glenn Hyland 20/8/2015
    !
    ! Note 2: this version of Introsort is stable (preserves the order of
    ! items of equal value) - Glenn Hyland 28/8/2015

    use gsw_mod_kinds

    implicit none

    integer, intent(inout) :: indx(:)
    integer, intent(in) :: lo, hi, depth_limit

    integer :: mid, p
    real (r8) :: pivot

    integer, parameter :: size_threshold = 16

    if (hi-lo .gt. size_threshold) then

        if (depth_limit .eq. 0) then
            call heapsort_index(indx(lo:hi))
            return
        end if

        mid = lo + (hi-lo)/2                   ! avoids potential overflow
        pivot = medianof3(indx,lo,mid,hi)
        call partition(indx,lo,hi,pivot,p)

        call introsort_index(indx,lo,p-1,depth_limit-1)
        call introsort_index(indx,p,hi,depth_limit-1)

    else

        call insertionsort_index(indx,lo,hi)

    end if
    return

    end subroutine introsort_index

    !--------------------------------------------------------------------------

    pure subroutine partition (indx, lo, hi, pivot, p)

    implicit none

    integer, intent(inout) :: indx(:)
    integer, intent(in) :: lo, hi
    real (r8), intent(in) :: pivot
    integer, intent(out) :: p

    integer :: left, right, t

    left = lo - 1
    right = hi + 1
    do while (left .lt. right)
        right = right - 1
        do while (pivot .lt. rarray(indx(right)))
            right = right - 1
        end do
        left = left + 1
        do while (rarray(indx(left)) .lt. pivot)
            left = left + 1
        end do
        if (left < right) then
            t = indx(left); indx(left) = indx(right); indx(right) = t
        end if
    end do
    if (left .eq. right) then
        p = left + 1
    else 
        p = left
    end if
    return
 
    end subroutine partition

    !--------------------------------------------------------------------------

    pure function medianof3 (indx, lo, mid, hi)

    implicit none

    integer, intent(in) :: indx(:), lo, mid, hi

    real (r8) :: medianof3

    if (rarray(indx(mid)) .lt. rarray(indx(lo))) then
        if (rarray(indx(hi)) .lt. rarray(indx(mid))) then
              medianof3 = rarray(indx(mid))
        else
            if (rarray(indx(hi)) .lt. rarray(indx(lo))) then
                medianof3 = rarray(indx(hi))
            else
                medianof3 = rarray(indx(lo))
            end if
        end if
    else
        if (rarray(indx(hi)) .lt. rarray(indx(mid))) then
            if (rarray(indx(hi)) .lt. rarray(indx(lo))) then
                medianof3 = rarray(indx(lo))
            else
                medianof3 = rarray(indx(hi))
            end if
        else
            medianof3 = rarray(indx(mid))
        end if
    end if
    return

    end function medianof3

    !--------------------------------------------------------------------------

    pure subroutine insertionsort_index (indx, lo, hi)

    ! Insertion sort algorithm
    !
    ! Note: this version returns an array of sorted indices. So the sorted
    ! data is real_array(indx) - Glenn Hyland 20/8/2015

    implicit none

    integer, intent(inout) :: indx(:)
    integer, intent(in) :: lo, hi

    integer :: i, j, t

    do i = lo+1, hi
        j = i - 1
        t = indx(i)
        do while (j .ge. lo .and. rarray(indx(j)) .gt. rarray(t))
            indx(j+1) = indx(j)
            j = j - 1
        end do
        indx(j+1) = t
    end do
    return

    end subroutine insertionsort_index

    !--------------------------------------------------------------------------

    pure subroutine heapsort_index (indx)
 
    ! Heapsort algorithm
    !
    ! Note 1: this version returns an array of sorted indices. So the sorted
    ! data is real_array(indx) - Glenn Hyland 20/8/2015
    !
    ! Note 2: this version of Heapsort is stable (preserves the order of items
    ! of equal value) through the use of the lexicographic key comparison
    ! function "less_than" - Glenn Hyland 28/8/2015

    integer, intent(inout) :: indx(0:)

    integer :: start, n, bottom, t
 
    n = size(indx)
    do start = (n-2)/2, 0, -1
       call siftdown(indx,start,n);
    end do
 
    do bottom = n-1, 1, -1
       t = indx(0); indx(0) = indx(bottom); indx(bottom) = t;
       call siftdown(indx,0,bottom)
    end do
    return
 
    end subroutine heapsort_index
 
    !--------------------------------------------------------------------------

    pure subroutine siftdown (indx, start, bottom)
 
    integer, intent(inout) :: indx(0:)
    integer, intent(in) :: start, bottom

    integer :: child, root, t
 
    root = start
    do while(root*2 + 1 < bottom)

       child = root*2 + 1
       if (child+1 < bottom) then
          if (less_than(indx(child),indx(child+1))) child = child + 1
       end if
 
       if (less_than(indx(root),indx(child))) then
          t = indx(child); indx(child) = indx(root); indx(root) = t
          root = child
       else
          return
       end if  
    end do      
    return
 
    end subroutine siftdown
 
    !--------------------------------------------------------------------------

    pure logical function less_than (indx1, indx2)

    ! Lexicographic key comparison function to ensure sorting stability so
    ! the order of items of equal value is preserved.

    integer, intent(in) :: indx1, indx2

    if (rarray(indx1) .lt. rarray(indx2)) then
        less_than = .true.
    else if (rarray(indx1) .eq. rarray(indx2)) then
        if (indx1 .lt. indx2) then
            less_than = .true.
        else
            less_than = .false.
        end if
    else
        less_than = .false.
    end if
    return

    end function less_than

end function

!--------------------------------------------------------------------------
!==========================================================================
pure function gsw_util_xinterp1 (x, y, n, x0)
!==========================================================================
!
! Linearly interpolate a real array   
!
! x      : x array (must be monotonic)               
! y      : y array     
! n      : length of X and Y arrays
! x0     : value to be interpolated
!
! xinterp1 : Linearly interpolated value
!--------------------------------------------------------------------------

use gsw_mod_toolbox, only : gsw_util_indx

use gsw_mod_kinds

implicit none

integer, intent(in) :: n
real (r8), intent(in) :: x0
real (r8), dimension(n), intent(in) :: x, y

real (r8) :: gsw_util_xinterp1

integer :: k

real (r8) :: r

call gsw_util_indx(x,n,x0,k)
r = (x0-x(k))/(x(k+1)-x(k))
gsw_util_xinterp1 = y(k) + r*(y(k+1)-y(k))

return
end function

!--------------------------------------------------------------------------
!==========================================================================
elemental function gsw_z_from_p (p, lat, geo_strf_dyn_height, &
                                                   sea_surface_geopotental)
!==========================================================================
!
!  Calculates height from sea pressure using the computationally-efficient
!  75-term expression for specific volume in terms of SA, CT and p 
!  (Roquet et al., 2015).  Dynamic height anomaly, geo_strf_dyn_height, if
!  provided, must be computed with its p_ref = 0 (the surface).  Also if
!  provided, sea_surface_geopotental is the geopotential at zero sea 
!  pressure. This function solves Eqn.(3.32.3) of IOC et al. (2010).  
!
!  Note. Height z is NEGATIVE in the ocean. i.e. Depth is -z.  
!   Depth is not used in the GSW computer software library.  
!
!  p    =  sea pressure                                            [ dbar ]
!          ( i.e. absolute pressure - 10.1325 dbar )
!  lat  =  latitude in decimal degrees north                [ -90 ... +90 ]
!
! OPTIONAL:
!  geo_strf_dyn_height = dynamic height anomaly                 [ m^2/s^2 ]
!    Note that the refernce pressure, p_ref, of geo_strf_dyn_height must be 
!    zero (0) dbar.
!  sea_surface_geopotental = geopotential at zero sea pressure  [ m^2/s^2 ]
!
!  z  =  height                                                       [ m ]
!
!==========================================================================

use gsw_mod_toolbox, only : gsw_enthalpy_sso_0

use gsw_mod_teos10_constants, only : deg2rad, gamma

use gsw_mod_kinds

implicit none

real (r8), intent(in) :: p, lat
real (r8), intent(in), optional :: geo_strf_dyn_height
real (r8), intent(in), optional :: sea_surface_geopotental

real (r8) :: gsw_z_from_p

real (r8) :: a, b, c, g, sin2, gsdh, ssg

if (present(geo_strf_dyn_height)) then
    gsdh = geo_strf_dyn_height
else
    gsdh = 0.0_r8
end if

if (present(sea_surface_geopotental)) then
    ssg = sea_surface_geopotental
else
    ssg = 0.0_r8
end if

g = gamma    ! If the graviational acceleration were to be regarded as 
             ! being depth-independent, which is often the case in 
             ! ocean models, then gamma would be set to be zero here,
	     ! and the code below works perfectly well.

sin2 = sin(lat*deg2rad)**2
b = 9.780327_r8*(1.0_r8 + (5.2792e-3_r8 + (2.32e-5_r8*sin2))*sin2) 
a = -0.5_r8*g*b
c = gsw_enthalpy_sso_0(p) - (gsdh + ssg)

gsw_z_from_p = -2.0_r8*c/(b + sqrt(b*b - 4.0_r8*a*c))

return
end function

!--------------------------------------------------------------------------
